#include "yac_config.h"

PROGRAM dummy_io

#ifdef YAC_MPI_ENABLED
  USE mpi
#endif

  USE mo_yac_finterface, ONLY: yac_finit, yac_fget_localcomm,      &
       &                       yac_fdef_elements, yac_fdef_points, &
       &                       yac_fdef_mask, yac_fdef_field,      &
       &                       yac_fsearch, yac_ffinalize,         &
       &                       YAC_LOCATION_CELL

  IMPLICIT NONE

  INTEGER, PARAMETER :: pd =  12
  INTEGER, PARAMETER :: rd = 307

  INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(pd,rd) !< double precision
  INTEGER, PARAMETER :: wp = dp                        !< selected working precision

  INTEGER, PARAMETER :: no_of_fields = 8
  INTEGER, PARAMETER :: max_char_length = 132

  INTEGER, PARAMETER :: nbr_subdomain_ids = 1

  INTEGER, PARAMETER :: nbr_cells = 2
  INTEGER, PARAMETER :: nbr_vertices = 4

  CHARACTER(LEN=max_char_length) :: field_name(no_of_fields)
  CHARACTER(LEN=max_char_length) :: xml_filename
  CHARACTER(LEN=max_char_length) :: xsd_filename
  CHARACTER(LEN=max_char_length) :: grid_name
  CHARACTER(LEN=max_char_length) :: comp_name

  INTEGER :: i, igrid, ierror

  INTEGER :: comp_ids(1)
  INTEGER :: cell_point_ids(2)
  INTEGER :: cell_mask_ids(1)
  INTEGER :: domain_id(2)
  INTEGER :: subdomain_id
  INTEGER :: subdomain_ids(nbr_subdomain_ids)

  INTEGER :: glb_index(nbr_cells)
  INTEGER :: owner_local(nbr_cells)
  INTEGER :: nbr_vertices_per_cell

  REAL(wp), ALLOCATABLE :: buffer_lon(:)
  REAL(wp), ALLOCATABLE :: buffer_lat(:)
  INTEGER, ALLOCATABLE  :: cell_to_vertex(:,:)

  INTEGER, ALLOCATABLE  :: cell_mask(:)
  INTEGER, ALLOCATABLE  :: field_id(:)

  INTEGER :: local_comm, npes, rank

#ifdef YAC_MPI_ENABLED
  CALL mpi_init (ierror)
#endif

  ! Initialise the coupler
  xml_filename = "dummy_coupling.xml"
  xsd_filename = "coupling.xsd"
  CALL yac_finit ( xml_filename, xsd_filename )

  ! Inform the coupler about what we are
  comp_name = "dummy_io"
  CALL yac_fdef_comp ( comp_name, comp_ids(1) )

  CALL yac_fget_localcomm ( local_comm, comp_ids(1) )

#ifdef YAC_MPI_ENABLED
  CALL mpi_comm_rank ( local_comm, rank, ierror )
  CALL mpi_comm_size ( local_comm, npes, ierror )
#else
  rank = 0
  npes = 1
#endif

  WRITE ( 6 , * ) TRIM(comp_name), " rank ", rank, ": local size is ", npes

  ! Here we define two grids (although numrically identical) to mimick
  ! two different Output server groups

  DO igrid = 1, 2

     ! Announce one subdomain (patch) to the coupler
     IF ( igrid == 1 ) THEN
        grid_name = "ocean_grid"
     ELSE
        grid_name = "atmos_grid"
     ENDIF

     WRITE ( 6 , * ) "Calling yac_fdef_subdomain ", igrid
     FLUSH(6)

     CALL yac_fdef_subdomain ( comp_ids(1), grid_name, subdomain_id )

     subdomain_ids(1) = subdomain_id

     ALLOCATE(buffer_lon(nbr_vertices))
     ALLOCATE(buffer_lat(nbr_vertices))
     ALLOCATE(cell_to_vertex(3,nbr_cells))

     nbr_vertices_per_cell = 3

     ! Define vertices

     !      1
     !     / \
     !    / o \
     !   /     \
     !  2-------3   Eq.
     !   \     /
     !    \ o /
     !     \ /
     !      4

     buffer_lon(1) =  0.0; buffer_lat(1) =  1.0
     buffer_lon(2) = -1.0; buffer_lat(2) =  0.0
     buffer_lon(3) =  1.0; buffer_lat(3) =  0.0
     buffer_lon(4) =  0.0; buffer_lat(4) = -1.0

     ! Connectivity
     cell_to_vertex(1,1) = 1; cell_to_vertex(2,1) = 2; cell_to_vertex(3,1) = 3 ! cell 1
     cell_to_vertex(1,2) = 2; cell_to_vertex(2,2) = 4; cell_to_vertex(3,2) = 3 ! cell 2

     ! Description of elements, here as unstructured grid
     CALL yac_fdef_elements (         &
          & subdomain_id,             &
          & nbr_vertices,             &
          & nbr_cells,                &
          & nbr_vertices_per_cell,    &
          & buffer_lon,               &
          & buffer_lat,               &
          & cell_to_vertex )

     ! Center points in cells (needed e.g. for patch recovery and nearest neighbour)

     buffer_lon(1) = 0.0; buffer_lat(1) =  0.5
     buffer_lon(2) = 0.0; buffer_lat(2) = -0.5

     CALL yac_fdef_points (           &
          & subdomain_id,             &
          & nbr_cells,                &
          & YAC_LOCATION_CELL,        &
          & buffer_lon,               &
          & buffer_lat,               &
          & cell_point_ids(igrid) )

     DEALLOCATE (buffer_lon, buffer_lat, cell_to_vertex)

     ! Decomposition information

     DO i = 1, nbr_cells
        glb_index(i) = i
        owner_local(i) = -1
     ENDDO

     CALL yac_fdef_index_location (   &
          & subdomain_id,             &
          & nbr_cells,                &
          & YAC_LOCATION_CELL,        &
          & glb_index,                &
          & owner_local )

     ! Connect subdomains
     CALL yac_fconnect_subdomains (   &
          & comp_ids(1),              &
          & nbr_subdomain_ids,        &
          & subdomain_ids,            &
          & domain_id(igrid) )

     ! Mask generation
     ALLOCATE(cell_mask(nbr_cells))
     DO i = 1, nbr_cells
        cell_mask(i) = 1
     ENDDO

     CALL yac_fdef_mask (          &
          & nbr_cells,             &
          & cell_mask,             &
          & cell_point_ids(igrid), &
          & cell_mask_ids(1) )

     DEALLOCATE (cell_mask)

  ENDDO

  field_name(1) = "atmos_out1" ! output field
  field_name(2) = "atmos_out2" ! output field
  field_name(3) = "atmos_out3" ! output field
  field_name(4) = "atmos_out4" ! output field

  field_name(5) = "ocean_out1" ! output field
  field_name(6) = "ocean_out2" ! output field
  field_name(7) = "ocean_out3" ! output field
  field_name(8) = "ocean_out4" ! output field

  ALLOCATE(field_id(no_of_fields))

  ! atmosphere output

  DO i = 1, no_of_fields-4
     CALL yac_fdef_field (     &
          & field_name(i),     &
          & comp_ids(1),       &
          & domain_id(2),      &
          & cell_point_ids(2), &
          & cell_mask_ids,     &
          & 1,                 &
          & field_id(i) )
  ENDDO

  ! ocean output

  DO i = no_of_fields-3, no_of_fields
     CALL yac_fdef_field (     &
          & field_name(i),     &
          & comp_ids(1),       &
          & domain_id(1),      &
          & cell_point_ids(1), &
          & cell_mask_ids,     &
          & 1,                 &
          & field_id(i) )
  ENDDO

  CALL yac_fsearch ( 1, comp_ids, no_of_fields, field_id, ierror )

  CALL yac_ffinalize

#ifdef YAC_MPI_ENABLED
  CALL mpi_finalize (ierror)
#endif

END PROGRAM dummy_io
