package de.mpg.mpimet.yacgui.xml;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import de.mpg.mpimet.yacgui.utils.Default;
import de.mpg.mpimet.yacgui.utils.Utils;

/**
 * Java representation of main {@code <coupling>} xml element.
 * 
 * @author Maxim Yastremsky
 */
public class Coupling {

	/** componentId : Component */
	private Map<Integer, Component> components;

	/** transientId : Transient */
	private Map<Integer, Transient> transients;

	/** gridId : Grid */
	private Map<Integer, Grid> grids;

	private Dates dates;

	private TimestepUnit timestepUnit;
	
	private boolean stdoutRedirect;
	
	private boolean rootRedirect;

	/** "component1Id-component2Id" : Couple */
	private Map<String, Couple> couples;

	public Coupling() {
		components = new HashMap<Integer, Component>();
		transients = new HashMap<Integer, Transient>();
		grids = new HashMap<Integer, Grid>();
		
		dates = Default.DATES;
		timestepUnit = Default.TIMESTEP_UNIT;
		couples = new HashMap<String, Couple>();
	}

	public Coupling(boolean stdoutRedirect, boolean rootRedirect, Map<Integer, Component> components,
			Map<Integer, Transient> transients, Map<Integer, Grid> grids,
			Dates dates, TimestepUnit timestepUnit, Map<String, Couple> couples) {
		this.components = components;
		this.transients = transients;
		this.grids = grids;
		this.dates = dates;
		this.timestepUnit = timestepUnit;
		this.couples = couples;
		this.stdoutRedirect = stdoutRedirect;
		this.rootRedirect = rootRedirect;
	}
	
	public boolean isEmpty() {
		return (components.isEmpty() 
				&& transients.isEmpty() 
				&& grids.isEmpty()
				&& dates.equals(Default.DATES)
				&& timestepUnit == Default.TIMESTEP_UNIT 
				&& couples.isEmpty());
	}

	/**
	 * Adds given Component to the Coupling and updates transient_refs and grid_refs fields of 
	 * transient_grid_ref according new sets of transients and grids.
	 * 
	 * <p>
	 * The Component object will be added to this Coupling's 
	 * <code>Map&lt;Integer, Component&gt;</code>. The Component's 
	 * transients and grids will be added to this Coupling's 
	 * <code>Map&lt;Integer, Transient&gt;</code> and 
	 * <code>Map&lt;Integer, Grid&gt;</code>, if they do not exist yet with 
	 * their respective transient standard name or alias name.
	 * 
	 * @param component	component to be added
	 * @return  component updated componet
	 */
	public Component addComponent(Component component) {
		String name = component.getName();
		String model = component.getModel();
		String simulated = component.getSimulated();
		
		Map<Integer, TransientGridRef> transientGridRefs = component.getTransientGridRefs();
		
		//we need to have another pool of already updated transientGridRefs to avoid updating them twice
		Map<Integer, TransientGridRef> updatedPool = new HashMap<>();
		
		//load new transients
		for (Transient t : component.getTransients().values()) {
			int oldTransientId = t.getId();
			String transientStandardName = t.getTransientStandardName();
			
			int newTransientId = 0;
			if (!transientsContain(transientStandardName)) {
				
				//generating new lowest available id
				while (this.transients.get(++newTransientId) != null);

				t.setId(newTransientId);
				this.transients.put(newTransientId, t);
			} else {
				for (Transient existingTr : transients.values()) {
					if (existingTr.getTransientStandardName().equals(transientStandardName)) {
						newTransientId = existingTr.getId();
						break;
					}
				}
			}
			
			if (oldTransientId != newTransientId) {
				for (Iterator<Map.Entry<Integer, TransientGridRef>> it = transientGridRefs.entrySet().iterator(); it.hasNext(); ) {
					Map.Entry<Integer, TransientGridRef> mapEntry = it.next();
					
					TransientGridRef transGridRef = mapEntry.getValue();

					if (transGridRef.getTransientRef() == oldTransientId) {
						transGridRef.setTransientRef(newTransientId);
						
						updatedPool.put(transGridRef.getId(), transGridRef);
						it.remove();
					}
				}
			}
		}
		
		//combine all transientsGridRefs again and make the same for grids
		updatedPool.putAll(transientGridRefs);
		transientGridRefs.clear();
		
		for (Grid g : component.getGrids().values()) {
			int oldGridId = g.getId();
			String aliasName = g.getAliasName();
			
			int newGridId = 0;
			if (!gridsContain(aliasName)) {
				
				//generate new lowest availabel id
				while(this.grids.get(++newGridId) != null);
				
				g.setId(newGridId);
				this.grids.put(newGridId, g);
			} else {
				for (Grid existingGrid : grids.values()) {
					if (existingGrid.getAliasName().equals(aliasName)) {
						newGridId = existingGrid.getId();
						break;
					}
				}
			}
			
			if (oldGridId != newGridId) {
				for (Iterator<Map.Entry<Integer, TransientGridRef>> it = updatedPool.entrySet().iterator(); it.hasNext();) {
					Map.Entry<Integer, TransientGridRef> mapEntry = it.next();
					
					TransientGridRef transGridRef = mapEntry.getValue();
					
					if (transGridRef.getGridRef() == oldGridId) {
						transGridRef.setGridRef(newGridId);
						
						transientGridRefs.put(transGridRef.getId(), transGridRef);
						it.remove();
					}
				}
			}
		}
		
		transientGridRefs.putAll(updatedPool);
		
		int newId = 1;
		while (components.get(newId) != null) {
			newId++;
		}
		Component componentToBeAdded = new Component(newId, name, model, simulated, transientGridRefs);
		components.put(newId, componentToBeAdded);
		return componentToBeAdded;
	}

	public void removeComponent(Component component) {
		boolean wasRemoved = components.remove(component.getId(), component);

		if (wasRemoved) {

			//remove all couples with this component
			for (Iterator<Map.Entry<String, Couple>> iter = couples.entrySet().iterator(); iter.hasNext();) {
				if (iter.next().getKey().contains(String.valueOf(component.getId()))) {
					iter.remove();
				}
			}

			//update transients list
			for (Iterator<Map.Entry<Integer, Transient>> it = transients.entrySet().iterator(); it.hasNext();) {
				boolean used = checkTransientUsageByComponents(it.next().getValue());
				
				if (!used) {
					it.remove();
				}
			}
			
			//update grids list
			for (Iterator<Map.Entry<Integer, Grid>> it = grids.entrySet().iterator(); it.hasNext();) {
				boolean used = checkGridUsageByComponents(it.next().getValue());
				
				if (!used) {
					it.remove();
				}
			}
		}
		
	}
	
	private boolean checkGridUsageByComponents(Grid g) {
		for (Integer componentId : components.keySet()) {
			Map<Integer, TransientGridRef> tGRefs = components.get(componentId).getTransientGridRefs();
			
			for (Integer transGridRefId : tGRefs.keySet()) {
				if (tGRefs.get(transGridRefId).getGridRef() == g.getId()) {
					return true;
				}
			}
		}
		return false;
	}

	private boolean checkTransientUsageByComponents(Transient t) {
		for (Integer componentId : components.keySet()) {
			Map<Integer, TransientGridRef> tGRefs = components.get(componentId).getTransientGridRefs(); 

			for (Integer transGridRefId : tGRefs.keySet()) {
				if (tGRefs.get(transGridRefId).getTransientRef() == t.getId()) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Adds a new couple to couplesMap and returns its id.
	 * 
	 * @param couple
	 * @return couple id
	 */
	public String addCouple(Couple couple) {
		int component1Id = couple.getComponent1Id();
		int component2Id = couple.getComponent2Id();
		
		String coupleId = Utils.getCoupleId(component1Id, component2Id); 
		couples.put(coupleId, couple);
		
		return coupleId;
	}

	public Map<Integer, Component> getComponentsMap() {
		return components;
	}

	/**
	 * @return transients as Map of <code>transientId:Transient</code>
	 */
	public Map<Integer, Transient> getTransients() {
		return transients;
	}

	public Map<Integer, Grid> getGrids() {
		return grids;
	}

	public Dates getDates() {
		return dates;
	}

	public TimestepUnit getTimestepUnit() {
		return timestepUnit;
	}

	public Map<String, Couple> getCouples() {
		return couples;
	}

	public void setComponents(Map<Integer, Component> components) {
		this.components = components;
	}

	public void setTransients(Map<Integer, Transient> transients) {
		this.transients = transients;
	}

	public void setGrids(Map<Integer, Grid> grids) {
		this.grids = grids;
	}

	public void setDates(Dates dates) {
		this.dates = dates;
	}

	public boolean isStdoutRedirect() {
		return stdoutRedirect;
	}

	public void setStdoutRedirect(boolean stdoutRedirect) {
		this.stdoutRedirect = stdoutRedirect;
	}

	public boolean isRootRedirect() {
		return rootRedirect;
	}

	public void setRootRedirect(boolean rootRedirect) {
		this.rootRedirect = rootRedirect;
	}

	public void setTimestepUnit(TimestepUnit timestepUnit) {
		this.timestepUnit = timestepUnit;
	}

	public void setCouples(Map<String, Couple> couples) {
		this.couples = couples;
	}

	private boolean transientsContain(String transientStandardName) {
		for (Transient t : transients.values()) {
			if (t.getTransientStandardName().equals(transientStandardName)) {
				return true;
			}
		}
		return false;
	}

	private boolean gridsContain(String aliasName) {
		for (Grid g : grids.values()) {
			if (g.getAliasName().equals(aliasName)) {
				return true;
			}
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((components == null) ? 0 : components.hashCode());
		result = prime * result + ((couples == null) ? 0 : couples.hashCode());
		result = prime * result + ((dates == null) ? 0 : dates.hashCode());
		result = prime * result + ((grids == null) ? 0 : grids.hashCode());
		result = prime * result + (rootRedirect ? 1231 : 1237);
		result = prime * result + (stdoutRedirect ? 1231 : 1237);
		result = prime * result
				+ ((timestepUnit == null) ? 0 : timestepUnit.hashCode());
		result = prime * result
				+ ((transients == null) ? 0 : transients.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Coupling other = (Coupling) obj;
		if (components == null) {
			if (other.components != null)
				return false;
		} else if (!components.equals(other.components))
			return false;
		if (couples == null) {
			if (other.couples != null)
				return false;
		} else if (!couples.equals(other.couples))
			return false;
		if (dates == null) {
			if (other.dates != null)
				return false;
		} else if (!dates.equals(other.dates))
			return false;
		if (grids == null) {
			if (other.grids != null)
				return false;
		} else if (!grids.equals(other.grids))
			return false;
		if (rootRedirect != other.rootRedirect)
			return false;
		if (stdoutRedirect != other.stdoutRedirect)
			return false;
		if (timestepUnit != other.timestepUnit)
			return false;
		if (transients == null) {
			if (other.transients != null)
				return false;
		} else if (!transients.equals(other.transients))
			return false;
		return true;
	}

	
}
