/**
 * @file communicator_local.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "communicator.h"
#include "communicator_local.h"
#include "utils.h"
#include "ensure_array_size.h"

static struct local_communicator ** local_comms = NULL;
static unsigned num_local_comms = 0;

struct local_comm_message {

   struct local_comm_message * next;
   union {
      struct {
         func_recv_callback f;
         func_recv_callback_cancel f_cancel;
      } recv;
      struct {
         func_send_callback f;
      } send;
   } callback;
   void * user_data;

   struct comm_message message;
};

struct comm_request_local {

  void (*waitall)(unsigned count, struct comm_request ** requests);
  unsigned (*testsome)(unsigned count, struct comm_request ** requests);
  void (*free_request)(struct comm_request ** request);
  void * buffer;
  unsigned message_complete;

  struct comm_request_local * next;
  struct comm_request_local ** p_active_requests;
  struct comm_request_local ** p_free_requests;
};

static char ** unique_tag_stings = NULL;
unsigned num_unique_tags = 0;
unsigned unique_tags_array_size = 0;

struct local_communicator {

   struct communicator_vtable * vtable;

   unsigned rank; // there is only one rank per local comm
                  // but multiple local comms can communicate with each other
                  // this is a little hack in order to simulate multiple processes

   struct local_communicator * parent_comm;

   struct local_comm_message * open_receives;
   struct local_comm_message * open_sends;
   struct local_comm_message * free_messages;

   struct comm_request_local * active_requests;
   struct comm_request_local * free_requests;

   unsigned name_hash;

   struct {
      char * identifier_string;
      unsigned tag;
   } * unique_tags;
   unsigned num_unique_tags;
   unsigned unique_tags_array_size;
};

struct temp_gatherv_data {

  unsigned * counts;
  void * user_data;
  enum yac_message_priority priority;
  unsigned num_open_recv;
  func_recv_callback gatherv_callback;
  void * buffer;
};

struct temp_allgather_data {
  unsigned count;
  unsigned tag;
  enum yac_message_priority priority;
  enum yac_comm_data_type data_type;
  unsigned recv_count;
  void * buffer;
  void * user_data;
  func_recv_callback allgather_callback;
};

struct temp_allgatherv_data {
  unsigned * counts;
  enum yac_message_priority priority;
  enum yac_comm_data_type data_type;
  void * buffer;
  void * user_data;
  func_recv_callback allgatherv_callback;
};

struct temp_alltoallv_data {
  func_recv_callback alltoallv_callback;
  void * user_data;
  enum yac_message_priority priority;
  enum yac_comm_data_type data_type;
  unsigned data_size, count, max_buffer_size;
  void * buffer;
};

struct temp_allreduce_data {
  void * user_data;
  func_recv_callback allreduce_callback;
  func_comm_op op;
};

static unsigned get_comm_size_local(struct communicator * comm);
static unsigned get_comm_rank_local(struct communicator * comm);
static void comm_bsend_local(void const * send_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned dest,
                             unsigned tag, struct communicator * comm);
static void comm_isend_local(void const * send_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned dest,
                             unsigned tag, struct communicator * comm,
                             struct comm_request ** request);
static void
comm_recv_callback_local(unsigned count, enum yac_comm_data_type data_type,
                         unsigned source, unsigned tag,
                         struct communicator * comm, void * user_data,
                         func_recv_callback recv_callback,
                         func_recv_callback_cancel recv_callback_cancel,
                         enum yac_message_priority priority);
static void comm_recv_local(void * recv_buffer, unsigned * count,
                            enum yac_comm_data_type data_type, unsigned source,
                            unsigned tag, struct communicator * comm);
static void comm_irecv_local(void * recv_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned source,
                             unsigned tag, struct communicator * comm,
                             struct comm_request ** request);
static unsigned comm_wait_local(struct communicator * comm, unsigned tag);
static void comm_waitall_local(unsigned count, struct comm_request ** requests);
static void comm_cancel_recv_local(struct communicator * comm, unsigned tag);
static unsigned get_unique_comm_tag_local(struct communicator * comm,
                                          char * identifier_string);
static void get_unique_comm_tags_local(struct communicator * comm,
                                       char const * unique_generator_string,
                                       char const ** tag_strings,
                                       unsigned num_tag_strings,
                                       unsigned * tags);
static void comm_abort_local(struct communicator * comm);
static void free_communicator_local(struct communicator * comm);
static int comm_pack_size_local(int count, enum yac_comm_data_type type,
                                struct communicator * comm);
static void comm_pack_local(void * in_buffer, int in_count,
                            enum yac_comm_data_type type, void * out_buffer,
                            int out_buffer_size, int * position,
                            struct communicator * comm);
static void comm_unpack_local(void * in_buffer, int in_size, int * position,
                              void * out_buffer, int out_count,
                              enum yac_comm_data_type type,
                              struct communicator * comm);
static void comm_translate_ranks_local(struct communicator * comm_a,
                                       struct communicator * comm_b,
                                       unsigned * ranks);
static void comm_bcast_callback_local(void * buffer, unsigned count,
                                      enum yac_comm_data_type data_type,
                                      unsigned root, unsigned tag,
                                      struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback bcast_callback,
                                      enum yac_message_priority priority);
static void comm_gatherv_callback_local(void * send_buffer, unsigned * counts,
                                        enum yac_comm_data_type data_type,
                                        unsigned root, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback gatherv_callback,
                                        enum yac_message_priority priority);
static void comm_allgather_callback_local(void * send_buffer, unsigned count,
                                          enum yac_comm_data_type data_type,
                                          unsigned tag,
                                          struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback allgather_callback,
                                          enum yac_message_priority priority);
static void comm_allgatherv_callback_local(void * send_buffer, unsigned * counts,
                                           enum yac_comm_data_type data_type,
                                           unsigned tag,
                                           struct communicator * comm,
                                           void * user_data,
                                           func_recv_callback allgatherv_callback,
                                           enum yac_message_priority priority);
static void comm_scatter_callback_local(void * send_buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        unsigned root, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback scatter_callback,
                                        enum yac_message_priority priority);
static void comm_scatterv_callback_local(void * send_buffer, unsigned * counts,
                                         enum yac_comm_data_type data_type,
                                         unsigned root, unsigned tag,
                                         struct communicator * comm,
                                         void * user_data,
                                         func_recv_callback scatterv_callback,
                                         enum yac_message_priority priority);
static void comm_alltoallv_callback_local(void * send_buffer, unsigned * counts,
                                          enum yac_comm_data_type data_type,
                                          unsigned tag, struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback alltoallv_callback,
                                          enum yac_message_priority priority);
static void comm_reduce_callback_local(void * buffer, unsigned count,
                                       enum yac_comm_data_type data_type,
                                       func_comm_op op, unsigned root,
                                       unsigned tag, struct communicator * comm,
                                       void * user_data,
                                       func_recv_callback allreduce_callback,
                                       enum yac_message_priority priority);
static void comm_allreduce_callback_local(void * buffer, unsigned count,
                                          enum yac_comm_data_type data_type,
                                          func_comm_op op, unsigned tag,
                                          struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback allreduce_callback,
                                          enum yac_message_priority priority);
static void comm_barrier_callback_local(unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback barrier_callback,
                                        enum yac_message_priority priority);
static void comm_check_local(struct communicator * comm);
static void comm_finalize_local();

static struct communicator_vtable
   local_communicator_vtable = {
      .free                = free_communicator_local,
      .get_size            = get_comm_size_local,
      .get_rank            = get_comm_rank_local,
      .bsend               = comm_bsend_local,
      .isend               = comm_isend_local,
      .recv_callback       = comm_recv_callback_local,
      .recv                = comm_recv_local,
      .irecv               = comm_irecv_local,
      .wait                = comm_wait_local,
      .cancel_recv         = comm_cancel_recv_local,
      .allgather           = NULL,
      .allgatherv          = NULL,
      .dup                 = NULL,
      .split               = NULL,
      .split_intercomm     = NULL,
      .get_remote_size     = NULL,
      .abort               = comm_abort_local,
      .get_unique_tag      = get_unique_comm_tag_local,
      .get_unique_tags     = get_unique_comm_tags_local,
      .pack_size           = comm_pack_size_local,
      .pack                = comm_pack_local,
      .unpack              = comm_unpack_local,
      .translate_ranks     = comm_translate_ranks_local,
      .bcast_callback      = comm_bcast_callback_local,
      .gatherv_callback    = comm_gatherv_callback_local,
      .allgather_callback  = comm_allgather_callback_local,
      .allgatherv_callback = comm_allgatherv_callback_local,
      .scatter_callback    = comm_scatter_callback_local,
      .scatterv_callback   = comm_scatterv_callback_local,
      .alltoallv_callback  = comm_alltoallv_callback_local,
      .reduce_callback     = comm_reduce_callback_local,
      .allreduce_callback  = comm_allreduce_callback_local,
      .barrier_callback    = comm_barrier_callback_local,
      .check               = comm_check_local
   };

static void remove_local_comm_message_from_list
   (struct local_comm_message ** list, struct local_comm_message * message) {

   if (message == *list) {

      *list = message->next;
   } else {

      struct local_comm_message * temp;

      temp = *list;

      while (temp->next != message)
         temp = temp->next;

      temp->next = message->next;
   }
}

static void add_local_comm_message_to_end_of_list(
   struct local_comm_message ** list, struct local_comm_message * message) {

   if (*list == NULL) {

      *list = message;
   } else {

      struct local_comm_message * temp;

      temp = *list;

      while (temp->next != NULL)
         temp = temp->next;

      temp->next = message;
   }
}

static unsigned get_data_size(enum yac_comm_data_type type) {
    switch (type) {
       case COMM_UINT: return sizeof(unsigned);
       case COMM_INT: return sizeof(int);
       case COMM_DBLE: return sizeof(double);
       case COMM_BYTE: return 1;
       case COMM_CHAR: return sizeof(char);
       case COMM_PACKED: return 1;
       case COMM_UINT64_T: return sizeof(uint64_t);
       default :
          yac_internal_abort_message(
             "ERROR: invalid datatype in comm_pack_size_local.",
             __FILE__, __LINE__);
          return 0;
    };
}

struct communicator * yac_named_local_comm_new(
   struct communicator * parent_comm, char const * name) {

   struct local_communicator * comm = xmalloc(1 * sizeof(*comm));

   comm->vtable = &local_communicator_vtable;
   comm->open_receives = NULL;
   comm->open_sends = NULL;
   comm->free_messages = NULL;
   comm->active_requests = NULL;
   comm->free_requests = NULL;
   comm->name_hash = yac_hash(name);
   comm->parent_comm = (struct local_communicator *)parent_comm;

   local_comms = xrealloc(local_comms, ++num_local_comms * sizeof(*local_comms));
   local_comms[num_local_comms-1] = comm;

   comm->rank = get_comm_size_local((struct communicator *)comm) - 1;

   comm->unique_tags = NULL;
   comm->num_unique_tags = 0;
   comm->unique_tags_array_size = 0;

   yac_add_comm_finalize_callback(comm_finalize_local);

   return (struct communicator *)comm;
}

struct communicator * yac_local_comm_new() {

   return yac_named_local_comm_new(NULL, "comm world");
}

static unsigned get_comm_size_local(struct communicator * comm) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   unsigned i;

   unsigned name_hash;

   name_hash = local_comm->name_hash;

   unsigned comm_size;

   comm_size = 0;

   for (i = 0; i < num_local_comms; ++i)
      if (local_comms[i]->name_hash == name_hash)
         comm_size++;

   return comm_size;
}

static unsigned get_comm_rank_local(struct communicator * comm) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   return local_comm->rank;
}

static struct local_comm_message * get_free_local_comm_message(
   struct local_communicator * comm) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   struct local_comm_message * msg;

   if (local_comm->free_messages == NULL) {

      msg = xmalloc(1 * sizeof(*msg));
   } else {

      msg = local_comm->free_messages;
      local_comm->free_messages = msg->next;
      msg->next = NULL;
   }

   return msg;
}

static void init_local_basic_comm_message(
  struct local_comm_message * msg, void * buffer, unsigned tag, unsigned count,
  enum yac_comm_data_type type, unsigned dst_rank, unsigned src_rank,
  void * user_data) {

   msg->message.buffer = buffer;
   msg->message.tag = tag;
   msg->message.count = count;
   msg->message.type = type;
   msg->message.dst_rank = dst_rank;
   msg->message.src_rank = src_rank;
   msg->user_data = user_data;
   msg->next = NULL;
}

static void init_local_recv_comm_message(
  struct local_comm_message * msg, void * buffer, unsigned tag, unsigned count,
  enum yac_comm_data_type type, unsigned dst_rank, unsigned src_rank,
  void * user_data, func_recv_callback recv_callback,
  func_recv_callback_cancel recv_callback_cancel) {

   init_local_basic_comm_message(msg, buffer, tag, count, type, dst_rank,
                                 src_rank, user_data);
   msg->callback.recv.f = recv_callback;
   msg->callback.recv.f_cancel = recv_callback_cancel;
}

static void init_local_send_comm_message(
  struct local_comm_message * msg, void * buffer, unsigned tag, unsigned count,
  enum yac_comm_data_type type, unsigned dst_rank, unsigned src_rank,
  void * user_data, func_send_callback send_callback) {

   init_local_basic_comm_message(msg, buffer, tag, count, type, dst_rank,
                                 src_rank, user_data);
   msg->callback.send.f = send_callback;
}

static void free_comm_message_local(struct local_comm_message * msg,
                                    struct local_communicator * comm) {

   free(msg->message.buffer);

   msg->next = comm->free_messages;
   comm->free_messages = msg;
}

static void send_internal(void const * send_buffer, unsigned count,
                          enum yac_comm_data_type data_type, unsigned dest,
                          unsigned tag, struct communicator * comm,
                          void * user_data, func_send_callback send_callback) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   // make a copy of the buffer
   void * buffer;
   unsigned buffer_size = get_data_size(data_type) * count;

   buffer = xmalloc(buffer_size);

   memcpy(buffer, send_buffer, buffer_size);

   // get the comm of the receiver rank

   struct local_communicator * recv_comm = NULL;
   unsigned i;

   for (i = 0; i < num_local_comms; ++i) {
      if (local_comm->name_hash ==
          local_comms[i]->name_hash &&
          local_comms[i]->rank == dest) {
         recv_comm = local_comms[i];
         break;
      }
   }

   if (recv_comm == NULL)
      yac_internal_abort_message("ERROR: Failed to identify a proper receiver.",
                                  __FILE__, __LINE__);

   struct local_comm_message * recv_msg, * send_msg;

   // generate a send message

   unsigned source = get_comm_rank_local(comm);

   send_msg = get_free_local_comm_message(local_comm);

   init_local_send_comm_message(send_msg, buffer, tag, count, data_type, dest,
                                source, user_data, send_callback);

   // check whether there is matching receive

   recv_msg = recv_comm->open_receives;

   while(recv_msg != NULL && ((recv_msg->message.tag != tag) ||
         (recv_msg->message.src_rank != source &&
          recv_msg->message.src_rank != (unsigned)COMM_ANYSOURCE)))
      recv_msg = recv_msg->next;

   // if there was a matching receive
   if (recv_msg != NULL) {

      // remove receive message from list of open receives
      remove_local_comm_message_from_list(&(recv_comm->open_receives),
                                          recv_msg);

      if (send_msg->message.count > recv_msg->message.count) {
         printf(
            "ERROR send_msg->message.count %u > recv_msg->message.count %u\n",
            send_msg->message.count, recv_msg->message.count);
        yac_internal_abort_message("ERROR: message truncated",
                                   __FILE__, __LINE__);
      }
      if (send_msg->message.type != recv_msg->message.type) {
         printf(
            "WARNING send_msg->message.type %d > recv_msg->message.type %d\n",
            (int)(send_msg->message.type), (int)(recv_msg->message.type));
        yac_internal_abort_message(
          "WARNING: datatype of messages does not match", __FILE__, __LINE__);
      }
      // deliver message
      if (recv_msg->callback.recv.f != NULL)
         recv_msg->callback.recv.f((struct communicator *)recv_comm,
                                   send_msg->message, recv_msg->user_data);
      if (send_msg->callback.send.f != NULL)
        send_msg->callback.send.f(comm, send_msg->message, send_msg->user_data);

      // free messages
      free_comm_message_local(send_msg, local_comm);
      free_comm_message_local(recv_msg, recv_comm);
   } else {

      struct local_comm_message * temp;

      temp = recv_comm->open_sends;

      if (temp == NULL) {

         recv_comm->open_sends = send_msg;
      } else {
         while(temp->next != NULL)
            temp = temp->next;

         temp->next = send_msg;
      }
   }
}

static void comm_bsend_local(void const * send_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned dest,
                             unsigned tag, struct communicator * comm) {

  send_internal(send_buffer, count, data_type, dest, tag, comm, NULL, NULL);
}

static void free_comm_request_local(struct comm_request ** request) {

  struct comm_request_local * request_local =
    (struct comm_request_local *)*request;

  if (*request == COMM_REQUEST_NULL) return;

  struct comm_request_local * curr_active_request =
    *(request_local->p_active_requests);

  if (curr_active_request == NULL)
    yac_internal_abort_message(
      "ERROR(free_comm_request_local): no active request found in comm",
      __FILE__, __LINE__);

  if (curr_active_request == request_local) {
    *(request_local->p_active_requests) = request_local->next;
  } else {
    while ((curr_active_request->next != NULL) &&
           (curr_active_request->next != request_local))
      curr_active_request = curr_active_request->next;
    if (curr_active_request->next != request_local)
      yac_internal_abort_message(
        "ERROR(free_comm_request_local): could not find request in active "
        "request list", __FILE__, __LINE__);
    curr_active_request->next = request_local->next;
  }

  request_local->buffer = NULL;
  request_local->next = *(request_local->p_free_requests);
  *(request_local->p_free_requests) = request_local;
  *request = (struct comm_request *)COMM_REQUEST_NULL;
}

static struct comm_request_local *
get_free_comm_request_local(struct local_communicator * local_comm) {

  struct comm_request_local * request = local_comm->free_requests;

  if (request == NULL) {
    request = xmalloc(1 * sizeof(*request));
    request->waitall = comm_waitall_local;
    request->testsome = NULL;
    request->free_request = free_comm_request_local;
    request->p_active_requests = &(local_comm->active_requests);
    request->p_free_requests = &(local_comm->free_requests);
  } else {
    local_comm->free_requests = request->next;
  }

  request->next = local_comm->active_requests;
  local_comm->active_requests = request;
  request->buffer = NULL;
  request->message_complete = 0;

  return request;
}

static void isend_callback(struct communicator * comm, struct comm_message msg,
                           void * user_data) {

  UNUSED(comm);
  UNUSED(msg);

  struct comm_request_local * request_local =
     (struct comm_request_local *)user_data;
  request_local->message_complete = 1;
}

static void comm_isend_local(void const * send_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned dest,
                             unsigned tag, struct communicator * comm,
                             struct comm_request ** request) {

  struct comm_request_local * user_data =
    get_free_comm_request_local((struct local_communicator*)comm);

  *request = (struct comm_request*)user_data;

  send_internal(send_buffer, count, data_type, dest, tag, comm, user_data,
                isend_callback);
}

static void comm_recv_callback_local(unsigned count,
                                     enum yac_comm_data_type data_type,
                                     unsigned source, unsigned tag,
                                     struct communicator * comm,
                                     void * user_data,
                                     func_recv_callback recv_callback,
                                     func_recv_callback_cancel recv_callback_cancel,
                                     enum yac_message_priority priority) {

   UNUSED(priority);

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   struct local_comm_message * send_msg;

   // check whether there is already a matching message in the send buffer
   send_msg = local_comm->open_sends;
   while (send_msg != NULL && ((send_msg->message.tag != tag) || 
          (send_msg->message.src_rank != source && source != (unsigned)COMM_ANYSOURCE)))
      send_msg = send_msg->next;

   // if there is a matching message
   if (send_msg != NULL) {

      // remove send message from list of open sends
      remove_local_comm_message_from_list(&(local_comm->open_sends), send_msg);

      if (send_msg->message.count > count)
        yac_internal_abort_message ( "ERROR: message truncated", __FILE__, __LINE__ );

      // call callback routine
      if (recv_callback != NULL)
       recv_callback(comm, send_msg->message, user_data);
      if (send_msg->callback.send.f != NULL)
        send_msg->callback.send.f(comm, send_msg->message, send_msg->user_data);

      // free message
      free_comm_message_local(send_msg, local_comm);

   } else {

      struct local_comm_message * recv_msg;

      recv_msg = get_free_local_comm_message(local_comm);

      unsigned dest;

      dest = get_comm_rank_local(comm);

      init_local_recv_comm_message(recv_msg, NULL, tag, count, data_type, dest,
                                   source, user_data, recv_callback,
                                   recv_callback_cancel);

      add_local_comm_message_to_end_of_list(&(local_comm->open_receives),
                                            recv_msg);
   }
}

static void comm_recv_local(void * recv_buffer, unsigned * count,
                            enum yac_comm_data_type data_type, unsigned source,
                            unsigned tag, struct communicator * comm) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

  struct local_comm_message * send_msg;

  // search for a matching message in the send buffer
  send_msg = local_comm->open_sends;
  while (send_msg != NULL && ((send_msg->message.tag != tag) ||
                              (send_msg->message.src_rank != source &&
                               source != (unsigned)COMM_ANYSOURCE)))
    send_msg = send_msg->next;

  // if there is a matching message
  if (send_msg != NULL) {

    if (send_msg->message.count > *count) {

      yac_internal_abort_message ( "ERROR: number of send messages exceeded", __FILE__, __LINE__ );

    }

    *count = send_msg->message.count;

    unsigned buffer_size = get_data_size(data_type) * send_msg->message.count;

    memcpy(recv_buffer, send_msg->message.buffer, buffer_size);

    // remove send message from list of open sends
    remove_local_comm_message_from_list(&(local_comm->open_sends), send_msg);

    // call callback routine
    if (send_msg->callback.send.f != NULL)
      send_msg->callback.send.f(comm, send_msg->message, send_msg->user_data);

    // free message
    free_comm_message_local(send_msg, local_comm);

  } else {

      yac_internal_abort_message ( "ERROR: invalid send message", __FILE__, __LINE__ );
  }
}

static void irecv_callback(struct communicator * comm, struct comm_message msg,
                           void * user_data) {

  UNUSED(comm);

  struct comm_request_local * request_local =
     (struct comm_request_local *)user_data;

  memcpy(request_local->buffer, msg.buffer,
         msg.count * get_data_size(msg.type));
  request_local->message_complete = 1;
}

static void comm_irecv_local(void * recv_buffer, unsigned count,
                             enum yac_comm_data_type data_type, unsigned source,
                             unsigned tag, struct communicator * comm,
                             struct comm_request ** request) {

  struct comm_request_local * user_data = 
    get_free_comm_request_local((struct local_communicator*)comm);

  user_data->buffer = recv_buffer;
  *request = (struct comm_request*)user_data;

  comm_recv_callback_local(count, data_type, source, tag, comm,
                           (void*)user_data, irecv_callback, NULL,
                           MSG_PRIO_HIGH);
}

static unsigned comm_wait_local(struct communicator * comm, unsigned tag) {

  struct local_communicator * local_comm = (struct local_communicator*)comm;

  struct local_comm_message * msg;

  msg = local_comm->open_receives;

  while (msg != NULL)
    if (msg->message.tag == tag)
      yac_internal_abort_message(
        "ERROR(comm_wait_local): There should not be any open receive messages.",
        __FILE__, __LINE__);
    else
      msg = msg->next;

  msg = local_comm->open_sends;

  while (msg != NULL)
    if (msg->message.tag == tag)
      yac_internal_abort_message(
        "ERROR(comm_wait_local): There should not be any open send message.",
        __FILE__, __LINE__ );
    else
      msg = msg->next;

  return 0;
}

static void comm_waitall_local(unsigned count,
                               struct comm_request ** requests) {

  struct comm_request_local ** requests_local =
     (struct comm_request_local **)requests;

  for (unsigned i = 0; i < count; ++i) {
    if (requests[i] != COMM_REQUEST_NULL) {
      if (requests_local[i]->message_complete == 0)
        yac_internal_abort_message(
          "ERROR(comm_waitall_local): There should not be any open send/recv "
          "message.", __FILE__, __LINE__);
      free_comm_request_local(requests + i);
      requests[i] = (struct comm_request *)COMM_REQUEST_NULL;
    }
  }
}

static void comm_cancel_recv_local(struct communicator * comm, unsigned tag) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   struct local_comm_message * msg, * temp;

   msg = local_comm->open_receives;

   while (msg != NULL && msg->message.tag == tag) {

      temp = msg->next;
      if (msg->callback.recv.f_cancel != NULL)
         msg->callback.recv.f_cancel(comm, msg->user_data);
      free_comm_message_local(msg, local_comm);
      msg = temp;
   }

   if (msg != local_comm->open_receives)
      local_comm->open_receives = msg;

   while (msg != NULL && msg->next != NULL) {
   
      if (msg->next->message.tag == tag) {

         temp = msg;
         msg = msg->next;
         temp->next = msg->next;
         if (msg->callback.recv.f_cancel != NULL)
            msg->callback.recv.f_cancel(comm, msg->user_data);
         free_comm_message_local(msg, local_comm);
         msg = temp->next;
      } else
         msg = msg->next;
   }
}

static unsigned get_unique_comm_tag_local(struct communicator * comm,
                                          char * identifier_string) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   unsigned i, j;

   unsigned tag;

   if (identifier_string != NULL) {

      for (i = 0; i < local_comm->num_unique_tags; ++i)
         if (local_comm->unique_tags[i].tag != -1 &&
             local_comm->unique_tags[i].identifier_string != NULL &&
             !strcmp(identifier_string, local_comm->unique_tags[i].identifier_string))
            break;

      if (i != local_comm->num_unique_tags)
         return local_comm->unique_tags[i].tag;

      tag = yac_hash(identifier_string);

      for (i = 0; i < local_comm->num_unique_tags; ++i)
         if (local_comm->unique_tags[i].tag == tag)
            break;

      if (i != local_comm->num_unique_tags) {
         yac_internal_abort_message("ERROR: tag is not unique\n", __FILE__, __LINE__);
         return -1;
      }

      if (tag == -1) {
         yac_internal_abort_message("ERROR: generate tag was invalid\n", __FILE__, __LINE__);
         return -1;
      }

   } else {

      if (local_comm->num_unique_tags > 0) {

        for (i = 0; i < 100; ++i) {

           tag = rand();

           for (j = 0; j < local_comm->num_unique_tags; ++j)
              if (local_comm->unique_tags[j].tag == tag)
                 break;

           if (j > 0 && j == local_comm->num_unique_tags) break;
        }

        if (i == 100) {
           yac_internal_abort_message("ERROR: could not generate a unique tag", __FILE__, __LINE__);
           return -1;
        }
      } else
        tag = rand();
   }

   ENSURE_ARRAY_SIZE(local_comm->unique_tags, local_comm->unique_tags_array_size,
                     local_comm->num_unique_tags + 1);

   local_comm->unique_tags[local_comm->num_unique_tags].tag = tag;

   if (identifier_string != NULL) {
      size_t string_length = strlen(identifier_string) + 1;
      local_comm->unique_tags[local_comm->num_unique_tags].identifier_string =
         xmalloc(string_length);
      strcpy(local_comm->unique_tags[local_comm->num_unique_tags].identifier_string,
             identifier_string);
   } else {
      local_comm->unique_tags[local_comm->num_unique_tags].identifier_string = NULL;
   }

   local_comm->num_unique_tags++;

   return tag;
}

static void get_unique_comm_tags_local(struct communicator * comm,
                                       char const * unique_generator_string,
                                       char const ** tag_strings,
                                       unsigned num_tag_strings,
                                       unsigned * tags) {

  UNUSED(comm);
  UNUSED(unique_generator_string);

  for (unsigned i = 0; i < num_tag_strings; ++i) {
    unsigned j;
    for (j = 0; j < num_unique_tags; ++j) {

      if (!strcmp(tag_strings[i], unique_tag_stings[j])) {
        tags[i] = j;
        break;
      }
    }
    if (j == num_unique_tags) {
      ENSURE_ARRAY_SIZE(unique_tag_stings, unique_tags_array_size,
                        num_unique_tags + 1);
      unsigned tag_string_length = strlen(tag_strings[i]);
      unique_tag_stings[j] = xmalloc(tag_string_length + 1);
      memcpy(unique_tag_stings[j], tag_strings[i], tag_string_length + 1);
      tags[i] = j;
      ++num_unique_tags;
    }
  }
}

static void comm_abort_local(struct communicator * comm) {

   fprintf(stderr, "comm_abort_local was called by rank %u\n",
           yac_get_comm_rank(comm));
   yac_internal_abort_message ( "ERROR: not implemented.", __FILE__, __LINE__ );
}

static void free_communicator_local(struct communicator * comm) {

   struct local_communicator * local_comm = (struct local_communicator*)comm;

   struct local_comm_message *curr_msg, *next_msg;
   curr_msg = local_comm->free_messages;

   while (curr_msg != NULL) {
      next_msg = curr_msg->next;
      free(curr_msg);
      curr_msg = next_msg;
   }
   
   curr_msg = local_comm->open_receives;

   while (curr_msg != NULL) {
      next_msg = curr_msg->next;
      free(curr_msg);
      curr_msg = next_msg;
   }

   curr_msg = local_comm->open_sends;

   while (curr_msg != NULL) {
      next_msg = curr_msg->next;
      free(curr_msg->message.buffer);
      free(curr_msg);
      curr_msg = next_msg;
   }

   struct comm_request_local *curr_request, *next_request;

   curr_request = local_comm->active_requests;

   while (curr_request != NULL) {
      next_request = curr_request->next;
      free(curr_request);
      curr_request = next_request;
   }

   curr_request = local_comm->free_requests;

   while (curr_request != NULL) {
      next_request = curr_request->next;
      free(curr_request);
      curr_request = next_request;
   }

   for (unsigned i = 0; i < local_comm->num_unique_tags; ++i)
      free(local_comm->unique_tags[i].identifier_string);

   free(local_comm->unique_tags);

   free(comm);

   unsigned i = 0;
   for (; i < num_local_comms;)
     if (local_comms[i++] == local_comm) break;

   for (; i < num_local_comms; ++i)
     local_comms[i-1] = local_comms[i];

   --num_local_comms;
   if (num_local_comms == 0) {
     free(local_comms);
     local_comms = NULL;
   }
}

static int comm_pack_size_local(int count, enum yac_comm_data_type type,
                                struct communicator * comm) {

   UNUSED(comm);
   return get_data_size(type) * count;
}

static void comm_pack_local(void * in_buffer, int in_count,
                            enum yac_comm_data_type type, void * out_buffer,
                            int out_buffer_size, int * position,
                            struct communicator * comm) {

   int buffer_size;

   buffer_size = comm_pack_size_local(in_count, type, comm);

   if (*position < 0)
      yac_internal_abort_message("ERROR: invalid position value in comm_pack_local (*position < 0).",
                                 __FILE__, __LINE__);

   if (*position > out_buffer_size)
      yac_internal_abort_message("ERROR: invalid position value in comm_pack_local (*position > out_buffer_size).",
                                  __FILE__, __LINE__);

   if (out_buffer_size - *position < buffer_size)
      yac_internal_abort_message("ERROR: out_buffer too small in comm_pack_local.",
                                 __FILE__, __LINE__);

   memcpy((char*)out_buffer + *position, in_buffer, buffer_size);

   *position += buffer_size;
}

static void comm_unpack_local(void * in_buffer, int in_size, int * position,
                              void * out_buffer, int out_count,
                              enum yac_comm_data_type type,
                              struct communicator * comm) {

   int buffer_size;

   buffer_size = comm_pack_size_local(out_count, type, comm);

   if (*position < 0)
      yac_internal_abort_message("ERROR: invalid position value in comm_unpack_local "
                                 "(*position < 0).\n", __FILE__, __LINE__);

   if (in_size < buffer_size) {
      printf ( "ERROR in_size %d < buffer_size %d\n", in_size, buffer_size);
      yac_internal_abort_message("ERROR: in_buffer too small in comm_unpack_local.\n",
                                 __FILE__, __LINE__);
   }

   memcpy((char*)out_buffer, (char*)in_buffer + *position, buffer_size);

   *position += buffer_size;
}

static void comm_translate_ranks_local(struct communicator * comm_a,
                                       struct communicator * comm_b,
                                       unsigned * ranks) {

   struct local_communicator * comm_local_a =
      (struct local_communicator*)comm_a;
   struct local_communicator * comm_local_b =
      (struct local_communicator*)comm_b;

   if (comm_local_b->parent_comm == NULL)
      yac_internal_abort_message("ERROR: invalid parent_comm (NULL).\n",
                                 __FILE__, __LINE__);

   unsigned parent_name_hash = comm_local_b->parent_comm->name_hash;
   unsigned child_name_hash = comm_local_b->name_hash;

   if (parent_name_hash != comm_local_a->name_hash)
      yac_internal_abort_message("ERROR: invalid comm_local_a (does not match parent of "
                                 "comm_local_b).\n", __FILE__, __LINE__);

   for (unsigned i = 0; i < num_local_comms; ++i) {

      struct local_communicator * curr_comm_local = local_comms[i];

      if (curr_comm_local->name_hash == child_name_hash) {

         if (curr_comm_local->parent_comm == NULL)
            if (comm_local_b->parent_comm == NULL)
               yac_internal_abort_message("ERROR: invalid parent_comm (NULL).\n",
                                          __FILE__, __LINE__);

         struct local_communicator * curr_parent_comm_local =
            curr_comm_local->parent_comm;

         if (curr_parent_comm_local->name_hash != parent_name_hash)
            yac_internal_abort_message("ERROR: invalid parent comm (does not match parent of "
                                       "comm_local_b).\n", __FILE__, __LINE__);

         ranks[curr_comm_local->rank] = curr_parent_comm_local->rank;
      }
   }
}

static void comm_bcast_callback_local(void * buffer, unsigned count,
                                      enum yac_comm_data_type data_type,
                                      unsigned root, unsigned tag,
                                      struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback bcast_callback,
                                      enum yac_message_priority priority) {

  struct local_communicator * comm_local = (struct local_communicator*)comm;
  unsigned comm_size = get_comm_size_local(comm);

  if (root >= comm_size)
    yac_internal_abort_message("ERROR: Invalid root rank.", __FILE__, __LINE__);

  if (comm_local->rank == root)
    for(unsigned i = 0; i < comm_size; ++i)
      comm_bsend_local(buffer, count, data_type, i, tag, comm);

  comm_recv_callback_local(count, data_type, root, tag, comm, user_data,
                           bcast_callback, NULL, priority);
}

static void gatherv_callback_local(struct communicator * comm,
                                   struct comm_message msg, void * user_data) {

  struct temp_gatherv_data * temp_data =
    (struct temp_gatherv_data*)user_data;
  unsigned comm_size = get_comm_size_local(comm);

  unsigned offset = 0;

  for (unsigned i = 0; i < msg.src_rank; ++i) offset += temp_data->counts[i];

  unsigned type_data_size = get_data_size(msg.type);
  memcpy((char*)(temp_data->buffer) + offset * type_data_size,
         msg.buffer, msg.count * type_data_size);

  temp_data->num_open_recv--;

  if (!temp_data->num_open_recv) {

    unsigned total_count = 0;

    for (unsigned i = 0; i < comm_size; ++i)
      total_count += temp_data->counts[i];

    msg.buffer = temp_data->buffer;
    msg.count = total_count;
    msg.src_rank = msg.dst_rank;

    if (temp_data->gatherv_callback)
      temp_data->gatherv_callback(comm, msg, temp_data->user_data);

    free(temp_data->counts);
    free(temp_data->buffer);
    free(temp_data);
  }
}

static void comm_gatherv_callback_local(void * send_buffer, unsigned * counts,
                                        enum yac_comm_data_type data_type,
                                        unsigned root, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback gatherv_callback,
                                        enum yac_message_priority priority) {

  struct local_communicator * comm_local = (struct local_communicator*)comm;
  unsigned comm_size = get_comm_size_local(comm);
  unsigned comm_rank = get_comm_rank_local(comm);

  if (root >= comm_size)
    yac_internal_abort_message("ERROR: Invalid root rank.", __FILE__, __LINE__);

  comm_bsend_local(send_buffer, counts[comm_rank], data_type, root, tag, comm);

  if (root == comm_local->rank) {

    unsigned total_count = 0;

    for (unsigned i = 0; i < comm_size; ++i) total_count += counts[i];

    struct temp_gatherv_data * temp_data = xmalloc(1 * sizeof(*temp_data));

    temp_data->counts = xmalloc(comm_size * sizeof(*(temp_data->counts)));
    memcpy(temp_data->counts, counts, comm_size * sizeof(*(temp_data->counts)));
    temp_data->user_data = user_data;
    temp_data->priority = priority;
    temp_data->num_open_recv = comm_size;
    temp_data->gatherv_callback = gatherv_callback;
    temp_data->buffer = xmalloc(total_count * get_data_size(data_type));

    for (unsigned i = 0; i < comm_size; ++i)
      comm_recv_callback_local(counts[i], data_type, i, tag, comm, temp_data,
                               gatherv_callback_local, NULL, priority);
  } else {

    struct comm_message msg;

    msg.buffer = NULL;
    msg.type = data_type;
    msg.count = 0;
    msg.tag = tag;
    msg.dst_rank = comm_rank;
    msg.src_rank = comm_rank;

    if (gatherv_callback) gatherv_callback(comm, msg, user_data);
  }
}

static void final_allgather_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct temp_allgather_data * temp_data =
    (struct temp_allgather_data*)user_data;

  msg.src_rank = msg.dst_rank;

  if (temp_data->allgather_callback != NULL)
    temp_data->allgather_callback(comm, msg, temp_data->user_data);

  free(temp_data->buffer);
  free(temp_data);
}

static void first_phase_allgather_callback(struct communicator * comm,
                                           struct comm_message msg,
                                           void * user_data) {

  struct temp_allgather_data * temp_data =
    (struct temp_allgather_data*)user_data;

  if (temp_data->count != msg.count)
    yac_internal_abort_message(
      "ERROR: invalid count in first_phase_allgather_callback.",
      __FILE__, __LINE__);

  if (temp_data->data_type != msg.type)
    yac_internal_abort_message(
      "ERROR: invalid data_type in first_phase_allgather_callback.",
      __FILE__, __LINE__);

  unsigned data_size = msg.count * get_data_size(msg.type);

  memcpy(((char*)temp_data->buffer) + data_size * msg.src_rank, msg.buffer,
         data_size);

  temp_data->recv_count++;

  if (temp_data->recv_count == get_comm_size_local(comm)) {

    comm_bcast_callback_local(temp_data->buffer, temp_data->count *
                              get_comm_size_local(comm), temp_data->data_type,
                              0, temp_data->tag, comm, temp_data,
                              final_allgather_callback, temp_data->priority);

  } else {

    comm_recv_callback_local(temp_data->count, temp_data->data_type,
                             COMM_ANYSOURCE, temp_data->tag, comm,
                             temp_data, first_phase_allgather_callback, NULL,
                             temp_data->priority);
  }
}

static void comm_allgather_callback_local(void * send_buffer, unsigned count,
                                          enum yac_comm_data_type data_type,
                                          unsigned tag,
                                          struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback allgather_callback,
                                          enum yac_message_priority priority) {

  // gather data on rank 0
  comm_bsend_local(send_buffer, count, data_type, 0, tag, comm);

  unsigned comm_size = get_comm_size_local(comm);

  if (get_comm_rank_local(comm) == 0) {

    struct temp_allgather_data * temp_data = xmalloc(sizeof(*temp_data));
    temp_data->count = count;
    temp_data->data_type = data_type;
    temp_data->tag = tag;
    temp_data->priority = priority;
    temp_data->buffer = xmalloc(get_data_size(data_type) * count * comm_size);
    temp_data->recv_count = 0;
    temp_data->user_data = user_data;
    temp_data->allgather_callback = allgather_callback;

    comm_recv_callback_local(count, data_type, COMM_ANYSOURCE, tag, comm,
                             temp_data, first_phase_allgather_callback, NULL,
                             priority);

  } else {

    struct temp_allgather_data * temp_data = xmalloc(sizeof(*temp_data));
    temp_data->buffer = NULL;
    temp_data->user_data = user_data;
    temp_data->allgather_callback = allgather_callback;

    comm_bcast_callback_local(NULL, count * comm_size, data_type, 0, tag, comm,
                              temp_data, final_allgather_callback, priority);
  }
}

static void final_allgatherv_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct temp_allgatherv_data * temp_data =
    (struct temp_allgatherv_data*)user_data;

  msg.src_rank = msg.dst_rank;

  if (temp_data->allgatherv_callback != NULL)
    temp_data->allgatherv_callback(comm, msg, temp_data->user_data);

  free(temp_data->counts);
  free(temp_data->buffer);
  free(temp_data);
}

static void first_phase_allgatherv_callback(struct communicator * comm,
                                            struct comm_message msg,
                                            void * user_data) {

  struct temp_allgatherv_data * temp_data =
    (struct temp_allgatherv_data*)user_data;

  unsigned comm_size = get_comm_size_local(comm);

  if (temp_data->counts[msg.src_rank] != msg.count)
    yac_internal_abort_message(
      "ERROR: invalid count in first_phase_allgathers_callback.",
      __FILE__, __LINE__);

  if (temp_data->data_type != msg.type)
    yac_internal_abort_message(
      "ERROR: invalid data_type in first_phase_allgather_callback.",
      __FILE__, __LINE__);

  unsigned data_size = get_data_size(msg.type);

  char * buffer_p = (char*)(temp_data->buffer);

  for (unsigned i = 0; i < msg.src_rank; ++i)
    buffer_p += temp_data->counts[i] * data_size;

  memcpy(buffer_p, msg.buffer, data_size * msg.count);

  // if all messages have been received
  if (msg.src_rank == comm_size - 1) {

    unsigned count = 0;

    for (unsigned i = 0; i < comm_size; ++i) count += temp_data->counts[i];

    comm_bcast_callback_local(temp_data->buffer, count, temp_data->data_type,
                              0, msg.tag, comm, temp_data,
                              final_allgatherv_callback, temp_data->priority);

  } else {

    unsigned src_rank = msg.src_rank + 1;

    comm_recv_callback_local(temp_data->counts[src_rank], temp_data->data_type,
                             src_rank, msg.tag, comm, temp_data,
                             first_phase_allgatherv_callback, NULL,
                             temp_data->priority);
  }
}

static void comm_allgatherv_callback_local(void * send_buffer, unsigned * counts,
                                           enum yac_comm_data_type data_type,
                                           unsigned tag,
                                           struct communicator * comm,
                                           void * user_data,
                                           func_recv_callback allgatherv_callback,
                                           enum yac_message_priority priority) {

  unsigned comm_rank = get_comm_rank_local(comm);
  unsigned comm_size = get_comm_size_local(comm);

  unsigned count = 0;

  for (unsigned i = 0; i < comm_size; ++i) count += counts[i];

  if (comm_rank == 0) {

    struct temp_allgatherv_data * temp_data = xmalloc(sizeof(*temp_data));
    temp_data->counts = xmalloc(comm_size * sizeof(*(temp_data->counts)));
    memcpy(temp_data->counts, counts, comm_size * sizeof(*(temp_data->counts)));
    temp_data->data_type = data_type;
    temp_data->priority = priority;
    temp_data->buffer = xmalloc(get_data_size(data_type) * count);
    temp_data->user_data = user_data;
    temp_data->allgatherv_callback = allgatherv_callback;

    struct comm_message msg;
    msg.buffer = send_buffer;
    msg.type = data_type;
    msg.count = counts[0];
    msg.tag = tag;
    msg.dst_rank = 0;
    msg.src_rank = 0;

    first_phase_allgatherv_callback(comm, msg, temp_data);

  } else {

    struct temp_allgatherv_data * temp_data = xmalloc(sizeof(*temp_data));
    temp_data->counts = NULL;
    temp_data->buffer = NULL;
    temp_data->user_data = user_data;
    temp_data->allgatherv_callback = allgatherv_callback;

    // gather data on rank 0
    comm_bsend_local(send_buffer, counts[comm_rank], data_type, 0, tag, comm);

    comm_bcast_callback_local(NULL, count * get_comm_size_local(comm),
                              data_type, 0, tag, comm, temp_data,
                              final_allgatherv_callback, priority);
  }
}

static void comm_scatter_callback_local(void * send_buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        unsigned root, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback scatter_callback,
                                        enum yac_message_priority priority) {

  unsigned comm_rank = get_comm_rank_local(comm);
  unsigned comm_size = get_comm_size_local(comm);

  if (comm_rank == root) {

    for (unsigned i = 0; i < comm_size; ++i) {

      comm_bsend_local(send_buffer, count, data_type, i, tag, comm);
      send_buffer = (char*)send_buffer + count * get_data_size(data_type);
    }
  }

  comm_recv_callback_local(count, data_type, root, tag, comm,
                           user_data, scatter_callback, NULL, priority);
}

static void comm_scatterv_callback_local(void * send_buffer, unsigned * counts,
                                         enum yac_comm_data_type data_type,
                                         unsigned root, unsigned tag,
                                         struct communicator * comm,
                                         void * user_data,
                                         func_recv_callback scatterv_callback,
                                         enum yac_message_priority priority) {

  unsigned comm_rank = get_comm_rank_local(comm);
  unsigned comm_size = get_comm_size_local(comm);

  if (comm_rank == root) {

    for (unsigned i = 0; i < comm_size; ++i) {

      comm_bsend_local(send_buffer, counts[i], data_type, i, tag, comm);
      send_buffer = (char*)send_buffer + counts[i] * get_data_size(data_type);
    }
  }

  comm_recv_callback_local(counts[comm_rank], data_type, root, tag, comm,
                           user_data, scatterv_callback, NULL, priority);
}

static void recv_alltoallv_count_callback(struct communicator * comm,
                                          struct comm_message msg,
                                          void * user_data);

static void recv_alltoallv_data_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_alltoallv_data * alltoallv_user_data =
    (struct temp_alltoallv_data *)user_data;

  char * buffer = alltoallv_user_data->buffer;

  ENSURE_ARRAY_SIZE(buffer, alltoallv_user_data->max_buffer_size,
                    alltoallv_user_data->count + msg.count *
                    alltoallv_user_data->data_size);

  alltoallv_user_data->buffer = buffer;

  memcpy(buffer + alltoallv_user_data->count *
         alltoallv_user_data->data_size, msg.buffer, msg.count *
         alltoallv_user_data->data_size);

  alltoallv_user_data->count += msg.count;

  if (msg.src_rank != get_comm_size_local(comm) - 1) {

    comm_recv_callback_local(1, COMM_UINT, msg.src_rank + 1, msg.tag, comm,
                             alltoallv_user_data, recv_alltoallv_count_callback,
                             NULL, alltoallv_user_data->priority);
  } else {

    msg.src_rank = msg.dst_rank;
    msg.count = alltoallv_user_data->count;
    msg.buffer = alltoallv_user_data->buffer;

    if (alltoallv_user_data->alltoallv_callback != NULL)
      alltoallv_user_data->alltoallv_callback(comm, msg,
                                              alltoallv_user_data->user_data);

    free(alltoallv_user_data->buffer);
    free(alltoallv_user_data);
  }
}

static void recv_alltoallv_count_callback(struct communicator * comm,
                                          struct comm_message msg,
                                          void * user_data) {

  struct temp_alltoallv_data * alltoallv_user_data =
    (struct temp_alltoallv_data *)user_data;

  comm_recv_callback_local(*(unsigned*)(msg.buffer),
                           alltoallv_user_data->data_type, msg.src_rank,
                           msg.tag, comm, alltoallv_user_data,
                           recv_alltoallv_data_callback, NULL,
                           alltoallv_user_data->priority);
}

static void comm_alltoallv_callback_local(void * send_buffer, unsigned * counts,
                                          enum yac_comm_data_type data_type,
                                          unsigned tag, struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback alltoallv_callback,
                                          enum yac_message_priority priority) {

  unsigned data_size = get_data_size(data_type);

  for (unsigned i = 0; i < get_comm_size_local(comm); ++i) {

    comm_bsend_local(counts + i, 1, COMM_UINT, i, tag, comm);
    comm_bsend_local(send_buffer, counts[i], data_type, i, tag, comm);
    send_buffer = (char*)send_buffer + data_size * counts[i];
  }

  struct temp_alltoallv_data * alltoallv_user_data =
    xmalloc(1 * sizeof(*alltoallv_user_data));

  alltoallv_user_data->alltoallv_callback = alltoallv_callback;
  alltoallv_user_data->user_data = user_data;
  alltoallv_user_data->priority = priority;
  alltoallv_user_data->data_type = data_type;
  alltoallv_user_data->data_size = data_size;
  alltoallv_user_data->buffer = NULL;
  alltoallv_user_data->count = 0;
  alltoallv_user_data->max_buffer_size = 0;

  comm_recv_callback_local(1, COMM_UINT, 0, tag, comm, alltoallv_user_data,
                           recv_alltoallv_count_callback, NULL, priority);

  return;
}

static void allgather_allreduce_data_callback(struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

  struct temp_allreduce_data * temp_data =
    (struct temp_allreduce_data *)user_data;
  unsigned comm_size = get_comm_size_local(comm);
  unsigned count = msg.count / comm_size;
  unsigned data_size = get_data_size(msg.type);
  func_comm_op op = temp_data->op;

  char * buffer = xmalloc(count * data_size);

  memcpy(buffer, msg.buffer, count * data_size);

  for (unsigned i = 1; i < comm_size; ++i)
    op(((char*)(msg.buffer) + i * count * data_size), buffer, count, msg.type);

  msg.count = count;
  msg.buffer = buffer;

  if (temp_data->allreduce_callback != NULL)
    temp_data->allreduce_callback(comm, msg, temp_data->user_data);

  free(temp_data);
  free(buffer);
}

static void comm_reduce_callback_local(void * buffer, unsigned count,
                                       enum yac_comm_data_type data_type,
                                       func_comm_op op, unsigned root,
                                       unsigned tag, struct communicator * comm,
                                       void * user_data,
                                       func_recv_callback reduce_callback,
                                       enum yac_message_priority priority) {

  UNUSED(root);

  comm_allreduce_callback_local(buffer, count, data_type, op, tag, comm,
                                user_data, reduce_callback, priority);
}

static void comm_allreduce_callback_local(void * buffer, unsigned count,
                                          enum yac_comm_data_type data_type,
                                          func_comm_op op, unsigned tag,
                                          struct communicator * comm,
                                          void * user_data,
                                          func_recv_callback allreduce_callback,
                                          enum yac_message_priority priority) {

  struct temp_allreduce_data * temp_data = xmalloc(1 * sizeof(*temp_data));

  temp_data->user_data = user_data;
  temp_data->allreduce_callback = allreduce_callback;
  temp_data->op = op;

  // allgather data
  comm_allgather_callback_local(buffer, count, data_type, tag, comm, temp_data,
                                allgather_allreduce_data_callback, priority);
}

static void comm_barrier_callback_local(unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback barrier_callback,
                                        enum yac_message_priority priority) {

  comm_allgather_callback_local(NULL, 0, COMM_BYTE, tag, comm, user_data,
                                barrier_callback, priority);
}

static void comm_check_local(struct communicator * comm) {
  UNUSED(comm);
  return;
}

static void comm_finalize_local() {

   for (unsigned i = 0; i < num_unique_tags; ++i)
      free(unique_tag_stings[i]);
   free(unique_tag_stings);
   unique_tag_stings = NULL;
   num_unique_tags = 0;
   unique_tags_array_size = 0;
   free(local_comms);
   local_comms = NULL;
   num_local_comms = 0;
}

int yac_is_local_comm(struct communicator * comm) {

   return comm->vtable == &local_communicator_vtable;
}

