/**
 * @file domain.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "utils.h"
#include "domain.h"
#include "communicator.h"
#include "dep_list.h"
#include "points.h"
#include "geometry.h"
#include "ensure_array_size.h"

// transforms the is_on_rank and global_ids array into a dependency list, which
// contains for each rank the local ids of the halo points on that rank
// this is stored in global_mapping.halo_points
static struct dep_list * init_halo_data(
   unsigned const * is_on_rank, size_t num_total_ranks, unsigned data_size) {

   size_t * num_halo_points_per_rank =
      xcalloc(num_total_ranks, sizeof(*num_halo_points_per_rank));

   size_t num_total_halo_poins = 0;

   // count the number of halo points per rank
   for (unsigned i = 0; i < data_size; ++i) {

      if (is_on_rank[i] != (unsigned)-1) {

         num_halo_points_per_rank[is_on_rank[i]]++;
         num_total_halo_poins++;
      }
   }

   unsigned *rank_offsets = xmalloc(num_total_ranks * sizeof(*rank_offsets));

   unsigned accum = 0;

   // compute the offsets for each rank (prescan)
   for (size_t i = 0; i < num_total_ranks; ++i) {
     rank_offsets[i] = accum;
     accum += num_halo_points_per_rank[i];
   }

   unsigned * rank_to_point_dependency;

   rank_to_point_dependency = xmalloc(num_total_halo_poins *
                                      sizeof(*rank_to_point_dependency));

   // store the local ids for the halo points
   for (unsigned i = 0; i < data_size; ++i) {

      if (is_on_rank[i] != (unsigned)-1) {

         rank_to_point_dependency[rank_offsets[is_on_rank[i]]++] = i;
      }
   }

   free(rank_offsets);

   // generate the dependency list
   return yac_dep_list_new(num_total_ranks, num_halo_points_per_rank,
                           rank_to_point_dependency);
}

static
void init_subdomain_unstruct_data(struct subdomain * subdomain,
                                  struct grid * grid,
                                  int * global_corner_ids,
                                  unsigned * global_corner_ranks,
                                  int * global_cell_ids,
                                  unsigned * global_cell_ranks,
                                  int * global_edge_ids,
                                  unsigned * global_edge_ranks) {

   subdomain->grid = grid;
   subdomain->corner.global_ids = global_corner_ids;
   subdomain->cell.global_ids   = global_cell_ids;
   subdomain->edge.global_ids   = global_edge_ids;
   subdomain->global_ids_by_user = 1;
   subdomain->grid_by_user = 1;
   subdomain->corner.halo_points = NULL;
   subdomain->cell.halo_points = NULL;
   subdomain->edge.halo_points = NULL;
   subdomain->corner.global_ranks = global_corner_ranks;
   subdomain->cell.global_ranks = global_cell_ranks;
   subdomain->edge.global_ranks = global_edge_ranks;
}

void yac_init_subdomain_unstruct(struct subdomain * subdomain,
                                 struct grid * grid,
                                 struct communicator * comm,
                                 int * global_corner_ids,
                                 unsigned * corner_is_on_rank,
                                 int * global_cell_ids,
                                 unsigned * cell_is_on_rank,
                                 int * global_edge_ids,
                                 unsigned * edge_is_on_rank) {

   size_t comm_size = (size_t)yac_get_comm_size(comm);

   init_subdomain_unstruct_data(
      subdomain, grid, global_corner_ids, corner_is_on_rank,
      global_cell_ids, cell_is_on_rank, global_edge_ids, edge_is_on_rank);

   // if the global ids of the corners are provided
   if (global_corner_ids != NULL && corner_is_on_rank != NULL)
      subdomain->corner.halo_points =
         init_halo_data(
            corner_is_on_rank, comm_size, yac_get_num_grid_corners(grid));

   // if the global ids of the cells are provided
   if (global_cell_ids != NULL && cell_is_on_rank != NULL)
      subdomain->cell.halo_points =
         init_halo_data(
            cell_is_on_rank, comm_size, yac_get_num_grid_cells(grid));

   // if the global ids of the edges are provided
   if (global_edge_ids != NULL && edge_is_on_rank != NULL)
      subdomain->edge.halo_points =
         init_halo_data(
            edge_is_on_rank, comm_size, yac_get_num_grid_edges(grid));

   yac_get_grid_bounding_circle(grid, &(subdomain->extent));
}

unsigned yac_get_num_total_halo_points(struct domain * domain,
                                       unsigned subdomain_idx,
                                       enum yac_location data_location) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   struct dep_list * halo_points;

   switch (data_location) {

      case (CORNER):
         halo_points = subdomain->corner.halo_points;
         break;
      case (CELL):
         halo_points = subdomain->cell.halo_points;
         break;
      case (EDGE):
         halo_points = subdomain->edge.halo_points;
         break;
      default:
         yac_internal_abort_message(
            "ERROR: invalid data location (get_num_halo_points)\n",
            __FILE__, __LINE__);
   };

   if (yac_get_num_elements(halo_points) == 0)
      return 0;

   return (unsigned)yac_get_total_num_dependencies(halo_points);
}

struct communicator * yac_get_domain_comm(struct domain * domain) {

   return domain->comm;
}

void yac_get_halo_points(struct domain * domain, unsigned subdomain_idx,
                         enum yac_location data_location, int * global_ids,
                         unsigned * local_ids, unsigned * num_points,
                         unsigned rank) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   struct global_mapping mapping;

   switch (data_location) {

      case (CORNER):
         mapping = subdomain->corner;
         break;
      case (CELL):
         mapping = subdomain->cell;
         break;
      case (EDGE):
         mapping = subdomain->edge;
         break;
      default:
         yac_internal_abort_message("ERROR: invalid data location "
                                    "(yac_get_halo_points)\n", __FILE__,
                                    __LINE__);
         exit(EXIT_FAILURE);
   };

   unsigned const * halo_points =
      yac_get_dependencies_of_element(mapping.halo_points, (size_t)rank);
   size_t num_halo_points =
      yac_get_num_dependencies_of_element(mapping.halo_points, (size_t)rank);

   *num_points = (unsigned)num_halo_points;

   if (local_ids != NULL)
     memcpy(local_ids, halo_points, num_halo_points * sizeof(*local_ids));

   for (size_t i = 0; i < num_halo_points; ++i) {

      if (global_ids != NULL)
         global_ids[i] = mapping.global_ids[halo_points[i]];
   }
}

void yac_get_non_halo_points(struct domain * domain, unsigned subdomain_idx,
                             enum yac_location data_location, int * global_ids,
                             unsigned * local_ids) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   struct global_mapping mapping;

   switch (data_location) {

      case (CORNER):
         mapping = subdomain->corner;
         break;
      case (CELL):
         mapping = subdomain->cell;
         break;
      case (EDGE):
         mapping = subdomain->edge;
         break;
      default:
         yac_internal_abort_message("ERROR: invalid data location "
                                    "(yac_get_non_halo_points)\n", __FILE__,
                                    __LINE__);
         exit(EXIT_FAILURE);
   };

   unsigned num_points = yac_get_subdomain_data_size(domain, subdomain_idx,
                                                     data_location);

   if (global_ids != NULL)
      for (unsigned i = 0, j = 0; i < num_points; ++i)
         if (mapping.global_ranks[i] == (unsigned)-1)
            global_ids[j++] = mapping.global_ids[i];
   if (local_ids != NULL)
      for (unsigned i = 0, j = 0; i < num_points; ++i)
         if (mapping.global_ranks[i] == (unsigned)-1)
            local_ids[j++] = i;
}

//void init_subdomain_reg2d(struct subdomain * subdomain,
//                          struct grid * grid;
//                          unsigned global_offsets[2]);

static
void init_domain_data(struct domain * domain, struct communicator * comm,
                      struct subdomain ** subdomains, unsigned num_local_subdomains,
                      char * domain_name) {

   domain->num_subdomains = num_local_subdomains;
   if (num_local_subdomains == 0) {
      domain->subdomains = NULL;
   } else {
      domain->subdomains = xmalloc(num_local_subdomains * sizeof(domain->subdomains[0]));
      memcpy(domain->subdomains, subdomains, num_local_subdomains * sizeof(domain->subdomains[0]));
   }
   domain->comm = comm;
   if (domain_name != NULL) {
      domain->name = xmalloc ((strlen(domain_name) + 1) * sizeof(domain->name[0]));
      strncpy(domain->name, domain_name, strlen(domain_name) + 1);
   } else {
      domain->name = NULL;
   }
   domain->subdomains_by_user = 1;
}

void yac_init_domain(struct domain * domain, struct communicator * comm,
                     struct subdomain ** subdomains, unsigned num_local_subdomains,
                     char * domain_name) {

   init_domain_data(domain, comm, subdomains, num_local_subdomains, domain_name);

}

unsigned yac_get_num_subdomains(struct domain * domain) {

   return domain->num_subdomains;
}

unsigned yac_get_subdomain_data_size(struct domain * domain, unsigned subdomain_idx,
                                     enum yac_location data_location) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   switch (data_location) {
      case CORNER:
         return yac_get_num_grid_corners(subdomain->grid);
      case CELL:
         return yac_get_num_grid_cells(subdomain->grid);
      case EDGE:
         return yac_get_num_grid_edges(subdomain->grid);
      default:
         return -1;
   };
}

struct grid * yac_get_subdomain_grid(struct domain domain, unsigned subdomain_idx) {

   if (subdomain_idx >= domain.num_subdomains) {

      yac_internal_abort_message ( "ERROR: subdomain index too large", __FILE__, __LINE__ );

   }

   return domain.subdomains[subdomain_idx]->grid;
}

int const * yac_get_global_ids(struct domain * domain, unsigned subdomain_idx,
                               enum yac_location data_location) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   switch (data_location) {
      case CORNER:
         return subdomain->corner.global_ids;
      case CELL:
         return subdomain->cell.global_ids;
      case EDGE:
         return subdomain->edge.global_ids;
      default:
         yac_internal_abort_message ( "ERROR: data location must be one of CORNER/EDGE/CELL.", __FILE__, __LINE__ );
         break;
   };

   // routine will never reach this points
   return NULL;
}

unsigned const * yac_get_global_ranks(struct domain * domain,
                                      unsigned subdomain_idx,
                                      enum yac_location data_location) {

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   switch (data_location) {
      case CORNER:
         return subdomain->corner.global_ranks;
      case CELL:
         return subdomain->cell.global_ranks;
      case EDGE:
         return subdomain->edge.global_ranks;
      default:
         yac_internal_abort_message ( "ERROR: data location must be one of CORNER/EDGE/CELL.", __FILE__, __LINE__ );
         break;
   };

   // routine will never reach this points
   return NULL;
}

unsigned yac_is_halo_value(struct domain * domain, unsigned subdomain_idx,
                           enum yac_location location, unsigned local_id,
                           unsigned * rank) {

   if (subdomain_idx >= domain->num_subdomains)
      yac_internal_abort_message("ERROR: subdomain index too large",
                                 __FILE__, __LINE__);

   struct subdomain * subdomain;

   subdomain = domain->subdomains[subdomain_idx];

   unsigned * global_ranks;

   switch (location) {
      case CORNER:
         global_ranks = subdomain->corner.global_ranks;
         break;
      case CELL:
         global_ranks = subdomain->cell.global_ranks;
         break;
      case EDGE:
         global_ranks = subdomain->edge.global_ranks;
         break;
      default:
         yac_internal_abort_message(
            "ERROR: data location must be one of CORNER/EDGE/CELL.",
            __FILE__, __LINE__);
         exit(EXIT_FAILURE);
   }

   if (global_ranks[local_id] == (unsigned)-1) return 0;

   if (rank != NULL) *rank = global_ranks[local_id];
   return 1;
}

void yac_match_global_ids(struct domain * domain, unsigned subdomain_idx,
                          enum yac_location data_location,
                          int const * global_ids, unsigned num_global_ids,
                          int * matching_ids, unsigned * matching_ids_idx,
                          unsigned * num_matching_ids, unsigned * local_ids,
                          unsigned include_remote_points) {

   int const * local_global_ids;
   unsigned const * local_global_ranks;
   unsigned local_num_global_ids;

   local_global_ids = yac_get_global_ids(domain, subdomain_idx, data_location);
   local_global_ranks = yac_get_global_ranks(domain, subdomain_idx,
                                             data_location);
   local_num_global_ids = yac_get_subdomain_data_size(domain, subdomain_idx,
                                                      data_location);

   if (local_global_ids == NULL)
     yac_internal_abort_message ( "ERROR: Invalid global/local ID", __FILE__, __LINE__ );

   *num_matching_ids = 0;

   if (num_global_ids == 0) return;

   // for all local global ids
   for (unsigned j = 0, i = 0; j < local_num_global_ids; ++j) {

      for (; global_ids[i] < local_global_ids[j] && i < num_global_ids-1; ++i);
      for (; global_ids[i] > local_global_ids[j] && i > 0; --i);
      for (; i > 0 && global_ids[i] == global_ids[i-1]; --i);

      if (global_ids[i] == local_global_ids[j]) {
         if (include_remote_points || (local_global_ranks[j] == (unsigned)-1)) {
            do {
               if (matching_ids != NULL)
                  matching_ids[*num_matching_ids] = global_ids[i];
               if (matching_ids_idx != NULL)
                  matching_ids_idx[*num_matching_ids] = i;
               if (local_ids != NULL)
                  local_ids[*num_matching_ids] = j;
               ++*num_matching_ids;
            } while (i < num_global_ids-1 &&
                     global_ids[++i] == local_global_ids[j]);
         }
      }
   }
}

void yac_enddef_domain(struct domain * domain) {

  UNUSED(domain);
  return;
}

void yac_get_subdomain_extents(struct domain domain,
                               struct bounding_circle * extents) {

   for (unsigned i = 0; i < domain.num_subdomains; ++i)
      extents[i] = domain.subdomains[i]->extent;
}

static
void free_subdomain(struct subdomain * subdomain) {

   yac_dep_list_delete(subdomain->corner.halo_points);
   yac_dep_list_delete(subdomain->cell.halo_points);
   yac_dep_list_delete(subdomain->edge.halo_points);

   if (!subdomain->global_ids_by_user) {

      free(subdomain->corner.global_ranks);
      free(subdomain->cell.global_ranks);
      free(subdomain->edge.global_ranks);

      free(subdomain->corner.global_ids);
      free(subdomain->cell.global_ids);
      free(subdomain->edge.global_ids);
   }

   if (!subdomain->grid_by_user)
      yac_delete_grid(subdomain->grid);
}

void yac_free_domain(struct domain * domain) {

   unsigned i;

   for (i = 0; i < domain->num_subdomains; ++i)
      free_subdomain(domain->subdomains[i]);

   if(!domain->subdomains_by_user && domain->num_subdomains)
      free(domain->subdomains[0]);

   free(domain->subdomains);

   free(domain->name);
}

void yac_get_matching_domain_neighbours(struct domain domain,
                                        struct bounding_circle extent,
                                        unsigned ** ranks,
                                        unsigned * curr_ranks_array_size,
                                        unsigned * num_ranks,
                                        unsigned offset) {

  unsigned comm_size = yac_get_comm_size(domain.comm);
  unsigned comm_rank = yac_get_comm_rank(domain.comm);

  unsigned * rank_mask = xcalloc(comm_size, sizeof(*rank_mask));

  unsigned * local_ids = NULL;
  size_t curr_local_ids_array_size = 0;
  size_t num_matching_cells;

  for (unsigned i = 0; i < domain.num_subdomains; ++i) {

    yac_get_matching_grid_cells(domain.subdomains[i]->grid, extent, NULL,
                                0, &local_ids, &curr_local_ids_array_size,
                                &num_matching_cells, 0);

    for (size_t j = 0; j < num_matching_cells; ++j)
      if (domain.subdomains[i]->cell.global_ranks[local_ids[j]] != -1)
        rank_mask[domain.subdomains[i]->cell.global_ranks[local_ids[j]]] = 1;
  }

  rank_mask[comm_rank] = 0;

  unsigned tmp_num_ranks = 0;
  for (unsigned i = 0; i < comm_size; ++i)
    tmp_num_ranks += rank_mask[i];

  ENSURE_ARRAY_SIZE(*ranks, *curr_ranks_array_size, offset + tmp_num_ranks);
  *num_ranks = 0;

  for (unsigned i = 0; i < comm_size; ++i) {
    if (rank_mask[i] != 0) {
      (*ranks)[offset + *num_ranks] = i;
      ++*num_ranks;
    }
  }

  free(local_ids);
  free(rank_mask);
}

static void pack_global_mapping_data(
   struct global_mapping mapping, unsigned num_points,
   unsigned * selected_local_ids, void * buffer, int buffer_size,
   int * position, struct communicator * comm) {

   if (mapping.global_ids == NULL) num_points = 0;

   yac_comm_pack(&num_points, 1, COMM_UINT,
                 buffer, buffer_size, position, comm);
   for (unsigned i = 0; i < num_points; ++i)
      yac_comm_pack(
         mapping.global_ids + selected_local_ids[i], 1, COMM_INT,
         buffer, buffer_size, position, comm);
   for (unsigned i = 0; i < num_points; ++i)
      yac_comm_pack(
         mapping.global_ranks + selected_local_ids[i], 1, COMM_UINT,
         buffer, buffer_size, position, comm);
}

static inline int global_mapping_data_pack_size(
  struct global_mapping mapping, unsigned num_points,
  struct communicator * comm) {

  if (mapping.global_ids == NULL) num_points = 0;

  int pack_size = yac_comm_pack_size(num_points, COMM_INT, comm);
  pack_size    += yac_comm_pack_size(num_points + 1, COMM_UINT, comm);
  return pack_size;
}

static void pack_subdomain(
   struct subdomain * subdomain,
   unsigned * local_cell_ids, unsigned num_local_cell_ids, void ** buffer,
   int * buffer_size, int * position, struct communicator * comm) {

   unsigned * selected_local_cell_ids;
   unsigned * selected_local_edge_ids;
   unsigned * selected_local_corner_ids;
   // generate a grid that contains all local cell ids but is probably smaller than the
   // original one
   struct grid * subgrid =
      yac_generate_subgrid(subdomain->grid, local_cell_ids, num_local_cell_ids,
                           &selected_local_cell_ids, &selected_local_corner_ids,
                           &selected_local_edge_ids);

   unsigned num_subgrid_corners = yac_get_num_grid_corners(subgrid);
   unsigned num_subgrid_cells = yac_get_num_grid_cells(subgrid);
   unsigned num_subgrid_edges = yac_get_num_grid_edges(subgrid);

   int pack_size = yac_grid_pack_size(subgrid, comm);
   pack_size += global_mapping_data_pack_size(subdomain->corner,
                                              num_subgrid_corners, comm);
   pack_size += global_mapping_data_pack_size(subdomain->cell,
                                              num_subgrid_cells, comm);
   pack_size += global_mapping_data_pack_size(subdomain->edge,
                                              num_subgrid_edges, comm);

   ENSURE_BYTE_ARRAY_SIZE(*buffer, *buffer_size, *position + pack_size);

   // pack the subgrid into the buffers
   yac_pack_grid(subgrid, *buffer, *buffer_size, position, comm);

   pack_global_mapping_data(
      subdomain->corner, num_subgrid_corners,
      selected_local_corner_ids, *buffer, *buffer_size, position, comm);
   pack_global_mapping_data(
      subdomain->cell, num_subgrid_cells,
      selected_local_cell_ids, *buffer, *buffer_size, position, comm);
   pack_global_mapping_data(
      subdomain->edge, num_subgrid_edges,
      selected_local_edge_ids, *buffer, *buffer_size, position, comm);

   yac_delete_grid(subgrid);
   free(selected_local_cell_ids);
   free(selected_local_corner_ids);
   free(selected_local_edge_ids);
}

static void pack_domain_name(
  struct domain * domain, void ** buffer, int * buffer_size, int * position,
  struct communicator * comm) {

  uint64_t domain_name_len = (uint64_t)strlen(domain->name);

  int pack_size = yac_comm_pack_size((int)domain_name_len, COMM_CHAR, comm);
  pack_size += yac_comm_pack_size(1, COMM_UINT64_T, comm);

  ENSURE_BYTE_ARRAY_SIZE(*buffer, *buffer_size, *position + pack_size);

  yac_comm_pack(&domain_name_len, 1, COMM_UINT64_T,
                *buffer, *buffer_size, position, comm);
  yac_comm_pack(domain->name, (int)domain_name_len, COMM_CHAR,
                *buffer, *buffer_size, position, comm);
}

int yac_pack_domain(
  struct domain * domain, struct bounding_circle * extents,
  unsigned num_extents, void ** buffer, int * buffer_size, int * position,
  struct communicator * comm) {

  int flag_has_overlap = 0, old_position = *position;
  pack_domain_name(domain, buffer, buffer_size, position, comm);

  unsigned * local_cell_ids = NULL;
  size_t local_cell_ids_array_size = 0;

  int flag = 1;
  int int_pack_size = yac_comm_pack_size(1, COMM_INT, comm);

  // for all subdomains
  for (unsigned subdomain_idx = 0; subdomain_idx < domain->num_subdomains;
       ++subdomain_idx) {

    size_t num_local_cell_ids = 0;

    // for all extents
    for (unsigned i = 0; i < num_extents; ++i) {

      // if current extent and the extent of the current subdomain don't overlap
      if (!yac_extents_overlap(&(domain->subdomains[subdomain_idx]->extent),
                               extents+i)) continue;

      size_t curr_num_local_cell_ids;

      // get local ids of all cells of the current subdomain that overlap with
      // the current extent
      yac_get_matching_grid_cells(domain->subdomains[subdomain_idx]->grid,
                                  extents[i], NULL, 0, &local_cell_ids,
                                  &local_cell_ids_array_size,
                                  &curr_num_local_cell_ids, num_local_cell_ids);

      num_local_cell_ids += curr_num_local_cell_ids;
    }

    if (num_local_cell_ids == 0) continue;

    flag_has_overlap = 1;

    // sort local ids
    yac_quicksort_index((int*)local_cell_ids, num_local_cell_ids, NULL);

    // remove duplicated local ids
    yac_remove_duplicates_uint(local_cell_ids, &num_local_cell_ids);

    // pack (flag == 1) into buffer -> a subdomain is following
    ENSURE_BYTE_ARRAY_SIZE(*buffer, *buffer_size, *position + int_pack_size);
    yac_comm_pack(&flag, 1, COMM_INT, *buffer, *buffer_size, position, comm);

    // pack subdomain
    pack_subdomain(
      domain->subdomains[subdomain_idx], local_cell_ids,
      (unsigned)num_local_cell_ids, buffer, buffer_size, position, comm);
  }

  // pack (flag == 0) into buffer -> a subdomain is following
  ENSURE_BYTE_ARRAY_SIZE(*buffer, *buffer_size, *position + int_pack_size);
  flag = 0;
  yac_comm_pack(&flag, 1, COMM_INT, *buffer, *buffer_size, position, comm);

  free(local_cell_ids);

  if (!flag_has_overlap) *position = old_position;

  return flag_has_overlap;
}

static void unpack_global_mapping_data(
  void * buffer, int buffer_size, int * position,
  struct global_mapping * mapping, struct communicator * comm) {

  unsigned num_points;
  yac_comm_unpack(
    buffer, buffer_size, position, &num_points, 1, COMM_UINT, comm);

  if (num_points == 0) {
    mapping->global_ids = NULL;
    mapping->global_ranks = NULL;
    return;
  }

  mapping->global_ids = xmalloc(num_points * sizeof(*(mapping->global_ids)));
  mapping->global_ranks = xmalloc(num_points * sizeof(*(mapping->global_ranks)));

  yac_comm_unpack(buffer, buffer_size, position, mapping->global_ids,
                  num_points, COMM_INT, comm);
  yac_comm_unpack(buffer, buffer_size, position, mapping->global_ranks,
                  num_points, COMM_UINT, comm);
}

static void unpack_subdomain(
   struct subdomain * subdomain, void * buffer, int buffer_size,
   int * position, struct communicator * comm) {

   struct grid * grid = yac_unpack_grid(buffer, buffer_size, position, comm);

   struct global_mapping mapping[3];

   for (size_t i = 0; i < 3; ++i)
      unpack_global_mapping_data(
         buffer, buffer_size, position, &(mapping[i]), comm);

   yac_init_subdomain_unstruct(
      subdomain, grid, comm,
      mapping[0].global_ids, mapping[0].global_ranks,
      mapping[1].global_ids, mapping[1].global_ranks,
      mapping[2].global_ids, mapping[2].global_ranks);

   subdomain->global_ids_by_user = 0;
   subdomain->grid_by_user = 0;
}

static char * unpack_domain_name(
  void * buffer, int buffer_size, int * position, struct communicator * comm) {

  uint64_t domain_name_len;
  yac_comm_unpack(buffer, buffer_size, position, &domain_name_len,
                  1, COMM_UINT64_T, comm);

  char * domain_name = xmalloc((domain_name_len + 1) * sizeof(*domain_name));
  yac_comm_unpack(buffer, buffer_size, position, domain_name,
                  (int)domain_name_len, COMM_CHAR, comm);
  domain_name[domain_name_len] = '\0';

  return domain_name;
}

void yac_unpack_domain(
  struct domain * domain, void * buffer, int buffer_size, int * position,
  struct communicator * comm) {

  char * domain_name = unpack_domain_name(buffer, buffer_size, position, comm);

  struct subdomain * subdomains = NULL;
  size_t subdomains_size = 0;
  size_t num_subdomains = 0;

  while (1) {

    int flag;
    yac_comm_unpack(buffer, buffer_size, position, &flag, 1, COMM_INT, comm);
    // if there is no more subdomain in the buffer
    if (!flag) break;

    // unpack subdomain
    ENSURE_ARRAY_SIZE(subdomains, subdomains_size, num_subdomains + 1);
    unpack_subdomain(
       subdomains + num_subdomains, buffer, buffer_size, position, comm);
    ++num_subdomains;
  }

  struct subdomain ** p_subdomains = NULL;
  if (num_subdomains != 0) {

    subdomains = xrealloc(subdomains, num_subdomains * sizeof(*subdomains));
    p_subdomains = xmalloc(num_subdomains * sizeof(*p_subdomains));
    for (size_t i = 0; i < num_subdomains; ++i)
      p_subdomains[i] = subdomains + i;
  }

  init_domain_data(
    domain, comm, p_subdomains, (unsigned)num_subdomains, NULL);
  domain->name = domain_name;

  free(p_subdomains);

  domain->subdomains_by_user = 0;
}

const char * yac_get_domain_name(struct domain * domain) {

  return domain->name;
}

void yac_get_domain_neighbours(struct domain domain, enum yac_location location,
                               unsigned ** ranks, unsigned * num_ranks) {

  unsigned num_subdomains = domain.num_subdomains;
  unsigned comm_rank = yac_get_comm_rank(domain.comm);
  unsigned comm_size = yac_get_comm_size(domain.comm);

  unsigned * tmp_ranks = xcalloc(comm_size, sizeof(*tmp_ranks));
  unsigned tmp_num_ranks = 0;

  for (unsigned i = 0; i < num_subdomains; ++i) {

    struct dep_list * halo_points;
    struct subdomain * subdomain = domain.subdomains[i];

    switch (location) {
      case (CORNER):
        halo_points = subdomain->corner.halo_points;
        break;
      case (CELL):
        halo_points = subdomain->cell.halo_points;
        break;
      case (EDGE):
        halo_points = subdomain->edge.halo_points;
        break;
      default:
        yac_internal_abort_message("ERROR: invalid data location (get_num_halo_points)\n",
                                   __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    };

    if (yac_get_total_num_dependencies(halo_points) == 0) continue;

    size_t halo_points_num_elements = yac_get_num_elements(halo_points);
    for (size_t j = 0; j < halo_points_num_elements; ++j)
      if (yac_get_num_dependencies_of_element(halo_points, j) > 0)
         tmp_ranks[j] = 1;
  }

  tmp_ranks[comm_rank] = 0;

  for (unsigned i = 0; i < comm_size; ++i)
    if (tmp_ranks[i]) tmp_ranks[tmp_num_ranks++] = i;

  *num_ranks = tmp_num_ranks;

  if (tmp_num_ranks == 0) {
    *ranks = NULL;
    free(tmp_ranks);
    return;
  } else {
    *ranks = xrealloc(tmp_ranks, tmp_num_ranks * sizeof(*tmp_ranks));
  }
}

