/**
 * @file domain.h
 * @brief Structs and interfaces to handle domains and subdomains 
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef DOMAIN_H
#define DOMAIN_H

/** \example test_domain.c
 * This contains some examples on how to use struct domain and struct subdomain.
 */

#include "grid.h"
#include "communicator.h"
#include "points.h"
#include "dep_list.h"
#include "geometry.h"

/**
 * contains information global mapping of the local points, which points are part of the
 * halo, and on which processes these are located on
 */
struct global_mapping {

   //! contains for each point in the subdomain the global id
   int * global_ids;
   //! contains for each point in the subdomain the rank
   unsigned * global_ranks;
   //! contains the halo points that are required from each rank
   struct dep_list * halo_points;
};

/**
 * subdomain(block) containing a local part of the global grid and its mapping into the
 * global domain
 */
struct subdomain {

   //! local grid (contains halo data)
   struct grid * grid;
   //! contains mapping information for corners
   struct global_mapping corner;
   //! contains mapping information for cells
   struct global_mapping cell;
   //! contains mapping information for edges
   struct global_mapping edge;

   //! are the global id arrays provided by the user or automatically generated (e.g. by \ref unpack_subdomain)
   unsigned global_ids_by_user;
   //! is the grid provided by the user or automatically generated (e.g. by \ref unpack_subdomain)
   unsigned grid_by_user;

   //! extent of the subdomain
   struct bounding_circle extent;
};

/**
 * contains information on the part of the global data %domain that is processed by the
 * local process
 */
struct domain {

   //! name of the global %domain (all processes that are part of the same global %domain have to provide the same name)
   char * name;

   //! communicator containing all processes that have a part of the global %domain
   struct communicator * comm;

   //! array with pointers to the %subdomains(blocks) of the local data of the global %domain
   struct subdomain ** subdomains;
   //! number of %subdomains
   unsigned num_subdomains;
   //! are the subdomain provided by the user or automatically generate (e.g. by \ref yac_unpack_domain)
   unsigned subdomains_by_user;
};

/**
 * initiates a subdomain making no assumption on the underlying grid data
 * @param[out] subdomain %subdomain to be initialised
 * @param[in] grid pointer to valid %grid structure
 * @param[in] comm communicator containing all processes that are part of the same global %domain
 * @param[in] global_corner_ids global ids for all corners in %grid
 * @param[in] corner_is_on_rank contains a the rank of the process in comm on which the respective point can be found on (local points have a rank == -1)
 * @param[in] global_cell_ids global ids for all cells in %grid
 * @param[in] cell_is_on_rank contains a the rank of the process in comm on which the respective point can be found on (local points have a rank == -1)
 * @param[in] global_edge_ids global ids for all edges in %grid
 * @param[in] edge_is_on_rank contains a the rank of the process in comm on which the respective point can be found on (local points have a rank == -1)
 *
 * \remark the position of each point in the global_*_ids and *_is_on_rank array matches its respective local id
 * \remark if not needed that user can provide NULL for global_*_ids and its matching *_is_on_rank array
 * \remark the routine makes no copy of the arrays associated with the global_*_ids and *_is_on_rank arrays, pointers to global_*_ids are stored internally for later use
 */
void yac_init_subdomain_unstruct(struct subdomain * subdomain,
                                 struct grid * grid,
                                 struct communicator * comm,
                                 int * global_corner_ids,
                                 unsigned * corner_is_on_rank,
                                 int * global_cell_ids,
                                 unsigned * cell_is_on_rank,
                                 int * global_edge_ids,
                                 unsigned * edge_is_on_rank);

//void init_subdomain_reg2d(struct subdomain * subdomain,
//                          struct grid * grid;
//                          unsigned global_offsets[2]);

/**
 * initiates a domain (a call to enddef_domain is required to finialise the setting up of the %domain)
 * @param[out] domain %domain to be initialised
 * @param[in] comm communicator containing all processes that are part of the same global %domain (the other processes must provide the same communicator)
 * @param[in] subdomains array containing pointers to all subdomains that the %domain is comprised of
 * @param[in] num_subdomains number of subdomains
 * @param[in] domain_name name of the global %domain (all other processes being part of the same %domain need to provide the same name)
 * @see enddef_domain
 */
void yac_init_domain(struct domain * domain, struct communicator * comm,
                     struct subdomain ** subdomains, unsigned num_subdomains,
                     char * domain_name);

/**
 * gets the number of subdomains
 * @param[in] domain
 * @return number of subdomains in %domain
 */
unsigned yac_get_num_subdomains(struct domain * domain);

/**
 * gets the communicator of the given domain
 * @param[in] domain
 * @return communicator of the domain
 */
struct communicator * yac_get_domain_comm(struct domain * domain);

/**
 * gets the number of points of a subdomain
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the number of corners, cells, or edges
 * @return number of points
 */
unsigned yac_get_subdomain_data_size(struct domain * domain, unsigned subdomain_idx,
                                     enum yac_location data_location);

/**
 * gets the grid data structure of a subdomain
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @return grid of the specified subdomain
 */
struct grid * yac_get_subdomain_grid(struct domain domain, unsigned subdomain_idx);

/**
 * gets a array containing all global ids of a subdomain
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the global ids of the corners, cells, or edges
 * @return array containing global ids
 */
int const * yac_get_global_ids(struct domain * domain, unsigned subdomain_idx,
                               enum yac_location data_location);

/**
 * gets a array containing the ranks for all points of a subdomain
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the ranks of the corners, cells, or edges
 * @return array containing ranks
 */
unsigned const * yac_get_global_ranks(struct domain * domain,
                                      unsigned subdomain_idx,
                                      enum yac_location data_location);

/**
 * checks whether a point is part of a halo
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] location location of the point
 * @param[in] local_id local id of the point to be checked
 * @param[out] rank rank of process on which the point is located
 * @return 0 if the point is no halo point
 *
 * \remark rank == NULL is valid, in this case no rank is returned
 */
unsigned yac_is_halo_value(struct domain * domain, unsigned subdomain_idx,
                           enum yac_location location, unsigned local_id, unsigned * rank);

/**
 * matches a set of given global ids with the ones available in a subdomains
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the points of the corners, cells, or edges
 * @param[in] global_ids array of global ids this routine is supposed to search for in the subdomain (the routine assumes that the global ids are sorted in ascending order)
 * @param[in] num_global_ids number of global ids in global_ids
 * @param[out] matching_ids global ids of the local domain that match an id in global_ids
 * @param[out] matching_ids_idx indices in global_ids of the matching global ids
 * @param[out] num_matching_ids number of matching global ids found in the %subdomain
 * @param[out] local_ids local ids of the matching points
 * @param[in] include_remote_points indicates whether this routine is supposed to take halo points into account
 *
 * \remark the user needs to provide the arrays for matching_ids, matching_ids_idx, and local_ids (these need to be big enough)
 * \remark the user can provide NULL for matching_ids, matching_ids_idx, and local_ids, in the case the respective information is not returned
 */
void yac_match_global_ids(struct domain * domain, unsigned subdomain_idx,
                          enum yac_location data_location, int const * global_ids,
                          unsigned num_global_ids, int * matching_ids,
                          unsigned * matching_ids_idx, unsigned * num_matching_ids,
                          unsigned * local_ids, unsigned include_remote_points);

/**
 * finalises the setting up of a domain
 * (this routine blocks until the communication with processes having neighbour subdomains is finished)
 * @param[in] domain
 * @see yac_init_domain
 */
void yac_enddef_domain(struct domain * domain);

/**
 * gets the number of halo points of a subdomain
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the halo points of the corners, cell, or edges
 * @return number of halo points
 */
unsigned yac_get_num_total_halo_points(struct domain * domain,
                                       unsigned subdomain_idx,
                                       enum yac_location data_location);

/**
 * gets the halo points of a subdomain for a given rank
 * @param[in] domain
 * @param[in] subdomain_idx index of the respective %subdomain
 * @param[in] data_location indicates whether the user wants the halo points of the corners, cell, or edges
 * @param[out] global_ids array containing the global ids of points the are located on the process with the given rank
 * @param[out] local_ids array containing the local ids of points the are located on the process with the given rank
 * @param[out] num_points number of entries in global_ids and or local_ids
 * @param[in] rank the rank of a process in the domain communicator
 * @see yac_get_num_total_halo_points
 *
 * \remark the user is responsible for providing arrays that are big enough for global_ids and local_ids
 * \remark the user can provide NULL for global_ids and local_ids, in that case the respective information is not returned
 */
void yac_get_halo_points(struct domain * domain, unsigned subdomain_idx,
                         enum yac_location data_location,
                         int * global_ids, unsigned * local_ids,
                         unsigned * num_points, unsigned rank);

/**
 * gets the non-halo points of a subdomain for a given rank
 * @param[in] domain
 * @param[in] subdomain_idx  index of the respective %subdomain
 * @param[in] data_location  indicates whether the user wants the non-halo
 *                           points of the corners, cell, or edges
 * @param[out] global_ids    array containing the global ids of points that are
 *                           located on the local process
 * @param[out] local_ids     array containing the local ids of points that are
 *                           located on the local process
 * @see yac_get_num_total_halo_points
 *
 * \remark the user is responsible for providing arrays that are big enough for
 *         global_ids or local_ids
 * \remark NULL is a valid value for global_ids or local_ids
 */
void yac_get_non_halo_points(struct domain * domain, unsigned subdomain_idx,
                             enum yac_location data_location, int * global_ids,
                             unsigned * local_ids);

/**
 * gets the extents of a subdomain
 * @param[in] domain
 * @param[out] extents extents of all subdomains (the user should provide an array big enough for all subdomains)
 */
void yac_get_subdomain_extents(struct domain domain,
                               struct bounding_circle * extents);

/**
 * packs a domain into a double and int buffer based on a provided extent
 * @param[in]     domain
 * @param[in]     extents     all cell of the domain that overlap with the
 *                            provided extents will be included (may also
 *                            contain additional cells)
 * @param[in]     num_extents number of entries in num_extents
 * @param[in,out] buffer      pointer to buffer into which the domain will be
 *                            packed
 * @param[in,out] buffer_size current size of the pack buffer
 * @param[in,out] position    current write position in the buffer
 * @param[in]     comm        communicator that contains all processes that have
 *                            a part of the same global %domain
 * @return 1 if the domain has an actual overlap with the provided extent
 * @see \ref yac_unpack_domain
 */
int yac_pack_domain(struct domain * domain, struct bounding_circle * extents,
                    unsigned num_extents, void ** buffer, int * buffer_size,
                    int * position, struct communicator * comm);

/**
 * unpacks a domain that was previously packed by \ref yac_pack_domain
 * @param[out]    domain
 * @param[in]     buffer      unpack buffer
 * @param[in]     buffer_size size of unpack buffer
 * @param[in,out] position    current read position in the buffer
 * @param[in]     comm        communicator that contains all processes that have
 *                            a part of the same global %domain
 * @see \ref yac_pack_domain
 */
void yac_unpack_domain(struct domain * domain, void * buffer, int buffer_size,
                       int * position, struct communicator * comm);

/**
 * gets the name of the domain
 * @param[in] domain
 * @return name of the domain
 */
const char * yac_get_domain_name(struct domain * domain);

/**
 * return a list of all ranks of processes that own cells/corners/edges that are
 * halo cells/corners/edges of the provided domain
 * @param[in] domain
 * @param[in] location
 * @param[out] ranks ranks of neighbour processes
 * @param[out] num_ranks number of ranks returned by this routine
 * @remark the user is responsible for freeing the ranks array
 */
void yac_get_domain_neighbours(struct domain domain, enum yac_location location,
                               unsigned ** ranks, unsigned * num_ranks);

/**
 * returns a list of all ranks of processes that own cells that are halo cells
 * of the provided domain and overlap with the provided bounding circle
 * @param[in] domain
 * @param[in] extent bounding circle
 * @param[in,out] ranks ranks of neighbour processes
 * @param[in,out] curr_ranks_array_size size of ranks array (if the ranks array
 *                                      is reallocated the size will be updated
 *                                      as well)
 * @param[out] num_ranks number of ranks returned by this routine
 * @param[in] offset number of elements already in ranks (these elements will
 *                   not be overwritten and results will be appended)
 */
void yac_get_matching_domain_neighbours(struct domain domain,
                                        struct bounding_circle extent,
                                        unsigned ** ranks,
                                        unsigned * curr_ranks_array_size,
                                        unsigned * num_ranks,
                                        unsigned offset);

/**
 * frees memory associated the a %domain
 * @param[in] domain
 */
void yac_free_domain(struct domain * domain);

#endif // DOMAIN_H

