/**
 * @file grid_curve2d.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "grid.h"
#include "grid_curve2d.h"
#include "grid_reg2d_common.h"
#include "utils.h"
#include "dep_list.h"
#include "geometry.h"
#include "ensure_array_size.h"
#include "sphere_part.h"

// routine declarations

static struct grid * copy_grid_curve2d(struct grid * grid);
static void get_2d_grid_extent_curve2d(struct grid * grid, double (* extent)[2]);
static void get_grid_cell_curve2d(struct grid * grid, unsigned cell_index,
                                  struct grid_cell * cell);
static void get_grid_cell2_curve2d(struct grid * grid, unsigned cell_index,
                                   struct grid_cell * cell,
                                   struct bounding_circle * bnd_circle);
static unsigned get_size_x_coords_curve2d(struct grid * grid);
static unsigned get_size_y_coords_curve2d(struct grid * grid);
static void set_x_coords_curve2d(struct grid * grid, double * x_coords);
static void set_y_coords_curve2d(struct grid * grid, double * y_coords);
static double const * get_x_coords_curve2d(struct grid * grid);
static double const * get_y_coords_curve2d(struct grid * grid);
static unsigned get_size_cell_grid_x_coords_curve2d(struct grid * grid);
static unsigned get_size_cell_grid_y_coords_curve2d(struct grid * grid);
static unsigned get_num_grid_cells_curve2d(struct grid * grid);
static unsigned get_num_grid_corners_curve2d(struct grid * grid);
static unsigned get_num_cell_corners_curve2d(struct grid * grid, unsigned cell_index);
static unsigned get_num_corner_cells_curve2d(struct grid * grid, unsigned corner_index);
static unsigned get_num_grid_edges_curve2d(struct grid * grid);
static unsigned get_num_corner_edges_curve2d(struct grid * grid, unsigned corner_index);
static unsigned get_num_cell_edges_curve2d(struct grid * grid, unsigned cell_index);
static unsigned const * get_corner_edges_curve2d(struct grid * grid, unsigned corner_index);
static unsigned const * get_cell_edge_indices_curve2d(struct grid * grid,
                                                      unsigned cell_index);
static enum yac_edge_type get_edge_type_curve2d(struct grid * grid, unsigned edge_index);
static unsigned const * get_cell_corner_indices_curve2d(struct grid * grid,
                                                        unsigned cell_index);
static unsigned const * get_corner_cell_indices_curve2d(struct grid * grid,
                                                        unsigned corner_index);
static unsigned const * get_cell_x_coord_indices_curve2d(struct grid * grid,
                                                         unsigned cell_index);
static unsigned const * get_cell_y_coord_indices_curve2d(struct grid * grid,
                                                         unsigned cell_index);
static unsigned get_corner_x_coord_index_curve2d(struct grid * grid, unsigned corner_index);
static unsigned get_corner_y_coord_index_curve2d(struct grid * grid, unsigned corner_index);
static double get_corner_x_coord_curve2d(struct grid * grid, unsigned corner_index);
static double get_corner_y_coord_curve2d(struct grid * grid, unsigned corner_index);
static int get_aux_grid_cell_curve2d(struct grid * grid, unsigned corner_index,
                                     unsigned * cell_indices, enum yac_edge_type * edge_type);
static struct dep_list * get_cell_neigh_dep_list_curve2d(struct grid * grid);
static void get_boundary_corners_curve2d(struct grid * grid, unsigned * bnd_corners,
                                         unsigned * num_bnd_corners);
static struct grid * generate_cell_grid_curve2d(struct grid * grid,
                                                double * coordinates_x, 
                                                double * coordinates_y);
static struct grid * generate_edge_grid_curve2d(struct grid * grid,
                                                double * coordinates_x, 
                                                double * coordinates_y);
static struct grid * generate_subgrid_curve2d(struct grid * grid,
                                              unsigned * selected_local_cell_ids,
                                              unsigned num_local_cells,
                                              unsigned ** local_cell_ids,
                                              unsigned ** local_corner_ids,
                                              unsigned ** local_edge_ids);
static void pack_grid_curve2d(struct grid *, void *, int, int *,
                              struct communicator *);
static int grid_pack_size_curve2d(struct grid *, struct communicator *);
static struct grid_search * get_grid_search_curve2d(struct grid * grid);
static int cell_contains_non_gc_edge_curve2d(struct grid * grid, unsigned local_cell_id);
static void delete_grid_curve2d(struct grid * grid);

static struct grid_vtable curve2d_grid_vtable = {

   .copy                        = copy_grid_curve2d,
   .get_2d_extent               = get_2d_grid_extent_curve2d,
   .get_grid_cell               = get_grid_cell_curve2d,
   .get_grid_cell2              = get_grid_cell2_curve2d,
   .get_size_x_coords           = get_size_x_coords_curve2d,
   .get_size_y_coords           = get_size_y_coords_curve2d,
   .get_x_coords                = get_x_coords_curve2d,
   .get_y_coords                = get_y_coords_curve2d,
   .set_x_coords                = set_x_coords_curve2d,
   .set_y_coords                = set_y_coords_curve2d,
   .get_size_cell_grid_x_coords = get_size_cell_grid_x_coords_curve2d,
   .get_size_cell_grid_y_coords = get_size_cell_grid_y_coords_curve2d,
   .get_num_grid_cells          = get_num_grid_cells_curve2d,
   .get_num_grid_corners        = get_num_grid_corners_curve2d,
   .get_num_cell_corners        = get_num_cell_corners_curve2d,
   .get_num_corner_cells        = get_num_corner_cells_curve2d,
   .get_num_grid_edges          = get_num_grid_edges_curve2d,
   .get_num_corner_edges        = get_num_corner_edges_curve2d,
   .get_num_cell_edges          = get_num_cell_edges_curve2d,
   .get_corner_edges            = get_corner_edges_curve2d,
   .get_cell_edge_indices       = get_cell_edge_indices_curve2d,
   .get_edge_type               = get_edge_type_curve2d,
   .get_cell_corner_indices     = get_cell_corner_indices_curve2d,
   .get_corner_cell_indices     = get_corner_cell_indices_curve2d,
   .get_cell_x_coord_indices    = get_cell_x_coord_indices_curve2d,
   .get_cell_y_coord_indices    = get_cell_y_coord_indices_curve2d,
   .get_corner_x_coord          = get_corner_x_coord_curve2d,
   .get_corner_y_coord          = get_corner_y_coord_curve2d,
   .get_corner_x_coord_index    = get_corner_x_coord_index_curve2d,
   .get_corner_y_coord_index    = get_corner_y_coord_index_curve2d,
   .get_aux_grid_cell           = get_aux_grid_cell_curve2d,
   .get_cell_neigh_dep_list     = get_cell_neigh_dep_list_curve2d,
   .get_boundary_corners        = get_boundary_corners_curve2d,
   .generate_cell_grid          = generate_cell_grid_curve2d,
   .generate_edge_grid          = generate_edge_grid_curve2d,
   .generate_subgrid            = generate_subgrid_curve2d,
   .pack_grid                   = pack_grid_curve2d,
   .grid_pack_size              = grid_pack_size_curve2d,
   .get_grid_search             = get_grid_search_curve2d,
   .cell_contains_non_gc_edge   = cell_contains_non_gc_edge_curve2d,
   .grid_delete                 = delete_grid_curve2d
};

struct curve2d_grid {
   struct grid_vtable * vtable;
   enum yac_grid_type grid_type;

   struct dep_list * cell_to_neigh; //!< dependency list containing cell neighbourhood
                                    //!< information (automatically generated once it is
                                    //!< needed)

   double * cell_corners_x,       //!< latitude data
          * cell_corners_y,       //!< longitude data
          (*cell_corners_xyz)[3]; //!< 3d coordinates

   unsigned cell_corners_x_by_user; //!< indicates whether cell_corners_x was provided by
                                    //!< the user or automatically generated by a grid
                                    //!< routine
   unsigned cell_corners_y_by_user; //!< indicates whether cell_corners_y was provided by
                                    //!< the user or automatically generated by a grid
                                    //!< routine

   unsigned num_cells[2];
   unsigned cyclic[2];

   struct grid_search * grid_search;
};

static unsigned const * get_cell_x_coord_indices_curve2d(struct grid * grid,
                                                         unsigned cell_index) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   static unsigned cell_corners[4];
   unsigned cell_x_index, cell_y_index;
   unsigned num_coords_in_x;
   unsigned offset;

   cell_y_index = cell_index / curve2d_grid->num_cells[0];
   cell_x_index = cell_index - cell_y_index * curve2d_grid->num_cells[0];
   num_coords_in_x = curve2d_grid->num_cells[0] +
                     ((curve2d_grid->cyclic[0])?0:1);

   offset = cell_x_index + cell_y_index * num_coords_in_x;

   cell_corners[0] = offset;
   if (curve2d_grid->cyclic[0] &&
       cell_x_index + 1 == curve2d_grid->num_cells[0]) {
      cell_corners[1] = offset + 1 - num_coords_in_x;
      cell_corners[2] = offset + 1;
   } else {
      cell_corners[1] = offset + 1;
      cell_corners[2] = offset + 1 +num_coords_in_x;
   }
   cell_corners[3] = offset     +num_coords_in_x;

   return cell_corners;
}

static unsigned const * get_cell_y_coord_indices_curve2d(struct grid * grid,
                                                         unsigned cell_index) {

   return get_cell_x_coord_indices_curve2d(grid, cell_index);
}

static double get_corner_x_coord_curve2d(struct grid * grid, unsigned corner_index) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->cell_corners_x[corner_index];
}

static double get_corner_y_coord_curve2d(struct grid * grid, unsigned corner_index) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->cell_corners_y[corner_index];
}

static unsigned get_corner_x_coord_index_curve2d(struct grid * grid, unsigned corner_index) {

   UNUSED(grid);

   return corner_index;
}

static unsigned get_corner_y_coord_index_curve2d(struct grid * grid, unsigned corner_index) {

   UNUSED(grid);

   return corner_index;
}

static int get_aux_grid_cell_curve2d(
  struct grid * grid, unsigned corner_index, unsigned * cell_indices,
  enum yac_edge_type * edge_type) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  edge_type[0] = GREAT_CIRCLE;
  edge_type[1] = GREAT_CIRCLE;
  edge_type[2] = GREAT_CIRCLE;
  edge_type[3] = GREAT_CIRCLE;

  return
    yac_get_aux_grid_cell_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic[0], corner_index,
      cell_indices);
}

static inline unsigned compute_total_num_coords(
  unsigned * num_cells, unsigned * cyclic) {

  return (num_cells[0] + ((cyclic[0])?0:1)) * (num_cells[1] + 1);
}

static unsigned get_size_x_coords_curve2d(struct grid * grid) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;
  return
    compute_total_num_coords(curve2d_grid->num_cells, curve2d_grid->cyclic);
}

static unsigned get_size_y_coords_curve2d(struct grid * grid) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;
  return
    compute_total_num_coords(curve2d_grid->num_cells, curve2d_grid->cyclic);
}

static double const * get_x_coords_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->cell_corners_x;
}

static double const * get_y_coords_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->cell_corners_y;
}

static void set_x_coords_curve2d(struct grid * grid, double * x_coords) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (!curve2d_grid->cell_corners_x_by_user)
      free(curve2d_grid->cell_corners_x);

   free(curve2d_grid->cell_corners_xyz);
   curve2d_grid->cell_corners_xyz = NULL;

   curve2d_grid->cell_corners_x_by_user = 1 == 1;
   curve2d_grid->cell_corners_x = x_coords;
}

static void set_y_coords_curve2d(struct grid * grid, double * y_coords) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (!curve2d_grid->cell_corners_y_by_user)
      free(curve2d_grid->cell_corners_y);

   free(curve2d_grid->cell_corners_xyz);
   curve2d_grid->cell_corners_xyz = NULL;

   curve2d_grid->cell_corners_y_by_user = 1 == 1;
   curve2d_grid->cell_corners_y = y_coords;
}

static unsigned get_size_cell_grid_x_coords_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->num_cells[0] * curve2d_grid->num_cells[1];
}

static unsigned get_size_cell_grid_y_coords_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->num_cells[0] * curve2d_grid->num_cells[1];
}

static unsigned get_num_grid_cells_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return curve2d_grid->num_cells[0] * curve2d_grid->num_cells[1];
}

static unsigned get_num_grid_corners_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return (curve2d_grid->num_cells[0]+((curve2d_grid->cyclic[0])?0:1)) * 
          (curve2d_grid->num_cells[1]+1);
}

static unsigned get_num_cell_corners_curve2d(struct grid * grid, unsigned cell_index) {

   UNUSED(grid);
   UNUSED(cell_index);

   return 4;
}

static unsigned get_num_cell_edges_curve2d(struct grid * grid, unsigned cell_index) {

   UNUSED(grid);
   UNUSED(cell_index);

   return 4;
}

static unsigned get_num_grid_edges_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   return (curve2d_grid->num_cells[0] + ((curve2d_grid->cyclic[0])?0:1)) *
          curve2d_grid->num_cells[1] + curve2d_grid->num_cells[0] *
          (curve2d_grid->num_cells[1] + 1);
}

static unsigned get_num_corner_edges_curve2d(
  struct grid * grid, unsigned corner_index) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  return
    yac_get_num_corner_edges_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic[0], corner_index);
}

static unsigned get_num_corner_cells_curve2d(struct grid * grid, unsigned corner_index) {

   return 1 << (get_num_corner_edges_curve2d(grid, corner_index) - 2);
}

static void generate_cell_neigh_dep_curve2d(struct grid * grid) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  curve2d_grid->cell_to_neigh =
    yac_generate_cell_neigh_dep_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic);
}

static void get_boundary_corners_curve2d(struct grid * grid, unsigned * bnd_corners,
                                         unsigned * num_bnd_corners) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  yac_get_boundary_corners_reg2d_common(
    curve2d_grid->num_cells, curve2d_grid->cyclic[0], bnd_corners,
    num_bnd_corners);
}

static struct dep_list * get_cell_neigh_dep_list_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (curve2d_grid->cell_to_neigh == NULL)
      generate_cell_neigh_dep_curve2d(grid);
   return curve2d_grid->cell_to_neigh;
}

static void generate_coordinates_xyz(struct curve2d_grid * curve2d_grid) {

   unsigned num_grid_corners =
      get_num_grid_corners_curve2d((struct grid *)curve2d_grid);

   curve2d_grid->cell_corners_xyz =
      xmalloc(num_grid_corners * sizeof(*(curve2d_grid->cell_corners_xyz)));

   for (unsigned i = 0; i < num_grid_corners; ++i) {

      LLtoXYZ(curve2d_grid->cell_corners_x[i], curve2d_grid->cell_corners_y[i],
              curve2d_grid->cell_corners_xyz[i]);
   }
}

static void get_grid_cell_curve2d(struct grid * grid, unsigned cell_index,
                                  struct grid_cell * cell) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (curve2d_grid->cell_corners_xyz == NULL)
      generate_coordinates_xyz(curve2d_grid);

   double const tol = 1e-12;

   size_t num_corners = (size_t)get_num_cell_corners_curve2d(grid, cell_index);

   // if the memory for the coordintes needs to be reallocated
   if (num_corners > cell->array_size) {
      cell->coordinates_x = xrealloc (cell->coordinates_x, num_corners * sizeof(cell->coordinates_x[0]));
      cell->coordinates_y = xrealloc (cell->coordinates_y, num_corners * sizeof(cell->coordinates_y[0]));
      cell->coordinates_xyz = xrealloc (cell->coordinates_xyz, num_corners * sizeof(cell->coordinates_xyz[0]));
      cell->edge_type = xrealloc (cell->edge_type, num_corners * sizeof(cell->edge_type[0]));
      cell->array_size = num_corners;
   }

   cell->num_corners = num_corners;

   // set the corners for of the cell
   unsigned const * x_indices, * y_indices;

   x_indices = yac_get_cell_x_coord_indices(grid, cell_index);
   y_indices = yac_get_cell_y_coord_indices(grid, cell_index);

   for (size_t i = 0; i < 4; ++i) {
      cell->coordinates_x[i] = curve2d_grid->cell_corners_x[x_indices[i]];
      cell->coordinates_y[i] = curve2d_grid->cell_corners_y[y_indices[i]];
      cell->coordinates_xyz[i][0] = curve2d_grid->cell_corners_xyz[x_indices[i]][0];
      cell->coordinates_xyz[i][1] = curve2d_grid->cell_corners_xyz[x_indices[i]][1];
      cell->coordinates_xyz[i][2] = curve2d_grid->cell_corners_xyz[x_indices[i]][2];
   }

   // set the edge type for the cell
   for (size_t i = 0; i < cell->num_corners-1; ++i)
      if (fabs(get_angle(cell->coordinates_x[i], cell->coordinates_x[i+1])) < tol)
         cell->edge_type[i] = LON_CIRCLE;
      else
         cell->edge_type[i] = GREAT_CIRCLE;

   size_t last_idx = cell->num_corners-1;
   if (fabs(get_angle(cell->coordinates_x[last_idx],
                      cell->coordinates_x[0])) < tol)
      cell->edge_type[last_idx] = LON_CIRCLE;
   else
      cell->edge_type[last_idx] = GREAT_CIRCLE;
}

static void get_grid_cell2_curve2d(struct grid * grid, unsigned cell_index,
                                   struct grid_cell * cell,
                                   struct bounding_circle * bnd_circle) {

  get_grid_cell_curve2d(grid, cell_index, cell);
  yac_get_cell_bounding_circle(*cell, bnd_circle);
}

static unsigned const * get_corner_edges_curve2d(struct grid * grid, unsigned corner_index) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  return
    yac_get_corner_edges_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic[0], corner_index);
}

static unsigned const * get_cell_edge_indices_curve2d(struct grid * grid,
                                                      unsigned cell_index) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (!curve2d_grid->cyclic[0]) {

      unsigned row = cell_index / curve2d_grid->num_cells[0];
      unsigned column = cell_index - curve2d_grid->num_cells[0] * row;

      return
        yac_get_cell_edge_indices_reg2d_common_non_cyclic(
          curve2d_grid->num_cells, row, column);

   } else {

      unsigned row = cell_index / curve2d_grid->num_cells[0];
      unsigned column = cell_index - curve2d_grid->num_cells[0] * row;

      return
        yac_get_cell_edge_indices_reg2d_common_cyclic(
          curve2d_grid->num_cells, row, column);

   } // if (!curve2d_grid->cyclic[0])
}

static enum yac_edge_type get_edge_type_curve2d(struct grid * grid, unsigned edge_index) {

   UNUSED(grid);
   UNUSED(edge_index);

   return GREAT_CIRCLE;
}

static unsigned const * get_cell_corner_indices_curve2d(
  struct grid * grid, unsigned cell_index) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  return
    yac_get_cell_corner_indices_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic[0], cell_index);
}

static unsigned const * get_corner_cell_indices_curve2d(struct grid * grid,
                                                        unsigned corner_index) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  return
    yac_get_corner_cell_indices_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic[0], corner_index);
}

static void get_2d_grid_extent_curve2d(struct grid * grid, double (* extent)[2]) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   unsigned i;

   double const tol = 1e-12;

   //--------------------------------
   // compute the extent of the grid
   //--------------------------------

   if (get_num_grid_corners_curve2d(grid) == 0) return;
   if (extent == NULL) return;

   extent[1][0] = extent[1][1] = curve2d_grid->cell_corners_y[0];

   for (i = 1; i < get_size_y_coords_curve2d(grid); ++i) {

      if (extent[1][0] > curve2d_grid->cell_corners_y[i]) extent[1][0] = curve2d_grid->cell_corners_y[i];
      if (extent[1][1] < curve2d_grid->cell_corners_y[i]) extent[1][1] = curve2d_grid->cell_corners_y[i];
   }

   if (curve2d_grid->cyclic[0]) {

      extent[0][0] = 0;
      extent[0][1] = 2*M_PI;
   } else {

      unsigned * checked_corners;

      checked_corners = xcalloc(get_num_grid_corners_curve2d(grid),
                                sizeof (checked_corners[0]));

      unsigned * corner_stack;
      unsigned corner_stack_size;
      
      corner_stack_size = 0;
      corner_stack = xmalloc(get_num_grid_corners_curve2d(grid) * sizeof(corner_stack[0]));

      corner_stack[corner_stack_size++] = 0;
      extent[0][0] = extent[0][1] = curve2d_grid->cell_corners_x[yac_get_corner_x_coord_index(grid, 0)];

      unsigned curr_corner;
      unsigned const * curr_neigh_corners;
      unsigned curr_num_neigh_corners;

      while (corner_stack_size > 0) {

         curr_corner = corner_stack[--corner_stack_size];
         curr_num_neigh_corners = get_num_corner_edges_curve2d(grid, curr_corner);
         curr_neigh_corners = get_corner_edges_curve2d(grid, curr_corner);

         double edge_angle;
         double edge_x_coord[2];

         edge_x_coord[0] =curve2d_grid->cell_corners_x[
            get_corner_x_coord_index_curve2d(grid, curr_corner)];

         for (i = 0; i < curr_num_neigh_corners; ++i) {

            // process every edge only once
            if (curr_corner > curr_neigh_corners[i]) continue;
            else if (!checked_corners[curr_neigh_corners[i]]) {
               corner_stack[corner_stack_size++] = curr_neigh_corners[i];
               checked_corners[curr_neigh_corners[i]] = 1 == 1;
            }

            edge_x_coord[1] = curve2d_grid->cell_corners_x[
               get_corner_x_coord_index_curve2d(grid, curr_neigh_corners[i])];
            edge_angle = fabs(get_angle(edge_x_coord[0], edge_x_coord[1]));

            // if the current edge crosses the up boundary of the current extent
            if (fabs(get_angle(edge_x_coord[0], extent[0][1])) <= edge_angle + tol &&
                fabs(get_angle(edge_x_coord[1], extent[0][1])) <= edge_angle + tol &&
                get_angle(extent[0][1], edge_x_coord[1]) <= 0)
               extent[0][1] = edge_x_coord[1];

            // if the current edge crosses the up boundary of the current extent
            if (fabs(get_angle(edge_x_coord[0], extent[0][0])) <= edge_angle + tol &&
                fabs(get_angle(edge_x_coord[1], extent[0][0])) <= edge_angle + tol &&
                get_angle(extent[0][0], edge_x_coord[1]) >= 0)
               extent[0][0] = edge_x_coord[1];

            // if we already have 2*Pi (360 degree)
            if (get_angle(extent[0][1], extent[0][0]) <= tol) {
               extent[0][0] = 0;
               extent[0][1] = 2.0 * M_PI;
               corner_stack_size = 0;
               break;
            }
         }
      }

      free(corner_stack);
      free(checked_corners);
   }

   //----------------------
   // check for pole cells
   //----------------------

   struct grid_cell cell;

   yac_init_grid_cell(&cell);

   // test whether the grid data includes the pole
   for (i = 0; i < get_num_grid_cells_curve2d(grid); ++i) {
        
      // get the current cell
      get_grid_cell_curve2d(grid, i, &cell);

      // check the cell
      if (yac_cell_covers_pole((unsigned)(cell.num_corners),
                               cell.coordinates_x, cell.coordinates_y)) {

         if (cell.coordinates_y[0] > 0)
            extent[1][1] = M_PI_2;
         else
            extent[1][0] = -M_PI_2;
      }
   } // i

   yac_free_grid_cell(&cell);
}

static struct grid * generate_cell_grid_curve2d(struct grid * grid,
                                                double * coordinates_x, 
                                                double * coordinates_y) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   unsigned num_cells[2] = {curve2d_grid->num_cells[0]-1,
                            curve2d_grid->num_cells[1]-1};

   if (curve2d_grid->cyclic[0])
      num_cells[0]++;

   return yac_curve2d_grid_new(coordinates_x, coordinates_y, num_cells,
                               curve2d_grid->cyclic);
}

static struct grid * generate_edge_grid_curve2d(struct grid * grid,
                                                double * coordinates_x, 
                                                double * coordinates_y) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  return
    yac_generate_edge_grid_reg2d_common(
      curve2d_grid->num_cells, curve2d_grid->cyclic,
      coordinates_x, coordinates_y);
}

static struct grid * copy_grid_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   struct curve2d_grid * copy = xmalloc(1 * sizeof(*copy));

   copy->vtable = &curve2d_grid_vtable;
   copy->grid_type = YAC_CURVE2D;
   copy->cell_to_neigh = NULL;

   unsigned x_coords_size = get_size_x_coords_curve2d(grid) * sizeof(*copy->cell_corners_x);
   unsigned y_coords_size = get_size_y_coords_curve2d(grid) * sizeof(*copy->cell_corners_y);

   copy->cell_corners_x = xmalloc(x_coords_size);
   copy->cell_corners_y = xmalloc(y_coords_size);
   memcpy(copy->cell_corners_x, curve2d_grid->cell_corners_x, x_coords_size);
   memcpy(copy->cell_corners_y, curve2d_grid->cell_corners_y, y_coords_size);
   copy->cell_corners_xyz = NULL;

   copy->cell_corners_x_by_user = 1 == 0;
   copy->cell_corners_y_by_user = 1 == 0;

   copy->num_cells[0] = curve2d_grid->num_cells[0];
   copy->num_cells[1] = curve2d_grid->num_cells[1];
   copy->cyclic[0] = curve2d_grid->cyclic[0];
   copy->cyclic[1] = curve2d_grid->cyclic[1];

   copy->grid_search = NULL;

   return (struct grid *)copy;
}

static void pack_grid_curve2d(
  struct grid * grid, void * buffer, int buffer_size, int * position,
  struct communicator * comm) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  unsigned total_num_coords =
    compute_total_num_coords(curve2d_grid->num_cells, curve2d_grid->cyclic);

  yac_comm_pack(curve2d_grid->num_cells, 2, COMM_UINT,
                buffer, buffer_size, position, comm);
  yac_comm_pack(curve2d_grid->cyclic, 2, COMM_UINT,
                buffer, buffer_size, position, comm);

  yac_comm_pack(curve2d_grid->cell_corners_x, total_num_coords,
                COMM_DBLE, buffer, buffer_size, position, comm);
  yac_comm_pack(curve2d_grid->cell_corners_y, total_num_coords,
                COMM_DBLE, buffer, buffer_size, position, comm);
}

struct grid * yac_unpack_curve2d_grid(
  void * buffer, int buffer_size, int * position, struct communicator * comm) {

  unsigned num_cells[2], cyclic[2];
  yac_comm_unpack(buffer, buffer_size, position, num_cells,
                  2, COMM_UINT, comm);
  yac_comm_unpack(buffer, buffer_size, position, cyclic,
                  2, COMM_UINT, comm);

  unsigned total_num_coords = compute_total_num_coords(num_cells, cyclic);

  double * cell_corners_x = xmalloc(total_num_coords * sizeof(*cell_corners_x));
  double * cell_corners_y = xmalloc(total_num_coords * sizeof(*cell_corners_y));
  yac_comm_unpack(buffer, buffer_size, position, cell_corners_x,
                  total_num_coords, COMM_DBLE, comm);
  yac_comm_unpack(buffer, buffer_size, position, cell_corners_y,
                  total_num_coords, COMM_DBLE, comm);

  struct curve2d_grid * curve2d_grid =
    (struct curve2d_grid *)yac_curve2d_grid_new(
      cell_corners_x, cell_corners_y, num_cells, cyclic);

  curve2d_grid->cell_corners_x_by_user = 1 == 0;
  curve2d_grid->cell_corners_y_by_user = 1 == 0;

  return (struct grid *)curve2d_grid;
}

static int grid_pack_size_curve2d(
  struct grid * grid, struct communicator * comm) {

  struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

  unsigned total_num_coords =
    compute_total_num_coords(curve2d_grid->num_cells, curve2d_grid->cyclic);

  int pack_size = yac_comm_pack_size(4, COMM_UINT, comm);
  pack_size    += yac_comm_pack_size(total_num_coords, COMM_DBLE, comm);
  pack_size    += yac_comm_pack_size(total_num_coords, COMM_DBLE, comm);

  return pack_size;
}

static void delete_grid_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   yac_dep_list_delete(curve2d_grid->cell_to_neigh);
   if (!curve2d_grid->cell_corners_x_by_user)
      free(curve2d_grid->cell_corners_x);
   if (!curve2d_grid->cell_corners_y_by_user)
      free(curve2d_grid->cell_corners_y);
   free(curve2d_grid->cell_corners_xyz);

   if (curve2d_grid->grid_search != NULL)
      yac_delete_grid_search(curve2d_grid->grid_search);

   free(curve2d_grid);
}

static struct grid * generate_subgrid_curve2d(struct grid * grid,
                                              unsigned * selected_local_cell_ids,
                                              unsigned num_local_cells,
                                              unsigned ** local_cell_ids,
                                              unsigned ** local_corner_ids,
                                              unsigned ** local_edge_ids) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (num_local_cells == 0)
      yac_internal_abort_message ( "ERROR: cells need to be defined.", __FILE__, __LINE__ );

   // compute the index range of the subgrid

   unsigned num_lon_per_lat;

   num_lon_per_lat = curve2d_grid->num_cells[0] + ((curve2d_grid->cyclic[0])?0:1);

   unsigned index_range[2][2];

   index_range[0][0] = index_range[0][1] = get_cell_x_coord_indices_curve2d(grid, selected_local_cell_ids[0])[0]%num_lon_per_lat;
   index_range[1][0] = index_range[1][1] = get_cell_y_coord_indices_curve2d(grid, selected_local_cell_ids[0])[0]/num_lon_per_lat;

   unsigned i, j, k, l; 

   for (i = 0; i < num_local_cells; ++i) {

      unsigned const * cell_x_coords, * cell_y_coords;

      cell_x_coords = get_cell_x_coord_indices_curve2d(grid, selected_local_cell_ids[i]);
      cell_y_coords = get_cell_y_coord_indices_curve2d(grid, selected_local_cell_ids[i]);

      for (j = 0; j < 4; ++j) {

         if (index_range[0][0] > cell_x_coords[j]%num_lon_per_lat)
             index_range[0][0] = cell_x_coords[j]%num_lon_per_lat;
         if (index_range[0][1] < cell_x_coords[j]%num_lon_per_lat)
             index_range[0][1] = cell_x_coords[j]%num_lon_per_lat;
         if (index_range[1][0] > cell_y_coords[j]/num_lon_per_lat)
             index_range[1][0] = cell_y_coords[j]/num_lon_per_lat;
         if (index_range[1][1] < cell_y_coords[j]/num_lon_per_lat)
             index_range[1][1] = cell_y_coords[j]/num_lon_per_lat;
      }
   }

   unsigned index_range_extent[2];

   index_range_extent[0] = index_range[0][1] - index_range[0][0] + 1;
   index_range_extent[1] = index_range[1][1] - index_range[1][0] + 1;

   // generate the subgrid

   unsigned num_cells[2], cyclic[2];

   cyclic[0] = (curve2d_grid->num_cells[0] == index_range_extent[0]) &&
                curve2d_grid->cyclic[0];
   cyclic[1] = 0;

   num_cells[0] = index_range_extent[0] - ((cyclic[0])?0:1);
   num_cells[1] = index_range_extent[1] - 1;

   double * cell_x_coords, * cell_y_coords;
   unsigned cell_coords_size;

   cell_coords_size = (num_cells[0] + ((cyclic[0])?0:1)) * (num_cells[1] + 1);
   cell_x_coords = xmalloc(cell_coords_size * sizeof(*cell_x_coords));
   cell_y_coords = xmalloc(cell_coords_size * sizeof(*cell_y_coords));

   unsigned sub_num_lon_per_lat;

   sub_num_lon_per_lat = num_cells[0] + ((cyclic[0])?0:1);

   for (i = 0; i < num_cells[1] + 1; ++i) {
      memcpy(cell_x_coords + i*sub_num_lon_per_lat,
            curve2d_grid->cell_corners_x + index_range[1][0] * num_lon_per_lat + index_range[0][0],
            sub_num_lon_per_lat * sizeof(*cell_x_coords));
      memcpy(cell_y_coords + i*sub_num_lon_per_lat,
            curve2d_grid->cell_corners_y + (index_range[1][0] + i) * num_lon_per_lat + index_range[0][0],
            sub_num_lon_per_lat * sizeof(*cell_y_coords));
   }

   struct grid * subgrid = yac_curve2d_grid_new(cell_x_coords, cell_y_coords, num_cells, cyclic);

   ((struct curve2d_grid *)subgrid)->cell_corners_x_by_user = 1 == 0;
   ((struct curve2d_grid *)subgrid)->cell_corners_y_by_user = 1 == 0;

   //generate the local id mapping between grid and subgrid
   *local_cell_ids = xmalloc(get_num_grid_cells_curve2d(subgrid) * sizeof(**local_cell_ids));
   *local_corner_ids = xmalloc(get_num_grid_corners_curve2d(subgrid) * sizeof(**local_corner_ids));
   *local_edge_ids = xmalloc(get_num_grid_edges_curve2d(subgrid) * sizeof(**local_edge_ids));

   unsigned num_grid_edges;
   unsigned * edge_mask;
   unsigned const * cell_edges;

   num_grid_edges = get_num_grid_edges_curve2d(grid);
   edge_mask = xcalloc(num_grid_edges, sizeof(*edge_mask));

   k = 0;

   for (i = index_range[1][0]; i < index_range[1][1]; ++i) {

      for (j = index_range[0][0]; j < index_range[0][1] + ((cyclic[0])?1:0); ++j) {

         (*local_cell_ids)[k] = j + i * (num_lon_per_lat - ((curve2d_grid->cyclic[0])?0:1));

         cell_edges = get_cell_edge_indices_curve2d(grid, (*local_cell_ids)[k++]);

         for (l = 0; l < 4; ++l)
            edge_mask[cell_edges[l]] = 1;
      }
   }

   k = 0;

   for (i = 0; i < num_grid_edges; ++i)
      if (edge_mask[i]) (*local_edge_ids)[k++] = i;

   free(edge_mask);

   if (k != get_num_grid_edges_curve2d(subgrid))
      yac_internal_abort_message ( "ERROR: Number of edges does not match.", __FILE__, __LINE__ );

   k = 0;

   for (i = index_range[1][0]; i <= index_range[1][1]; ++i) {

      for (j = index_range[0][0]; j <= index_range[0][1]; ++j) {

         (*local_corner_ids)[k++] = j + i * num_lon_per_lat;
      }
   }

   return subgrid;
}

struct grid * yac_curve2d_grid_new(double * coordinates_x, double * coordinates_y,
                                   unsigned const num_cells[2], unsigned const cyclic[2]) {

   struct curve2d_grid * curve2d_grid = xmalloc(1 * sizeof(*curve2d_grid));

   curve2d_grid->vtable = &curve2d_grid_vtable;
   curve2d_grid->grid_type = YAC_CURVE2D;

   curve2d_grid->cell_to_neigh = NULL;

   curve2d_grid->cell_corners_x = coordinates_x;
   curve2d_grid->cell_corners_y = coordinates_y;
   curve2d_grid->cell_corners_xyz = NULL;

   curve2d_grid->num_cells[0] = num_cells[0];
   curve2d_grid->num_cells[1] = num_cells[1];
   curve2d_grid->cyclic[0] = cyclic[0];
   curve2d_grid->cyclic[1] = cyclic[1];

   curve2d_grid->cell_corners_x_by_user = 1 == 1;
   curve2d_grid->cell_corners_y_by_user = 1 == 1;

   curve2d_grid->grid_search = NULL;

   return (struct grid *)curve2d_grid;
}

static struct grid_search * get_grid_search_curve2d(struct grid * grid) {

   struct curve2d_grid * curve2d_grid = (struct curve2d_grid *)grid;

   if (curve2d_grid->grid_search == NULL)
      curve2d_grid->grid_search = yac_sphere_part_search_new(grid);

   return curve2d_grid->grid_search;
}

static int cell_contains_non_gc_edge_curve2d(
  struct grid * grid, unsigned local_cell_id) {

  UNUSED(grid);
  UNUSED(local_cell_id);

  return 0;
}

