/**
 * @file interpolation_method_conserv.c
 *
 * \brief implementation of the 1st order conservative remapping method
 *
 * \section interpolation_method_conserv Communication schema for conservative
 *          remapping
 *
 * Prerequisite for this algorithm is the availability of a list of target
 * cells with its associated overlapping source cell on the local rank
 * plus information about ranks which may contain additional source cell
 * overlapping the target.
 *
 * Intersection cells between source and target cells are determined
 * and the area is calculated for each of these cells.
 *
 * The source processes exchange information, and for each target cell
 * a master is identified, which collects the partial_areas from other source
 * ranks where an overlap with the particular target cell was detected.
 * Information is sent back from the master ranks to the servant ranks,
 * whether target cells can be processed or not. Target cells that can be
 * handled by the master ranks are eliminated on the server ranks, and the
 * next interpolation method is called.
 *
 * The weights are normalised and corrected such that the sum of all weights
 * per target cell equals 1.0.
 *
 *  \msc
 *    width = "800";
 *    "Master Sources","Servant Sources";
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="build target and overlapping source cells",
 *                       URL="\ref do_search_interpolation_method_conserv"];
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="construct source and target cells"];
 *    "Master Sources" box "Servant Sources" [label="exclude source cells that are located in halos"];
 *    "Master Sources" box "Servant Sources" [label="calculate partial areas"];
 *    "Master Sources" box "Servant Sources" [label="identify master and servant ranks"];
 *    "Master Sources" box "Servant Sources" [label="determine send and receive information"];
 *    "Servant Sources" box "Servant Sources" [label="post receive_callback_accepted for answer from master ranks"];
 *    "Master Sources"<-"Servant Sources" [label="partial areas and provenance data"];
 *    "Master Sources" box "Master Sources" [label="post receive_callback_partial_areas for partial areas and provenance data"];
 *    "Servant Sources" box "Servant Sources" [label="post 2nd set of callback receive message"];
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="continue with second part",
 *                       URL="\ref do_search_interpolation_method_conserv_continue_2"];
 *    |||;
 *    "Master Sources" box "Master Sources" [label="check if target cell is fully covered"];
 *    "Master Sources"->"Servant Sources" [label="list of cells that can be processed by the master"];
 *    "Master Sources" box "Servant Sources" [label="reorder weights according to increasing global cell ID"];
 *    "Servant Sources" box "Servant Sources" [label="erase target cells from local list"];
 *    |||;
 *  \endmsc
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <stdint.h>

#include "grid.h"
#include "interpolation_method.h"
#include "interpolation_method_conserv.h"
#include "interpolation_plan.h"
#include "clipping.h"
#include "search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"
#include "area.h"

// -----------------------------------------------------------------------------

static void
do_search_interpolation_method_conserv ( struct interpolation_method ** method,
                                         struct interpolation_plan_element ** plan_stack,
                                         int * global_ids, double (*coordinates_xyz)[3],
                                         enum yac_location tgt_location,
                                         unsigned num_cells,
                                         unsigned * num_interpolated_cells,
                                         struct interpolation_method_data_request ** data_request );

static void
enddef_interpolation_method_conserv ( struct interpolation_method ** method );

static void
get_source_exchange_data_interpolation_method_conserv ( struct interpolation_method ** method,
                                                        struct interpolation_method_data_request *** data_request );

static void
delete_interpolation_method_conserv ( struct interpolation_method * method );

static struct global_search *
get_global_search ( struct interpolation_method * method );

static void
set_finish_callback ( struct interpolation_method * method,
                      void (*callback)(void * user_data),
                      void * user_data);

// -----------------------------------------------------------------------------

static struct interpolation_method_vtable
interpolation_method_conserv_vtable = {
  .do_search                = do_search_interpolation_method_conserv,
  .enddef                   = enddef_interpolation_method_conserv,
  .get_source_exchange_data = get_source_exchange_data_interpolation_method_conserv,
  .delete                   = delete_interpolation_method_conserv,
  .get_global_search        = get_global_search,
  .set_finish_callback      = set_finish_callback};

struct interpolation_method_conserv_tags {
  unsigned alltoallv_uncertain_tgt;
  unsigned alltoallv_master_ranks;
  unsigned header1_exchange;
  unsigned header2_exchange;
  unsigned partial_areas_exchange;
  unsigned accepts_exchange;
};

struct interpolation_method_conserv {
  struct interpolation_method_vtable * vtable;
  void (*finish_callback)(void * user_data);
  void * finish_user_data;
  struct global_search * search;
  struct domain * domain;
  struct points *** points;
  struct mask *** masks;
  unsigned num_pointsets;
  unsigned points_idx;
  struct interpolation_method_data_request ** data_request;
  struct interpolation_method_conserv_tags tags;
  int enforced_conserv;
  int partial_coverage;
  enum yac_interpolation_method_conserv_normalisation normalisation;
  unsigned use_source_mask;
  int order; // order of the interpolation (supported 1 or 2)
};

//! contains information on a single overlap between a source and a target cell
struct overlap {
  int subdomain_id;     //!< global source cell subdomain ID
  int global_cell_id;   //!< global source cell ID
  int local_cell_id;    //!< local source cell ID
  int rank;             //!< source cell rank
  double area;          //!< area associated with cell
  double weight;        //!< source cell partial weight
  double barycenter[3]; //!< barycenter of the overlap (only used by 2nd order)
};

//! contains all information the is relevant for the interpolation of a single
//! target cell
struct weight_provenance {
  int target_cell_id;           //!< id of the target cell
  double coordinate_xyz[3];     //!< coordinate of the target cell
  int is_master;                //!< determines whether the local process is the
                                //!< master for the target cell
  int master_rank;              //!< rank of the source process that is
                                //!< responsible for interpolating this
                                //!< target cell
  int nbr_servant_ranks;        //!< number of source processes (not including
                                //!< the master) that have cells, which have
                                //!< an overlap with the target cell (this is
                                //!< only available on the respective master)
  unsigned * servant_ranks;     //!< ranks of source processes (not including
                                //!< the master) that have cells, which have
                                //!< an overlap with the target cell (this is
                                //!< only available on the respective master)
  size_t nbr_overlaps;          //!< total number of overlaps
  struct overlap * overlaps;    //!< array containing information on all
                                //!< overlaps
  int cell_can_be_interpolated; //!< flag that indicates whether the target
                                //!< cell can be interpolated
  double target_cell_area;      //!< area of the target cell in km^2
};

struct send_buffer {
  unsigned num_target_cells;    //!< number of target cells to be sent
  unsigned buffer_array_size;   //!< size of buffer
  unsigned buffer_size;         //!< size of buffer
  int position;                 //!< position of packed buffer
  int * ibuffer;                //!< target cell ids and number of partial areas
                                //!< per target cell
  char * cbuffer;               //!< buffer for packing
                                //!< - source cell ids associated to the partial
                                //!< areas
                                //!< - partial areas in km^2
                                //!< - target cell ids
                                //!< - number of partial areas per target cell
};

struct temp_recv_partial_areas_data {

  struct search_data_conserv * search_data;
  unsigned num_target_cells;
};

struct temp_uncertain_tgt {

  unsigned src_rank;       // which rank sent this point
                           // or later the master rank
  int master_rank;         // the final master rank
  int global_id;           // global id of the target point
  unsigned num_src_points; // how many points did the rank found for this tgt
};

struct weight_vector_data_3d {
  double weight[3];
  union {
    unsigned local;
    int global;
  } id;
};

struct weight_vector_3d {
  struct weight_vector_data_3d * data;
  size_t n;
};

struct weight_vector_3d_subdomain {
  struct weight_vector_3d * weight_vector;
  int global_id;
  unsigned subdomain_id;
};

struct weight_vector_data {
  double weight;
  int global_id;
  unsigned subdomain_id;
  unsigned local_id;
};

struct temp_remote_cell {
  int global_id;
  unsigned rank;
};

struct remote_cell {
  int global_id;         //! global id of the remote cell
  unsigned rank;         //! rank of the owner of the cell
                         //! (can also be the local rank)
  unsigned subdomain_id; //! -1 for remote and 0 to num_subdomains-1 for
                         //! local points
  unsigned local_id;     //! subdomain_id == -1 -> position of global id in
                         //!                          remote_global_ids array
                         //! subdomain_id != -1 -> local id in local data
};

struct remote_gradients {
  struct weight_vector_3d * gradients;
  int * global_src_cell_ids;
  size_t num_gradients;
  struct weight_vector_data_3d * weight_vector_data_3d_buffer;
  size_t weight_vector_data_3d_buffer_size;
  struct remote_cell * remote_cells;
  size_t num_remote_cells;
};

struct remote_overlap {

  size_t tgt_cell_idx;
  size_t overlap_idx;
  int global_id;
  unsigned local_id;
};

struct search_data_conserv {

  struct interpolation_method ** method_stack;
  struct interpolation_plan_element ** plan_stack;

  int * global_ids_buffer;                        //!< global ids of target cells that are to be
                                                  //!< interpolated by the local process
  double (*coordinates_xyz_buffer)[3];            //!< cartesian (centre) coordinates of the target cells
  unsigned num_cells;                             //!< number of target cells
  unsigned * num_interpolated_cells;              //!< number of cells that are interpolated by
                                                  //!< the conservative remapping interpolation method (this is
                                                  //!< determined in \ref do_search_interpolation_method_conserv_continue_2
                                                  //!< ( \ref remove_cells ))
  unsigned num_partial_areas_messages_to_receive; //!< number of open message for source cell exchange
  unsigned num_accepts_messages_to_receive;       //!< number of open message for rejected target cells

  struct weight_provenance * weight_provenance;   //!< partial areas, weights and metadata calculated based on intersections

  int * target_cell_ids_to_remove;                //!< array containing all global target cell ids of cells that have to be removed
  int num_target_cell_ids_to_remove;              //!< number of ids in target_cell_ids_to_remove

  // data required for 2nd order
  struct weight_vector_3d ** src_cell_gradients; //!< weights for the gradient of the source cells
  struct remote_gradients * remote_src_cell_gradients;
  size_t remote_src_cell_gradients_size;
  struct weight_vector_data_3d * weight_vector_data_buffer;
};

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas ( struct communicator * comm,
                              struct comm_message message,
                              void * user_data );

static void
recv_callback_accepts ( struct communicator * comm, struct comm_message message,
                        void * user_data );

// -----------------------------------------------------------------------------

static void
generate_tag_string ( char * buf, size_t buffer_size,
                      char const * tag_base_str,
                      char const * source_comp_name,
                      char const * target_comp_name,
                      char const * source_domain_name,
                      char const * target_domain_name,
                      char const * stack_name,
                      int points_idx,
                      int unique_interpolation_method_id,
                      int order,
                      int enforced_conserv,
                      int partial_coverage,
                      enum yac_interpolation_method_conserv_normalisation normalisation,
                      unsigned use_source_mask,
                      char const * tag_str) {

  int ret_val;
  ret_val = snprintf ( buf, buffer_size, "%s%s%s%s%s%s%d%d%d%d%d%d%u%s", tag_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, stack_name, points_idx,
                       unique_interpolation_method_id, order, enforced_conserv,
                       partial_coverage, (int)normalisation, use_source_mask,
                       tag_str);

  if ( ret_val < 0 || (size_t)ret_val >= buffer_size )
    yac_internal_abort_message ( "ERROR(generate_tag_string): truncation", __FILE__, __LINE__ );
}

// -----------------------------------------------------------------------------

static void
generate_interpolation_method_conserv_tag ( struct interpolation_method_conserv * method,
                                            char const * stack_name ) {

  char const * str_base_str                = "interpolation method conserv";
  char const * str_generator_string        = "generator string";
  char const * str_alltoallv_uncertain_tgt = "alltoallv uncertain tgt";
  char const * str_alltoallv_master_ranks  = "alltoallv master ranks";
  char const * str_header1_exchange        = "header1 exchange";
  char const * str_header2_exchange        = "header2 exchange";
  char const * str_partial_areas_exchange  = "partial_areas exchange";
  char const * str_accepts_exchange        = "accepts exchange";

  struct global_search * search =
    get_global_search((struct interpolation_method *)method);

  struct communicator * comm = yac_get_domain_comm(yac_get_local_domain(search));

  char const * source_comp_name = yac_get_local_component_name(search);
  char const * target_comp_name = yac_get_remote_component_name(search);

  char const * source_domain_name = yac_get_local_domain_name(search);
  char const * target_domain_name = yac_get_remote_domain_name(search);

  unsigned max_tag_strlen = strlen(str_alltoallv_uncertain_tgt);

  unsigned max_str_buf_size = max_tag_strlen;
  max_str_buf_size += strlen(str_base_str);
  max_str_buf_size += strlen(source_comp_name);
  max_str_buf_size += strlen(target_comp_name);
  max_str_buf_size += strlen(source_domain_name);
  max_str_buf_size += strlen(target_domain_name);
  max_str_buf_size += strlen(stack_name) + 10 // for points_idx
                                         + 10 // for unique_method_id
                                         + 1  // for order
                                         + 1  // for enforced_conserv (boolean)
                                         + 1  // for partial_cov (boolean)
                                         + 10 // for normalisation
                                         + 10 // for use_source_mask
                                         + 1; //  for '\0'

   char * unique_generator_string = xmalloc(max_str_buf_size);
   generate_tag_string ( unique_generator_string, max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_generator_string );

   char * tag_strings[6];
   tag_strings[0] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[0], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_alltoallv_uncertain_tgt );

   tag_strings[1] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[1], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_alltoallv_master_ranks );

   tag_strings[2] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[2], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_header1_exchange );

   tag_strings[3] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[3], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_header2_exchange );

   tag_strings[4] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[4], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_partial_areas_exchange );

   tag_strings[5] = xmalloc(max_str_buf_size);
   generate_tag_string ( tag_strings[5], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method(
                           (struct interpolation_method *)method),
                         method->order,
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_accepts_exchange );

   unsigned tags[6];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 6, tags);

   free(unique_generator_string);

   method->tags.alltoallv_uncertain_tgt = tags[0];
   method->tags.alltoallv_master_ranks = tags[1];
   method->tags.header1_exchange = tags[2];
   method->tags.header2_exchange = tags[3];
   method->tags.partial_areas_exchange = tags[4];
   method->tags.accepts_exchange = tags[5];

   for (unsigned i = 0; i < 6; ++i) free(tag_strings[i]);
}

// -----------------------------------------------------------------------------

struct interpolation_method * yac_new_interpolation_method_conserv_mask(
  struct global_search * search, struct domain * domain,
  char const * stack_name, struct points *** points, struct mask *** masks,
  unsigned num_pointsets, unsigned points_idx, int order, int enforced_conserv,
  int partial_coverage,
  enum yac_interpolation_method_conserv_normalisation normalisation,
  unsigned use_source_mask) {

  // ensure that the right point data is available

  if (num_pointsets == 0)
    yac_internal_abort_message(
      "ERROR: missing data points.", __FILE__, __LINE__);
  if (points_idx >= num_pointsets)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_conserv_mask): invalid points_idx.",
      __FILE__, __LINE__);
  if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_conserv_mask): no point set data "
      "available", __FILE__, __LINE__);
  if (points[0][points_idx]->location != CELL)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_conserv_mask): data points have wrong "
      "location type (needs to be CELL).", __FILE__, __LINE__);
  if ((order < 1) || (order > 2))
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_conserv_mask): "
      "invalid interpolation order", __FILE__, __LINE__);

  struct interpolation_method_conserv * method = xmalloc(1 * sizeof(*method));

  method->vtable = &interpolation_method_conserv_vtable;
  method->search = search;
  method->finish_callback = NULL;
  method->finish_user_data = NULL;
  method->order = order;
  method->enforced_conserv = enforced_conserv;
  method->partial_coverage = partial_coverage;
  method->normalisation = normalisation;
  method->points_idx = points_idx;
  method->use_source_mask = use_source_mask;
  method->domain = domain;
  method->points = points;
  method->masks = masks;
  method->num_pointsets = num_pointsets;

  if ((method->order == 2) && (normalisation != CONSERV_DESTAREA))
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_conserv_mask): 2nd order conservative "
      "interpolation can only be used in combination with CONSERV_DESTAREA",
      __FILE__, __LINE__);

  yac_init_interpolation_method((struct interpolation_method *)method, search);

  generate_interpolation_method_conserv_tag(method, stack_name);

  return (struct interpolation_method *)method;
}

struct interpolation_method * yac_new_interpolation_method_conserv(
  struct global_search * search, struct domain * domain,
  char const * stack_name, struct points *** points, struct mask *** masks,
  unsigned num_pointsets, unsigned points_idx, int order, int enforced_conserv,
  int partial_coverage,
  enum yac_interpolation_method_conserv_normalisation normalisation ) {

  unsigned use_source_mask = 1;

  return
    yac_new_interpolation_method_conserv_mask(
      search, domain, stack_name, points, masks, num_pointsets, points_idx,
      order, enforced_conserv, partial_coverage, normalisation,
      use_source_mask);
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

/**
 * Removes cells from the list of target cells (even if they could
 * be interpolated by the local process). This is because they are going
 * to be interpolated by another source process which might later
 * request some data from the local process to do the interpolation.
 */
static void remove_cells(struct search_data_conserv * data) {

  struct weight_provenance * weight_provenance = data->weight_provenance;
  size_t new_num_cells = 0, num_cells = data->num_cells;
  int * global_ids_to_be_removed = data->target_cell_ids_to_remove;
  size_t num_cells_to_be_removed = data->num_target_cell_ids_to_remove;

  // sort global ids (weight_provenance is sorted by the global ids of the
  // target cells at this point)
  qsort(global_ids_to_be_removed, num_cells_to_be_removed,
        sizeof(*global_ids_to_be_removed), compare_int);

  for (size_t i = 0, j = 0; i < num_cells; ++i) {

    int curr_global_id = weight_provenance[i].target_cell_id;
    while ((j < num_cells_to_be_removed) &&
           (global_ids_to_be_removed[j] < curr_global_id)) ++j;

    if ((j == num_cells_to_be_removed) ||
        (global_ids_to_be_removed[j] != curr_global_id)) {

      if (i != new_num_cells)
        weight_provenance[new_num_cells] = weight_provenance[i];
      ++new_num_cells;

    } else {
      if (weight_provenance[i].is_master)
        free(weight_provenance[i].servant_ranks);
      free(weight_provenance[i].overlaps);
    }
  }

  free(global_ids_to_be_removed);

  // write back the data that was reduced by the cells that are interpolated by
  // another source process

  data->num_cells = new_num_cells;
}

// -----------------------------------------------------------------------------

static int compare_remote_cells_global_id(const void * a, const void * b) {

  int ret = ((struct temp_remote_cell*)a)->global_id -
            ((struct temp_remote_cell*)b)->global_id;
  if (ret) return ret;
  return (((struct temp_remote_cell*)a)->rank >
          ((struct temp_remote_cell*)b)->rank) -
         (((struct temp_remote_cell*)a)->rank <
          ((struct temp_remote_cell*)b)->rank);
}

static void remove_duplicated_remote_cells(
  struct temp_remote_cell * remote_cells, size_t * n) {

  size_t n_ = *n, new_n = 1;

  if (n_ == 0) return;

  qsort(remote_cells, n_, sizeof(*remote_cells),
        compare_remote_cells_global_id);

  int prev_global_id = remote_cells[0].global_id;

  for (size_t i = 1; i < n_; ++i) {

    if (remote_cells[i].global_id != prev_global_id) {

      prev_global_id = remote_cells[i].global_id;
      remote_cells[new_n] = remote_cells[i];
      new_n++;
    }
  }

  *n = new_n;
}

/**
 * Check for source cells that are required for the interpolation, which are
 * located on other source processes or on another local subdomain (which the
 * target cell is not assigned to).
 *
 * @param[out] num_remote_global_ids  number of entries in remote_global_ids
 * @param[out] remote_global_ids      global ids of source cells, required for
 *                                    the interpolation, that are located on
 *                                    remote source domains
 * @param[out] remote_global_id_ranks rank for process that has the respective
 *                                    remote global ids
 * @param[in]  data                   basic search data
 *                                    (see \ref search_data_conserv)
 */
static void
check_for_remote_global_ids(unsigned * num_remote_global_ids,
                            int ** remote_global_ids,
                            int ** remote_global_id_ranks,
                            struct search_data_conserv * data ) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(data->method_stack[0]);
  struct domain * domain = method_conserv->domain;
  int comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));
  int is_2nd_order = method_conserv->order == 2;

  struct weight_provenance * weight_provenance = data->weight_provenance;

  size_t num_interpolated_cells = (size_t)(*(data->num_interpolated_cells));
  size_t num_remote_global_ids_ = 0;
  unsigned num_global_ids_in_remote_gradients = 0;

  // count number of remote overlaps and remote components of gradients of
  // local cells
  for (size_t i = 0; i < num_interpolated_cells; ++i) {

    size_t curr_num_overlaps = (size_t)(weight_provenance[i].nbr_overlaps);
    struct overlap * curr_overlaps = weight_provenance[i].overlaps;

    for (size_t k = 0; k < curr_num_overlaps; ++k) {

      // if the current overlap is with a remote cell
      if ( curr_overlaps[k].rank != comm_rank ) {

        num_remote_global_ids_++;

        // if we use 2nd order conservative -> check components of gradients
        // for remote cells
        if (is_2nd_order)
          num_global_ids_in_remote_gradients +=
            data->remote_src_cell_gradients->gradients[
              curr_overlaps[k].local_cell_id].n;

      // the current overlap is with a local cell
      } else {

        // if we use 2nd order conservative -> check components of gradients
        // for remote cells
        if (is_2nd_order) {

          unsigned subdomain_id = curr_overlaps[k].subdomain_id;
          unsigned local_id = curr_overlaps[k].local_cell_id;
          unsigned const * ranks =
            yac_get_global_ranks(domain, subdomain_id, CELL);

          struct weight_vector_3d * gradient =
            data->src_cell_gradients[subdomain_id] + local_id;

          for (size_t j = 0; j < gradient->n; ++j)
            if (ranks[gradient->data[j].id.local] != (unsigned)-1)
              num_remote_global_ids_++;
        }
      }
    }
  }

  int * global_ids_from_remote_gradients = NULL;

  // if we use 2nd order conservative -> get all unique global ids from
  // remote gradients
  if (is_2nd_order && (data->remote_src_cell_gradients_size > 0)) {

    struct weight_vector_3d * remote_gradients =
      remote_gradients = data->remote_src_cell_gradients->gradients;
    global_ids_from_remote_gradients =
      xmalloc(num_global_ids_in_remote_gradients *
              sizeof(*global_ids_from_remote_gradients));

    // get all global ids from used remote gradients
    for (size_t i = 0, j = 0; i < num_interpolated_cells; ++i) {

      size_t curr_num_overlaps = (size_t)(weight_provenance[i].nbr_overlaps);
      struct overlap * curr_overlaps = weight_provenance[i].overlaps;

      for (size_t k = 0; k < curr_num_overlaps; ++k) {

        // if the overlapping source cell is a remote cell
        if (curr_overlaps[k].rank != comm_rank) {

          struct weight_vector_3d * gradient =
            remote_gradients + curr_overlaps[k].local_cell_id;

          for (size_t l = 0; l < gradient->n; ++l, ++j)
            global_ids_from_remote_gradients[j] = gradient->data[l].id.global;
        }
      }
    }

    // remove duplicated global ids
    qsort(global_ids_from_remote_gradients, num_global_ids_in_remote_gradients,
          sizeof(*global_ids_from_remote_gradients), compare_int);
    yac_remove_duplicates_int(
      global_ids_from_remote_gradients, &num_global_ids_in_remote_gradients);
  }

  struct temp_remote_cell * temp_remote_cells =
    xmalloc((num_remote_global_ids_ + num_global_ids_in_remote_gradients) *
            sizeof(*temp_remote_cells));

   num_remote_global_ids_ = 0;

  // if we use 2nd order conservative -> check components of gradients
  // for remote cells
  if (is_2nd_order && (num_global_ids_in_remote_gradients > 0)) {

    struct remote_cell * remote_cells =
      data->remote_src_cell_gradients->remote_cells;

    // add all entries from global_ids_from_remote_gradients of cells that are
    // only available from remote processes by checking the entries in
    // remote_cells
    for (size_t i = 0, j = 0; i < (size_t)num_global_ids_in_remote_gradients;
         ++i) {

      int curr_global_id = global_ids_from_remote_gradients[i];
      while(remote_cells[j].global_id != curr_global_id) ++j;

      unsigned rank = remote_cells[j].rank;

      if (rank != comm_rank) {
        temp_remote_cells[num_remote_global_ids_].global_id = curr_global_id;
        temp_remote_cells[num_remote_global_ids_].rank = rank;
        ++num_remote_global_ids_;
      }
    }
  }
  free(global_ids_from_remote_gradients);

  // get all global ids and ranks of all remote overlaps and from remote
  // components of gradients of local cells
  for (size_t i = 0; i < num_interpolated_cells; ++i) {

    size_t curr_num_overlaps = (size_t)(weight_provenance[i].nbr_overlaps);
    struct overlap * curr_overlaps = weight_provenance[i].overlaps;

    for (size_t k = 0; k < curr_num_overlaps; ++k) {

      // if the current overlap is with a remote cell
      if ( curr_overlaps[k].rank != comm_rank ) {

        temp_remote_cells[num_remote_global_ids_].global_id =
          curr_overlaps[k].global_cell_id;
        temp_remote_cells[num_remote_global_ids_].rank = curr_overlaps[k].rank;
        ++num_remote_global_ids_;

      // the current overlap is with a local cell
      } else {

        // if we use 2nd order conservative -> check components of gradients
        // for remote cells
        if (is_2nd_order) {

          unsigned subdomain_id = curr_overlaps[k].subdomain_id;
          unsigned local_id = curr_overlaps[k].local_cell_id;
          int const * global_ids =
            yac_get_global_ids(domain, subdomain_id, CELL);
          unsigned const * ranks =
            yac_get_global_ranks(domain, subdomain_id, CELL);

          struct weight_vector_3d * gradient =
            data->src_cell_gradients[subdomain_id] + local_id;

          for (size_t l = 0; l < gradient->n; ++l) {

            unsigned curr_local_id = gradient->data[l].id.local;
            unsigned rank = ranks[curr_local_id];

            if (rank != (unsigned)-1) {
              temp_remote_cells[num_remote_global_ids_].global_id =
                global_ids[curr_local_id];
              temp_remote_cells[num_remote_global_ids_].rank = rank;
              ++num_remote_global_ids_;
            }
          }
        }
      }
    }
  }

  remove_duplicated_remote_cells(temp_remote_cells, &num_remote_global_ids_);

  int * remote_global_ids_ =
    xmalloc(num_remote_global_ids_ * sizeof(*remote_global_ids_));
  int * remote_global_id_ranks_ =
    xmalloc(num_remote_global_ids_ * sizeof(*remote_global_id_ranks_));

  for (size_t i = 0; i < num_remote_global_ids_; ++i) {
    remote_global_ids_[i] = temp_remote_cells[i].global_id;
    remote_global_id_ranks_[i] = temp_remote_cells[i].rank;
  }
  free(temp_remote_cells);

  *remote_global_ids = remote_global_ids_;
  *remote_global_id_ranks = remote_global_id_ranks_;
  *num_remote_global_ids = num_remote_global_ids_;
}

static int compare_weight_vector_data_weight(
  void const * a, void const * b) {

  struct weight_vector_data const * weight_a =
    (struct weight_vector_data const *)a;
  struct weight_vector_data const * weight_b =
    (struct weight_vector_data const *)b;

  int ret = weight_a->global_id - weight_b->global_id;
  if (ret) return ret;
  double abs_weight_a = fabs(weight_a->weight);
  double abs_weight_b = fabs(weight_b->weight);
  ret = (abs_weight_a > abs_weight_b) - (abs_weight_a < abs_weight_b);
  if (ret) return ret;
  return (weight_a->weight > weight_b->weight) -
         (weight_a->weight < weight_b->weight);
}

static int compare_weight_vector_data(
  void const * a, void const * b) {

  struct weight_vector_data const * weight_a =
    (struct weight_vector_data const *)a;
  struct weight_vector_data const * weight_b =
    (struct weight_vector_data const *)b;

  return weight_a->global_id - weight_b->global_id;
}

static void compact_weight_vector_data(
  struct weight_vector_data * weights, size_t * n) {

  size_t n_ = *n;

  if (n_ <= 1) return;

  // sort weights by global_id then by weight
  qsort(weights, n_, sizeof(*weights), compare_weight_vector_data_weight);

  size_t new_n = 1;
  struct weight_vector_data * prev_weight_data = weights;
  struct weight_vector_data * curr_weight_data = weights + 1;
  for (size_t i = 1; i < n_; ++i, ++curr_weight_data) {

    // if both weights refer to the same source point (by global_id); they can
    // be local or remote (subdomain_id == (unsigned)-1)
    if (!compare_weight_vector_data(prev_weight_data, curr_weight_data)) {
      prev_weight_data->weight += curr_weight_data->weight;
      // the following if ensures that local points are prefered before remote
      // ones
      if (prev_weight_data->subdomain_id > curr_weight_data->subdomain_id) {
        prev_weight_data->subdomain_id = curr_weight_data->subdomain_id;
        prev_weight_data->local_id = curr_weight_data->local_id;
      }
    } else {
      ++new_n;
      ++prev_weight_data;
      *prev_weight_data = *curr_weight_data;
    }
  }

  *n = new_n;
}

static size_t get_2nd_order_weights(
  struct overlap * overlap, struct weight_vector_data * weights,
  struct weight_vector_3d ** src_cell_gradients, int const ** global_ids,
  unsigned const ** ranks, struct remote_cell * remote_cells,
  double target_cell_area) {

  int subdomain_id = overlap->subdomain_id;
  int local_cell_id = overlap->local_cell_id;

  weights[0].weight       = overlap->weight;
  weights[0].global_id    = overlap->global_cell_id;
  weights[0].subdomain_id = subdomain_id;
  weights[0].local_id     = local_cell_id;

  struct weight_vector_3d * src_cell_gradient =
    src_cell_gradients[subdomain_id] + local_cell_id;

  size_t N = src_cell_gradient->n;
  struct weight_vector_data_3d * gradient_weights = src_cell_gradient->data - 1;
  double * overlap_barycenter = overlap->barycenter;

  for (size_t n = 1; n <= N; ++n) {
    weights[n].weight       =
      (gradient_weights[n].weight[0] * overlap_barycenter[0] +
       gradient_weights[n].weight[1] * overlap_barycenter[1] +
       gradient_weights[n].weight[2] * overlap_barycenter[2]) *
      (overlap->area / target_cell_area);
  }

  if (subdomain_id == -1) {
    for (size_t n = 1; n <= N; ++n) {
      struct remote_cell * remote_cell =
        remote_cells + gradient_weights[n].id.local;
      weights[n].global_id    = remote_cell->global_id;
      weights[n].subdomain_id = remote_cell->subdomain_id;
      weights[n].local_id     = remote_cell->local_id;
    }
  } else {
    int const * subdomain_global_ids = global_ids[subdomain_id];
    unsigned const * subdomain_ranks = ranks[subdomain_id];
    for (size_t n = 1; n <= N; ++n) {
      unsigned local_id = gradient_weights[n].id.local;
      weights[n].global_id    = subdomain_global_ids[local_id];
      if (subdomain_ranks[local_id] != (unsigned)-1) {
        weights[n].subdomain_id = -1;
        weights[n].local_id     = -1;
      } else {
        weights[n].subdomain_id = subdomain_id;
        weights[n].local_id     = local_id;
      }
    }
  }

  return 1 + N;
}

static void get_src_indices_1st_order(
  unsigned (**src_indices)[3], unsigned ** num_src_per_tgt, double ** weights,
  int comm_rank, unsigned num_interpolated_cells,
  struct weight_provenance * weight_provenance, int * remote_global_ids,
  unsigned points_idx) {

  size_t max_num_total_weights = 0;
  for (unsigned n = 0; n < num_interpolated_cells; ++n )
    max_num_total_weights += weight_provenance[n].nbr_overlaps;

  unsigned (*src_indices_)[3] =
    xmalloc(max_num_total_weights * sizeof(*src_indices_));
  unsigned * num_src_per_tgt_ =
    xcalloc(num_interpolated_cells, sizeof(*num_src_per_tgt_));
  double * weights_ = xmalloc(max_num_total_weights * sizeof(*weights_));

  *src_indices = src_indices_;
  *num_src_per_tgt = num_src_per_tgt_;
  *weights = weights_;

  size_t num_remote_overlaps = 0;

  // compute the number of remote overlaps
  for (unsigned n = 0; n < num_interpolated_cells; ++n) {

    size_t nbr_overlaps = weight_provenance[n].nbr_overlaps;
    struct overlap * overlaps = weight_provenance[n].overlaps;

    for (size_t i = 0; i < nbr_overlaps; ++i) {

      struct overlap * curr_overlap = overlaps + i;

      if ((fabs(curr_overlap->weight) > 1.e-16) &&
          (curr_overlap->rank != comm_rank)) num_remote_overlaps++;
    }
  }

  int * remote_global_ids_to_check =
    xmalloc(num_remote_overlaps * sizeof(*remote_global_ids_to_check));
  unsigned * remote_global_ids_to_check_pos =
    xmalloc(num_remote_overlaps * sizeof(*remote_global_ids_to_check_pos));
  num_remote_overlaps = 0;

  // loop over all target cells that can be interpolated by this process
  for (unsigned n = 0, num_src_indices = 0; n < num_interpolated_cells; ++n) {

    size_t nbr_overlaps = weight_provenance[n].nbr_overlaps;
    struct overlap * overlaps = weight_provenance[n].overlaps;

    // store requests to other source processes if necessary
    for (size_t i = 0; i < nbr_overlaps; ++i) {

      struct overlap * curr_overlap = overlaps + i;

      if (fabs(curr_overlap->weight) < 1.e-16) continue;

      if (curr_overlap->rank != comm_rank ) {

        // source cell weights come from a remote rank
        remote_global_ids_to_check[num_remote_overlaps] =
          curr_overlap->global_cell_id;
        remote_global_ids_to_check_pos[num_remote_overlaps] = num_src_indices;
        num_remote_overlaps++;

        src_indices_[num_src_indices][0] = 0;
        src_indices_[num_src_indices][1] = points_idx;
        // src_indices_[num_src_indices][2] = index;

      } else {

        // we have the source cell data in our own subdomain
        src_indices_[num_src_indices][0] = curr_overlap->subdomain_id + 1; // source subdomain ID + 1
        src_indices_[num_src_indices][1] = points_idx;
        src_indices_[num_src_indices][2] = curr_overlap->local_cell_id;    // local source cell ID
      }

      weights_[num_src_indices] = curr_overlap->weight;
      num_src_per_tgt_[n]++;
      num_src_indices++;
    }
  }

  // look up the local ids for the remote global ids
  yac_quicksort_index(
    remote_global_ids_to_check, num_remote_overlaps,
    (int*)remote_global_ids_to_check_pos);

  for (size_t i = 0, j = 0; i < num_remote_overlaps; ++i) {

    int curr_global_id = remote_global_ids_to_check[i];
    while (curr_global_id != remote_global_ids[j]) ++j;
    assert(curr_global_id == remote_global_ids[j]);

    src_indices_[remote_global_ids_to_check_pos[i]][2] = (unsigned)j;
  }
  free(remote_global_ids_to_check);
  free(remote_global_ids_to_check_pos);
}

static void get_src_indices_2nd_order(
  unsigned (**src_indices)[3], unsigned ** num_src_per_tgt, double ** weights,
  unsigned num_interpolated_cells, struct weight_provenance * weight_provenance,
  struct weight_vector_3d ** src_cell_gradients, int * remote_global_ids,
  struct remote_cell * remote_cells, unsigned points_idx,
  struct domain * domain) {

  size_t max_num_total_weights = 0;
  size_t max_num_weights = 0;

  // weights for gradient
  for (unsigned n = 0; n < num_interpolated_cells; ++n ) {
    size_t curr_num_weights = 0;
    curr_num_weights += weight_provenance[n].nbr_overlaps;
    for (size_t i = 0; i < weight_provenance[n].nbr_overlaps; ++i) {
      int subdomain_id = weight_provenance[n].overlaps[i].subdomain_id;
      int local_cell_id = weight_provenance[n].overlaps[i].local_cell_id;
      curr_num_weights += src_cell_gradients[subdomain_id][local_cell_id].n;
    }
    if (curr_num_weights > max_num_weights) max_num_weights = curr_num_weights;
    max_num_total_weights += curr_num_weights;
  }

  unsigned (*src_indices_)[3] =
    xmalloc(max_num_total_weights * sizeof(*src_indices_));
  unsigned * num_src_per_tgt_ =
    xmalloc(num_interpolated_cells * sizeof(*num_src_per_tgt_));
  double * weights_ = xmalloc(max_num_total_weights * sizeof(*weights_));

  *src_indices = src_indices_;
  *num_src_per_tgt = num_src_per_tgt_;
  *weights = weights_;

  unsigned num_subdomains = yac_get_num_subdomains(domain);
  int const ** global_ids = xmalloc(num_subdomains * sizeof(*global_ids));
  unsigned const ** ranks = xmalloc(num_subdomains * sizeof(*ranks));
  for (unsigned i = 0; i < num_subdomains; ++i) {
    global_ids[i] = yac_get_global_ids(domain, i, CELL);
    ranks[i] = yac_get_global_ranks(domain, i, CELL);
  }

  struct weight_vector_data * weight_buffer =
    xmalloc(max_num_weights * sizeof(*weight_buffer));
  int * remote_global_ids_to_check =
    xmalloc(max_num_weights * sizeof(*remote_global_ids_to_check));
  int * remote_global_ids_to_check_pos =
    xmalloc(max_num_weights * sizeof(*remote_global_ids_to_check_pos));

  // loop over all target cells that can be interpolated by this process
  for ( unsigned n = 0, num_src_indices = 0; n < num_interpolated_cells; ++n ) {

    size_t nbr_overlaps = weight_provenance[n].nbr_overlaps;
    struct overlap * overlaps = weight_provenance[n].overlaps;
    size_t weight_buffer_offset = 0;

    // gather all weights
    for (size_t i = 0; i < nbr_overlaps; ++i) {

      struct overlap * curr_overlap = overlaps + i;

      if (fabs(curr_overlap->weight) < 1.e-16) continue;

      size_t num_weights =
        get_2nd_order_weights(curr_overlap, weight_buffer+weight_buffer_offset,
                              src_cell_gradients, global_ids, ranks,
                              remote_cells,
                              weight_provenance[n].target_cell_area);
      weight_buffer_offset += num_weights;
    }

    // compact weights and sort them by global_id
    compact_weight_vector_data(weight_buffer, &weight_buffer_offset);

    unsigned (*curr_src_indices)[3] = src_indices_ + num_src_indices;
    double * curr_weights = weights_ + num_src_indices;

    size_t num_remote_global_ids_to_check = 0;

    // store weights in src_indices
    for (size_t i = 0; i < weight_buffer_offset; ++i) {

      unsigned subdomain_id = weight_buffer[i].subdomain_id;
      unsigned local_id = weight_buffer[i].local_id;

      if (local_id == (unsigned)-1) {

        // source cell weights come from a remote rank
        remote_global_ids_to_check[num_remote_global_ids_to_check] =
          weight_buffer[i].global_id;
        remote_global_ids_to_check_pos[num_remote_global_ids_to_check] = (int)i;
        ++num_remote_global_ids_to_check;

        curr_src_indices[i][0] = 0;
        curr_src_indices[i][1] = points_idx;
        // curr_src_indices[i][2] = index;

      } else {

        // we have the source cell data in our own subdomain
        curr_src_indices[i][0] = subdomain_id + 1; // source subdomain ID + 1
        curr_src_indices[i][1] = points_idx;
        curr_src_indices[i][2] = local_id; // local source cell ID
      }

      curr_weights[i] = weight_buffer[i].weight;
    }

    // look up the local ids of the remote global ids
    yac_quicksort_index(
      remote_global_ids_to_check, num_remote_global_ids_to_check,
      remote_global_ids_to_check_pos);
    for (size_t i = 0, j = 0; i < num_remote_global_ids_to_check; ++i) {

      int curr_global_id = remote_global_ids_to_check[i];
      while (curr_global_id != remote_global_ids[j]) ++j;
      assert(curr_global_id == remote_global_ids[j]);

      curr_src_indices[remote_global_ids_to_check_pos[i]][2] = (unsigned)j;
    }

    num_src_per_tgt_[n] = (unsigned)weight_buffer_offset;
    num_src_indices += (unsigned)weight_buffer_offset;
  }

  free(remote_global_ids_to_check);
  free(remote_global_ids_to_check_pos);
  free(ranks);
  free(global_ids);
  free(weight_buffer);
}

// -----------------------------------------------------------------------------

/**
 * generates the interpolation_plan_element based on the collected search data
 *
 * @param[in]  data                  basic search data
 *                                   (see \ref search_data_conserv)
 * @param[in]  remote_global_ids     global ids of source cells, required for
 *                                   the interpolation, that are located on
 *                                   remote source domains
 * @param[in]  num_remote_global_ids number of entries in remote_global_ids
 */

static struct interpolation_plan_element *
generate_interpolation_plan_conserv ( struct search_data_conserv * data,
                                      int * remote_global_ids,
                                      unsigned num_remote_global_ids ) {

  if (*(data->num_interpolated_cells) == 0) return NULL;

  struct weight_provenance * weight_provenance = data->weight_provenance;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(data->method_stack[0]);
  unsigned points_idx = method_conserv->points_idx;
  int comm_rank =
    yac_get_comm_rank(
      yac_get_domain_comm(yac_get_local_domain(method_conserv->search)));

  unsigned (*src_indices)[3];
  unsigned * num_src_per_tgt;
  double * weights;

  switch(method_conserv->order) {
    case (1):
      get_src_indices_1st_order(&src_indices, &num_src_per_tgt, &weights,
                                comm_rank, *(data->num_interpolated_cells),
                                weight_provenance, remote_global_ids,
                                points_idx);
      break;
    case(2): {
      struct remote_cell * remote_cells;
      if (data->remote_src_cell_gradients_size == 0) remote_cells = NULL;
      else remote_cells = data->remote_src_cell_gradients->remote_cells;

      get_src_indices_2nd_order(&src_indices, &num_src_per_tgt, &weights,
                                *(data->num_interpolated_cells),
                                weight_provenance, data->src_cell_gradients,
                                remote_global_ids, remote_cells, points_idx,
                                method_conserv->domain);
      break;
    }
    default:
      yac_internal_abort_message(
        "ERROR: invalid interpolation order\n", __FILE__, __LINE__);
  };

  unsigned num_pointsets = method_conserv->num_pointsets;

  unsigned * num_remote_global_ids_;

  num_remote_global_ids_ =
    xcalloc(num_pointsets, sizeof(*num_remote_global_ids_));
  num_remote_global_ids_[points_idx] = num_remote_global_ids;

  // generate a plan to execute the interpolation
  struct interpolation_plan_element * plan;

  plan = xmalloc(1 * sizeof(*plan));

  yac_init_interpolation_plan_element_wpointsum(
    plan, *(data->num_interpolated_cells), &((*src_indices)[0]),
    num_src_per_tgt, weights, num_remote_global_ids_, num_pointsets);

  free(num_remote_global_ids_);
  free(num_src_per_tgt);
  free(src_indices);
  free(weights);

  return plan;
}

static void convert_remote_src_cell_gradients(
  struct remote_gradients * remote_gradients,
  struct weight_vector_3d ** gradients,
  int * remote_global_ids, unsigned num_remote_global_ids) {

  if (num_remote_global_ids == 0) return;

  //-----------------------------------------------------
  // set the gradients for each remote cell
  // (for some remote cells we might not have a gradient)
  //-----------------------------------------------------
  struct weight_vector_3d * gradients_ =
    xmalloc(num_remote_global_ids * sizeof(**gradients));
  *gradients = gradients_;
  for (unsigned i = 0; i < num_remote_global_ids; ++i)
    gradients_[i] = (struct weight_vector_3d){.data = NULL, .n = 0};

  size_t num_remote_gradients = remote_gradients->num_gradients;
  int * remote_gradient_global_ids = remote_gradients->global_src_cell_ids;

  for (size_t i = 0, j = 0; i < num_remote_gradients; ++i) {

    int curr_remote_gradient_global_id = remote_gradient_global_ids[i];

    while (remote_global_ids[j] < curr_remote_gradient_global_id) ++j;


    // in case the global id of the current remote gradient is not in the list
    // of required remote cells, we do not need the respective gradient and can
    // skip it
    if (remote_global_ids[j] != curr_remote_gradient_global_id) continue;

    gradients_[j] = remote_gradients->gradients[i];
  }

  //----------------------------------------------------------------------------
  // convert the global_id of the weights in the remote gradients into local ids
  // which refere to remote_cells
  //----------------------------------------------------------------------------

  struct remote_cell * remote_cells = remote_gradients->remote_cells;
  size_t num_global_ids_to_convert =
    remote_gradients->weight_vector_data_3d_buffer_size;
  struct weight_vector_data_3d * weight_vector_data_3d_buffer =
    remote_gradients->weight_vector_data_3d_buffer;
  int * global_ids_to_convert =
    xmalloc(num_global_ids_to_convert * sizeof(*global_ids_to_convert));
  int * global_ids_to_convert_pos =
    xmalloc(num_global_ids_to_convert * sizeof(*global_ids_to_convert_pos));
  for (size_t i = 0; i < num_global_ids_to_convert; ++i) {
    global_ids_to_convert[i] = weight_vector_data_3d_buffer[i].id.global;
    global_ids_to_convert_pos[i] = (int)i;
  }

  yac_quicksort_index(global_ids_to_convert, num_global_ids_to_convert,
                      global_ids_to_convert_pos);

  for (size_t i = 0, j = 0; i < num_global_ids_to_convert; ++i) {

    int curr_global_id_to_convert = global_ids_to_convert[i];

    // check whether the current global id is available locally

    while (remote_cells[j].global_id < curr_global_id_to_convert) ++j;

    assert(remote_cells[j].global_id == curr_global_id_to_convert);

    weight_vector_data_3d_buffer[global_ids_to_convert_pos[i]].id.local =
      (unsigned)j;
  }

  free(global_ids_to_convert_pos);
  free(global_ids_to_convert);
}

static int
compare_remote_overlap_global_id(const void * a, const void * b) {

  struct remote_overlap * g_a = (struct remote_overlap *)a;
  struct remote_overlap * g_b = (struct remote_overlap *)b;

  return (g_a->global_id > g_b->global_id) -
         (g_a->global_id < g_b->global_id);
}

static void get_required_remote_overlaps(
  struct weight_provenance * weight_provenance, size_t num_cells,
  struct remote_overlap ** remote_overlaps, size_t * num_remote_overlaps) {

  size_t num_remote_overlaps_ = 0;

  // count number of remote overlaps
  for (size_t i = 0; i < num_cells; ++i)
    for (size_t j = 0; j < weight_provenance[i].nbr_overlaps; ++j)
      if (weight_provenance[i].overlaps[j].subdomain_id == (unsigned)-1)
        num_remote_overlaps_++;

  struct remote_overlap * remote_overlaps_ =
    xmalloc(num_remote_overlaps_ * sizeof(*remote_overlaps_));

  // get remote overlaps
  for (size_t i = 0, offset = 0; i < num_cells; ++i) {

    for (size_t j = 0; j < weight_provenance[i].nbr_overlaps; ++j) {

      if (weight_provenance[i].overlaps[j].subdomain_id == (unsigned)-1) {

        remote_overlaps_[offset].tgt_cell_idx = i;
        remote_overlaps_[offset].overlap_idx = j;
        remote_overlaps_[offset].global_id =
          weight_provenance[i].overlaps[j].global_cell_id;
        remote_overlaps_[offset].local_id = (unsigned)-1;
        offset++;
      }
    }
  }

  // sort remote overlaps by their global_id
  qsort(remote_overlaps_, num_remote_overlaps_, sizeof(*remote_overlaps_),
        compare_remote_overlap_global_id);

  *remote_overlaps = remote_overlaps_;
  *num_remote_overlaps = num_remote_overlaps_;
}

static int
compare_remote_overlap_cell_idx(const void * a, const void * b) {

  struct remote_overlap * g_a = (struct remote_overlap *)a;
  struct remote_overlap * g_b = (struct remote_overlap *)b;

  int ret = (g_a->tgt_cell_idx > g_b->tgt_cell_idx) -
            (g_a->tgt_cell_idx < g_b->tgt_cell_idx);
  if (ret) return ret;
  return (g_a->overlap_idx > g_b->overlap_idx) -
         (g_a->overlap_idx < g_b->overlap_idx);
}

static void set_local_ids_in_remote_overlaps(
  int * remote_global_ids, struct remote_overlap * remote_overlaps,
  size_t num_remote_overlaps, struct weight_provenance * weight_provenance) {

  // translate global_ids to local_ids
  for (size_t i = 0, j = 0; i < num_remote_overlaps; ++i) {

    int curr_global_id = remote_overlaps[i].global_id;
    while (remote_global_ids[j] < curr_global_id) ++j;
    remote_overlaps[i].local_id = j;
  }

  // sorte remote_overlaps by tgt_cell_idx and overlap_idx
  qsort(remote_overlaps, num_remote_overlaps, sizeof(*remote_overlaps),
        compare_remote_overlap_cell_idx);

  // write back local_ids
  for (size_t i = 0; i < num_remote_overlaps; ++i)
    weight_provenance[remote_overlaps[i].tgt_cell_idx].
      overlaps[remote_overlaps[i].overlap_idx].local_cell_id =
        remote_overlaps[i].local_id;
}

static void remove_unused_remote_gradients(
  struct search_data_conserv * data, struct remote_overlap * remote_overlaps,
  size_t num_remote_overlaps) {

  if (data->remote_src_cell_gradients_size == 0) return;

  struct weight_provenance * weight_provenance = data->weight_provenance;
  struct remote_gradients * remote_gradients = data->remote_src_cell_gradients;
  struct weight_vector_3d * gradients = remote_gradients->gradients;
  int * global_ids = remote_gradients->global_src_cell_ids;
  size_t num_gradients = remote_gradients->num_gradients;
  struct weight_vector_data_3d * weight_vector_data_3d_buffer =
    remote_gradients->weight_vector_data_3d_buffer;

  size_t new_num_gradients = 0;
  size_t new_weight_vector_data_3d_buffer_size = 0;

  // remove unused remote gradients
  for (size_t i = 0, j = 0; (i < num_remote_overlaps) && (j < num_gradients);) {

    while (remote_overlaps[i].global_id != global_ids[j]) ++j;

    assert(j < num_gradients);
    assert(remote_overlaps[i].global_id == global_ids[j]);

    do {
      weight_provenance[remote_overlaps[i].tgt_cell_idx].
        overlaps[remote_overlaps[i].overlap_idx].local_cell_id =
        new_num_gradients;
      ++i;
    } while ((i < num_remote_overlaps) &&
             (remote_overlaps[i].global_id == global_ids[j]));

    if (j != new_num_gradients) {
      gradients[new_num_gradients].data =
        weight_vector_data_3d_buffer + new_weight_vector_data_3d_buffer_size;
      memmove(gradients[new_num_gradients].data, gradients[j].data,
              gradients[j].n * sizeof(*weight_vector_data_3d_buffer));
      gradients[new_num_gradients].n = gradients[j].n;
      global_ids[new_num_gradients] = global_ids[j];
    }

    new_num_gradients++;
    new_weight_vector_data_3d_buffer_size += gradients[j].n;
  }

  // update
  if (new_num_gradients != num_gradients) {
    weight_vector_data_3d_buffer =
      xrealloc(weight_vector_data_3d_buffer,
               new_weight_vector_data_3d_buffer_size *
               sizeof(*weight_vector_data_3d_buffer));
    remote_gradients->weight_vector_data_3d_buffer =
      weight_vector_data_3d_buffer;
    for (size_t i = 0, offset = 0; i < new_num_gradients; ++i) {
      gradients[i].data = weight_vector_data_3d_buffer + offset;
      offset += gradients[i].n;
    }
    remote_gradients->gradients =
      xrealloc(gradients, new_num_gradients * sizeof(*gradients));
    remote_gradients->global_src_cell_ids =
      xrealloc(global_ids, new_num_gradients * sizeof(*global_ids));
    remote_gradients->num_gradients = new_num_gradients;
    remote_gradients->weight_vector_data_3d_buffer_size =
      new_weight_vector_data_3d_buffer_size;
  }
}

static void set_local_ids_in_remote_cells(
  struct search_data_conserv * data, int * remote_global_ids,
  unsigned num_remote_global_ids) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(data->method_stack[0]);
  struct domain * domain = method_conserv->domain;
  int comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

  struct remote_cell * remote_cells =
    data->remote_src_cell_gradients->remote_cells;
  size_t num_remote_cells =
    data->remote_src_cell_gradients->num_remote_cells;

  int * global_ids = xmalloc(num_remote_cells * sizeof(*global_ids));

  for (size_t i = 0; i < num_remote_cells; ++i)
    global_ids[i] = remote_cells[i].global_id;

  size_t num_local_points = 0;

  unsigned j = 0;
  for (size_t i = 0; i < num_remote_cells; ++i) {

    int curr_global_id = global_ids[i];

    if (remote_cells[i].rank == comm_rank) {
      global_ids[num_local_points] = curr_global_id;
      ++num_local_points;
    } else {

      while ((j < num_remote_global_ids) &&
             (remote_global_ids[j] < curr_global_id)) ++j;

      // if the current remote point is actually not beeing used on the
      // local process
      if ((j == num_remote_global_ids) ||
          (remote_global_ids[j] != curr_global_id))
        continue;

      remote_cells[i].subdomain_id = (unsigned)-1;
      remote_cells[i].local_id = j;
    }
  }

  unsigned num_subdomains = yac_get_num_subdomains(domain);

  unsigned * matching_ids_idx =
    xmalloc(num_local_points * sizeof(*matching_ids_idx));
  unsigned num_matching_ids = 0;
  unsigned * local_ids = xmalloc(num_local_points * sizeof(*local_ids));

  // get the local ids of all points the are available locally
  for (unsigned i = 0; i < num_subdomains; ++i) {
    yac_match_global_ids(
      domain, i, CELL, global_ids, (unsigned)num_local_points, NULL,
      matching_ids_idx, &num_matching_ids, local_ids, 0);

    size_t k = 0;
    for (j = 0; j < num_matching_ids; ++j) {
      int curr_global_id = global_ids[matching_ids_idx[j]];
      while (remote_cells[k].global_id < curr_global_id) ++k;
      remote_cells[k].subdomain_id = i;
      remote_cells[k].local_id = local_ids[j];
    }
  }

  free(local_ids);
  free(matching_ids_idx);
  free(global_ids);
}

static void extract_target_cells(struct search_data_conserv * data) {

  double (*coordinates_xyz)[3] = data->coordinates_xyz_buffer;
  int * global_ids = data->global_ids_buffer;
  struct weight_provenance * weight_provenance = data->weight_provenance;

  size_t num_cells = data->num_cells;

  if (coordinates_xyz == NULL) {

    for (size_t i = 0; i < num_cells; ++i)
      global_ids[i] = weight_provenance[i].target_cell_id;

  } else {

    for (size_t i = 0; i < num_cells; ++i) {
      global_ids[i] = weight_provenance[i].target_cell_id;
      coordinates_xyz[i][0] = weight_provenance[i].coordinate_xyz[0];
      coordinates_xyz[i][1] = weight_provenance[i].coordinate_xyz[1];
      coordinates_xyz[i][2] = weight_provenance[i].coordinate_xyz[2];
    }
  }
}

// -----------------------------------------------------------------------------

static int compare_weight_provenance_can_be_interpolated(
  const void * a, const void * b) {

  struct weight_provenance const * wp_a = (struct weight_provenance const *)a;
  struct weight_provenance const * wp_b = (struct weight_provenance const *)b;

  int ret = wp_b->cell_can_be_interpolated - wp_a->cell_can_be_interpolated;

  if (ret != 0) return ret;
  else return wp_a->target_cell_id - wp_b->target_cell_id;
}

/**
 * finalising the search for conservative remapping
 *
 * call remove, reorder, generate plan, generate data request,
 * call next method, set_finish_callback
 *
 * in addition it could do the weight correction as well
 *
 */

static void
do_search_final_conserv ( struct search_data_conserv * data ) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(data->method_stack[0]);

  // remove all cells that are interpolated by other source processes

  remove_cells(data);

  // if not all cells can be interpolated

  unsigned num_cells = data->num_cells;
  unsigned num_interpolated_cells = data->num_interpolated_cells[0];

  if (num_interpolated_cells != num_cells) {

    // The criteria for cell_can_be_interpolated needs to be revised
    // Note that masks of source cells are not yet communicated to the master
    // nor are they used or stored.

    // move cells to the end for further treatment that cannot be interpolated
    qsort(data->weight_provenance, num_cells, sizeof(*data->weight_provenance),
          compare_weight_provenance_can_be_interpolated);
  }

  //--------------------------------------------------------------------
  // once we know which target points we actually have to interpolate,
  // we can remove information a about remote gradients that is not
  // needed anymore
  //--------------------------------------------------------------------

  struct remote_overlap * remote_overlaps = NULL;
  size_t num_remote_overlaps = 0;
  if (method_conserv->order == 2) {
    get_required_remote_overlaps(
      data->weight_provenance, num_interpolated_cells, &remote_overlaps,
      &num_remote_overlaps);

    remove_unused_remote_gradients(data, remote_overlaps, num_remote_overlaps);
  }

  //--------------------------------------------------------------------
  // Check for source cells required for the interpolation that are not
  // in the subdomain, which the target cells are assigned to. (can be
  // on another local subdomain or on a remote source domain)
  //--------------------------------------------------------------------

  int * remote_global_ids = NULL;
  int * remote_global_id_ranks = NULL;
  unsigned num_remote_global_ids = 0;

  check_for_remote_global_ids(
    &num_remote_global_ids, &remote_global_ids, &remote_global_id_ranks, data );

  //--------------------------------------------------------------------------
  // Copy the gradients received from remote processes into
  // data->src_cell_gradients and convert the global ids of their weights into
  // local ones based on remote_cells
  //--------------------------------------------------------------------------
  struct weight_vector_data_3d * weight_vector_data_3d_buffer = NULL;
  if (method_conserv->order == 2) {
    if (data->remote_src_cell_gradients_size > 0) {
      convert_remote_src_cell_gradients(
        data->remote_src_cell_gradients, &(data->src_cell_gradients[-1]),
        remote_global_ids, num_remote_global_ids);
      // update local ids of all remote overlaps
      set_local_ids_in_remote_overlaps(
        remote_global_ids, remote_overlaps,
        num_remote_overlaps, data->weight_provenance);
      weight_vector_data_3d_buffer =
        data->remote_src_cell_gradients->weight_vector_data_3d_buffer;
      free(data->remote_src_cell_gradients->global_src_cell_ids);
      free(data->remote_src_cell_gradients->gradients);

      set_local_ids_in_remote_cells(
        data, remote_global_ids, num_remote_global_ids);
    }
    free(remote_overlaps);

  }

  if (!method_conserv->partial_coverage &&
      (method_conserv->normalisation == CONSERV_DESTAREA)) {

    // normalise the partial areas (weights) of the intersections with the target area
    // in other words: minimise the error and take care that the sum of weights is as
    // close as possible to 1.0

    size_t nmaxWeights = 0;
    double * weight = NULL;

    for (unsigned n = 0; n < data->num_cells; ++n ) {

      if (data->weight_provenance[n].cell_can_be_interpolated) {
         ENSURE_ARRAY_SIZE(
           weight, nmaxWeights, data->weight_provenance[n].nbr_overlaps );

         for ( size_t i = 0; i < data->weight_provenance[n].nbr_overlaps; ++i )
           weight[i] = data->weight_provenance[n].overlaps[i].weight;

         yac_correct_weights(data->weight_provenance[n].nbr_overlaps, weight);

         for ( size_t i = 0; i < data->weight_provenance[n].nbr_overlaps; ++i )
            data->weight_provenance[n].overlaps[i].weight = weight[i];
      }
    }
    free(weight);
  }

  //----------------------------------------------------------
  // generate a interpolation plan based on the collected data
  //----------------------------------------------------------

  data->plan_stack[0] =
    generate_interpolation_plan_conserv(
      data, remote_global_ids, num_remote_global_ids );
  if (method_conserv->order == 2) {
    struct domain * domain = method_conserv->domain;
    unsigned num_subdomains = yac_get_num_subdomains(domain);
    struct weight_vector_3d ** src_cell_gradients =
      data->src_cell_gradients - 1;
    free(data->weight_vector_data_buffer);
    for (unsigned i = 0; i <= num_subdomains; ++i) free(src_cell_gradients[i]);
    free(src_cell_gradients);
    free(weight_vector_data_3d_buffer);
    if (data->remote_src_cell_gradients_size > 0)
      free(data->remote_src_cell_gradients->remote_cells);
    free(data->remote_src_cell_gradients);
  }

  // advance interpolation plan stack if an interpolation plan was generated

  if (data->plan_stack[0] != NULL) data->plan_stack++;

  unsigned num_pointsets = method_conserv->num_pointsets;
  unsigned points_idx = method_conserv->points_idx;

  // structure of data_request is given and cannot be modified

  method_conserv->data_request[0] =
    xmalloc(num_pointsets * sizeof(*(method_conserv->data_request[0])));

  method_conserv->data_request[0][points_idx].num_values = num_remote_global_ids;
  method_conserv->data_request[0][points_idx].global_ids = remote_global_ids;
  method_conserv->data_request[0][points_idx].ranks = remote_global_id_ranks;

  for (unsigned i = 0; i < num_pointsets; ++i) {
    if (i != points_idx) {
      method_conserv->data_request[0][i].num_values = 0;
      method_conserv->data_request[0][i].global_ids = NULL;
      method_conserv->data_request[0][i].ranks = NULL;
    }
  }

  // copy coordinates_xyz and global_ids from weight_provenance back into the
  // original arrays
  extract_target_cells(data);

  // free weight_provenance data
  for (unsigned n = 0; n < data->num_cells; ++n ) {
    if (data->weight_provenance[n].is_master)
      free(data->weight_provenance[n].servant_ranks);
    free(data->weight_provenance[n].overlaps);
  }
  free(data->weight_provenance);


  // call the do search interpolation_method for the next method
  data->num_cells -= num_interpolated_cells;
  if (data->coordinates_xyz_buffer != NULL)
    data->coordinates_xyz_buffer += num_interpolated_cells;
  if (data->method_stack[1] != NULL)
    data->method_stack[1]->vtable->do_search(
      data->method_stack+1, data->plan_stack,
      data->global_ids_buffer + num_interpolated_cells,
      data->coordinates_xyz_buffer, CELL, data->num_cells,
      data->num_interpolated_cells + 1, method_conserv->data_request + 1);
  free(data);

  if (method_conserv->finish_callback != NULL)
    method_conserv->finish_callback(method_conserv->finish_user_data);
}

// ------------------------------------------------

struct compact_remote_gradients_data {
  int global_id;
  size_t pos[2];
};

static int compare_compact_remote_gradients_data_global_id(
  const void * a, const void * b) {

  int ret = ((struct compact_remote_gradients_data*)a)->global_id -
            ((struct compact_remote_gradients_data*)b)->global_id;
  if (ret) return ret;
  ret = (((struct compact_remote_gradients_data*)a)->pos[0] >
         ((struct compact_remote_gradients_data*)b)->pos[0]) -
        (((struct compact_remote_gradients_data*)a)->pos[0] <
         ((struct compact_remote_gradients_data*)b)->pos[0]);
  if (ret) return ret;
  return (((struct compact_remote_gradients_data*)a)->pos[1] >
          ((struct compact_remote_gradients_data*)b)->pos[1]) -
         (((struct compact_remote_gradients_data*)a)->pos[1] <
          ((struct compact_remote_gradients_data*)b)->pos[1]);
}

static void compact_remote_gradients(
  struct remote_gradients * remote_gradients, size_t num_remote_gradients) {

  size_t total_num_gradients = 0;

  for (size_t i = 0; i < num_remote_gradients; ++i)
    total_num_gradients += remote_gradients[i].num_gradients;

  struct compact_remote_gradients_data * temp_data =
    xmalloc(total_num_gradients * sizeof(*temp_data));

  for (size_t i = 0; i < num_remote_gradients; ++i) {

    int * curr_global_src_cell_ids = remote_gradients[i].global_src_cell_ids;
    size_t curr_num_gradients = remote_gradients[i].num_gradients;

    for (size_t j = 0; j < curr_num_gradients; ++j) {
      temp_data[j].global_id = curr_global_src_cell_ids[j];
      temp_data[j].pos[0] = i;
      temp_data[j].pos[1] = j;
    }
    temp_data += curr_num_gradients;
  }
  temp_data -= total_num_gradients;

  qsort(temp_data, total_num_gradients, sizeof(*temp_data),
        compare_compact_remote_gradients_data_global_id);

  // remote duplicated global ids
  size_t new_num_gradients;
  int prev_global_id;
  if (total_num_gradients > 0) {
    new_num_gradients = 1;
    prev_global_id = temp_data[0].global_id;
  } else {
    new_num_gradients = 0;
    prev_global_id = -1;
  }
  for (size_t i = 1; i < total_num_gradients; ++i) {

    if (temp_data[i].global_id != prev_global_id) {
      prev_global_id = temp_data[i].global_id;
      temp_data[new_num_gradients] = temp_data[i];
      new_num_gradients++;
    }
  }

  struct weight_vector_3d * new_gradients =
    xmalloc(new_num_gradients * sizeof(*new_gradients));
  int * new_global_src_cell_ids =
    xmalloc(new_num_gradients * sizeof(*new_global_src_cell_ids));
  size_t new_weight_vector_data_3d_buffer_size = 0;

  for (size_t i = 0; i < new_num_gradients; ++i) {

    struct compact_remote_gradients_data * curr_temp_data = temp_data + i;
    size_t * curr_pos = &(curr_temp_data->pos[0]);

    new_gradients[i] = remote_gradients[curr_pos[0]].gradients[curr_pos[1]];
    new_global_src_cell_ids[i] = curr_temp_data->global_id;
    new_weight_vector_data_3d_buffer_size += new_gradients[i].n;
  }

  struct weight_vector_data_3d * new_weight_vector_data_3d_buffer =
    xmalloc(new_weight_vector_data_3d_buffer_size *
            sizeof(*new_weight_vector_data_3d_buffer));

  for (size_t i = 0, offset = 0; i < new_num_gradients; ++i) {

    struct weight_vector_data_3d * curr_weight_vector_data_3d =
      new_weight_vector_data_3d_buffer + offset;
    size_t curr_num_weights = new_gradients[i].n;

    memcpy(curr_weight_vector_data_3d, new_gradients[i].data,
           curr_num_weights * sizeof(*new_weight_vector_data_3d_buffer));
    new_gradients[i].data = curr_weight_vector_data_3d;

    offset += curr_num_weights;
  }

  size_t new_num_remote_cells = 0;
  for (size_t i = 0; i < num_remote_gradients; ++i)
    new_num_remote_cells += remote_gradients[i].num_remote_cells;

  struct temp_remote_cell * new_remote_cells =
    xmalloc(new_num_remote_cells * sizeof(*new_remote_cells));

  for (size_t i = 0, offset = 0; i < num_remote_gradients; ++i) {
    size_t curr_num_remote_cells = remote_gradients[i].num_remote_cells;
    struct remote_cell * curr_remote_cells = remote_gradients[i].remote_cells;
    for (size_t j = 0; j < curr_num_remote_cells; ++j) {
      new_remote_cells[offset + j].global_id = curr_remote_cells[j].global_id;
      new_remote_cells[offset + j].rank      = curr_remote_cells[j].rank;
    }
    offset += curr_num_remote_cells;
  }

  qsort(new_remote_cells, new_num_remote_cells, sizeof(*new_remote_cells),
        compare_remote_cells_global_id);

  {
    size_t new_num_remote_cells_;

    if (new_num_remote_cells > 0) {
      new_num_remote_cells_ = 1;
      prev_global_id = new_remote_cells[0].global_id;
    } else {
      new_num_remote_cells_ = 0;
      prev_global_id = -1;
    }

    for (size_t i = 1; i < new_num_remote_cells; ++i) {

      if (prev_global_id != new_remote_cells[i].global_id) {

        prev_global_id = new_remote_cells[i].global_id;
        new_remote_cells[new_num_remote_cells_] = new_remote_cells[i];
        ++new_num_remote_cells_;
      }
    }
    new_num_remote_cells = new_num_remote_cells_;
  }

  // free original remote gradient data
  for (size_t i = 0; i < num_remote_gradients; ++i) {

    free(remote_gradients[i].gradients);
    free(remote_gradients[i].global_src_cell_ids);
    free(remote_gradients[i].weight_vector_data_3d_buffer);
    free(remote_gradients[i].remote_cells);
  }

  struct remote_cell * remote_cells =
    xmalloc(new_num_remote_cells * sizeof(*remote_cells));
  for (size_t i = 0; i < new_num_remote_cells; ++i) {
    remote_cells[i].global_id = new_remote_cells[i].global_id;
    remote_cells[i].rank = new_remote_cells[i].rank;
  }
  free(new_remote_cells);

  remote_gradients[0].gradients = new_gradients;
  remote_gradients[0].global_src_cell_ids = new_global_src_cell_ids;
  remote_gradients[0].num_gradients = new_num_gradients;
  remote_gradients[0].weight_vector_data_3d_buffer =
    new_weight_vector_data_3d_buffer;
  remote_gradients[0].weight_vector_data_3d_buffer_size =
    new_weight_vector_data_3d_buffer_size;
  remote_gradients[0].remote_cells = remote_cells;
  remote_gradients[0].num_remote_cells = new_num_remote_cells;

  free(temp_data);
}

// ------------------------------------------------

static void compute_remote_overlap_local_id(
  struct remote_gradients * remote_src_cell_gradients,
  struct weight_provenance * weight_provenance, size_t num_cells) {

  struct remote_overlap * remote_overlaps;
  size_t num_remote_overlaps;

  get_required_remote_overlaps(
    weight_provenance, num_cells, &remote_overlaps, &num_remote_overlaps);

  set_local_ids_in_remote_overlaps(
    remote_src_cell_gradients->global_src_cell_ids,
    remote_overlaps, num_remote_overlaps, weight_provenance);

  free(remote_overlaps);
}

static int
compare_overlaps(const void * a, const void * b) {

  struct overlap * w_a = (struct overlap *)a;
  struct overlap * w_b = (struct overlap *)b;

  return (w_a->global_cell_id > w_b->global_cell_id) -
         (w_a->global_cell_id < w_b->global_cell_id);
}

static void
do_search_interpolation_method_conserv_continue_2(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);
  struct weight_provenance * weight_provenance =
    user_data->weight_provenance;
  struct communicator * comm =
    yac_get_domain_comm(method_conserv->domain);
  unsigned comm_size = yac_get_comm_size(comm);

  // 2nd part
  //
  // once all partial areas have been received by
  // comm_recv_callback_partial_areas the partial area list shall be complete.
  // now we can check whether the source cell information is complete for the
  // target cells. The master ranks send back a list of target cells, which can
  // be processed (are accepted) by the master. Thus, these can be erased on the
  // servant side (see recv_callback_accepts)

  if ((method_conserv->order == 2) &&
      (user_data->remote_src_cell_gradients_size > 0)) {

    user_data->remote_src_cell_gradients -=
      user_data->remote_src_cell_gradients_size;

    // compact remote gradient data and sort them by global ids
    compact_remote_gradients(user_data->remote_src_cell_gradients,
                             user_data->remote_src_cell_gradients_size);

    // for all remote overlaps, compute the local id in reference to
    // remote_src_cell_gradients based on their global_id
    compute_remote_overlap_local_id(
      user_data->remote_src_cell_gradients, weight_provenance,
      (size_t)(user_data->num_cells));
  }

  user_data->num_interpolated_cells[0] = 0;

  int * send_rank_to_msg_idx = xcalloc(comm_size, sizeof(*send_rank_to_msg_idx));

  for (unsigned n = 0; n < user_data->num_cells; ++n)
    if (weight_provenance[n].is_master)
      for (int m = 0; m < weight_provenance[n].nbr_servant_ranks; ++m)
        send_rank_to_msg_idx[weight_provenance[n].servant_ranks[m]] = 1;

  // count number of sends
  unsigned num_sends = 0;
  for (unsigned i = 0; i < comm_size; ++i)
    if (send_rank_to_msg_idx[i]) ++num_sends;

  unsigned * send_rank_list = xmalloc(num_sends * sizeof(*send_rank_list));
  num_sends = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    unsigned tmp = send_rank_to_msg_idx[i];
    send_rank_to_msg_idx[i] = (int)num_sends;
    if (tmp) {
      send_rank_list[num_sends] = i;
      ++num_sends;
    }
  }

  struct send_buffer * send_msg = xmalloc(num_sends * sizeof(*send_msg));

  for (unsigned i = 0; i < num_sends; ++i) {
    send_msg[i].ibuffer          = NULL;
    send_msg[i].buffer_size      = 0;
    send_msg[i].num_target_cells = 0;
  }

  for (unsigned n = 0; n < user_data->num_cells; ++n) {

    if (weight_provenance[n].is_master) {

      double target_cell_area_diff = weight_provenance[n].target_cell_area;
      double area_tol = weight_provenance[n].target_cell_area * 1e-8;

      for ( size_t m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
        target_cell_area_diff -= weight_provenance[n].overlaps[m].area;

#ifdef YAC_DEBUG_CONSERV
      if ( weight_provenance[n].target_cell_id == -1 ) {
        printf(
          "cell %d area tol %e area diff %e num overlaps %uz target area %e \n",
          weight_provenance[n].target_cell_id, area_tol, target_cell_area_diff,
          weight_provenance[n].nbr_overlaps,
          weight_provenance[n].target_cell_area);
        for ( size_t i = 0; i < weight_provenance[n].nbr_overlaps; ++i )
          printf(" %uz : cell id %d has area %.16e \n", i,
                   weight_provenance[n].overlaps[i].global_cell_id,
                   weight_provenance[n].overlaps[i].area );
      }
#endif

      int flag_2nd_order = 1;

      // check wether all required gradients are available
      if (method_conserv->order == 2) {

        struct weight_vector_3d ** src_cell_gradients =
          user_data->src_cell_gradients;
        for (size_t m = 0; m < weight_provenance[n].nbr_overlaps; m++) {
          int subdomain_id = weight_provenance[n].overlaps[m].subdomain_id;
          int local_cell_id = weight_provenance[n].overlaps[m].local_cell_id;
          if (((subdomain_id != -1) &&
               (src_cell_gradients[subdomain_id][local_cell_id].n == 0)) ||
              ((subdomain_id == -1) &&
               (user_data->remote_src_cell_gradients->gradients[
                  local_cell_id].n == 0))) {
            flag_2nd_order = 0;
            break;
          }
        }
      }

      // if the target cell is completely covered by non-masked source cells or
      // partial coverage is allowed
      if (((fabs(target_cell_area_diff) < area_tol) ||
           (method_conserv->partial_coverage &&
            weight_provenance[n].nbr_overlaps > 0)) && flag_2nd_order) {

        // reorder overlaps in accordance with increasing source cell id

        unsigned reorder = 0;

        for ( size_t i = 1; i < weight_provenance[n].nbr_overlaps; ++i )
          if ( weight_provenance[n].overlaps[i].global_cell_id <
               weight_provenance[n].overlaps[i-1].global_cell_id ) reorder = 1;

        if ( reorder == 1 )
          qsort(weight_provenance[n].overlaps, weight_provenance[n].nbr_overlaps,
                sizeof(*weight_provenance[n].overlaps), compare_overlaps);

        double norm_factor;

        switch (method_conserv->normalisation) {

          case(CONSERV_DESTAREA):
            norm_factor = 1.0 / weight_provenance[n].target_cell_area;
            break;
          case(CONSERV_FRACAREA):
          {
            double fracarea = 0;
            for ( size_t m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
              fracarea += weight_provenance[n].overlaps[m].area;
            norm_factor = 1.0 / fracarea;
            break;
          }
          default:
            norm_factor = 0.0;
            yac_internal_abort_message(
              "ERROR: invalid normalisation option in conservative remapping\n",
              __FILE__, __LINE__);
        }

        // compute the weights
        for ( size_t m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
          weight_provenance[n].overlaps[m].weight =
            weight_provenance[n].overlaps[m].area * norm_factor;

        user_data->num_interpolated_cells[0]++;

        for (int m = 0; m < weight_provenance[n].nbr_servant_ranks; ++m) {

          int index =
            send_rank_to_msg_idx[weight_provenance[n].servant_ranks[m]];

          ENSURE_ARRAY_SIZE(
            send_msg[index].ibuffer, send_msg[index].buffer_size,
            send_msg[index].num_target_cells + 1);

          // store target cell IDs which can be erased on the servant side in send buffer

          send_msg[index].ibuffer[send_msg[index].num_target_cells++] =
            weight_provenance[n].target_cell_id;

        }

        // flag target cell IDs as being interpolated

        weight_provenance[n].cell_can_be_interpolated = 1;

      } else {

//#define YAC_DEBUG_CONSERV
#ifdef YAC_DEBUG_CONSERV

        {
          int global_id = -1; // provide global cell id for the cell that you
                              // want to investigate

          if (global_id == -1 ||
              weight_provenance[n].target_cell_id == global_id) {

            struct source_search_results const * source_target_relation =
              yac_get_search_results_for_tgt_cell(
                method_conserv->search, weight_provenance[n].target_cell_id);

            if (source_target_relation != NULL) {

              struct grid_cell tgt_cell;
              yac_init_grid_cell(&tgt_cell);
              yac_get_grid_cell(yac_get_subdomain_grid(
                *source_target_relation->tgt_domain,
                source_target_relation->tgt_subdomain_idx),
                source_target_relation->local_tgt_cell_id, &tgt_cell);

              fprintf(stderr, "Target cell %d could not be interpolated\n"
                      "target area = %.16f\n"
                      "target_cell_area_diff = %.16f\n",
                      weight_provenance[n].target_cell_id,
                      weight_provenance[n].target_cell_area,
                      target_cell_area_diff);
              print_grid_cell(stderr, tgt_cell, "target cell");
              yac_free_grid_cell(&tgt_cell);

              struct grid_cell src_cell;
              yac_init_grid_cell(&src_cell);
              for (size_t i = 0; i < weight_provenance[n].nbr_overlaps; ++i) {
                fprintf(stderr, "%uz global src cell id %d overlap area %.16f\n", i,
                        weight_provenance[n].overlaps[i].global_cell_id,
                        weight_provenance[n].overlaps[i].area);

                if (weight_provenance[n].overlaps[i].subdomain_id != -1) {

                  yac_get_grid_cell(yac_get_subdomain_grid(*method_conserv->domain,
                    weight_provenance[n].overlaps[i].subdomain_id),
                    weight_provenance[n].overlaps[i].local_cell_id, &src_cell);
                  print_grid_cell(stderr, src_cell, "source cell");
                } else {
                  fputs("source cell:\n"
                        "coordinates not available on this process\n", stderr);
                }
              }
              yac_free_grid_cell(&src_cell);
            }
          }
        }
#endif

        weight_provenance[n].cell_can_be_interpolated = 0;

        // compute the weights
        for ( size_t m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
          weight_provenance[n].overlaps[m].weight = -1.0;
      }
    }
  }

  // send target IDs that can be treated

  for (unsigned i = 0; i < num_sends; ++i) {

    // send data before sending header, this does not give a
    // conflict as messages are identified by their tag.

    if ( send_msg[i].num_target_cells > 0 ) {
      yac_comm_bsend(
        send_msg[i].ibuffer, send_msg[i].num_target_cells, COMM_INT,
        send_rank_list[i], method_conserv->tags.accepts_exchange, comm);

      free(send_msg[i].ibuffer);

    }

    // send header
    yac_comm_bsend(
      &send_msg[i].num_target_cells, 1, COMM_UINT, send_rank_list[i],
      method_conserv->tags.header2_exchange, comm);
  }

  free(send_rank_to_msg_idx);
  free(send_rank_list);
  free(send_msg);
}

static inline void orthogonalise_weight_vector(
  double M[3][3], struct weight_vector_3d * G_i,
  struct weight_vector_data_3d * buffer) {

  // This routine computes: (I_3 - C_i * C_i^-1) * G_i
  //
  // O(g_i) = g_i - C_i * (C_i^-1 * g_i)
  // where: C_i is the centeroid of a source cell
  //        g_i is the gradient of the source field in C_i
  //        O(g_i) is the projection of g_i into the plane perpendicular to C_i
  // g_i = G_i * f
  // where: G_i is the weight matrix to compute g_i
  //        f is the source field vector
  // => O(g_i) = (I_3 - C_i * C_i^-1) * G_i * f
  // where: I_3 is the identity matrix of size 3 x 3
  //
  // M = I_3 - C_i * C_i^-1

  struct weight_vector_data_3d * G_i_data = G_i->data;

  size_t N = G_i->n;
  for (size_t i = 0; i < N; ++i) {
    buffer[i].id = G_i_data[i].id;
    for (size_t j = 0; j < 3; ++j) buffer[i].weight[j] = 0.0;
  }

  for (size_t n = 0; n < N; ++n)
    for (size_t i = 0; i < 3; ++i)
      for (size_t j = 0; j < 3; ++j)
        buffer[n].weight[i] += G_i_data[n].weight[j] * M[i][j];

  memcpy(G_i->data, buffer, N * sizeof(*buffer));
}

static size_t get_edge_neigbours(
  struct grid * grid, unsigned cell_index, struct dep_list * neigh_list,
  unsigned * edge_neighs) {

  size_t num_edges = yac_get_num_cell_edges(grid, cell_index);
  unsigned * cell_corner_indices =
    xmalloc(num_edges * sizeof(*cell_corner_indices));
  {
    unsigned const * temp_cell_corner_indices =
      yac_get_cell_corner_indices(grid, cell_index);
    for (size_t i = 0; i < num_edges; ++i)
      cell_corner_indices[i] = temp_cell_corner_indices[i];
  }

  size_t num_neighs =
    yac_get_num_dependencies_of_element(neigh_list, (size_t)cell_index);
  unsigned const * neighs =
    yac_get_dependencies_of_element(neigh_list, (size_t)cell_index);

  size_t num_found_edges = 0;

  unsigned curr_edge_corner_indices[2] =
    {(unsigned)-1, cell_corner_indices[num_edges-1]};
  for (size_t i = 0; i < num_edges; ++i) {
    curr_edge_corner_indices[0] = curr_edge_corner_indices[1];
    curr_edge_corner_indices[1] = cell_corner_indices[i];
    for (size_t j = 0; j < num_neighs; ++j) {
      unsigned neigh = neighs[j];
      unsigned const * neigh_cell_corner_indices =
        yac_get_cell_corner_indices(grid, neigh);
      size_t neigh_num_edges = yac_get_num_cell_edges(grid, neigh);
      unsigned neigh_edge_corner_indices[2] =
        {(unsigned)-1, neigh_cell_corner_indices[neigh_num_edges-1]};
      for (size_t k = 0; k < neigh_num_edges; ++k) {
        neigh_edge_corner_indices[0] = neigh_edge_corner_indices[1];
        neigh_edge_corner_indices[1] = neigh_cell_corner_indices[k];
        if (((curr_edge_corner_indices[0] == neigh_edge_corner_indices[0]) &&
             (curr_edge_corner_indices[1] == neigh_edge_corner_indices[1])) ||
            ((curr_edge_corner_indices[0] == neigh_edge_corner_indices[1]) &&
             (curr_edge_corner_indices[1] == neigh_edge_corner_indices[0]))) {

          edge_neighs[num_found_edges] = neigh;
          j = num_neighs;
          ++num_found_edges;
          break;
        }
      }
    }
  }
  free(cell_corner_indices);
  edge_neighs[num_found_edges] = edge_neighs[0];
  return num_found_edges;
}

static void inline compute_cell_barycenter(
  struct grid_cell cell, double * barycenter) {

  if (cell.num_corners == 0) return;
  barycenter[0] = cell.coordinates_xyz[0][0];
  barycenter[1] = cell.coordinates_xyz[0][1];
  barycenter[2] = cell.coordinates_xyz[0][2];
  for (size_t i = 1; i < cell.num_corners; ++i) {
    barycenter[0] += cell.coordinates_xyz[i][0];
    barycenter[1] += cell.coordinates_xyz[i][1];
    barycenter[2] += cell.coordinates_xyz[i][2];
  }
  normalise_vector(barycenter);
}

static unsigned * get_global_search_to_domain_ranks(
  struct interpolation_method_conserv * method_conserv) {

  struct communicator * domain_comm =
    yac_get_domain_comm(method_conserv->domain);
  struct communicator * global_search_comm =
    yac_get_communicator(method_conserv->search);

  unsigned domain_comm_size = yac_get_comm_size(domain_comm);
  unsigned global_search_comm_size = yac_get_comm_size(global_search_comm);

  unsigned * domain_to_global_search_ranks =
    xmalloc(domain_comm_size * sizeof(*domain_to_global_search_ranks));
  unsigned * global_search_to_domain_ranks =
    xmalloc(global_search_comm_size * sizeof(*global_search_to_domain_ranks));

  yac_comm_translate_ranks(
    global_search_comm, domain_comm, domain_to_global_search_ranks);

  for (unsigned i = 0; i < domain_comm_size; ++i)
    global_search_to_domain_ranks[domain_to_global_search_ranks[i]] = i;

  free(domain_to_global_search_ranks);

  return global_search_to_domain_ranks;
}

static void init_weight_provenance_data(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  struct domain * domain = method_conserv->domain;
  unsigned num_subdomains = yac_get_num_subdomains(domain);

  struct communicator * comm =
    yac_get_domain_comm(method_conserv->domain);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned * global_search_to_domain_ranks =
    get_global_search_to_domain_ranks(method_conserv);

  struct grid_cell * SourceCell = NULL;
  struct grid_cell  TargetCell;
  size_t nmaxSourceCells = 0;
  size_t initialisedSourceCells = 0;
  double * areas = NULL;
  size_t nmaxAreas = 0;
  double (*barycenters)[3] = NULL;
  size_t barycenters_array_size = 0;

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;
  double (*coordinates_xyz)[3] = user_data->coordinates_xyz_buffer;
  int * global_ids = user_data->global_ids_buffer;

  double (**src_cell_centroids)[3];
  double (**src_cell_barycenters)[3];
  double ** total_overlap_areas;
  struct weight_vector_3d ** src_cell_gradients;

  // set up basic data for 2nd order
  if (method_conserv->order == 2) {

    src_cell_centroids = xmalloc(num_subdomains * sizeof(*src_cell_centroids));
    src_cell_barycenters = xmalloc(num_subdomains * sizeof(*src_cell_barycenters));
    total_overlap_areas = xmalloc(num_subdomains * sizeof(*total_overlap_areas));
    src_cell_gradients =
      xcalloc((num_subdomains + 1), sizeof(*src_cell_gradients));
    src_cell_gradients++; // at positions -1 we will store the source cell
                          // gradients from remote processes

    for (unsigned i = 0; i < num_subdomains; ++i) {

      struct grid * subdomain_grid = yac_get_subdomain_grid(*domain, i);
      unsigned num_grid_cells = yac_get_num_grid_cells(subdomain_grid);

      double (*curr_src_cell_centroids)[3] =
        xmalloc(num_grid_cells * sizeof(*curr_src_cell_centroids));
      double (*curr_src_cell_barycenters)[3] =
        xmalloc(num_grid_cells * sizeof(*curr_src_cell_barycenters));
      double * curr_total_overlap_areas =
        xmalloc(num_grid_cells * sizeof(*curr_total_overlap_areas));
      struct weight_vector_3d * curr_src_cell_gradients =
        xmalloc(num_grid_cells * sizeof(*curr_src_cell_gradients));
      src_cell_centroids[i] = curr_src_cell_centroids;
      src_cell_barycenters[i] = curr_src_cell_barycenters;
      total_overlap_areas[i] = curr_total_overlap_areas;
      src_cell_gradients[i] = curr_src_cell_gradients;

      for (unsigned j = 0; j < num_grid_cells; ++j) {
        curr_src_cell_centroids[j][0] = 0.0;
        curr_src_cell_centroids[j][1] = 0.0;
        curr_src_cell_centroids[j][2] = 0.0;
        curr_total_overlap_areas[j] = 0.0;
        curr_src_cell_gradients[j].data = NULL;
        curr_src_cell_gradients[j].n = 0;
      }
    }

  } else {
    src_cell_centroids = NULL;
    src_cell_barycenters = NULL;
    src_cell_gradients = NULL;
  }

  user_data->src_cell_gradients = src_cell_gradients;
  user_data->weight_vector_data_buffer = NULL;

  yac_init_grid_cell ( &TargetCell );

  // Loop over all target cells

  for (unsigned n = 0; n < num_cells; ++n ) {

    if ((n % 1024) == 0) yac_comm_check(comm);

    weight_provenance[n].target_cell_id = global_ids[n];
    if (coordinates_xyz != NULL) {
      weight_provenance[n].coordinate_xyz[0] = coordinates_xyz[n][0];
      weight_provenance[n].coordinate_xyz[1] = coordinates_xyz[n][1];
      weight_provenance[n].coordinate_xyz[2] = coordinates_xyz[n][2];
    }
    weight_provenance[n].cell_can_be_interpolated = 0;

    struct source_search_results const * source_target_relation =
      yac_get_search_results_for_tgt_cell(method_conserv->search, global_ids[n]);

    // if there are no search results for a requested cell, we do not know how
    // the other source processes handle this cell -> we cannot deal with it
    if (source_target_relation == NULL)
      yac_internal_abort_message(
        "ERROR: there are no search results for a requested cell\n",
        __FILE__, __LINE__);

    // get the target cell
    yac_get_grid_cell(
      yac_get_subdomain_grid(
        *source_target_relation->tgt_domain,
        source_target_relation->tgt_subdomain_idx ),
      source_target_relation->local_tgt_cell_id, &TargetCell);

    // initialise the list of source cells

    size_t nSourceCells = 0;

    for (unsigned i = 0; i < num_subdomains; ++i )
      nSourceCells +=
        (size_t)
          (source_target_relation->num_overlapping_src_cells_per_subdomain[i]);

    ENSURE_ARRAY_SIZE(SourceCell, nmaxSourceCells, nSourceCells);
    for (; initialisedSourceCells < nSourceCells; ++initialisedSourceCells)
      yac_init_grid_cell(SourceCell + initialisedSourceCells);

    // allocate memory for weights and associated provenance data
    // probably we only have to store provenance info if we do not
    // stay local. If we are reaching the halo we allocated more
    // memory than needed. But doing this already here avoid some
    // additional costly checking in loops further down.

    struct overlap * curr_overlaps =
      xmalloc(nSourceCells * sizeof(*(weight_provenance[n].overlaps)));
    weight_provenance[n].overlaps = curr_overlaps;

    // loop over subdomains and collect geographical information plus provenance
    // data for each cell

    nSourceCells = 0;
    for (unsigned src_subdomain_idx = 0; src_subdomain_idx < num_subdomains;
         ++src_subdomain_idx) {

      struct grid * grid = yac_get_subdomain_grid(*domain, src_subdomain_idx);
      int const * global_src_ids =
        yac_get_global_ids(domain, src_subdomain_idx, CELL);
      const unsigned num_overlapping_src_cells =
        source_target_relation->
          num_overlapping_src_cells_per_subdomain[src_subdomain_idx];
      const unsigned * overlapping_src_cells =
        source_target_relation->overlapping_src_cells[src_subdomain_idx];
      struct mask *** masks = method_conserv->masks;
      unsigned points_idx = method_conserv->points_idx;

      int use_mask = method_conserv->use_source_mask &&
                     (masks != NULL) && (masks[src_subdomain_idx] != NULL) &&
                     (masks[src_subdomain_idx][points_idx] != NULL);

      for (unsigned i = 0; i < num_overlapping_src_cells; ++i) {

        unsigned curr_local_id = overlapping_src_cells[i];

        // Perform mask handling for source cells already here and
        // do not consider halo cells for the weights to avoid duplication
        if (yac_is_halo_value(
              domain, src_subdomain_idx, CELL, curr_local_id, NULL) ||
            (use_mask &&
             masks[src_subdomain_idx][points_idx]-> mask[curr_local_id]))
          continue;

        // add cell for weight calculation
        yac_get_grid_cell(
          grid, overlapping_src_cells[i], SourceCell + nSourceCells);
        curr_overlaps[nSourceCells].local_cell_id  = curr_local_id;
        curr_overlaps[nSourceCells].subdomain_id   = src_subdomain_idx;
        curr_overlaps[nSourceCells].global_cell_id = global_src_ids[curr_local_id];
        curr_overlaps[nSourceCells].rank           = comm_rank;
        ++nSourceCells;
      }
    }

    // set up prerequisites for communication among source processes
    //
    // check whether we are alone or if we have to share information
    // with other sources for this particular target cell.
    // in case we have to share information we have to determine
    // master and servants for each target cells

    weight_provenance[n].master_rank       = (int)comm_rank;
    weight_provenance[n].is_master         = 1;
    weight_provenance[n].nbr_servant_ranks = 0;
    weight_provenance[n].servant_ranks     = NULL;

    if ( source_target_relation->num_other_src_ranks > 0 ) {

      unsigned num_other_src_ranks =
        source_target_relation->num_other_src_ranks;
      unsigned * other_src_ranks =
        source_target_relation->other_src_ranks;
      unsigned * servant_ranks =
        xmalloc(num_other_src_ranks * sizeof(*servant_ranks));
      unsigned master_rank = comm_rank;

      // the preliminary master is the one with the lowest rank, we later switch
      // to the one with the most source cells overlapping this target cell
      for (unsigned i = 0; i < num_other_src_ranks; ++i) {
        unsigned curr_rank = global_search_to_domain_ranks[other_src_ranks[i]];
        servant_ranks[i] = curr_rank;
        if (curr_rank < master_rank) master_rank = curr_rank;
      }

      weight_provenance[n].nbr_servant_ranks = num_other_src_ranks;
      weight_provenance[n].servant_ranks = servant_ranks;
      weight_provenance[n].master_rank = (int)master_rank;
      weight_provenance[n].is_master = master_rank == comm_rank;
    }

    // determine the partial areas. the sequence of returned areas corresponds
    // to the sequence of ingoing SourceCells (and the sequence of source cell
    // ids).

    ENSURE_ARRAY_SIZE ( areas, nmaxAreas, nSourceCells );

    if (method_conserv->order == 2) {
      ENSURE_ARRAY_SIZE ( barycenters, barycenters_array_size, nSourceCells );
      yac_compute_concave_overlap_info ( nSourceCells, SourceCell, TargetCell,
                                         coordinates_xyz[n], areas,
                                         barycenters);
    } else
      yac_compute_concave_overlap_areas ( nSourceCells, SourceCell, TargetCell,
                                          coordinates_xyz[n], areas);

    double target_cell_area = yac_huiliers_area(TargetCell);

    //! \todo how to we handle the case when the area of the target cell is
    //!       around the area tolerance of the computation
    // if (target_cell_area <= area_tol * 100)
      // abort_message("ERROR: target cell is too small\n", __FILE__, __LINE__);

    // scale area tolerance to target cell area
    weight_provenance[n].target_cell_area = target_cell_area;

    // delete cells that have no overlap with the target cell
    size_t new_nSourceCells = 0;

    for (size_t i = 0; i < nSourceCells; ++i) {
      if (areas[i] > 0) {

        struct overlap * curr_overlap =
          weight_provenance[n].overlaps + new_nSourceCells;

        if (new_nSourceCells != i)
          *curr_overlap = weight_provenance[n].overlaps[i];

        curr_overlap->area = areas[i];
        curr_overlap->weight = 0.0;
        if (barycenters != NULL) {
          curr_overlap->barycenter[0] = barycenters[i][0];
          curr_overlap->barycenter[1] = barycenters[i][1];
          curr_overlap->barycenter[2] = barycenters[i][2];
        } else {
          curr_overlap->barycenter[0] = 0.0;
          curr_overlap->barycenter[1] = 0.0;
          curr_overlap->barycenter[2] = 0.0;
        }
        new_nSourceCells++;
      }
    }

    if (new_nSourceCells != nSourceCells)
      weight_provenance[n].overlaps =
        xrealloc(weight_provenance[n].overlaps,
                 new_nSourceCells * sizeof(*(weight_provenance[n].overlaps)));

    if (method_conserv->order == 2) {

      for (size_t i = 0; i < new_nSourceCells; ++i) {

        struct overlap * curr_overlap = weight_provenance[n].overlaps + i;

        unsigned src_cell_subdomain_id = curr_overlap->subdomain_id;
        unsigned local_src_cell_id = curr_overlap->local_cell_id;
        double * curr_src_cell_centroid =
          src_cell_centroids[src_cell_subdomain_id][local_src_cell_id];
        double * curr_barycenter = curr_overlap->barycenter;
        double curr_area = curr_overlap->area;
        total_overlap_areas[src_cell_subdomain_id][local_src_cell_id] +=
          curr_area;
        curr_src_cell_centroid[0] += curr_area * curr_barycenter[0];
        curr_src_cell_centroid[1] += curr_area * curr_barycenter[1];
        curr_src_cell_centroid[2] += curr_area * curr_barycenter[2];
      }
    }

    // store current size of weight_provenance[n].overlaps

    weight_provenance[n].nbr_overlaps = new_nSourceCells;
  }

  // compute centroids of source cells
  // C_i = N(S_(U_k) (A_k*C_k))
  // where: N(C) = C*(C*C)^-0.5 // normalisation
  //        U_k all supermesh cells overlaping with the respective source cell
  //        A_k area of supermesh cell
  //        C_k barycenter of supermesh cell (normalisation of the sum of all
  //                                          vertices of the cell)
  if (method_conserv->order == 2) {
    struct grid_cell * curr_src_cell = &TargetCell;
    for (unsigned i = 0; i < num_subdomains; ++i) {

      double * curr_total_overlap_areas = total_overlap_areas[i];
      double (*curr_src_cell_centroids)[3] = src_cell_centroids[i];
      double (*curr_src_cell_barycenters)[3] = src_cell_barycenters[i];

      struct grid * grid = yac_get_subdomain_grid(*domain, i);
      unsigned num_grid_cells = yac_get_num_grid_cells(grid);

      for (unsigned j = 0; j < num_grid_cells; ++j) {

        yac_get_grid_cell(grid, j, curr_src_cell);
        double src_cell_area = yac_huiliers_area(*curr_src_cell);
        double area_tol = src_cell_area * 1e-8;

        compute_cell_barycenter(*curr_src_cell, curr_src_cell_barycenters[j]);

        double cell_area_diff =
          fabs(src_cell_area - curr_total_overlap_areas[j]);

        // if the whole source cell is covered by target cells
        if (area_tol > cell_area_diff) {
          normalise_vector(curr_src_cell_centroids[j]);
        } else {
          curr_src_cell_centroids[j][0] = 0.0;
          curr_src_cell_centroids[j][1] = 0.0;
          curr_src_cell_centroids[j][2] = 0.0;
        }
      }
      free(curr_total_overlap_areas);
    }
    free(total_overlap_areas);
  }

  // compute gradient in the centroid for each source cell
  // g_i = O_i(g'_i)
  // where: O_i(g) = g - C_i*(C_i*g) // makes g orthogonal to C_i
  // g'_i = (A_(C_k))^-1 * S_ijk(((f_j + f_k) * 0.5 - f_i) *
  //                         (C_j x C_k) * |C_j x C_k|^-1 *
  //                         asin(|C_j x C_k|))
  // where: g'_i is the estimated centroid gradient
  //        A_(C_k) is the area of the cell generated by connecting the
  //                centroids of all neighbour cells
  //        S_ijk is the sum of all edges of the previously described cell in
  //              counterclockwise order
  //        f_i, f_j, and f_k are the mean values of the field over the area of
  //                          the respective source cell area
  // asin(|C_j x C_k|) / |C_j x C_k| ~ 1.0
  // => g'_i = S_ijk(((f_j + f_k) * 0.5 - f_i) * (C_j x C_k)) / A_(C_k)
  //
  // g_i = G_i * f
  // G_i = S_ijk(((I_j + I_k) * 0.5 - I_i) * (C_j x C_k)) / A_(C_k)
  // where: G_i is the gradient weight matrix
  //        I_x is a vector of the same size as f, which is all zero except at
  //            position x, where it is one
  //        f is the source field vector
  if (method_conserv->order == 2) {

    size_t max_num_total_gradient_weights = 0;
    size_t max_num_edge_neighs = 0;

    for (unsigned i = 0; i < num_subdomains; ++i) {

      struct grid * grid = yac_get_subdomain_grid(*domain, i);
      unsigned num_grid_cells = yac_get_num_grid_cells(grid);

      double (*curr_src_cell_centroids)[3] = src_cell_centroids[i];

      // for all source grid cells, for which we computed a centroid, we will
      // also have to compute the gradient
      for (size_t j = 0; j < num_grid_cells; ++j) {

        // if we have not computed a centroid for the current source cell
        if ((curr_src_cell_centroids[j][0] < 0.5) &&
            (curr_src_cell_centroids[j][1] < 0.5) &&
            (curr_src_cell_centroids[j][2] < 0.5)) continue;

        size_t num_cell_edge_neighs = (size_t)yac_get_num_cell_corners(grid, j);

        if (num_cell_edge_neighs > max_num_edge_neighs)
          max_num_edge_neighs = num_cell_edge_neighs;
        max_num_total_gradient_weights += num_cell_edge_neighs + 1;
      }
    }

    unsigned * edge_neighs =
      xmalloc((max_num_edge_neighs + 1) * sizeof(*edge_neighs));

    struct weight_vector_data_3d * weight_vector_data_buffer =
      xmalloc(max_num_total_gradient_weights *
              sizeof(*weight_vector_data_buffer));
    size_t weight_vector_data_buffer_offset = 0;
    struct weight_vector_data_3d * orth_buffer =
      xmalloc((max_num_edge_neighs + 1) * sizeof(*orth_buffer));

    user_data->weight_vector_data_buffer = weight_vector_data_buffer;

    for (size_t i = 0; i < max_num_total_gradient_weights; ++i)
      for (size_t j = 0; j < 3; ++j)
        weight_vector_data_buffer[i].weight[j] = 0.0;

    for (unsigned i = 0; i < num_subdomains; ++i) {

      struct grid * grid = yac_get_subdomain_grid(*domain, i);
      unsigned num_grid_cells = yac_get_num_grid_cells(grid);
      struct dep_list * neigh_list = yac_get_cell_neigh_dep_list(grid);

      struct mask *** masks = method_conserv->masks;
      unsigned points_idx = method_conserv->points_idx;
      int use_mask = method_conserv->use_source_mask &&
                     (masks != NULL) && (masks[i] != NULL) &&
                     (masks[i][points_idx] != NULL);
      unsigned * mask;
      if (use_mask) mask = masks[i][points_idx]-> mask;

      double (*curr_src_cell_centroids)[3] = src_cell_centroids[i];
      double (*curr_src_cell_barycenters)[3] = src_cell_barycenters[i];
      struct weight_vector_3d * curr_src_cell_gradients = src_cell_gradients[i];

      // for all source grid cells, for which we computed a centroid, we will
      // also have to compute the gradient
      for (size_t j = 0; j < num_grid_cells; ++j) {

        size_t num_cell_edge_neighs = (size_t)yac_get_num_cell_corners(grid, j);

        // if we have not computed a centroid for the current source cell
        if ((curr_src_cell_centroids[j][0] < 0.5) &&
            (curr_src_cell_centroids[j][1] < 0.5) &&
            (curr_src_cell_centroids[j][2] < 0.5)) continue;

        // get all source cells that have a shared edge with the current cell.
        // if we do not have enough neighbours, we cannot compute the gradient
        // (edge_neighs[num_cell_edge_neighs] = edge_neighs[0])
        if (num_cell_edge_neighs !=
            get_edge_neigbours(grid, j, neigh_list, edge_neighs))
          continue;

        // check whether any of the required neighbours is masked
        if (use_mask) {
          size_t k;
          for (k = 0; k < num_cell_edge_neighs; ++k)
            if (mask[edge_neighs[k]]) break;
          if (k != num_cell_edge_neighs) continue;
        }

        struct weight_vector_3d * G_i = curr_src_cell_gradients + j;

        G_i->data = weight_vector_data_buffer +
                    weight_vector_data_buffer_offset;
        weight_vector_data_buffer_offset += num_cell_edge_neighs + 1;
        G_i->n = num_cell_edge_neighs + 1;
        G_i->data[0].id.local = j;
        for (size_t k = 0; k < num_cell_edge_neighs; ++k)
          G_i->data[k+1].id.local = edge_neighs[k];

        // area of the cell whose corners are the cell centroids of the edge
        // neighbour cells
        double A_C_k = 0.0;

        struct grid_cell centroid_triangle = {
          .coordinates_x = NULL, .coordinates_y = NULL,
          .coordinates_xyz = (double[3][3]){{0}},
          .edge_type =
            (enum yac_edge_type[]) {GREAT_CIRCLE,GREAT_CIRCLE,GREAT_CIRCLE},
          .num_corners = 3, .array_size = 0};
        centroid_triangle.coordinates_xyz[0][0] = curr_src_cell_centroids[j][0];
        centroid_triangle.coordinates_xyz[0][1] = curr_src_cell_centroids[j][1];
        centroid_triangle.coordinates_xyz[0][2] = curr_src_cell_centroids[j][2];

        // the following matrix is used by the routine
        // orthogonalise_weight_vector
        // M = I_3 - C_i * C_i^-1
        double M[3][3];
        for (size_t k = 0; k < 3; ++k)
          for (size_t l = 0; l < 3; ++l)
            M[k][l] = - centroid_triangle.coordinates_xyz[0][k] *
                        centroid_triangle.coordinates_xyz[0][l];
        for (size_t k = 0; k < 3; ++k)
          M[k][k] += 1.0;

        // We split the cell that is comprised of the barycenters of the edge
        // neigbours into triangles, which has the centroid of the current cell
        // as one corner. The sum of the areas of these triangles is A_C_K.
        for (size_t k = 0; k < num_cell_edge_neighs; ++k) {

          double * neigh_cell_barycenters[2] = {
            curr_src_cell_barycenters[edge_neighs[k]],
            curr_src_cell_barycenters[edge_neighs[k+1]]};

          centroid_triangle.coordinates_xyz[1][0] = neigh_cell_barycenters[0][0];
          centroid_triangle.coordinates_xyz[1][1] = neigh_cell_barycenters[0][1];
          centroid_triangle.coordinates_xyz[1][2] = neigh_cell_barycenters[0][2];
          centroid_triangle.coordinates_xyz[2][0] = neigh_cell_barycenters[1][0];
          centroid_triangle.coordinates_xyz[2][1] = neigh_cell_barycenters[1][1];
          centroid_triangle.coordinates_xyz[2][2] = neigh_cell_barycenters[1][2];

          A_C_k += yac_huiliers_area(centroid_triangle);

          // C_j x C_k
          double C_j_x_C_k[3];
          crossproduct_ld(
            neigh_cell_barycenters[0], neigh_cell_barycenters[1], C_j_x_C_k);
          // this factor ensures that C_j_x_C_k points outwards of the centroid
          // cell
          double factor =
            -copysign(1.0, C_j_x_C_k[0] * centroid_triangle.coordinates_xyz[0][0] +
                           C_j_x_C_k[1] * centroid_triangle.coordinates_xyz[0][1] +
                           C_j_x_C_k[2] * centroid_triangle.coordinates_xyz[0][2]);

          for (size_t l = 0; l < 3; ++l) C_j_x_C_k[l] *= factor;

          // -I_i * (C_j x C_k)
          G_i->data[0].weight[0] -= C_j_x_C_k[0];
          G_i->data[0].weight[1] -= C_j_x_C_k[1];
          G_i->data[0].weight[2] -= C_j_x_C_k[2];

          for (size_t l = 0; l < 3; ++l) C_j_x_C_k[l] *= 0.5;

          // 0.5 * I_j * (C_j x C_k)
          G_i->data[k+1].weight[0] += C_j_x_C_k[0];
          G_i->data[k+1].weight[1] += C_j_x_C_k[1];
          G_i->data[k+1].weight[2] += C_j_x_C_k[2];

          // 0.5 * I_k * (C_j x C_k)
          size_t k_ = (k + 1) % num_cell_edge_neighs + 1;
          G_i->data[k_].weight[0] += C_j_x_C_k[0];
          G_i->data[k_].weight[1] += C_j_x_C_k[1];
          G_i->data[k_].weight[2] += C_j_x_C_k[2];
        }

        // ((I_j + I_k) * 0.5 - I_i) * (C_j x C_k) / A_(C_k)
        double inv_A_C_K = 1.0 / A_C_k;
        for (size_t k = 0; k <=num_cell_edge_neighs; ++k)
          for (size_t l = 0; l < 3; ++l)
            G_i->data[k].weight[l] *= inv_A_C_K;

        orthogonalise_weight_vector(M, G_i, orth_buffer);
      }
      free(curr_src_cell_centroids);
      free(curr_src_cell_barycenters);
    }
    free(src_cell_centroids);
    free(src_cell_barycenters);
    free(edge_neighs);
    free(orth_buffer);
  }

  for (unsigned i = 0; i < initialisedSourceCells; ++i )
    yac_free_grid_cell ( &SourceCell[i] );
  yac_free_grid_cell ( &TargetCell );

  free(SourceCell);
  free(global_search_to_domain_ranks);
  free(barycenters);
  free(areas);
}

static int compare_weight_vector_3d_subdomain(void const * a, void const * b) {

  return (((struct weight_vector_3d_subdomain*)a)->weight_vector >
          ((struct weight_vector_3d_subdomain*)b)->weight_vector) -
         (((struct weight_vector_3d_subdomain*)a)->weight_vector <
          ((struct weight_vector_3d_subdomain*)b)->weight_vector);
}

static void remove_duplicated_weight_vector_3d_subdomain(
  struct weight_vector_3d_subdomain * src_cell_gradients, size_t * n) {

  size_t n_ = *n, new_n = 1;

  if (n_ == 0) return;

  qsort(src_cell_gradients, n_, sizeof(*src_cell_gradients),
        compare_weight_vector_3d_subdomain);

  struct weight_vector_3d * prev_weight_vector =
    src_cell_gradients[0].weight_vector;

  for (size_t i = 1; i < n_; ++i) {

    if (src_cell_gradients[i].weight_vector != prev_weight_vector) {

      prev_weight_vector = src_cell_gradients[i].weight_vector;
      src_cell_gradients[new_n] = src_cell_gradients[i];
      new_n++;
    }
  }

  *n = new_n;
}

// Determines which data is required by other processes und sends this data.
// In addition, it also sets up the respective receives.
static void
do_search_interpolation_method_conserv_continue(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);
  struct domain * domain = method_conserv->domain;

  struct communicator * comm =
    yac_get_domain_comm(method_conserv->domain);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  // from here on we work with the weight_provenance data, which does not
  // contain any halo cells

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;

  struct weight_vector_3d_subdomain ** src_cell_gradients_per_send = NULL;
  size_t * num_src_cell_gradients_per_send = NULL;
  size_t * num_src_cell_gradient_weights_per_send = NULL;
  struct weight_vector_3d_subdomain * src_cell_gradient_buffer = NULL;

  int * send_rank_to_msg_idx = xcalloc(comm_size, sizeof(*send_rank_to_msg_idx));
  int * recv_rank_flag_list = xcalloc(comm_size, sizeof(*recv_rank_flag_list));

  // determine the send and receive operations
  for (unsigned n = 0; n < num_cells; ++n ) {
    // if multiple source processes have the current target cell
    if (weight_provenance[n].nbr_servant_ranks != 0) {
      if (weight_provenance[n].is_master) {
        for (int i = 0; i < weight_provenance[n].nbr_servant_ranks; ++i)
          recv_rank_flag_list[weight_provenance[n].servant_ranks[i]] = 1;
      } else {
        send_rank_to_msg_idx[weight_provenance[n].master_rank] = 1;
      }
    }
  }

  unsigned num_receives = 0;
  unsigned num_sends = 0;
  // count number of sends and receives
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_rank_to_msg_idx[i]) ++num_sends;
    if (recv_rank_flag_list[i]) ++num_receives;
  }
  free(recv_rank_flag_list);

  unsigned * send_rank_list = xmalloc(num_sends * sizeof(*send_rank_list));
  num_sends = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    unsigned tmp = send_rank_to_msg_idx[i];
    send_rank_to_msg_idx[i] = num_sends;
    if (tmp) {
      send_rank_list[num_sends] = i;
      ++num_sends;
    }
  }

  // store data that is to be passed to do_search_final_conserv (via
  // recv_callback_partial_areas)
  user_data->num_partial_areas_messages_to_receive = num_receives;
  user_data->num_accepts_messages_to_receive = num_sends;
  if (method_conserv->order == 2) {
    user_data->remote_src_cell_gradients =
      xmalloc(num_receives * sizeof(*(user_data->remote_src_cell_gradients)));
    user_data->remote_src_cell_gradients_size = (size_t)num_receives;
  } else {
    user_data->remote_src_cell_gradients = NULL;
    user_data->remote_src_cell_gradients_size = 0;
  }

  // if we do not require any weights from other source processes, we have to
  // call do_search_interpolation_method_conserv_continue_2 here in order for it
  // to not interfere with a call to do_search_final_conserv from
  // recv_callback_accepts
  if (num_receives == 0)
    do_search_interpolation_method_conserv_continue_2(user_data);

  // fill send buffer

  struct send_buffer * send_msg = xmalloc(num_sends * sizeof(*send_msg));
  for (unsigned i = 0; i < num_sends; ++i ) {
    send_msg[i].num_target_cells   = 0;
    send_msg[i].buffer_size        = 0;
    send_msg[i].buffer_array_size  = 0;
    send_msg[i].position           = 0;
    send_msg[i].cbuffer            = NULL;
  }

  if ( num_sends > 0 ) {

    // determine buffer size (for order 1 data)

    for (unsigned n = 0; n < num_cells; ++n ) {

      if (!weight_provenance[n].is_master) {

        struct send_buffer * curr_send_msg =
          send_msg + send_rank_to_msg_idx[weight_provenance[n].master_rank];

        curr_send_msg->num_target_cells++;

        int nbr_overlaps = (int)(weight_provenance[n].nbr_overlaps);

        curr_send_msg->buffer_size +=
          yac_comm_pack_size(2, COMM_INT, comm);
        curr_send_msg->buffer_size +=
          yac_comm_pack_size(nbr_overlaps, COMM_INT, comm);
        curr_send_msg->buffer_size +=
          yac_comm_pack_size(nbr_overlaps, COMM_DBLE, comm);
      }
    }

    // determine buffer size (for order 2 data) and gather source cell gradients
    // required for each target cell

    if (method_conserv->order == 2) {

      num_src_cell_gradients_per_send =
        xcalloc(num_sends, sizeof(*num_src_cell_gradients_per_send));
      num_src_cell_gradient_weights_per_send =
        xcalloc(num_sends, sizeof(*num_src_cell_gradient_weights_per_send));

      // compute maximum number of gradients and gradient weights per send
      // operation
      size_t total_num_gradients_per_tgt = 0;
      for (unsigned n = 0; n < num_cells; ++n ) {

        if (!weight_provenance[n].is_master) {

          size_t nbr_overlaps = weight_provenance[n].nbr_overlaps;
          num_src_cell_gradients_per_send[
            send_rank_to_msg_idx[weight_provenance[n].master_rank]] +=
              nbr_overlaps;
          total_num_gradients_per_tgt += nbr_overlaps;
        }
      }

      src_cell_gradients_per_send =
        xmalloc(num_sends * sizeof(*src_cell_gradients_per_send));
      src_cell_gradient_buffer =
        xmalloc(total_num_gradients_per_tgt * sizeof(*src_cell_gradient_buffer));

      size_t offset = 0;
      for (unsigned i = 0; i < num_sends; ++i) {
        src_cell_gradients_per_send[i] = src_cell_gradient_buffer + offset;
        offset += num_src_cell_gradients_per_send[i];
        num_src_cell_gradients_per_send[i] = 0;
      }

      struct weight_vector_3d ** src_cell_gradients =
        user_data->src_cell_gradients;

      // get all gradients required for each send
      for (unsigned n = 0; n < num_cells; ++n ) {

        if (!weight_provenance[n].is_master) {

          int index = send_rank_to_msg_idx[weight_provenance[n].master_rank];

          size_t nbr_overlaps = weight_provenance[n].nbr_overlaps;
          struct overlap * overlaps = weight_provenance[n].overlaps;
          struct weight_vector_3d_subdomain * curr_src_cell_gradients =
            src_cell_gradients_per_send[index] +
              num_src_cell_gradients_per_send[index];
          num_src_cell_gradients_per_send[index] += nbr_overlaps;

          send_msg[index].buffer_size +=
            yac_comm_pack_size(3 * nbr_overlaps, COMM_DBLE, comm); // barycenter

          for (size_t i = 0; i < nbr_overlaps; ++i) {
            curr_src_cell_gradients[i].weight_vector =
              src_cell_gradients[overlaps[i].subdomain_id] +
              overlaps[i].local_cell_id;
            curr_src_cell_gradients[i].subdomain_id = overlaps[i].subdomain_id;
            curr_src_cell_gradients[i].global_id = overlaps[i].global_cell_id;
          }
        }
      }

      // reduce amount of data to be sent, by removing duplicated gradients
      // (multiple target points might require the same gradient)
      for (unsigned i = 0; i < num_sends; ++i) {
        remove_duplicated_weight_vector_3d_subdomain(
          src_cell_gradients_per_send[i], num_src_cell_gradients_per_send + i);

        // number of gradients and gradient weights
        send_msg[i].buffer_size += yac_comm_pack_size(2, COMM_UINT64_T, comm);

        for (size_t j = 0; j < num_src_cell_gradients_per_send[i]; ++j) {
          size_t n = src_cell_gradients_per_send[i][j].weight_vector->n;
          num_src_cell_gradient_weights_per_send[i] += n;
          send_msg[i].buffer_size +=
            yac_comm_pack_size(1, COMM_INT, comm); // global_id
          send_msg[i].buffer_size +=
            yac_comm_pack_size(1, COMM_UINT64_T, comm); // n
          send_msg[i].buffer_size +=
            n * yac_comm_pack_size(1, COMM_INT, comm); // weight global_id
          send_msg[i].buffer_size +=
            n * yac_comm_pack_size(1, COMM_UINT, comm); // rank
          send_msg[i].buffer_size +=
            n * yac_comm_pack_size(3, COMM_DBLE, comm); // weights
        }
      }
    }

    // allocate sufficient memory

    for (unsigned i = 0; i < num_sends; ++i)
      ENSURE_ARRAY_SIZE(send_msg[i].cbuffer, send_msg[i].buffer_array_size,
                        send_msg[i].buffer_size);

    // store data in send buffer

    for (unsigned n = 0; n < num_cells; ++n) {

      if (!weight_provenance[n].is_master) {

        struct send_buffer * curr_send_msg =
          send_msg + send_rank_to_msg_idx[weight_provenance[n].master_rank];

        yac_comm_pack(&(weight_provenance[n].target_cell_id), 1, COMM_INT,
                      curr_send_msg->cbuffer, curr_send_msg->buffer_size,
                      &(curr_send_msg->position), comm);

        int nbr_overlaps = (int)(weight_provenance[n].nbr_overlaps);
        yac_comm_pack(&nbr_overlaps, 1, COMM_INT,
                      curr_send_msg->cbuffer, curr_send_msg->buffer_size,
                      &(curr_send_msg->position), comm);

        for (int i = 0; i < nbr_overlaps; ++i ) {
            yac_comm_pack(&(weight_provenance[n].overlaps[i].global_cell_id), 1,
                          COMM_INT, curr_send_msg->cbuffer,
                          curr_send_msg->buffer_size,
                          &(curr_send_msg->position), comm);
            yac_comm_pack(&(weight_provenance[n].overlaps[i].area), 1,
                          COMM_DBLE, curr_send_msg->cbuffer,
                          curr_send_msg->buffer_size,
                          &(curr_send_msg->position), comm);
            if (method_conserv->order == 2)
              yac_comm_pack(&(weight_provenance[n].overlaps[i].barycenter[0]),
                            3, COMM_DBLE, curr_send_msg->cbuffer,
                            curr_send_msg->buffer_size,
                            &(curr_send_msg->position), comm);
        }
      }
    }

    if (method_conserv->order == 2) {

      // pack source cell gradients
      for (unsigned i = 0; i < num_sends; ++i) {

        struct weight_vector_3d_subdomain * curr_src_cell_gradients =
          src_cell_gradients_per_send[i];
        uint64_t curr_num_gradients =
          (uint64_t)num_src_cell_gradients_per_send[i];
        uint64_t curr_num_gradient_weights =
          (uint64_t)num_src_cell_gradient_weights_per_send[i];

        yac_comm_pack(&curr_num_gradients, 1, COMM_UINT64_T,
                      send_msg[i].cbuffer, send_msg[i].buffer_size,
                      &(send_msg[i].position), comm);
        yac_comm_pack(&curr_num_gradient_weights, 1, COMM_UINT64_T,
                      send_msg[i].cbuffer, send_msg[i].buffer_size,
                      &(send_msg[i].position), comm);

        for (uint64_t j = 0; j < curr_num_gradients; ++j) {

          struct weight_vector_3d * curr_src_cell_gradient =
            curr_src_cell_gradients[j].weight_vector;
          unsigned subdomain_id = curr_src_cell_gradients[j].subdomain_id;
          int global_id = curr_src_cell_gradients[j].global_id;
          uint64_t n = (uint64_t)curr_src_cell_gradient->n;
          int const * subdomain_global_ids =
            yac_get_global_ids(domain, subdomain_id, CELL);
          unsigned const * global_ranks =
            yac_get_global_ranks(domain, subdomain_id, CELL);

          yac_comm_pack(&global_id, 1, COMM_INT, send_msg[i].cbuffer,
                        send_msg[i].buffer_size, &(send_msg[i].position),
                        comm);
          yac_comm_pack(&n, 1, COMM_UINT64_T, send_msg[i].cbuffer,
                        send_msg[i].buffer_size, &(send_msg[i].position),
                        comm);

          for (uint64_t k = 0; k < n; ++k) {

            global_id =
              subdomain_global_ids[curr_src_cell_gradient->data[k].id.local];
            unsigned rank =
              global_ranks[curr_src_cell_gradient->data[k].id.local];
            if (rank == -1) rank = comm_rank;

            yac_comm_pack(&global_id, 1, COMM_INT, send_msg[i].cbuffer,
                          send_msg[i].buffer_size, &(send_msg[i].position),
                          comm);
            yac_comm_pack(&rank, 1, COMM_UINT, send_msg[i].cbuffer,
                          send_msg[i].buffer_size, &(send_msg[i].position),
                          comm);
            yac_comm_pack(curr_src_cell_gradient->data[k].weight, 3,
                          COMM_DBLE, send_msg[i].cbuffer,
                          send_msg[i].buffer_size, &(send_msg[i].position),
                          comm);
          }
        }
      }
    }

    // We expect an answer from the master rank in the second part in any case

    yac_comm_recv_callback ( 1, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header2_exchange,
                             comm, user_data, recv_callback_accepts, NULL,
                             MSG_PRIO_HIGH );
  }

  for (unsigned i = 0; i < num_sends; ++i) {

    unsigned send_rank = send_rank_list[i];

    struct send_buffer * curr_send_msg = send_msg + i;

    // send data before sending header, this does not give a
    // conflict as messages are identified by their tag.
    if ( curr_send_msg->buffer_size > 0 ) {


      yac_comm_bsend(
        curr_send_msg->cbuffer, curr_send_msg->buffer_size, COMM_PACKED,
        send_rank, method_conserv->tags.partial_areas_exchange, comm);

      free(curr_send_msg->cbuffer);
    }

    // send header
    unsigned nbr_items[2] = {curr_send_msg->buffer_size,
                             curr_send_msg->num_target_cells};
    yac_comm_bsend(nbr_items, 2, COMM_UINT, send_rank,
                   method_conserv->tags.header1_exchange, comm);
  }

  free(num_src_cell_gradient_weights_per_send);
  free(num_src_cell_gradients_per_send);
  free(src_cell_gradient_buffer);
  free(src_cell_gradients_per_send);
  free(send_rank_list);
  free(send_rank_to_msg_idx);
  free(send_msg);

  // receive first header to get the length of data, all
  // following header messages are received inside recv_callback_partial_areas
  // for a well defined source use receives_rank_list[0], note
  // that these are currently freed and not available.

  if ( num_receives > 0 ) {

    yac_comm_recv_callback ( 2, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header1_exchange,
                             comm, user_data, recv_callback_partial_areas, NULL,
                             MSG_PRIO_HIGH );
  }

  // do_search_final_conserv and
  // do_search_interpolation_method_conserv_continue_2 are called by the
  // callback function in case messages have to be processed from other source
  // ranks. If no messages are expected we call do_search_final_conserv here.

  if ( num_receives == 0 && num_sends == 0 )
    do_search_final_conserv ( user_data );
}

// Receives the final master ranks for target cells that are shared by multiple
// source processes.
static void alltoallv_master_ranks_callback(
  struct communicator * comm, struct comm_message msg, void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;

  int comm_rank = (int)(yac_get_comm_rank(comm));

  int * recv_buffer = (int*)(msg.buffer);
  unsigned num_tgt_points = msg.count / 2;

  int * tgt_global_id = xmalloc(num_tgt_points * sizeof(*tgt_global_id));
  int * master_ranks = xmalloc(num_tgt_points * sizeof(*master_ranks));

  for (unsigned i = 0; i < num_tgt_points; ++i) {

    tgt_global_id[i] = recv_buffer[2*i+0];
    master_ranks[i] = recv_buffer[2*i+1];
  }

  yac_quicksort_index(tgt_global_id, (size_t)num_tgt_points, master_ranks);

  struct weight_provenance * weight_provenance = search_data->weight_provenance;
  unsigned num_cells = search_data->num_cells;

  for (unsigned i = 0, j = 0; i < num_tgt_points; ++i) {

    int curr_tgt_global_id = tgt_global_id[i];
    int master_rank = master_ranks[i];

    // find the respective entry in the weight provenance data
    while (weight_provenance[j].target_cell_id < curr_tgt_global_id) ++j;

    assert(j < num_cells);

    weight_provenance[j].master_rank = master_rank;
    weight_provenance[j].is_master = master_rank == comm_rank;
  }

  free(master_ranks);
  free(tgt_global_id);

  for (unsigned n = 0; n < num_cells; ++n )
    if (!weight_provenance[n].is_master)
      free(weight_provenance[n].servant_ranks);

  // now servants and master are determined
  // continue with the interpolation
  do_search_interpolation_method_conserv_continue(user_data);
}

static int
compare_uncertain_tgt_global_id(const void * a, const void * b) {

  struct temp_uncertain_tgt * t_a = (struct temp_uncertain_tgt *)a;
  struct temp_uncertain_tgt * t_b = (struct temp_uncertain_tgt *)b;

  int ret;
  if ( (ret = (t_a->global_id > t_b->global_id) -
              (t_a->global_id < t_b->global_id) )) return ret;
  if ( (ret = (t_a->num_src_points < t_b->num_src_points) -
              (t_a->num_src_points > t_b->num_src_points) )) return ret;
  return (t_a->src_rank > t_b->src_rank) - (t_a->src_rank < t_b->src_rank);
}

static int
compare_uncertain_tgt_src_rank(const void * a, const void * b) {

  struct temp_uncertain_tgt * t_a = (struct temp_uncertain_tgt *)a;
  struct temp_uncertain_tgt * t_b = (struct temp_uncertain_tgt *)b;

  int ret;
  if ( (ret = (t_a->src_rank > t_b->src_rank) -
              (t_a->src_rank < t_b->src_rank) )) return ret;
  return (t_a->global_id > t_b->global_id) - (t_a->global_id < t_b->global_id);
}

// This routine determines the final master for target cells that are shared by
// multiple source process. The master is the source process with the most
// overlaps for a given target cell.
static void alltoallv_uncertain_tgt_callback(
  struct communicator * comm, struct comm_message msg, void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  int comm_rank = (int)(yac_get_comm_rank(comm));
  unsigned comm_size = yac_get_comm_size(comm);

  int * recv_buffer = (int*)(msg.buffer);
  unsigned offset = 0;

  struct temp_uncertain_tgt * tgt_points =
    xmalloc((msg.count / 2) * sizeof(*tgt_points));

  unsigned num_received_tgt = 0;

  // unpack data packed by routine determine_master_for_tgt_cells
  while (offset < msg.count) {

    unsigned curr_src_rank = recv_buffer[offset++];
    unsigned curr_num_tgt = recv_buffer[offset++];

    for (unsigned i = 0; i < curr_num_tgt; ++i) {

      tgt_points[num_received_tgt + i].src_rank = curr_src_rank;
      tgt_points[num_received_tgt + i].global_id = recv_buffer[offset++];
      tgt_points[num_received_tgt + i].num_src_points = recv_buffer[offset++];
    }
    num_received_tgt += curr_num_tgt;
  }

  // sort the received target points by their global ids, then their number of
  // source points (descending order) followed by their source rank
  qsort(tgt_points, num_received_tgt, sizeof(*tgt_points),
        compare_uncertain_tgt_global_id);

  struct weight_provenance * weight_provenance = search_data->weight_provenance;
  unsigned num_cells = search_data->num_cells;

  // determine the master rank
  int prev_global_id;
  int prev_master_rank = -1;
  if (num_received_tgt > 0) prev_global_id = tgt_points[0].global_id - 1;

  for (unsigned i = 0, j = 0; i < num_received_tgt; ++i) {

    int curr_global_id = tgt_points[i].global_id;

    // if we have already determined the master rank for this global id
    if (prev_global_id == curr_global_id) {
      tgt_points[i].master_rank = prev_master_rank;
      continue;
    }

    // find the respective entry in the weight provenance data
    while (weight_provenance[j].target_cell_id < curr_global_id) ++j;

    assert(j < num_cells);
    assert(weight_provenance[j].target_cell_id == curr_global_id);

    // if the local process is not the master for the current target point
    if ((size_t)(tgt_points[i].num_src_points) >
        weight_provenance[j].nbr_overlaps) {
      weight_provenance[j].master_rank = (int)(tgt_points[i].src_rank);
      weight_provenance[j].is_master = 0;
      prev_master_rank = (int)(tgt_points[i].src_rank);
    } else
      prev_master_rank = comm_rank;

    prev_global_id = curr_global_id;
    tgt_points[i].master_rank = prev_master_rank;
  }

  // sort the received target points by source rank followd by the global id
  qsort(tgt_points, num_received_tgt, sizeof(*tgt_points),
        compare_uncertain_tgt_src_rank);

  // return results
  int * send_list = xmalloc(2 * num_received_tgt * sizeof(*send_list));
  unsigned * send_size_per_rank =
    xcalloc(comm_size, sizeof(*send_size_per_rank));

  for (unsigned i = 0; i < num_received_tgt; ++i) {

    send_list[2*i+0] = tgt_points[i].global_id;
    send_list[2*i+1] = tgt_points[i].master_rank;
    send_size_per_rank[tgt_points[i].src_rank] += 2;
  }

  yac_comm_alltoallv_callback(send_list, send_size_per_rank, COMM_INT,
                              method_conserv->tags.alltoallv_master_ranks,
                              comm, user_data, alltoallv_master_ranks_callback,
                              MSG_PRIO_HIGH);

  free(send_list);
  free(send_size_per_rank);
  free(tgt_points);
}

// this routine send the number of overlaps for each target cell that is shared
// by multiple source processes to the preliminary master
static void
determine_master_for_tgt_cells(struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  struct communicator * comm =
    yac_get_domain_comm(method_conserv->domain);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;

  unsigned send_list_size = 0;
  unsigned * send_size_per_rank =
    xcalloc(comm_size, sizeof(send_size_per_rank));
  unsigned * send_list_offsets =
    xmalloc(comm_size * sizeof(send_list_offsets));

  // count number of target cells that are currently assigned to another source
  // process
  for (unsigned i = 0; i < num_cells; ++i) {
    if (!weight_provenance[i].is_master) {
      send_list_size++;
      send_size_per_rank[weight_provenance[i].master_rank] += 2;
    }
  }
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] > 0) {
      send_list_size++;
      send_size_per_rank[i] += 2; // for the local rank
    }
  }
  send_list_size *= 2;
  unsigned accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    send_list_offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  int * send_list = xmalloc(2 * send_list_size * sizeof(*send_list));
  // set local rank and number of target points in front of each message
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] > 0) {

      unsigned offset = send_list_offsets[i];

      send_list_offsets[i] += 2;
      send_list[offset+0] = comm_rank;
      send_list[offset+1] = send_size_per_rank[i] / 2 - 1;
    }
  }

  for (unsigned i = 0; i < num_cells; ++i) {
    if (!weight_provenance[i].is_master) {

      int master_rank = weight_provenance[i].master_rank;
      unsigned offset = send_list_offsets[master_rank];
      send_list_offsets[master_rank] += 2;
      send_list[offset+0] = weight_provenance[i].target_cell_id;
      send_list[offset+1] = (int)(weight_provenance[i].nbr_overlaps);
    }
  }

  // send uncertain target points to current masters
  yac_comm_alltoallv_callback(send_list, send_size_per_rank, COMM_INT,
                              method_conserv->tags.alltoallv_uncertain_tgt,
                              comm, user_data, alltoallv_uncertain_tgt_callback,
                              MSG_PRIO_HIGH);

  free(send_list);
  free(send_list_offsets);
  free(send_size_per_rank);
}

// -----------------------------------------------------------------------------

static void do_search_interpolation_method_conserv(
  struct interpolation_method ** method,
  struct interpolation_plan_element ** plan_stack, int * global_ids,
  double (*coordinates_xyz)[3], enum yac_location tgt_location,
  unsigned num_cells, unsigned * num_interpolated_cells,
  struct interpolation_method_data_request ** data_request ) {

  if (tgt_location != CELL)
    yac_internal_abort_message("ERROR(do_search_interpolation_method_conserv): "
                               "invalid target location (must be CELL)\n",
                               __FILE__, __LINE__);

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(method[0]);
  method_conserv->data_request = data_request;
  num_interpolated_cells[0] = 0;

  struct search_data_conserv * user_data = xmalloc(1 * sizeof(*user_data));

  // sort target points by global ids
  yac_sort_points(global_ids, coordinates_xyz, (size_t)num_cells);

  // store data that is to be passed to do_search_final_conserv (via
  // recv_callback_partial_areas)
  user_data->method_stack = method;
  user_data->plan_stack = plan_stack;
  user_data->global_ids_buffer = global_ids;
  user_data->coordinates_xyz_buffer = coordinates_xyz;
  user_data->num_cells = num_cells;
  user_data->num_interpolated_cells = num_interpolated_cells;
  user_data->num_partial_areas_messages_to_receive = 0;
  user_data->num_accepts_messages_to_receive = 0;
  user_data->weight_provenance =
    xmalloc(num_cells * sizeof(*(user_data->weight_provenance)));
  user_data->target_cell_ids_to_remove = NULL;
  user_data->num_target_cell_ids_to_remove = 0;

  init_weight_provenance_data(user_data);

  determine_master_for_tgt_cells(user_data);
}

// -----------------------------------------------------------------------------

static void
enddef_interpolation_method_conserv ( struct interpolation_method ** method ) {

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(method[0]);

  struct communicator * comm =
    yac_get_domain_comm(method_conserv->domain);

  while (yac_comm_wait(comm, method_conserv->tags.alltoallv_uncertain_tgt) ||
         yac_comm_wait(comm, method_conserv->tags.alltoallv_master_ranks) ||
         yac_comm_wait(comm, method_conserv->tags.header1_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.partial_areas_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.header2_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.accepts_exchange));

  if (method[1] != NULL)
    method[1]->vtable->enddef(method + 1);
}

// -----------------------------------------------------------------------------

static void
get_source_exchange_data_interpolation_method_conserv (
  struct interpolation_method ** method,
  struct interpolation_method_data_request *** data_request ) {

  struct interpolation_method_conserv * method_conserv;

  method_conserv = (struct interpolation_method_conserv*)(method[0]);

  *data_request = method_conserv->data_request;
  method_conserv->data_request = NULL;
}

// -----------------------------------------------------------------------------

static void
delete_interpolation_method_conserv(struct interpolation_method * method) {
  free(method);
}

// -----------------------------------------------------------------------------

static struct global_search *
get_global_search(struct interpolation_method * method) {

  return ((struct interpolation_method_conserv*)(method))->search;
}

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas_cont ( struct communicator * comm,
                                   struct comm_message msg,
                                   void * user_data ) {

  struct temp_recv_partial_areas_data * temp_data =
    (struct temp_recv_partial_areas_data *)user_data;

  unsigned num_target_cells   = temp_data->num_target_cells;
  struct search_data_conserv * search_data = temp_data->search_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  free(temp_data);

  int position = 0;
  size_t offset = 0;

  for (unsigned n = 0; n < num_target_cells; ++n ) {

    // stupid n^2 search to find matching m
    // something better is needed, probably a sorted lists

    int target_id;
    int nbr_overlaps;

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(target_id), 1, COMM_INT, comm);

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(nbr_overlaps), 1, COMM_INT, comm);

    unsigned m;

    for (m = 0; m < search_data->num_cells; ++m)
      if ( target_id == search_data->weight_provenance[m].target_cell_id ) break;

    if ( m == search_data->num_cells ) continue;

    // increase size of search_data->weight_provenance[m].overlaps

    search_data->weight_provenance[m].overlaps =
      xrealloc(search_data->weight_provenance[m].overlaps,
             (search_data->weight_provenance[m].nbr_overlaps +
             (size_t)(nbr_overlaps) ) *
             sizeof(*(search_data->weight_provenance[m].overlaps)) );

    offset = search_data->weight_provenance[m].nbr_overlaps;
    search_data->weight_provenance[m].nbr_overlaps += (size_t)nbr_overlaps;

    for (size_t i = 0; i < (size_t)nbr_overlaps; ++i ) {
      search_data->weight_provenance[m].overlaps[offset+i].subdomain_id = -1;
      search_data->weight_provenance[m].overlaps[offset+i].local_cell_id = -1;
      search_data->weight_provenance[m].overlaps[offset+i].rank = msg.src_rank;

      yac_comm_unpack(
        msg.buffer, msg.count, &position,
        &(search_data->weight_provenance[m].overlaps[offset+i].global_cell_id),
        1, COMM_INT, comm);

      yac_comm_unpack(
        msg.buffer, msg.count, &position,
        &(search_data->weight_provenance[m].overlaps[offset+i].area),
        1, COMM_DBLE, comm);

      if (method_conserv->order == 2)
        yac_comm_unpack(
          msg.buffer, msg.count, &position,
          &(search_data->weight_provenance[m].overlaps[offset+i].barycenter[0]),
          3, COMM_DBLE, comm);
    }
  }

  if (method_conserv->order == 2) {

    uint64_t num_src_cell_gradients;
    uint64_t total_weight_vector_data_3d;
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &num_src_cell_gradients, 1, COMM_UINT64_T, comm);
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &total_weight_vector_data_3d, 1, COMM_UINT64_T, comm);

    struct weight_vector_data_3d * weight_vector_data_3d_buffer =
      xmalloc(total_weight_vector_data_3d *
              sizeof(*weight_vector_data_3d_buffer));

    struct remote_gradients * curr_remote_gradients =
      search_data->remote_src_cell_gradients;
    search_data->remote_src_cell_gradients++;

    curr_remote_gradients->gradients =
      xmalloc(num_src_cell_gradients *
              sizeof(*(curr_remote_gradients->gradients)));
    curr_remote_gradients->num_gradients = (size_t)num_src_cell_gradients;
    curr_remote_gradients->weight_vector_data_3d_buffer =
      weight_vector_data_3d_buffer;
    curr_remote_gradients->weight_vector_data_3d_buffer_size =
      (size_t)total_weight_vector_data_3d;
    curr_remote_gradients->global_src_cell_ids =
      xmalloc(num_src_cell_gradients *
              sizeof(*(curr_remote_gradients->global_src_cell_ids)));

    struct temp_remote_cell * temp_remote_cells =
      xmalloc(total_weight_vector_data_3d * sizeof(*temp_remote_cells));
    size_t num_remote_cells = 0;

    for (uint64_t i = 0; i < num_src_cell_gradients; ++i) {

      yac_comm_unpack(
        msg.buffer, msg.count, &position,
        curr_remote_gradients->global_src_cell_ids + i, 1, COMM_INT, comm);

      uint64_t n;
      yac_comm_unpack(
        msg.buffer, msg.count, &position, &n, 1, COMM_UINT64_T, comm);

      struct weight_vector_data_3d * curr_weights =
        weight_vector_data_3d_buffer;
      weight_vector_data_3d_buffer += (size_t)n;

      struct weight_vector_3d * curr_weight_vector =
        curr_remote_gradients->gradients + i;
      curr_weight_vector->data = curr_weights;
      curr_weight_vector->n = (size_t)n;

      for (uint64_t j = 0; j < n; ++j, ++num_remote_cells) {
        int global_id;
        unsigned rank;
        yac_comm_unpack(msg.buffer, msg.count, &position, &global_id, 1,
                        COMM_INT, comm);
        yac_comm_unpack(msg.buffer, msg.count, &position, &rank, 1,
                        COMM_UINT, comm);
        yac_comm_unpack(msg.buffer, msg.count, &position,
                        &(curr_weights[j].weight[0]), 3, COMM_DBLE, comm);
        curr_weights[j].id.global = global_id;

        temp_remote_cells[num_remote_cells].global_id = global_id;
        temp_remote_cells[num_remote_cells].rank = rank;
      }
    }

    remove_duplicated_remote_cells(temp_remote_cells, &num_remote_cells);
    struct remote_cell * remote_cells =
      xmalloc(num_remote_cells * sizeof(*curr_remote_gradients->remote_cells));
    curr_remote_gradients->remote_cells = remote_cells;
    for (size_t i = 0; i < num_remote_cells; ++i) {
      remote_cells[i].global_id = temp_remote_cells[i].global_id;
      remote_cells[i].rank      = temp_remote_cells[i].rank;
    }
    free(temp_remote_cells);
    curr_remote_gradients->num_remote_cells = num_remote_cells;
  }

  // send the next header message

  // Even though the following part looks rather simple, you have to be really
  // careful when changing it!
  // do_search_final_conserv has to be called once all accept messages and all
  // partial_area messages have been received. It is only allowed to be called
  // ONCE by EITHER recv_callback_accepts or recv_callback_partial_areas. It is
  // possible that both routines are on the stack. So you have to ensure, that
  // only one call on the stack sees:
  // "search_data->num_accepts_messages_to_receive == 0 &&
  //  search_data->num_partial_areas_messages_to_receive == 0"
  // On the other hand you also have to ensure that when calling
  // comm_recv_callback the respective message counter
  // (num_accepts_messages_to_receive or num_partial_areas_messages_to_receive)
  // is correct. Because otherwise comm_recv_callback will be called to often,
  // which causes a deadlock in enddef_interpolation_method_conserv.

  if ( search_data->num_partial_areas_messages_to_receive > 1 ) {

    search_data->num_partial_areas_messages_to_receive--;

    yac_comm_recv_callback ( 2, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header1_exchange,
                             comm, search_data, recv_callback_partial_areas,
                             NULL, MSG_PRIO_HIGH );

  } else {

    // num_partial_areas_messages_to_receive must be decreased after
    // do_search_interpolation_method_conserv_continue_2 was called, because
    // otherwise it would be possible that recv_callback_accepts might call
    // do_search_final_conserv before
    // do_search_interpolation_method_conserv_continue_2 was finished, which is
    // mandatory

    do_search_interpolation_method_conserv_continue_2 ( search_data );

    search_data->num_partial_areas_messages_to_receive--;

    if ( search_data->num_accepts_messages_to_receive == 0 &&
         search_data->num_partial_areas_messages_to_receive == 0 )
      do_search_final_conserv ( search_data );
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas ( struct communicator * comm,
                              struct comm_message msg,
                              void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);
  unsigned packed_buffer_size = ((unsigned *)(msg.buffer))[0];

  struct temp_recv_partial_areas_data * temp_data =
    xmalloc(1 * sizeof(*temp_data));

  temp_data->num_target_cells = ((unsigned *)(msg.buffer))[1];
  temp_data->search_data = search_data;

  if ( packed_buffer_size > 0 ) {

    // receive owners (ranks) of source cell IDs, source cell IDs, and number of
    // vertices per source cell
    yac_comm_recv_callback(packed_buffer_size, COMM_PACKED, msg.src_rank,
                           method_conserv->tags.partial_areas_exchange, comm,
                           temp_data, recv_callback_partial_areas_cont, NULL,
                           MSG_PRIO_HIGH);
  } else {

    struct comm_message temp_msg = msg;

    temp_msg.buffer = xmalloc(0);
    temp_msg.type = COMM_PACKED;
    temp_msg.count = 0;
    temp_msg.tag = method_conserv->tags.partial_areas_exchange;
    recv_callback_partial_areas_cont(comm, temp_msg, temp_data);
    free(temp_msg.buffer);
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_accepts_cont ( struct communicator * comm,
                        struct comm_message message,
                        void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  unsigned num_target_cells = message.count;

  // copy list of target cell IDs that can be processed by the master

  if ( num_target_cells > 0 ) {

    search_data->target_cell_ids_to_remove =
      xrealloc(search_data->target_cell_ids_to_remove,
               (search_data->num_target_cell_ids_to_remove + num_target_cells) *
               sizeof(*(search_data->target_cell_ids_to_remove)));

    memcpy(search_data->target_cell_ids_to_remove +
           search_data->num_target_cell_ids_to_remove, message.buffer,
           num_target_cells * sizeof(*(search_data->target_cell_ids_to_remove)));

    search_data->num_target_cell_ids_to_remove += num_target_cells;
  }

  // send the next header message

  // Even though the following part looks rather simple, you have to be really
  // careful when changing it!
  // do_search_final_conserv has to be called once all accept messages and all
  // partial_area messages have been received. It is only allowed to be called
  // ONCE by EITHER recv_callback_accepts or recv_callback_partial_areas. It is
  // possible that both routines are on the stack. So you have to ensure, that
  // only one call on the stack sees:
  // "search_data->num_accepts_messages_to_receive == 0 &&
  //  search_data->num_partial_areas_messages_to_receive == 0"
  // On the other hand you also have to ensure that when calling
  // comm_recv_callback the respective message counter
  // (num_accepts_messages_to_receive or num_partial_areas_messages_to_receive)
  // is correct. Because otherwise comm_recv_callback will be called to often,
  // which causes a deadlock in enddef_interpolation_method_conserv.

  search_data->num_accepts_messages_to_receive--;

  if ( search_data->num_accepts_messages_to_receive > 0 ) {

    yac_comm_recv_callback ( 1, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header2_exchange,
                             comm, user_data, recv_callback_accepts, NULL,
                             MSG_PRIO_HIGH );

  } else if ( search_data->num_partial_areas_messages_to_receive == 0) {

    do_search_final_conserv ( search_data );
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_accepts ( struct communicator * comm,
                        struct comm_message message,
                        void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);
  unsigned num_target_cells = *(unsigned *)(message.buffer);

  // receive list of target cell IDs that can be processed by the master

  if ( num_target_cells > 0 ) {

    yac_comm_recv_callback(num_target_cells, COMM_INT, message.src_rank,
                           method_conserv->tags.accepts_exchange, comm,
                           user_data, recv_callback_accepts_cont, NULL,
                           MSG_PRIO_HIGH);
  } else {

    struct comm_message temp_msg = message;

    temp_msg.buffer = xmalloc(0);
    temp_msg.type = COMM_INT;
    temp_msg.count = 0;
    temp_msg.tag = method_conserv->tags.accepts_exchange;
    recv_callback_accepts_cont(comm, temp_msg, user_data);
    free(temp_msg.buffer);
  }
}

// -----------------------------------------------------------------------------

static void
set_finish_callback ( struct interpolation_method * method,
                      void (*callback)(void * user_data),
                      void * user_data ) {

  struct interpolation_method_conserv * method_conserv;

  method_conserv = (struct interpolation_method_conserv*)method;

  method_conserv->finish_callback = callback;
  method_conserv->finish_user_data = user_data;
}

