/**
 * @file interpolation_method_hcsbb.c
 * @brief implementation of the hybrid cubic spherical Bernstein-Bezier
 * patche interpolation method
 * (see: Xiaoyu Liu, Larry L. Schumaker, Hybrid Bézier patches on sphere-like
 *      surfaces, Journal of Computational and Applied Mathematics, Volume 73,
 *      Issues 1–2, 1996, Pages 157-172, ISSN 0377-0427,
 *      https://doi.org/10.1016/0377-0427(96)00041-6)
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "interpolation_method.h"
#include "interpolation_method_hcsbb.h"
#include "interpolation_plan.h"
#include "grid_search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"
#include "grid.h"
#include "points.h"
#include "nnn_search.h"
#include "yac_lapack_interface.h"

// the default values for HCSBB_GAUSS_NNN, HCSBB_D_NNN, and HCSBB_GAUSS_ORDER
// were determined empirically

//! number of nearest neighbours for each gauss integration point
#ifndef HCSBB_GAUSS_NNN
#define HCSBB_GAUSS_NNN (4)
#endif
//! number of points nearest source points used to generate the triangle, which
//! is used to estimate the derivative at a certain point
#ifndef HCSBB_D_NNN
#define HCSBB_D_NNN (7)
#endif

#ifndef HCSBB_GAUSS_ORDER
#define HCSBB_GAUSS_ORDER (6)
#endif
// static size_t const num_gauss_points[] = {1,1,3,4,6,7,12,13,16,19,25,27,33};
// -> HCSBB_NUM_GAUSS_POINTS = num_gauss_points[HCSBB_GAUSS_ORDER]
#if HCSBB_GAUSS_ORDER == 0
#define HCSBB_NUM_GAUSS_POINTS (1)
#elif HCSBB_GAUSS_ORDER == 1
#define HCSBB_NUM_GAUSS_POINTS (1)
#elif HCSBB_GAUSS_ORDER == 2
#define HCSBB_NUM_GAUSS_POINTS (3)
#elif HCSBB_GAUSS_ORDER == 3
#define HCSBB_NUM_GAUSS_POINTS (4)
#elif HCSBB_GAUSS_ORDER == 4
#define HCSBB_NUM_GAUSS_POINTS (6)
#elif HCSBB_GAUSS_ORDER == 5
#define HCSBB_NUM_GAUSS_POINTS (7)
#elif HCSBB_GAUSS_ORDER == 6
#define HCSBB_NUM_GAUSS_POINTS (12)
#elif HCSBB_GAUSS_ORDER == 7
#define HCSBB_NUM_GAUSS_POINTS (13)
#elif HCSBB_GAUSS_ORDER == 8
#define HCSBB_NUM_GAUSS_POINTS (16)
#elif HCSBB_GAUSS_ORDER == 9
#define HCSBB_NUM_GAUSS_POINTS (19)
#else
#define HCSBB_NUM_GAUSS_POINTS (-1)
#error "interpolation_method_hcsbb: the specified gauss order is currently not supported."
#endif

// number spherical Bernstein-Bezier coefficients
// n_c = = ((degree + 2) * (degree + 1)) / 2
// sbb_degree = 3
#define HCSBB_NUM_SBB_COEFFS (10)

static void do_search_interpolation_method_hcsbb(
  struct interpolation_method ** method,
  struct interpolation_plan_element ** plan_stack,
  int * global_ids, double (*coordinates_xyz)[3],
  enum yac_location tgt_location, unsigned num_points,
  unsigned * num_interpolated_points,
  struct interpolation_method_data_request ** data_request);
static void enddef_interpolation_method_hcsbb(
  struct interpolation_method ** method);
static void get_source_exchange_data_interpolation_method_hcsbb(
  struct interpolation_method ** method,
  struct interpolation_method_data_request *** data_request);
static void delete_interpolation_method_hcsbb(
  struct interpolation_method * method);
static struct global_search * get_global_search(
  struct interpolation_method * method);
static void set_finish_callback(
  struct interpolation_method * method, void (*callback)(void * user_data),
  void * user_data);

static struct interpolation_method_vtable
  interpolation_method_hcsbb_vtable = {
    .do_search                = do_search_interpolation_method_hcsbb,
    .enddef                   = enddef_interpolation_method_hcsbb,
    .get_source_exchange_data = get_source_exchange_data_interpolation_method_hcsbb,
    .delete                   = delete_interpolation_method_hcsbb,
    .get_global_search        = get_global_search,
    .set_finish_callback      = set_finish_callback
};

struct interpolation_method_hcsbb_tags {

  unsigned exchange_initial_results[2]; //!< tags for exchanging of the results
                                        //!< of the initial search that
                                        //!< determines which process takes
                                        //!< care of which target points
  unsigned exchange_results[2]; //!< tags for exchanging which target points are
                                //!< actually interpolated
};


struct interpolation_method_hcsbb {

   struct interpolation_method_vtable * vtable;
   void (*finish_callback)(void * user_data); //!< callback routine that is called once the interpolation is finished
   void * finish_user_data;                   //!< user data that is passed to finish_callback
   struct global_search * search;             //!< global search object used a the basis for the interpolation
   struct domain * domain;
   struct points *** points;
   struct mask *** masks;
   size_t num_pointsets;
   enum yac_location src_location;
   size_t points_idx;                          //!< index of the points set of field that is to be used for the interpolation
   struct interpolation_method_data_request ** data_request;
   struct interpolation_method_hcsbb_tags tags;
   size_t use_source_mask;
   struct nnn_search * nnn_search;
};

enum interp_status_flag {
  LOCAL = 0,  //!< assigned to the local process
  REMOTE = 1, //!< assigned to a remote process
  NOT = 2,    //!< cannot be interpolated by hcsbb
  REMOVE = 3, //!< interpolated by a remote process
};

enum interp_type_flag {
  ON_VERTEX = 0,   //!< there is a source point that exactly matches the target
                   //!< point
  ON_EDGE = 1,     //!< the target point is on an edge of the source point grid
  ON_TRIANGLE = 2, //!< the target point is within a triangle of the source
                   //!< point grid
};

struct src_point {
  int global_id;
  unsigned subdomain_id;
  unsigned local_id;
};

struct edge_pos {
  int global_ids[2];
  size_t orig_pos;
};

//! gauss integration point patch data (used to estimate derivatives)
//! (if num_src_points == 0, no valid gauss integration point patch could be
//!  generated)
struct gauss_nnn_patch {

  double bnd_triangle[3][3]; //!< bounding triangle

  //! source points used to estimate the derivative
  struct src_point * src_points;
  //! owner ranks of source points used to estimate the derivative
  unsigned * src_point_ranks;
  //! number of source points
  size_t num_src_points;
  //! after init_gauss_nnn_patch:
  //!  w_g ([HCSBB_NUM_SBB_COEFFS][HCSBB_NUM_SBB_COEFFS])
  //! after compute_gauss_nnn_patch:
  //! w_g * w_nnn ([num_src_points][HCSBB_NUM_SBB_COEFFS])
  double * weights;
};

struct derivative_data_src_point {

  //! source point at which a derivative is to be estimated
  //! (if it is not a valid source point -> subdomain_id = 0, local_id = -1)
  struct src_point src_point;
  //! point at which a derivative is to be estimated
  double coordinate_xyz[3];
  //! gauss integration point patch used to estimate the derivative
  struct gauss_nnn_patch * gauss_nnn_patch;
};

struct derivative_data_middle_point {

  //! point at which a derivative is to be estimated
  double coordinate_xyz[3];
  //! gauss integration point patch used to estimate the derivative
  struct gauss_nnn_patch * gauss_nnn_patch;
};

struct weight_vector_data {
  double weight;
  unsigned subdomain_id;
  unsigned local_id;
};

struct weight_vector_data_pos {
  struct weight_vector_data * weight;
  size_t pos;
};

struct weight_vector {

  struct weight_vector_data * data;
  size_t n;
};

struct edge_interp_data {

  //! c_300 = f(v1)
  //! c_030 = f(v2)
  //! c_210 = f(v1) + D_12 p(v1)/3
  //! c_120 = f(v2) + D_21 p(v2)/3
  //! we do not explicitly store c_300 and c_030
  //! the remaining coefficients are zero on the edge
  struct weight_vector c[2]; // c_210, c_120;

  //! derivative data of both vertices
  struct derivative_data_src_point * vertex_d_data[2];

  //! derivative data of edge middle point
  struct derivative_data_middle_point * middle_d_data;

  //! spherical barycentric coordinate of the middle point
  //! (b0 and b1 are identical)
  double sb_coord_middle_point;

  //! vector that is a tangent of the sphere in the middle points of the edge
  //! and is perpendicular to the edge
  double g[3];
};

struct triangle_interp_data {

  //! edges of the triangle (edge[0] is the on opposing corner src_points[0])
  struct edge_interp_data * edges[3];

  //! weights for alpha values used to compute c_111
  struct weight_vector c_111_a[3];
};

struct basic_patch_data {

  //! gauss integration point patches for the computation of derivatives
  struct gauss_nnn_patch * gauss_nnn_patches;
  size_t num_gauss_nnn_patches;

  //! derivative data for required vertices
  struct derivative_data_src_point * d_data_src_point;
  size_t d_data_src_point_count;

  //! derivative data for required middle points of required edges
  struct derivative_data_middle_point * d_data_middle_point;
  size_t d_data_middle_point_count;

  //! number of unique vertices required for the interpolation
  size_t num_vertices;

  //! interpolation data of edges required for the interpolation
  struct edge_interp_data * edges;
  size_t num_edges;
  struct weight_vector_data * edge_weight_vector_data_buffer;

  //! interpolation data of triangles required for the interpolation
  struct triangle_interp_data * triangles;
  size_t num_triangles;
};

/**
 * contains all temporary search specific data for one target point
 */
struct tgt_point_search_data {

  int global_id; //!< global id of the target point
  enum interp_status_flag status_flag;
  double coordinate_xyz[3];

  enum interp_type_flag type_flag;
  unsigned subdomain_id;
  unsigned src_point_local_ids[3];
  unsigned ranks[3]; // -1 if the point is local
  int src_point_global_ids[3];
  double sb_coords[3];

  //! depending on type flag interp_data contains a point to the required data
  union {
    struct edge_interp_data * edge;
    struct triangle_interp_data * triangle;
  } interp_data;
};

struct search_data {

  struct tgt_point_search_data * tgt_point_data;
  size_t num_tgt_points;               //!< total number of entries in
                                       //!< tgt_point_data
  struct weight_vector_data * weights; //!< weight for the interpolation of the
                                       //!< target points
  unsigned * num_weights_per_tgt;      //!< number of weights per target point
  size_t total_num_weights;            //!< total number of weights
};

//! collection of data that is required through out the nearest neighbourhood search
struct intermediate_method_data {

  struct search_data search_data; //!< search data for the target points that
                                  //!< are assigned to the process

  struct basic_patch_data patch_data;

  struct interpolation_method ** method_stack;
  struct interpolation_plan_element ** plan_stack;
  unsigned * num_interpolated_points;
  double (*coordinates_xyz)[3];
  int * global_ids;
  enum yac_location tgt_location;
};

struct data_request_point {
  int global_id;
  unsigned rank;
};

struct bnd_triangle_gid {
  double * bnd_triangle;
  int global_id;
};

static void generate_tag_string(char * buf,  size_t buffer_size,
                                char const * tag_base_str,
                                char const * source_comp_name,
                                char const * target_comp_name,
                                char const * source_domain_name,
                                char const * target_domain_name,
                                char const * stack_name,
                                unsigned points_idx,
                                unsigned use_source_mask,
                                int unique_interpolation_method_id,
                                char const * tag_str) {

  int ret_val;
  ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%u%u%d%s", tag_base_str,
                     source_comp_name, target_comp_name, source_domain_name,
                     target_domain_name, stack_name, points_idx,
                     use_source_mask, unique_interpolation_method_id, tag_str);

  if (ret_val < 0 || (size_t)ret_val >= buffer_size)
    yac_internal_abort_message(
      "ERROR(generate_tag_string): truncation", __FILE__, __LINE__);
}

static void generate_interpolation_method_hcsbb_tag(
  struct interpolation_method_hcsbb * method, char const * stack_name) {

   char const * str_base_str                  = "interpolation method hcsbb";
   char const * str_generator_string          = "generator string";
   char const * str_exchange_inital_results_a = "exchange initial results a";
   char const * str_exchange_inital_results_b = "exchange initial results b";
   char const * str_exchange_results_a        = "exchange results a";
   char const * str_exchange_results_b        = "exchange results b";

   struct global_search * search =
      get_global_search((struct interpolation_method *)method);

   struct communicator * comm = yac_get_domain_comm(method->domain);

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen = strlen(str_exchange_inital_results_a);

   unsigned max_str_buf_size = max_tag_strlen;
   max_str_buf_size += strlen(str_base_str);
   max_str_buf_size += strlen(source_comp_name);
   max_str_buf_size += strlen(target_comp_name);
   max_str_buf_size += strlen(source_domain_name);
   max_str_buf_size += strlen(target_domain_name);
   max_str_buf_size += strlen(stack_name) + 10 + 2 + 10 + 1; // (+ 10) for points_idx
                                                             // (+ 2)  for mask handling
                                                             // (+ 10) for unique_method_id
                                                             // (+ 1)  for '\0'

   char * unique_generator_string = xmalloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[4];

   tag_strings[0] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_inital_results_a);

   tag_strings[1] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_inital_results_b);

   tag_strings[2] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_a);

   tag_strings[3] = xmalloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_b);

   unsigned tags[4];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 4, tags);

   free(unique_generator_string);

   method->tags.exchange_initial_results[0] = tags[0];
   method->tags.exchange_initial_results[1] = tags[1];
   method->tags.exchange_results[0] = tags[2];
   method->tags.exchange_results[1] = tags[3];

   for (unsigned i = 0; i < 4; ++i) free(tag_strings[i]);
}

static void init_tgt_search_data(
  struct tgt_point_search_data * data, double coordinate_xyz[3],
  int global_id) {

  data->global_id = global_id;
  data->status_flag = NOT;
  data->coordinate_xyz[0] = coordinate_xyz[0];
  data->coordinate_xyz[1] = coordinate_xyz[1];
  data->coordinate_xyz[2] = coordinate_xyz[2];
}

static struct tgt_point_search_data * tgt_point_search_data_new(
  size_t num_points, double (*coordinates_xyz)[3], int * global_ids) {

  struct tgt_point_search_data * data = xmalloc(num_points * sizeof(*data));
  
  for (size_t i = 0; i < num_points; ++i)
    init_tgt_search_data(data+i, coordinates_xyz[i], global_ids[i]);
  
  return data;
}

static void free_search_data(struct search_data search_data) {

  free(search_data.tgt_point_data);
  free(search_data.weights);
  free(search_data.num_weights_per_tgt);
}

static void free_patch_data(struct basic_patch_data patch_data) {

  {
    struct gauss_nnn_patch * gauss_nnn_patches = patch_data.gauss_nnn_patches;
    size_t num_gauss_nnn_patches = patch_data.num_gauss_nnn_patches;
    for(size_t i = 0; i < num_gauss_nnn_patches; ++i) {
      free(gauss_nnn_patches[i].src_points);
      free(gauss_nnn_patches[i].src_point_ranks);
      free(gauss_nnn_patches[i].weights);
    }
    free(gauss_nnn_patches);
  }

  free(patch_data.d_data_src_point);
  free(patch_data.d_data_middle_point);

  free(patch_data.edges);
  free(patch_data.edge_weight_vector_data_buffer);

  for (size_t i = 0; i < patch_data.num_triangles; ++i)
    for (size_t j = 0; j < 3; ++j)
      free(patch_data.triangles[i].c_111_a[j].data);
  free(patch_data.triangles);
}

static void free_intermediate_method_data(
  struct intermediate_method_data * data) {

  free_search_data(data->search_data);
  free_patch_data(data->patch_data);
  free(data);
}

static void retrieve_tgt_point_data(
  struct tgt_point_search_data * tgt_point_data, size_t num_tgt_points,
  double (*coordinates_xyz)[3], int * global_ids) {

  for (size_t i = 0; i < num_tgt_points; ++i) {
    global_ids[i] = tgt_point_data[i].global_id;
    coordinates_xyz[i][0] = tgt_point_data[i].coordinate_xyz[0];
    coordinates_xyz[i][1] = tgt_point_data[i].coordinate_xyz[1];
    coordinates_xyz[i][2] = tgt_point_data[i].coordinate_xyz[2];
  }
}

/**
 * Marks points from the list of target points (even if they could
 * be interpolated by the local process). This is done because they are going
 * to be interpolated by another source process, which might later
 * request some data from the local process to do the interpolation.
 * @param[in] num_points_to_be_marked number of points that are interpolated
 *                                    by another source process
 * @param[in] global_ids_to_be_marked array with global ids of target points
 *                                    that are to be marked from the local list
 * @param[in,out] search_data         basic search data
 *                                    (see \ref intermediate_method_data)
 * @param[in] status_flag             use this flag for marking
 */
static void mark_tgt_points(
  size_t num_points_to_be_marked, int * global_ids_to_be_marked,
  struct search_data search_data, enum interp_status_flag status_flag) {

  if (num_points_to_be_marked == 0) return;

  // first sort all global ids that need to be marked
  yac_quicksort_index(global_ids_to_be_marked, num_points_to_be_marked, NULL);

  // for all target points that need to be marked
  for (size_t i = 0, j = 0; i < num_points_to_be_marked; ++i) {

    // find the matching entry in the local target points
    while ((j < search_data.num_tgt_points) &&
           (search_data.tgt_point_data[j].global_id <
            global_ids_to_be_marked[i])) ++j;

    // if no matching entry could be found
    if ((j == search_data.num_tgt_points) ||
        (search_data.tgt_point_data[j].global_id !=
        global_ids_to_be_marked[i]))
      yac_internal_abort_message(
        "ERROR(mark_tgt_points): invalid global target id\n",
        __FILE__, __LINE__);

    // mark the target point
    search_data.tgt_point_data[j].status_flag = status_flag;
  }
}

static inline int compare_tgt_point_data_global_id(
  const void * a,const void * b) {

  return ((struct tgt_point_search_data*)a)->global_id -
         ((struct tgt_point_search_data*)b)->global_id;
}

static inline int compare_tgt_point_data_flag_LOCAL(
  const void * a,const void * b) {

  enum interp_status_flag status_a =
    ((struct tgt_point_search_data*)a)->status_flag;
  enum interp_status_flag status_b =
    ((struct tgt_point_search_data*)b)->status_flag;

  if (status_a != status_b) {

    return (int)status_a - (int)status_b;

  } else {

    if (status_a == LOCAL) {

      int type_a = (int)(((struct tgt_point_search_data*)a)->type_flag);
      int type_b = (int)(((struct tgt_point_search_data*)b)->type_flag);

      if (type_a != type_b) return type_a - type_b;

      for (int i = 0; i <= type_a; ++i) {

        int src_global_id_a =
          ((struct tgt_point_search_data*)a)->src_point_global_ids[i];
        int src_global_id_b =
          ((struct tgt_point_search_data*)b)->src_point_global_ids[i];

        if (src_global_id_a != src_global_id_b)
          return src_global_id_a - src_global_id_b;
      }
    }

    return (((struct tgt_point_search_data*)a)->global_id -
            ((struct tgt_point_search_data*)b)->global_id);
  }
}

static inline int compare_tgt_point_data_flag_global_id(
  const void * a,const void * b) {

  enum interp_status_flag status_a =
    ((struct tgt_point_search_data*)a)->status_flag;
  enum interp_status_flag status_b =
    ((struct tgt_point_search_data*)b)->status_flag;

  if (status_a != status_b)
    return (int)status_a - (int)status_b;

  return ((struct tgt_point_search_data*)a)->global_id -
         ((struct tgt_point_search_data*)b)->global_id;
}

//! constructor for a hybrid cubic spherical Bernstein-Bezier patch
//! interpolation method
struct interpolation_method * yac_new_interpolation_method_hcsbb_mask(
  struct global_search * search, struct domain * domain,
  char const * stack_name, struct points *** points, struct mask *** masks,
  unsigned num_pointsets, unsigned points_idx, unsigned use_source_mask) {

  // ensure that the required point data is available
  if (num_pointsets == 0)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_hcsbb): missing data points.",
      __FILE__, __LINE__);

  if (points_idx >= num_pointsets)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_hcsbb): invalid points_idx.",
      __FILE__, __LINE__);

  if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
    yac_internal_abort_message(
      "ERROR(new_interpolation_method_hcsbb): no point set data available",
      __FILE__, __LINE__);

  enum yac_location src_location = points[0][points_idx]->location;

  // check that all points in the pointset have the same source location
  {
    unsigned num_subdomains = yac_get_num_subdomains(domain);

    for (unsigned subdomain_idx = 0; subdomain_idx < num_subdomains;
         ++subdomain_idx) {

      if (points[subdomain_idx] == NULL ||
          points[subdomain_idx][points_idx] == NULL)
        yac_internal_abort_message(
          "ERROR(yac_new_interpolation_method_hcsbb_mask): no point set data "
          "available", __FILE__, __LINE__);

      if (src_location != points[subdomain_idx][points_idx]->location)
        yac_internal_abort_message(
          "ERROR(yac_new_interpolation_method_hcsbb_mask): location of "
          "points do not match", __FILE__, __LINE__);
    }
  }

  struct interpolation_method_hcsbb * method_hcsbb =
    xmalloc(1 * sizeof(*method_hcsbb));

  method_hcsbb->vtable = &interpolation_method_hcsbb_vtable;
  method_hcsbb->search = search;
  method_hcsbb->domain = domain;
  method_hcsbb->points = points;
  method_hcsbb->masks = (use_source_mask)?masks:NULL;
  method_hcsbb->num_pointsets = num_pointsets;
  method_hcsbb->src_location = src_location;
  method_hcsbb->finish_callback = NULL;
  method_hcsbb->finish_user_data = NULL;
  method_hcsbb->points_idx = points_idx;
  method_hcsbb->use_source_mask = use_source_mask && (masks != NULL);

  yac_init_interpolation_method(
    (struct interpolation_method *)method_hcsbb, search);

  generate_interpolation_method_hcsbb_tag(method_hcsbb, stack_name);

  char * unique_hcsbb_string;
  {
    char const * str_base_str = "interpolation method hcsbb";
    char const * str_nnn_search = "nnn search";
    char const * source_comp_name = yac_get_local_component_name(search);
    char const * target_comp_name = yac_get_remote_component_name(search);
    char const * source_domain_name = yac_get_local_domain_name(search);
    char const * target_domain_name = yac_get_remote_domain_name(search);
    size_t str_buf_size = strlen(str_nnn_search);
    str_buf_size += strlen(str_base_str);
    str_buf_size += strlen(source_comp_name);
    str_buf_size += strlen(target_comp_name);
    str_buf_size += strlen(source_domain_name);
    str_buf_size += strlen(target_domain_name);
    str_buf_size += strlen(stack_name) + 10 + 2 + 10 + 1; // (+ 10) for points_idx
                                                          // (+ 2)  for mask handling
                                                          // (+ 10) for unique_method_id
                                                          // (+ 1)  for '\0'

    unique_hcsbb_string = xmalloc(str_buf_size);
    generate_tag_string(
      unique_hcsbb_string, str_buf_size, str_base_str, source_comp_name,
      target_comp_name, source_domain_name, target_domain_name, stack_name,
      points_idx, use_source_mask,
      yac_get_unique_id_interpolation_method(
        (struct interpolation_method *)method_hcsbb), str_nnn_search);
  }
  method_hcsbb->nnn_search =
   yac_nnn_search_new(
     domain, unique_hcsbb_string, points, masks, num_pointsets, points_idx);
  free(unique_hcsbb_string);

  return (struct interpolation_method *)method_hcsbb;
}

//! constructor for a hybrid cubic spherical Bernstein-Bezier patch
//! interpolation method
struct interpolation_method * yac_new_interpolation_method_hcsbb(
  struct global_search * search, struct domain * domain,
  char const * stack_name, struct points *** points, unsigned num_pointsets,
  unsigned points_idx) {

  unsigned use_source_mask = 0;

  return
    (struct interpolation_method *)yac_new_interpolation_method_hcsbb_mask(
      search, domain, stack_name, points, NULL, num_pointsets, points_idx,
      use_source_mask);
}

static int compare_src_point_edges(const void * a, const void * b) {

  struct src_point const * edge_a = (struct src_point const *)a;
  struct src_point const * edge_b = (struct src_point const *)b;

  int ret = edge_a[0].global_id - edge_b[0].global_id;
  if (ret) return ret;
  return edge_a[1].global_id - edge_b[1].global_id;
}

static int compare_src_points(const void * a, const void * b) {

  struct src_point const * point_a = (struct src_point const *)a;
  struct src_point const * point_b = (struct src_point const *)b;

  return point_a[0].global_id - point_b[0].global_id;
}

static int compare_3_int(const void * a, const void * b) {

  int const * a_ = (int const *)a;
  int const * b_ = (int const *)b;

  int ret = a_[0] - b_[0];
  if (ret) return ret;
  ret = a_[1] - b_[1];
  if (ret) return ret;
  return a_[2] - b_[2];
}

static void get_unique_interp_data(
  struct tgt_point_search_data * tgt_point_data, size_t num_edges,
  size_t num_triangles,
  int (**unique_edge_global_ids)[2], size_t * num_unique_edges,
  int (**unique_triangle_global_ids)[3], size_t * num_unique_triangles,
  struct src_point ** unique_edge_vertices, size_t * num_unique_edge_vertices) {

  struct src_point * unique_edge_vertices_ =
    xmalloc(2*(num_edges + 3*num_triangles)*sizeof(*unique_edge_vertices_));

  size_t tmp_num_edges = 0;

  struct src_point (*edges)[2] = (struct src_point (*)[2])unique_edge_vertices_;

  // get unique edges from target points that are directly on the source edges
  if (num_edges > 0) {
    int prev_global_ids[2] = {tgt_point_data[0].src_point_global_ids[0]-1, 0};
    for (size_t i = 0; i < num_edges; ++i) {

      struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data + i;
      int * curr_global_ids = &(curr_tgt_point_data->src_point_global_ids[0]);
      if ((prev_global_ids[0] != curr_global_ids[0]) ||
          (prev_global_ids[1] != curr_global_ids[1])) {

        struct src_point * curr_edge =
          (struct src_point *)(edges + tmp_num_edges);
        tmp_num_edges++;
        unsigned subdomain_id = curr_tgt_point_data->subdomain_id;
        unsigned * curr_local_ids =
          &(curr_tgt_point_data->src_point_local_ids[0]);
        unsigned * curr_ranks = &(curr_tgt_point_data->ranks[0]);
        // code local-remote status into subdomain_id while keeping actual
        // subdomain_id
        curr_edge[0].subdomain_id =
          (curr_ranks[0] == -1)?subdomain_id:(-1-subdomain_id);
        curr_edge[0].global_id = curr_global_ids[0];
        curr_edge[0].local_id = curr_local_ids[0];
        // code local-remote status into subdomain_id while keeping actual
        // subdomain_id
        curr_edge[1].subdomain_id =
          (curr_ranks[1] == -1)?subdomain_id:(-1-subdomain_id);
        curr_edge[1].global_id = curr_global_ids[1];
        curr_edge[1].local_id = curr_local_ids[1];
      }
    }
    tgt_point_data += num_edges;
  }

  edges += tmp_num_edges;

  size_t num_unique_edges_ = 0;

  // get edges and unique triangles from target points that are in source point
  // triangles
  {
    int (*triangle_global_ids_)[3] =
      xmalloc(num_triangles * sizeof(*triangle_global_ids_));

    for (size_t i = 0; i < num_triangles; ++i) {
      unsigned subdomain_id = tgt_point_data[i].subdomain_id;
      unsigned * ranks = &(tgt_point_data[i].ranks[0]);
      // code local-remote status into subdomain_id while keeping actual
      // subdomain_id
      unsigned subdomain_ids[3] =
        {(ranks[0] == -1)?subdomain_id:(-1-subdomain_id),
         (ranks[1] == -1)?subdomain_id:(-1-subdomain_id),
         (ranks[2] == -1)?subdomain_id:(-1-subdomain_id)};
      int * global_ids = &(triangle_global_ids_[i][0]);
      global_ids[0] =  tgt_point_data[i].src_point_global_ids[0];
      global_ids[1] =  tgt_point_data[i].src_point_global_ids[1];
      global_ids[2] =  tgt_point_data[i].src_point_global_ids[2];
      unsigned * local_ids = &(tgt_point_data[i].src_point_local_ids[0]);
      edges[3*i+0][0].subdomain_id = subdomain_ids[0];
      edges[3*i+0][0].global_id = global_ids[0];
      edges[3*i+0][0].local_id = local_ids[0];
      edges[3*i+0][1].subdomain_id = subdomain_ids[1];
      edges[3*i+0][1].global_id = global_ids[1];
      edges[3*i+0][1].local_id = local_ids[1];
      edges[3*i+1][0].subdomain_id = subdomain_ids[1];
      edges[3*i+1][0].global_id = global_ids[1];
      edges[3*i+1][0].local_id = local_ids[1];
      edges[3*i+1][1].subdomain_id = subdomain_ids[2];
      edges[3*i+1][1].global_id = global_ids[2];
      edges[3*i+1][1].local_id = local_ids[2];
      edges[3*i+2][0].subdomain_id = subdomain_ids[0];
      edges[3*i+2][0].global_id = global_ids[0];
      edges[3*i+2][0].local_id = local_ids[0];
      edges[3*i+2][1].subdomain_id = subdomain_ids[2];
      edges[3*i+2][1].global_id = global_ids[2];
      edges[3*i+2][1].local_id = local_ids[2];
    }

    tmp_num_edges += 3 * num_triangles;
    edges = (struct src_point (*)[2])unique_edge_vertices_;
    qsort(edges, tmp_num_edges, sizeof(*edges), compare_src_point_edges);
    qsort(triangle_global_ids_, num_triangles,
          sizeof(*triangle_global_ids_), compare_3_int);

    size_t num_unique_triangles_ = 0;

    // remove duplicated entries from triangle_global_ids_
    if (num_triangles > 0) {
      int prev_global_ids[3] = {triangle_global_ids_[0][0]-1,0,0};
      for (size_t i = 0; i < num_triangles; ++i) {

        int * curr_global_ids = &(triangle_global_ids_[i][0]);
        if ((prev_global_ids[0] != curr_global_ids[0]) ||
            (prev_global_ids[1] != curr_global_ids[1]) ||
            (prev_global_ids[2] != curr_global_ids[2])) {

          prev_global_ids[0] = curr_global_ids[0];
          prev_global_ids[1] = curr_global_ids[1];
          prev_global_ids[2] = curr_global_ids[2];
          triangle_global_ids_[num_unique_triangles_][0] = curr_global_ids[0];
          triangle_global_ids_[num_unique_triangles_][1] = curr_global_ids[1];
          triangle_global_ids_[num_unique_triangles_][2] = curr_global_ids[2];
          num_unique_triangles_++;
        }
      }
    }

    *unique_triangle_global_ids =
      xrealloc(triangle_global_ids_,
               num_unique_triangles_ * sizeof(*triangle_global_ids_));
    *num_unique_triangles = num_unique_triangles_;

    // remove duplicated entries from edges
    if (tmp_num_edges > 0) {
      int prev_global_ids[2] = {edges[0][0].global_id-1,0};
      for (size_t i = 0; i < tmp_num_edges; ++i) {

        int curr_global_ids[2] = {edges[i][0].global_id, edges[i][1].global_id};
        if ((prev_global_ids[0] != curr_global_ids[0]) ||
            (prev_global_ids[1] != curr_global_ids[1])) {

          prev_global_ids[0] = curr_global_ids[0];
          prev_global_ids[1] = curr_global_ids[1];
          edges[num_unique_edges_][0] = edges[i][0];
          edges[num_unique_edges_][1] = edges[i][1];
          num_unique_edges_++;
        }
      }
    }
  }

  int (*edge_global_ids_)[2] =
    xmalloc(num_unique_edges_ * sizeof(*edge_global_ids_));
  for (size_t i = 0; i < num_unique_edges_; ++i) {
    edge_global_ids_[i][0] = edges[i][0].global_id;
    edge_global_ids_[i][1] = edges[i][1].global_id;
  }

  *unique_edge_global_ids = edge_global_ids_;
  *num_unique_edges = num_unique_edges_;

  // get the unique edge vertices
  size_t num_unique_edge_vertices_ = 0;
  if (num_unique_edges_ > 0) {

    qsort(unique_edge_vertices_, 2 * num_unique_edges_,
          sizeof(*unique_edge_vertices_), compare_src_points);
    int prev_global_id = unique_edge_vertices_[0].global_id - 1;

    for (size_t i = 0; i < 2 * num_unique_edges_; ++i) {

      int curr_global_id = unique_edge_vertices_[i].global_id;
      if (prev_global_id != curr_global_id) {

        prev_global_id = curr_global_id;
        unique_edge_vertices_[num_unique_edge_vertices_++] =
          unique_edge_vertices_[i];
      }
    }
  }

  *unique_edge_vertices = 
    xrealloc(unique_edge_vertices_,
             num_unique_edge_vertices_ * sizeof(*unique_edge_vertices_));
  *num_unique_edge_vertices = num_unique_edge_vertices_;
}

static void insertion_sort_data_request_point(
  struct data_request_point * a, size_t * curr_a_length,
  struct data_request_point * b, size_t b_length) {

  size_t tmp_a_length = *curr_a_length;
  size_t num_elements_to_copy = 0;

  // count the number of unique elements that need to be copied from b to a
  for (size_t i = 0, j = 0; i < b_length;) {

    struct data_request_point curr_b = b[i++];

    // try to find curr_b in a
    while ((j < tmp_a_length) && (a[j].global_id < curr_b.global_id)) ++j;
    // skip all b that are equal to curr_b
    while ((i < b_length) && (b[i].global_id == curr_b.global_id)) ++i;

    if ((j >= tmp_a_length) || (a[j].global_id != curr_b.global_id)) {
      b[num_elements_to_copy++] = curr_b;
    }
  }

  *curr_a_length += num_elements_to_copy;

  // copy elements from back to front
  size_t i = num_elements_to_copy - 1,
         j = tmp_a_length - 1,
         k = tmp_a_length + num_elements_to_copy - 1;

  while ((i < num_elements_to_copy) && (j < tmp_a_length))
    a[k--] = (a[j].global_id > b[i].global_id)?a[j--]:b[i--];

  if (i < num_elements_to_copy) memcpy(a, b, (i+1)*sizeof(*a));
}

static int compare_edge_pos_global_id(void const * a, void const * b) {

  struct edge_pos const * edge_a = (struct edge_pos const *)a;
  struct edge_pos const * edge_b = (struct edge_pos const *)b;

  int ret = edge_a->global_ids[0] - edge_b->global_ids[0];
  if (ret) return ret;
  return edge_a->global_ids[1] - edge_b->global_ids[1];
}

static int compare_edge_pos_pos(void const * a, void const * b) {

  return ((struct edge_pos const *)a)->orig_pos -
         ((struct edge_pos const *)b)->orig_pos;
}

static void init_basic_patch_data(
  struct basic_patch_data * patch_data,
  struct tgt_point_search_data * tgt_point_data, size_t num_tgt_points,
  double (**required_points_xyz)[3], size_t * num_required_points,
  struct grid ** point_grids, unsigned num_subdomains) {

  // skip all target points that are directly on a source point
  while (num_tgt_points && (tgt_point_data->type_flag == ON_VERTEX)) {
    tgt_point_data++;
    num_tgt_points--;
  }

  size_t num_edges = 0, num_triangles;
  // count number of target points that are on an edge of a source point
  // triangle
  while ((num_edges != num_tgt_points) &&
         (tgt_point_data[num_edges].type_flag == ON_EDGE)) ++num_edges;
  num_triangles = num_tgt_points - num_edges;

  // get all edges and triangles required for the interpolation
  // (each triangle requires three edges, usually each triangle edges is shared
  //  between two triangles)
  int (*unique_edge_global_ids)[2];
  int (*unique_triangle_global_ids)[3];
  struct src_point * unique_edge_vertices;
  size_t num_unique_edges, num_unique_triangles, num_unique_edge_vertices;
  get_unique_interp_data(
    tgt_point_data, num_edges, num_triangles,
    &unique_edge_global_ids, &num_unique_edges,
    &unique_triangle_global_ids, &num_unique_triangles,
    &unique_edge_vertices, &num_unique_edge_vertices);

  double (*required_points_xyz_)[3] =
    xmalloc((num_unique_edge_vertices + num_unique_edges) *
            sizeof(*required_points_xyz_));
  *required_points_xyz = required_points_xyz_;
  *num_required_points = num_unique_edge_vertices + num_unique_edges;
  double (*required_vertices_xyz)[3] = required_points_xyz_;

  // we need to estimate derivatives at all unique edge vertices
  size_t d_data_src_point_count = num_unique_edge_vertices;
  struct derivative_data_src_point * d_data_src_point =
    xmalloc(d_data_src_point_count * sizeof(*d_data_src_point));
  patch_data->d_data_src_point = d_data_src_point;
  patch_data->d_data_src_point_count = d_data_src_point_count;

  // get coordinates of all unique edge vertices for the computation of
  // the derivative data
  for (size_t i = 0; i < num_unique_edge_vertices; ++i) {
    unsigned subdomain_id = unique_edge_vertices[i].subdomain_id;
    // if the current source point is on a remote process
    if (subdomain_id >= num_subdomains) {
      subdomain_id = - 1 - subdomain_id;
      // set subdomain_id that is used for all remote source points
      unique_edge_vertices[i].subdomain_id = -1;
    }
    struct grid * point_grid = point_grids[subdomain_id];
    double coordinate_x =
      yac_get_corner_x_coord(point_grid, unique_edge_vertices[i].local_id);
    double coordinate_y =
      yac_get_corner_y_coord(point_grid, unique_edge_vertices[i].local_id);
    LLtoXYZ(coordinate_x, coordinate_y, required_points_xyz_[i]);
    d_data_src_point[i].coordinate_xyz[0] = required_points_xyz_[i][0];
    d_data_src_point[i].coordinate_xyz[1] = required_points_xyz_[i][1];
    d_data_src_point[i].coordinate_xyz[2] = required_points_xyz_[i][2];
    d_data_src_point[i].gauss_nnn_patch = NULL;
    d_data_src_point[i].src_point = unique_edge_vertices[i];
  }
  required_points_xyz_ += num_unique_edge_vertices;

  // we need to estimate derivatives at all the middle points of each edge
  size_t d_data_middle_point_count = num_unique_edges;
  struct derivative_data_middle_point * d_data_middle_point =
    xmalloc(d_data_middle_point_count * sizeof(*d_data_middle_point));
  patch_data->d_data_middle_point = d_data_middle_point;
  patch_data->d_data_middle_point_count = d_data_middle_point_count;

  // add middle points of all unique edges and generate links between
  // edge_interp_data and derivative_data of its vertices
  patch_data->edges = xmalloc(num_unique_edges * sizeof(*(patch_data->edges)));
  patch_data->num_edges = num_unique_edges;
  {
    int * global_ids = xmalloc(2 * 1024 * sizeof(*global_ids));
    int * orig_pos = global_ids + 1024;
    for (size_t i = 0; i < 1024; ++i) orig_pos[i] = (int)i;

    struct edge_interp_data * edge_data = patch_data->edges;

    for (size_t i = 0; i < (num_unique_edges + 511) / 512; ++i) {

      size_t curr_num_edges = MIN(512, num_unique_edges - i * 512);
      int * curr_edge_global_ids = (int*)(unique_edge_global_ids + 512 * i);
      struct derivative_data_middle_point * curr_d_data_middle_point =
        d_data_middle_point + i * 512;

      // get global ids of the vertices of the current batch of edges
      memcpy(global_ids, curr_edge_global_ids,
             2 * curr_num_edges * sizeof(*global_ids));

      // sort them by global ids
      yac_quicksort_index(global_ids, 2 * curr_num_edges, orig_pos);

      // determine their index in the list of unique_edge_vertices
      for (size_t j = 0, k = 0; j < 2 * curr_num_edges; ++j) {
        int global_id = global_ids[j];
        while (unique_edge_vertices[k].global_id != global_id) ++k;
        global_ids[j] = (int)k;
      }

      // bring the global_ids array (now containing the vertex indices) into
      // its original order
      yac_quicksort_index(orig_pos, 2 * curr_num_edges, global_ids);

      for (size_t j = 0; j < curr_num_edges; ++j) {

        size_t curr_vertex_indices[2] =
          {(size_t)(global_ids[2*j+0]), (size_t)(global_ids[2*j+1])};

        edge_data[j].vertex_d_data[0] =
          d_data_src_point + curr_vertex_indices[0];
        edge_data[j].vertex_d_data[1] =
          d_data_src_point + curr_vertex_indices[1];

        double * vertex_coordinates_xyz[2] = {
          required_vertices_xyz[curr_vertex_indices[0]],
          required_vertices_xyz[curr_vertex_indices[1]]};
        double * middle_point = required_points_xyz_[j];
        middle_point[0] =
          vertex_coordinates_xyz[0][0] + vertex_coordinates_xyz[1][0];
        middle_point[1] =
          vertex_coordinates_xyz[0][1] + vertex_coordinates_xyz[1][1];
        middle_point[2] =
          vertex_coordinates_xyz[0][2] + vertex_coordinates_xyz[1][2];
        double sb_coord =
          1.0 / sqrt(middle_point[0] * middle_point[0] +
                     middle_point[1] * middle_point[1] +
                     middle_point[2] * middle_point[2]);
        middle_point[0] *= sb_coord;
        middle_point[1] *= sb_coord;
        middle_point[2] *= sb_coord;
        curr_d_data_middle_point[j].coordinate_xyz[0] = middle_point[0];
        curr_d_data_middle_point[j].coordinate_xyz[1] = middle_point[1];
        curr_d_data_middle_point[j].coordinate_xyz[2] = middle_point[2];
        curr_d_data_middle_point[j].gauss_nnn_patch = NULL;
        edge_data[j].middle_d_data = curr_d_data_middle_point + j;
        edge_data[j].sb_coord_middle_point = sb_coord;
        crossproduct_d(vertex_coordinates_xyz[0], vertex_coordinates_xyz[1],
                       &(edge_data[j].g[0]));
      }

      edge_data += curr_num_edges;
      required_points_xyz_ += curr_num_edges;
    }
    free(global_ids);
  }

  // generate links between the triangles and its edges
  struct triangle_interp_data * triangle_data =
    xmalloc(num_unique_triangles * sizeof(*triangle_data));
  patch_data->triangles = triangle_data;
  patch_data->num_triangles = num_unique_triangles;
  {
    struct edge_pos * tmp_edges = xmalloc(3 * 256 * sizeof(*tmp_edges));
    for (size_t i = 0; i < 3 * 256; ++i) tmp_edges[i].orig_pos = i;

    struct edge_interp_data * edge_data = patch_data->edges;

    for (size_t i = 0; i < (num_unique_triangles + 255)/256; ++i) {

      size_t curr_num_triangles = MIN(256, num_unique_triangles - i * 256);
      int (*curr_triangle_global_ids)[3] = unique_triangle_global_ids + i * 256;

      for (size_t j = 0; j < curr_num_triangles; ++j) {

        tmp_edges[3*j+0].global_ids[0] = curr_triangle_global_ids[j][1];
        tmp_edges[3*j+0].global_ids[1] = curr_triangle_global_ids[j][2];
        tmp_edges[3*j+1].global_ids[0] = curr_triangle_global_ids[j][0];
        tmp_edges[3*j+1].global_ids[1] = curr_triangle_global_ids[j][2];
        tmp_edges[3*j+2].global_ids[0] = curr_triangle_global_ids[j][0];
        tmp_edges[3*j+2].global_ids[1] = curr_triangle_global_ids[j][1];
      }

      qsort(tmp_edges, 3 * curr_num_triangles, sizeof(*tmp_edges),
            compare_edge_pos_global_id);

      for (size_t j = 0, k = 0; j < 3 * curr_num_triangles; ++j) {

        int global_ids[2] =
          {tmp_edges[j].global_ids[0], tmp_edges[j].global_ids[1]};
        while ((unique_edge_global_ids[k][0] != global_ids[0]) ||
               (unique_edge_global_ids[k][1] != global_ids[1])) ++k;
        tmp_edges[j].global_ids[0] = (int)k;
      }

      qsort(tmp_edges, 3 * curr_num_triangles, sizeof(*tmp_edges),
            compare_edge_pos_pos);

      for (size_t j = 0; j < curr_num_triangles; ++j) {

        triangle_data[j].edges[0] = edge_data + tmp_edges[3*j+0].global_ids[0];
        triangle_data[j].edges[1] = edge_data + tmp_edges[3*j+1].global_ids[0];
        triangle_data[j].edges[2] = edge_data + tmp_edges[3*j+2].global_ids[0];
      }

      triangle_data += curr_num_triangles;
    }
    free(tmp_edges);
  }

  // link tgt_point_data to the edge_data and triangle_data
  {
    struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data;
    struct edge_interp_data * edge_data = patch_data->edges;

    for (size_t i = 0, j = 0; i < num_edges; ++i) {

      int * curr_global_ids = curr_tgt_point_data[i].src_point_global_ids;
      while ((unique_edge_global_ids[j][0] != curr_global_ids[0]) ||
             (unique_edge_global_ids[j][1] != curr_global_ids[1])) ++j;

       curr_tgt_point_data[i].interp_data.edge = edge_data + j;
    }

    curr_tgt_point_data += num_edges;
    triangle_data = patch_data->triangles;

    for (size_t i = 0, j = 0; i < num_triangles; ++i) {

      int * curr_global_ids = curr_tgt_point_data[i].src_point_global_ids;
      while ((unique_triangle_global_ids[j][0] != curr_global_ids[0]) ||
             (unique_triangle_global_ids[j][1] != curr_global_ids[1]) ||
             (unique_triangle_global_ids[j][2] != curr_global_ids[2])) ++j;

       curr_tgt_point_data[i].interp_data.triangle = triangle_data + j;
    }
  }

  free(unique_edge_global_ids);
  free(unique_triangle_global_ids);
  free(unique_edge_vertices);
}

static int compare_data_request_points(const void * a, const void * b) {

  return ((struct data_request_point const *)a)->global_id -
         ((struct data_request_point const *)b)->global_id;
}

/**
 * analyses the results of the search and generates a data_request object, which
 * is required by the field_interpolation in order to be able to provide the
 * remote points needed for the local interpolation
 */
static void generate_data_request(
  struct intermediate_method_data * data, size_t num_interpolated_points) {

  struct tgt_point_search_data * tgt_point_data =
    data->search_data.tgt_point_data;
  struct basic_patch_data * patch_data = &(data->patch_data);
  struct gauss_nnn_patch * gauss_nnn_patches = patch_data->gauss_nnn_patches;
  size_t num_gauss_nnn_patches = patch_data->num_gauss_nnn_patches;
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb *)(data->method_stack[0]);
  size_t num_pointsets = method_hcsbb->num_pointsets;
  struct interpolation_method_data_request * data_request =
    xmalloc(num_pointsets * sizeof(*data_request));
  method_hcsbb->data_request[0] = data_request;

  for (size_t i = 0; i < num_pointsets; ++i) {
    data_request[i].num_values = 0;
    data_request[i].global_ids = NULL;
    data_request[i].ranks = NULL;
  }

  // select the data request that is actually required
  data_request += method_hcsbb->points_idx;

  struct data_request_point * data_request_points = NULL;
  size_t data_request_points_array_size = 0;
  size_t num_requested_points = 0;

  struct data_request_point * tmp_data_request_points =
    xmalloc(MAX(1024, 16 * HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN) *
            sizeof(*tmp_data_request_points));

  // get the remote src points that are directly associated to the respective
  // target points
  for (size_t i = 0; i < (num_interpolated_points + 255)/256; ++i) {

    size_t curr_num_tgt_points = MIN(256, num_interpolated_points - i * 256);
    size_t tmp_num_requested_points = 0;

    for (size_t j = 0; j < curr_num_tgt_points; ++j) {

      struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data + j;
      size_t num_src_points = (size_t)(curr_tgt_point_data->type_flag) + 1;

      for (size_t k = 0; k < num_src_points; ++k) {

        unsigned rank = curr_tgt_point_data->ranks[k];
        // if this is a remote point
        if (rank != (unsigned)-1) {

          tmp_data_request_points[tmp_num_requested_points].global_id =
            curr_tgt_point_data->src_point_global_ids[k];
          tmp_data_request_points[tmp_num_requested_points].rank = rank;
          tmp_num_requested_points++;
        }
      }
    }
    tgt_point_data += 256;

    if (tmp_num_requested_points > 0) {
      ENSURE_ARRAY_SIZE(
        data_request_points, data_request_points_array_size,
        num_requested_points + tmp_num_requested_points);

      qsort(tmp_data_request_points, tmp_num_requested_points,
            sizeof(*tmp_data_request_points), compare_data_request_points);
      insertion_sort_data_request_point(
        data_request_points, &num_requested_points,
        &(tmp_data_request_points[0]), tmp_num_requested_points);
    }
  }

  // get the remote src points that are indirectly associated to the respective
  // target points (required for the computation of the derivatives)
  for (size_t i = 0; i < (num_gauss_nnn_patches + 15)/16; ++i) {

    size_t curr_num_gauss_nnn_patches =
      MIN(16, num_gauss_nnn_patches - i * 16);
    size_t tmp_num_requested_points = 0;

    for (size_t j = 0; j < curr_num_gauss_nnn_patches; ++j) {

      struct gauss_nnn_patch * curr_gauss_nnn_patches = gauss_nnn_patches + j;
      struct src_point * curr_src_points = curr_gauss_nnn_patches->src_points;
      unsigned * curr_src_point_ranks = curr_gauss_nnn_patches->src_point_ranks;
      size_t curr_num_src_points = curr_gauss_nnn_patches->num_src_points;

      for (size_t k = 0; k < curr_num_src_points; ++k) {
        unsigned rank = curr_src_point_ranks[k];
        if (rank != (unsigned)-1) {
          tmp_data_request_points[tmp_num_requested_points].global_id =
            curr_src_points[k].global_id;
          tmp_data_request_points[tmp_num_requested_points].rank = rank;
          tmp_num_requested_points++;
        }
      }
    }

    gauss_nnn_patches += 16;

    if (tmp_num_requested_points > 0) {
      ENSURE_ARRAY_SIZE(
        data_request_points, data_request_points_array_size,
        num_requested_points + tmp_num_requested_points);

      qsort(tmp_data_request_points, tmp_num_requested_points,
            sizeof(*tmp_data_request_points), compare_data_request_points);
      insertion_sort_data_request_point(
        data_request_points, &num_requested_points,
        tmp_data_request_points, tmp_num_requested_points);
    }
  }

  free(tmp_data_request_points);

  data_request->num_values = (unsigned)num_requested_points;
  int * global_ids =
    xmalloc(num_requested_points * sizeof(*(data_request->global_ids)));
  int * ranks =
    xmalloc(num_requested_points * sizeof(*(data_request->ranks)));
  data_request->global_ids = global_ids;
  data_request->ranks = ranks;

  for (size_t i = 0; i < num_requested_points; ++i) {

    global_ids[i] = data_request_points[i].global_id;
    ranks[i] = (int)(data_request_points[i].rank);
  }

  free(data_request_points);
}

static struct interpolation_plan_element * generate_interpolation_plan_1nn(
  struct intermediate_method_data * data, size_t num_interpolated_points) {

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(data->method_stack[0]);
  unsigned points_idx = method_hcsbb->points_idx;
  struct tgt_point_search_data * tgt_point_data =
    data->search_data.tgt_point_data;
  unsigned num_pointsets = method_hcsbb->num_pointsets;

  unsigned * src_indices =
    xmalloc(3 * num_interpolated_points * sizeof(*src_indices));

  for (size_t i = 0; i < num_interpolated_points; ++i) {

    // for remote points subdomain_id is -1
    src_indices[3*i+0] = tgt_point_data[i].subdomain_id+1;
    src_indices[3*i+1] = points_idx;
    src_indices[3*i+2] = tgt_point_data[i].src_point_local_ids[0];
  }

  unsigned * num_remote_points =
     xcalloc(num_pointsets, sizeof(*num_remote_points));
  num_remote_points[points_idx] =
    method_hcsbb->data_request[0][points_idx].num_values;

  // generate a plan to execute the interpolation
  struct interpolation_plan_element * plan = xmalloc(1 * sizeof(*plan));

  yac_init_interpolation_plan_element_point(
    plan, num_interpolated_points, src_indices, num_remote_points,
    num_pointsets);

  free(num_remote_points);
  free(src_indices);

  return plan;
}

static struct interpolation_plan_element * generate_interpolation_plan_hcsbb(
  struct intermediate_method_data * data, size_t num_interpolated_points) {

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(data->method_stack[0]);
  unsigned points_idx = method_hcsbb->points_idx;
  unsigned num_pointsets = method_hcsbb->num_pointsets;
  struct weight_vector_data * weight_vector_data = data->search_data.weights;
  unsigned * num_weights_per_tgt = data->search_data.num_weights_per_tgt;
  size_t total_num_weights = data->search_data.total_num_weights;

  unsigned * src_indices = xmalloc(3 * total_num_weights * sizeof(*src_indices));
  double * weights = xmalloc(total_num_weights * sizeof(*weights));

  for (size_t i = 0; i < total_num_weights; ++i) {

    // for remote points subdomain_id is -1
    src_indices[3*i+0] = weight_vector_data[i].subdomain_id + 1;
    src_indices[3*i+1] = points_idx;
    src_indices[3*i+2] = weight_vector_data[i].local_id;
    weights[i] = weight_vector_data[i].weight;
  }

  unsigned * num_remote_points =
    xcalloc(num_pointsets, sizeof(*num_remote_points));
  num_remote_points[points_idx] =
    method_hcsbb->data_request[0][points_idx].num_values;

  // generate a plan to execute the interpolation
  struct interpolation_plan_element * plan = xmalloc(1 * sizeof(*plan));

  yac_init_interpolation_plan_element_wpointsum(
    plan, num_interpolated_points, src_indices, num_weights_per_tgt, weights,
    num_remote_points, num_pointsets);

  free(num_remote_points);
  free(weights);
  free(src_indices);

  return plan;
}

static struct interpolation_plan_element * generate_interpolation_plan(
  struct intermediate_method_data * data, size_t num_interpolated_points) {

  struct tgt_point_search_data * tgt_point_data =
    data->search_data.tgt_point_data;

  // if there is nothing to interpolate
  if (num_interpolated_points == 0) return NULL;

  // check whether there are target points with interp_type_flag != ON_VERTEX
  size_t i = 0;
  while ((i < num_interpolated_points) &&
         (tgt_point_data[i].type_flag == ON_VERTEX)) ++i;

  // if there is only a one to one mapping
  if (i == num_interpolated_points)
    return generate_interpolation_plan_1nn(data, num_interpolated_points);
  else
    return generate_interpolation_plan_hcsbb(data, num_interpolated_points);
}

static void compute_hcsbb_weights_vertex(
  struct tgt_point_search_data * tgt_point_data,
  struct weight_vector_data * weights, size_t * n) {

  weights[0].weight = 1.0;
  weights[0].subdomain_id =
    (tgt_point_data->ranks[0] == -1)?tgt_point_data->subdomain_id:-1;
  weights[0].local_id = tgt_point_data->src_point_local_ids[0];
  *n = 1;
}

// compute the spherical barycentric coordinates of the given vertices with
// respect to the given triangle
static void compute_sb_coords(
  double * sb_coords, unsigned num_vertices, double triangle[3][3]) {

  double A[3][3];
  lapack_int n = 3, nrhs = (lapack_int) num_vertices, lda = n, ldx = n, ipiv[3], info;
  memcpy(A, triangle, sizeof(A));

  // for a vertex v the spherical barycentric coordinates b are defined as
  // follows
  // A * b = v
  // where: A is the matrix consisting of the vertex coordinates of the three
  // corners of the triangle
  // we compute b by solving this linear system using LAPACK
  if (0 < (info = LAPACKE_dgesv(LAPACK_COL_MAJOR, n, nrhs, &A[0][0], lda,
                                ipiv, sb_coords, ldx)))
    yac_internal_abort_message(
      "ERROR: internal error (could not solve linear 3x3 system)\n",
      __FILE__, __LINE__);
}


// computes the spherical Bernstein basis polynomials
// for degree = 3 the polynomials B_ijk are ordered in the following order:
// B_003, B_012, B_021, B_030, B_102, B_111, B_120, B_201, B_210, B_300
static void compute_sbb_polynomials_gauss_3d(
  double sb_coords[HCSBB_NUM_GAUSS_POINTS][3],
  double sbb_polynomials[HCSBB_NUM_SBB_COEFFS][HCSBB_NUM_GAUSS_POINTS]) {

#define POW_0(x) (1.0)
#define POW_1(x) ((x))
#define POW_2(x) ((x)*(x))
#define POW_3(x) ((x)*(x)*(x))
#define FAC_0 (1.0)
#define FAC_1 (1.0)
#define FAC_2 (2.0)
#define FAC_3 (6.0)

  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[0][vertex_idx] =
      (FAC_3 / (FAC_0 * FAC_0 * FAC_3)) * (POW_0(sb_coords[vertex_idx][0]) *
                                           POW_0(sb_coords[vertex_idx][1]) *
                                           POW_3(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[1][vertex_idx] =
      (FAC_3 / (FAC_0 * FAC_1 * FAC_2)) * (POW_0(sb_coords[vertex_idx][0]) *
                                           POW_1(sb_coords[vertex_idx][1]) *
                                           POW_2(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[2][vertex_idx] =
      (FAC_3 / (FAC_0 * FAC_2 * FAC_1)) * (POW_0(sb_coords[vertex_idx][0]) *
                                           POW_2(sb_coords[vertex_idx][1]) *
                                           POW_1(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[3][vertex_idx] =
      (FAC_3 / (FAC_0 * FAC_3 * FAC_0)) * (POW_0(sb_coords[vertex_idx][0]) *
                                           POW_3(sb_coords[vertex_idx][1]) *
                                           POW_0(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[4][vertex_idx] =
      (FAC_3 / (FAC_1 * FAC_0 * FAC_2)) * (POW_1(sb_coords[vertex_idx][0]) *
                                           POW_0(sb_coords[vertex_idx][1]) *
                                           POW_2(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[5][vertex_idx] =
      (FAC_3 / (FAC_1 * FAC_1 * FAC_1)) * (POW_1(sb_coords[vertex_idx][0]) *
                                           POW_1(sb_coords[vertex_idx][1]) *
                                           POW_1(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[6][vertex_idx] =
      (FAC_3 / (FAC_1 * FAC_2 * FAC_0)) * (POW_1(sb_coords[vertex_idx][0]) *
                                           POW_2(sb_coords[vertex_idx][1]) *
                                           POW_0(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[7][vertex_idx] =
      (FAC_3 / (FAC_2 * FAC_0 * FAC_1)) * (POW_2(sb_coords[vertex_idx][0]) *
                                           POW_0(sb_coords[vertex_idx][1]) *
                                           POW_1(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[8][vertex_idx] =
      (FAC_3 / (FAC_2 * FAC_1 * FAC_0)) * (POW_2(sb_coords[vertex_idx][0]) *
                                           POW_1(sb_coords[vertex_idx][1]) *
                                           POW_0(sb_coords[vertex_idx][2]));
  for (size_t vertex_idx = 0; vertex_idx < HCSBB_NUM_GAUSS_POINTS; ++vertex_idx)
    sbb_polynomials[9][vertex_idx] =
      (FAC_3 / (FAC_3 * FAC_0 * FAC_0)) * (POW_3(sb_coords[vertex_idx][0]) *
                                           POW_0(sb_coords[vertex_idx][1]) *
                                           POW_0(sb_coords[vertex_idx][2]));

#undef POW_0
#undef POW_1
#undef POW_2
#undef POW_3
#undef FAC_0
#undef FAC_1
#undef FAC_2
#undef FAC_3
}

// computes the directional derivatives of the spherical Bernstein basis
// polynomials for degree = 3 in the given direction
// the polynomials B_ijk are ordered in the following order:
// B_003, B_012, B_021, B_030, B_102, B_111, B_120, B_201, B_210, B_300
static void compute_d_sbb_polynomials_3d(
  double bnd_triangle[3][3], double direction[3], double coordinate_xyz[3],
  double d_sbb_polynomials[HCSBB_NUM_SBB_COEFFS]) {

  double sb_coord_p[3] = {
      coordinate_xyz[0], coordinate_xyz[1], coordinate_xyz[2]};
  double sb_coord_d[3] = {direction[0], direction[1], direction[2]};

  // compute the spherical barycentric coordinates of the point and the
  // direction
  compute_sb_coords(&(sb_coord_p[0]), 1, bnd_triangle);
  compute_sb_coords(&(sb_coord_d[0]), 1, bnd_triangle);

#define POW_0(x) (1.0)
#define POW_1(x) ((x))
#define POW_2(x) ((x)*(x))
#define POW_3(x) ((x)*(x)*(x))
#define FAC_0 (1.0)
#define FAC_1 (1.0)
#define FAC_2 (2.0)
#define FAC_3 (6.0)

  d_sbb_polynomials[0] =
    // sb_coord_d[0] * (0.0 * FAC_3 / (FAC_0 * FAC_0 * FAC_3)) *
    // (POW_0(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_3(sb_coord_p[2])) +
    // sb_coord_d[1] * (0.0 * FAC_3 / (FAC_0 * FAC_0 * FAC_3)) *
    // (POW_0(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_3(sb_coord_p[2])) +
    sb_coord_d[2] * (3.0 * FAC_3 / (FAC_0 * FAC_0 * FAC_3)) *
    (POW_0(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_2(sb_coord_p[2]));
  d_sbb_polynomials[1] =
    // sb_coord_d[0] * (0.0 * FAC_3 / (FAC_0 * FAC_1 * FAC_2)) *
    // (POW_0(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_2(sb_coord_p[2])) +
    sb_coord_d[1] * (1.0 * FAC_3 / (FAC_0 * FAC_1 * FAC_2)) *
    (POW_0(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_2(sb_coord_p[2])) +
    sb_coord_d[2] * (2.0 * FAC_3 / (FAC_0 * FAC_1 * FAC_2)) *
    (POW_0(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_1(sb_coord_p[2]));
  d_sbb_polynomials[2] =
    // sb_coord_d[0] * (0.0 * FAC_3 / (FAC_0 * FAC_2 * FAC_1)) *
    // (POW_0(sb_coord_p[0]) * POW_2(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    sb_coord_d[1] * (2.0 * FAC_3 / (FAC_0 * FAC_2 * FAC_1)) *
    (POW_0(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    sb_coord_d[2] * (1.0 * FAC_3 / (FAC_0 * FAC_2 * FAC_1)) *
    (POW_0(sb_coord_p[0]) * POW_2(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[3] =
    // sb_coord_d[0] * (0.0 * FAC_3 / (FAC_0 * FAC_3 * FAC_0)) *
    // (POW_0(sb_coord_p[0]) * POW_3(sb_coord_p[1]) * POW_0(sb_coord_p[2])) +
    sb_coord_d[1] * (3.0 * FAC_3 / (FAC_0 * FAC_3 * FAC_0)) *
    (POW_0(sb_coord_p[0]) * POW_2(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
    // sb_coord_d[2] * (0.0 * FAC_3 / (FAC_0 * FAC_3 * FAC_0)) *
    // (POW_0(sb_coord_p[0]) * POW_3(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[4] =
    sb_coord_d[0] * (1.0 * FAC_3 / (FAC_1 * FAC_0 * FAC_2)) *
    (POW_0(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_2(sb_coord_p[2])) +
    // sb_coord_d[1] * (0.0 * FAC_3 / (FAC_1 * FAC_0 * FAC_2)) *
    // (POW_1(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_2(sb_coord_p[2])) +
    sb_coord_d[2] * (2.0 * FAC_3 / (FAC_1 * FAC_0 * FAC_2)) *
    (POW_1(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_1(sb_coord_p[2]));
  d_sbb_polynomials[5] =
    sb_coord_d[0] * (1.0 * FAC_3 / (FAC_1 * FAC_1 * FAC_1)) *
    (POW_0(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    sb_coord_d[1] * (1.0 * FAC_3 / (FAC_1 * FAC_1 * FAC_1)) *
    (POW_1(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    sb_coord_d[2] * (1.0 * FAC_3 / (FAC_1 * FAC_1 * FAC_1)) *
    (POW_1(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[6] =
    sb_coord_d[0] * (1.0 * FAC_3 / (FAC_1 * FAC_2 * FAC_0)) *
    (POW_0(sb_coord_p[0]) * POW_2(sb_coord_p[1]) * POW_0(sb_coord_p[2])) +
    sb_coord_d[1] * (2.0 * FAC_3 / (FAC_1 * FAC_2 * FAC_0)) *
    (POW_1(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
    // sb_coord_d[2] * (0.0 * FAC_3 / (FAC_1 * FAC_2 * FAC_0)) *
    // (POW_1(sb_coord_p[0]) * POW_2(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[7] =
    sb_coord_d[0] * (2.0 * FAC_3 / (FAC_2 * FAC_0 * FAC_1)) *
    (POW_1(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    // sb_coord_d[1] * (0.0 * FAC_3 / (FAC_2 * FAC_0 * FAC_1)) *
    // (POW_2(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_1(sb_coord_p[2])) +
    sb_coord_d[2] * (1.0 * FAC_3 / (FAC_2 * FAC_0 * FAC_1)) *
    (POW_2(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[8] =
    sb_coord_d[0] * (2.0 * FAC_3 / (FAC_2 * FAC_1 * FAC_0)) *
    (POW_1(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_0(sb_coord_p[2])) +
    sb_coord_d[1] * (1.0 * FAC_3 / (FAC_2 * FAC_1 * FAC_0)) *
    (POW_2(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
    // sb_coord_d[2] * (0.0 * FAC_3 / (FAC_2 * FAC_1 * FAC_0)) *
    // (POW_2(sb_coord_p[0]) * POW_1(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
  d_sbb_polynomials[9] =
    sb_coord_d[0] * (3.0 * FAC_3 / (FAC_3 * FAC_0 * FAC_0)) *
    (POW_2(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_0(sb_coord_p[2]));
    // sb_coord_d[1] * (0.0 * FAC_3 / (FAC_3 * FAC_0 * FAC_0)) *
    // (POW_3(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_0(sb_coord_p[2])) +
    // sb_coord_d[2] * (0.0 * FAC_3 / (FAC_3 * FAC_0 * FAC_0)) *
    // (POW_3(sb_coord_p[0]) * POW_0(sb_coord_p[1]) * POW_0(sb_coord_p[2]));

#undef POW_0
#undef POW_1
#undef POW_2
#undef POW_3
#undef FAC_0
#undef FAC_1
#undef FAC_2
#undef FAC_3
}

static void get_derivative_weights(
  struct gauss_nnn_patch * gauss_nnn_patch, double coordinate_xyz[3],
  double direction[3], struct weight_vector_data * weights,
  double mult_factor) {

  // compute w_p
  double d_sbb_vertex_polynomials[HCSBB_NUM_SBB_COEFFS];
  compute_d_sbb_polynomials_3d(
    gauss_nnn_patch->bnd_triangle, direction, coordinate_xyz,
    d_sbb_vertex_polynomials);

  // compute w_p * (w_g * w_nnn)
  struct src_point * src_points = gauss_nnn_patch->src_points;
  size_t num_src_points = gauss_nnn_patch->num_src_points;
  double * d_data_weights = gauss_nnn_patch->weights;
  for (size_t i = 0, k = 0; i < num_src_points; ++i) {

    double weight = 0.0;
    for (size_t j = 0; j < HCSBB_NUM_SBB_COEFFS; ++j, ++k)
      weight += d_sbb_vertex_polynomials[j] * d_data_weights[k];

    weights[i].weight = weight * mult_factor;
    weights[i].subdomain_id = src_points[i].subdomain_id;
    weights[i].local_id = src_points[i].local_id;
  }
}

static void compute_hcsbb_weights_edge(
  struct tgt_point_search_data * tgt_point_data,
  struct weight_vector_data * weights, size_t * n) {

  // spherical barycentric coordinates of the target point
  double * sb_coords = tgt_point_data->sb_coords;

  // compute the spherical Bernstein base polynomials for the given vertices
  // (since the target point is on the edge, we only need a subset of the
  //  polynomials, because the rest are zero anyway)
  // B_ijk(p) = d!/(i!*j!*k!) * b_0(p)^i * b_1(p)^j * b_2(p)^k
  double B_300 = sb_coords[0] * sb_coords[0] * sb_coords[0];
  double B_030 = sb_coords[1] * sb_coords[1] * sb_coords[1];
  double B_210 = 3.0 * sb_coords[0] * sb_coords[0] * sb_coords[1];
  double B_120 = 3.0 * sb_coords[0] * sb_coords[1] * sb_coords[1];

  // c_300 = f(v1)
  // c_030 = f(v2)
  // c_210 and c_120
  struct weight_vector * c = &(tgt_point_data->interp_data.edge->c[0]);

  // f(p) = SUM(c_ijk * B_ijk) // for 300, 030, 210, and 120

  unsigned subdomain_id = tgt_point_data->subdomain_id;
  unsigned * ranks = &(tgt_point_data->ranks[0]);
  unsigned * src_point_local_ids = &(tgt_point_data->src_point_local_ids[0]);

  // c_300 * B_300
  // weights[0].weight = B_300 + B_210;
  weights[0].weight = B_300;
  weights[0].subdomain_id = (ranks[0] == -1)?subdomain_id:-1;
  weights[0].local_id = src_point_local_ids[0];
  // c_030 * B_030
  // weights[1].weight = B_030 + B_120;
  weights[1].weight = B_030;
  weights[1].subdomain_id = (ranks[1] == -1)?subdomain_id:-1;
  weights[1].local_id = src_point_local_ids[1];
  // c_210 * B_210
  struct weight_vector_data * curr_weights = weights + 2;
  memcpy(curr_weights, c[0].data, c[0].n * sizeof(*curr_weights));
  for (size_t i = 0; i < c[0].n; ++i) curr_weights[i].weight *= B_210;
  // c_120 * B_120
  curr_weights += c[0].n;
  memcpy(curr_weights, c[1].data, c[1].n * sizeof(*curr_weights));
  for (size_t i = 0; i < c[1].n; ++i) curr_weights[i].weight *= B_120;

  *n = 2 + c[0].n + c[1].n;
}

static void evaluate_blending_function(
  double * sb_coords, double * A) {

  double b[3] = {sb_coords[1]*sb_coords[1] * sb_coords[2]*sb_coords[2],
                 sb_coords[0]*sb_coords[0] * sb_coords[2]*sb_coords[2],
                 sb_coords[0]*sb_coords[0] * sb_coords[1]*sb_coords[1]};
  double temp = 1.0 / (b[0] + b[1] + b[2]);

  for (size_t i = 0; i < 3; ++i) {

    if (fabs(b[i]) < 1e-9) A[i] = 0.0;
    else                   A[i] = b[i] * temp;
  }
}

static void compute_hcsbb_weights_triangle(
  struct tgt_point_search_data * tgt_point_data,
  struct weight_vector_data * weights, size_t * n) {

  // spherical barycentric coordinates of the target point
  double * sb_coords = tgt_point_data->sb_coords;

  // compute the spherical Bernstein base polynomials for the given vertices
  // B_ijk(p) = d!/(i!*j!*k!) * b_0(p)^i * b_1(p)^j * b_2(p)^k
  double B_vertex[3] =
    {sb_coords[0] * sb_coords[0] * sb_coords[0],  // B_300
     sb_coords[1] * sb_coords[1] * sb_coords[1],  // B_030
     sb_coords[2] * sb_coords[2] * sb_coords[2]}; // B_003
  double B_edge[3][2] =
    {{3.0 * sb_coords[1] * sb_coords[1] * sb_coords[2],   // B_021
      3.0 * sb_coords[1] * sb_coords[2] * sb_coords[2]},  // B_012
     {3.0 * sb_coords[0] * sb_coords[0] * sb_coords[2],   // B_201
      3.0 * sb_coords[0] * sb_coords[2] * sb_coords[2]},  // B_102
     {3.0 * sb_coords[0] * sb_coords[0] * sb_coords[1],   // B_210
      3.0 * sb_coords[0] * sb_coords[1] * sb_coords[1]}}; // B_120
  double B_111 = 6.0 * sb_coords[0] * sb_coords[1] * sb_coords[2];

  // f(p) = SUM(c_ijk * B_ijk) // for all (i+j+k) = 3

  // the three vertices coefficients (300, 030, and 003)
  unsigned subdomain_id = tgt_point_data->subdomain_id;
  unsigned * ranks = &(tgt_point_data->ranks[0]);
  for (size_t i = 0; i < 3; ++i) {
    weights[i].weight = B_vertex[i];
    weights[i].subdomain_id = (ranks[i] == -1)?subdomain_id:-1;
    weights[i].local_id = tgt_point_data->src_point_local_ids[i];
  }

  size_t n_ = 3;

  // the six edges coefficients (210, 120, 201, 102, 021, and 012)
  for (size_t edge_idx = 0; edge_idx < 3; ++edge_idx) {
    struct weight_vector * c =
      &(tgt_point_data->interp_data.triangle->edges[edge_idx]->c[0]);
    for (size_t i = 0; i < 2; ++i) {
      struct weight_vector_data * curr_weights = weights + n_;
      size_t curr_n = c[i].n;
      n_ += curr_n;
      memcpy(curr_weights, c[i].data, c[i].n * sizeof(*curr_weights));
      double B = B_edge[edge_idx][i];
      for (size_t j = 0; j < curr_n; ++j) curr_weights[j].weight *= B;
    }
  }

  // the middle vertex
  // c_111 = SUM(c_111_a[i] * A[i]) // for i = 0..2
  // where a is the blending function
  double A[3];
  evaluate_blending_function(sb_coords, &(A[0]));

  struct weight_vector * c_111_a =
    &(tgt_point_data->interp_data.triangle->c_111_a[0]);
  for (size_t i = 0; i < 3; ++i) {
    struct weight_vector_data * curr_weights = weights + n_;
    size_t curr_n = c_111_a[i].n;
    n_ += curr_n;
    memcpy(curr_weights, c_111_a[i].data, c_111_a[i].n * sizeof(*curr_weights));
    double B_111_x_A_i = B_111 * A[i];
    for (size_t j = 0; j < curr_n; ++j) curr_weights[j].weight *= B_111_x_A_i;
  }

  *n = n_;
}

static int compare_weight_vector_data_pos_weight(
  void const * a, void const * b) {

  struct weight_vector_data const * weight_a =
    ((struct weight_vector_data_pos const *)a)->weight;
  struct weight_vector_data const * weight_b =
    ((struct weight_vector_data_pos const *)b)->weight;

  int ret = (int)(weight_a->subdomain_id) - (int)(weight_b->subdomain_id);
  if (ret) return ret;
  ret = (int)(weight_a->local_id) - (int)(weight_b->local_id);
  if (ret) return ret;
  double abs_weight_a = fabs(weight_a->weight);
  double abs_weight_b = fabs(weight_b->weight);
  ret = (abs_weight_a > abs_weight_b) - (abs_weight_a < abs_weight_b);
  if (ret) return ret;
  return (weight_a->weight > weight_b->weight) -
         (weight_a->weight < weight_b->weight);
}

static int compare_weight_vector_data_pos_pos(
  void const * a, void const * b) {

  return ((struct weight_vector_data_pos const *)a)->pos -
         ((struct weight_vector_data_pos const *)b)->pos;
}

static int compare_weight_vector_data_point(
  void  const * a, void const * b) {

  struct weight_vector_data const * weight_a =
    (struct weight_vector_data const *)a;
  struct weight_vector_data const * weight_b =
    (struct weight_vector_data const *)b;

  int ret = (int)(weight_a->subdomain_id) -
            (int)(weight_b->subdomain_id);
  if (ret) return ret;
  return (int)(weight_a->local_id) - (int)(weight_b->local_id);
}

static void compact_weight_vector_data(
  struct weight_vector_data * weights, size_t * n,
  struct weight_vector_data_pos * buffer) {

  size_t n_ = *n;

  if (n_ <= 1) return;

  for (size_t i = 0; i < n_; ++i) {
    buffer[i].weight = weights + i;
    buffer[i].pos = i;
  }

  qsort(buffer, n_, sizeof(*buffer), compare_weight_vector_data_pos_weight);

  size_t new_n = 1;
  struct weight_vector_data_pos * prev_weight_data_pos = buffer;
  struct weight_vector_data * prev_weight_data = prev_weight_data_pos->weight;
  struct weight_vector_data_pos * curr_weight_data_pos = buffer + 1;
  for (size_t i = 1; i < n_; ++i, ++curr_weight_data_pos) {

    struct weight_vector_data * curr_weight_data = curr_weight_data_pos->weight;
    if (!compare_weight_vector_data_point(prev_weight_data, curr_weight_data)) {
      prev_weight_data->weight += curr_weight_data->weight;
      if (prev_weight_data_pos->pos > curr_weight_data_pos->pos)
        prev_weight_data_pos->pos = curr_weight_data_pos->pos;
    } else {
      ++new_n;
      ++prev_weight_data_pos;
      *prev_weight_data_pos = *curr_weight_data_pos;
      prev_weight_data = prev_weight_data_pos->weight;
    }
  }

  if (n_ == new_n) return;

  qsort(buffer, new_n, sizeof(*buffer), compare_weight_vector_data_pos_pos);

  for (size_t i = 0; i < new_n; ++i) weights[i] = *(buffer[i].weight);

  *n = new_n;
}

static void compute_hcsbb_weights(
  struct tgt_point_search_data * tgt_point_data,
  struct weight_vector_data * weights, unsigned * n,
  struct weight_vector_data_pos * compact_buffer) {

  size_t n_;

  switch(tgt_point_data->type_flag) {

    case (ON_VERTEX):
      compute_hcsbb_weights_vertex(tgt_point_data, weights, &n_);
      break;
    case (ON_EDGE):
      compute_hcsbb_weights_edge(tgt_point_data, weights, &n_);
      break;
    case (ON_TRIANGLE):
      compute_hcsbb_weights_triangle(tgt_point_data, weights, &n_);
      break;
    default:
      yac_internal_abort_message(
        "ERROR(compute_hcsbb_weights): invalid type_flag", __FILE__, __LINE__);
  }

  compact_weight_vector_data(weights, &n_, compact_buffer);
  *n = n_;
}

static size_t get_max_num_weights(
  struct tgt_point_search_data * tgt_point_data) {

  switch(tgt_point_data->type_flag) {

    case (ON_VERTEX):
      return 1;
    case (ON_EDGE):
      return 2 + tgt_point_data->interp_data.edge->c[0].n +
             tgt_point_data->interp_data.edge->c[1].n;
    case (ON_TRIANGLE): {
      struct triangle_interp_data * triangle_data =
        tgt_point_data->interp_data.triangle;
      return 3 +
             triangle_data->edges[0]->c[0].n + triangle_data->edges[0]->c[1].n +
             triangle_data->edges[1]->c[0].n + triangle_data->edges[1]->c[1].n +
             triangle_data->edges[2]->c[0].n + triangle_data->edges[2]->c[1].n +
             triangle_data->c_111_a[0].n +
             triangle_data->c_111_a[1].n +
             triangle_data->c_111_a[2].n;
    }
    default:
      yac_internal_abort_message(
        "ERROR(compute_hcsbb_weights): invalid type_flag", __FILE__, __LINE__);
      return 0;
  };
}

static void free_gauss_nnn_patches(
  struct gauss_nnn_patch * gauss_nnn_patches, size_t num_gauss_nnn_patches) {

  for (size_t i = 0; i < num_gauss_nnn_patches; ++i) {

    free(gauss_nnn_patches[i].src_points);
    free(gauss_nnn_patches[i].src_point_ranks);
    free(gauss_nnn_patches[i].weights);
  }

  free(gauss_nnn_patches);
}

static void compute_edge_coefficients(
  struct basic_patch_data * patch_data, struct communicator * comm) {

  struct edge_interp_data * edges = patch_data->edges;
  size_t num_edges = patch_data->num_edges;

  size_t edge_weight_vector_data_buffer_size = 0;
  // count total number of weight_vector_data elements required
  for (size_t i = 0; i < num_edges; ++i)
    if ((edges[i].vertex_d_data[0]->gauss_nnn_patch != NULL) &&
        (edges[i].vertex_d_data[1]->gauss_nnn_patch != NULL))
      edge_weight_vector_data_buffer_size +=
        2 + edges[i].vertex_d_data[0]->gauss_nnn_patch->num_src_points +
        edges[i].vertex_d_data[1]->gauss_nnn_patch->num_src_points;

  // 2 -> two corner weights
  // 2*(HCSBB_NUM_GAUSS_POINTS*HCSBB_GAUSS_NNN+1)
  //   -> maximum number of weights per edge coefficient
  struct weight_vector_data * curr_edge_weight_vector_data =
    xmalloc(edge_weight_vector_data_buffer_size *
            sizeof(*curr_edge_weight_vector_data));
  patch_data->edge_weight_vector_data_buffer = curr_edge_weight_vector_data;

  for (size_t i = 0; i < num_edges; ++i) {

    struct edge_interp_data * curr_edge = edges + i;
    struct derivative_data_src_point * curr_vertex_d_data[2] = {
      curr_edge->vertex_d_data[0], curr_edge->vertex_d_data[1]};

    if ((curr_vertex_d_data[0]->gauss_nnn_patch == NULL) ||
        (curr_vertex_d_data[1]->gauss_nnn_patch == NULL)) continue;

    double direction[3] = {
      curr_vertex_d_data[1]->coordinate_xyz[0] -
      curr_vertex_d_data[0]->coordinate_xyz[0],
      curr_vertex_d_data[1]->coordinate_xyz[1] -
      curr_vertex_d_data[0]->coordinate_xyz[1],
      curr_vertex_d_data[1]->coordinate_xyz[2] -
      curr_vertex_d_data[0]->coordinate_xyz[2]};

    // c_210 and c_120
    for (size_t j = 0; j < 2; ++j) {

      size_t n = curr_vertex_d_data[j]->gauss_nnn_patch->num_src_points + 1;

      get_derivative_weights(
        curr_vertex_d_data[j]->gauss_nnn_patch,
        curr_vertex_d_data[j]->coordinate_xyz, &(direction[0]),
        curr_edge_weight_vector_data, 1.0/3.0);

      curr_edge_weight_vector_data[n-1].weight = 1.0;
      curr_edge_weight_vector_data[n-1].subdomain_id =
        curr_edge->vertex_d_data[j]->src_point.subdomain_id;
      curr_edge_weight_vector_data[n-1].local_id =
        curr_edge->vertex_d_data[j]->src_point.local_id;
      curr_edge->c[j].data = curr_edge_weight_vector_data;
      curr_edge->c[j].n = n;

      direction[0] *= -1, direction[1] *= -1, direction[2] *= -1;

      curr_edge_weight_vector_data += n;
    }

    if (!(i&1024)) yac_comm_check(comm);
  }
}

static void compute_triangle_coefficients(
  struct basic_patch_data * patch_data, struct communicator * comm) {

  // maximum number of coefficients per c_111 alpha value
  // 6 edge coefficients (HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS + 1)
  // 2 corner coefficients (1)
  // 1 middle point coefficient (HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS)
  struct weight_vector_data * weight_vector_data_buffer =
    xmalloc((6 * (HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS + 1) + 2 * 1 +
             HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS) *
             sizeof(*weight_vector_data_buffer));
  struct weight_vector_data_pos * weight_vector_data_pos_buffer =
    xmalloc((6 * (HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS + 1) + 2 * 1 +
             HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS) *
             sizeof(*weight_vector_data_pos_buffer));

  struct triangle_interp_data * curr_triangle = patch_data->triangles;
  size_t num_triangles = patch_data->num_triangles;

  for (size_t i = 0; i < num_triangles; ++i, ++curr_triangle) {

    // z_w0/3 = B_020(w0) * (b_0(g0)*c_120 + b_1(g0)*c_030 + b_2(g0)*c_021) +
    //          B_011(w0) * (b_0(g0)*a_1     b_1(g0)*c_021 + b_2(g0)*c_012) +
    //          B_002(w0) * (b_0(g0)*c_102 + b_1(g0)*c_012 + b_2(g0)*c_003)
    // w0 = ||v_1+v_2||  -> middle point of edge 0
    // g0 = v_2 x v_1    -> vector, which is a tangent of the sphere in w0 and
    //                      perpendicular to w0
    // z_w0 = D_g0 f(w0) -> directional derivative of f in w0 in the direction
    //                      of g0
    // =>
    // a_0 = (D_g f(w0)/3 -
    //        B_020(w0) * (b_0(g0)*c_120 + b_1(g0)*c_030 + b_2(g0)*c_021) -
    //        B_011(w0) * (                b_1(g0)*c_021 + b_2(g0)*c_012) -
    //        B_002(w0) * (b_0(g0)*c_102 + b_1(g0)*c_012 + b_2(g0)*c_003)) /
    //       (B_011(w0) * b_0(g0))
    // a_1 = (D_g f(w1)/3 -
    //        B_200(w1) * (b_0(g1)*c_300 + b_1(g1)*c_210 + b_2(g1)*c_201) -
    //        B_101(w1) * (b_0(g1)*c_201 +                 b_2(g1)*c_102) -
    //        B_002(w1) * (b_0(g1)*c_102 + b_1(g1)*c_012 + b_2(g1)*c_003)) /
    //       (B_101(w1) * b_1(g1))
    // a_2 = (D_g f(w2)/3 -
    //        B_200(w2) * (b_0(g2)*c_300 + b_1(g2)*c_210 + b_2(g2)*c_201) -
    //        B_110(w2) * (b_0(g2)*c_210 + b_1(g2)*c_120                ) -
    //        B_020(w2) * (b_0(g2)*c_120 + b_1(g2)*c_030 + b_2(g2)*c_021)) /
    //       (B_110(w2) * b_2(g2))

    struct edge_interp_data * curr_edges[3] = {curr_triangle->edges[0],
                                               curr_triangle->edges[1],
                                               curr_triangle->edges[2]};
    struct src_point src_points[3] = {
      curr_edges[2]->vertex_d_data[0]->src_point,
      curr_edges[2]->vertex_d_data[1]->src_point,
      curr_edges[0]->vertex_d_data[1]->src_point};

    if ((curr_edges[0]->vertex_d_data[0]->gauss_nnn_patch == NULL) ||
        (curr_edges[0]->vertex_d_data[1]->gauss_nnn_patch == NULL) ||
        (curr_edges[1]->vertex_d_data[0]->gauss_nnn_patch == NULL) ||
        (curr_edges[1]->vertex_d_data[1]->gauss_nnn_patch == NULL) ||
        (curr_edges[2]->vertex_d_data[0]->gauss_nnn_patch == NULL) ||
        (curr_edges[2]->vertex_d_data[1]->gauss_nnn_patch == NULL) ||
        (curr_edges[0]->middle_d_data->gauss_nnn_patch == NULL) ||
        (curr_edges[1]->middle_d_data->gauss_nnn_patch == NULL) ||
        (curr_edges[2]->middle_d_data->gauss_nnn_patch == NULL)) {

      for (size_t j = 0; j < 3; ++j) {
        curr_triangle->c_111_a[0].data = NULL;
        curr_triangle->c_111_a[0].n = 0;
      }
      continue;
    }

    // remark the corners of the triangle are sorted by global ids, the same is
    // true for the edge data
    struct weight_vector_data c_3_data[3] = {
      {.weight = 1.0,
       .subdomain_id = src_points[0].subdomain_id,
       .local_id = src_points[0].local_id},
      {.weight = 1.0,
       .subdomain_id = src_points[1].subdomain_id,
       .local_id = src_points[1].local_id},
      {.weight = 1.0,
       .subdomain_id = src_points[2].subdomain_id,
       .local_id = src_points[2].local_id}};
    struct weight_vector c_3[3] = {{.data = &(c_3_data[0]), .n = 1},
                                   {.data = &(c_3_data[1]), .n = 1},
                                   {.data = &(c_3_data[2]), .n = 1}};
    struct weight_vector * c_300 = &(c_3[0]);
    struct weight_vector * c_030 = &(c_3[1]);
    struct weight_vector * c_003 = &(c_3[2]);
    struct weight_vector * c_021 = &(curr_edges[0]->c[0]);
    struct weight_vector * c_012 = &(curr_edges[0]->c[1]);
    struct weight_vector * c_201 = &(curr_edges[1]->c[0]);
    struct weight_vector * c_102 = &(curr_edges[1]->c[1]);
    struct weight_vector * c_210 = &(curr_edges[2]->c[0]);
    struct weight_vector * c_120 = &(curr_edges[2]->c[1]);

    double corner_coordinate_xyz[3][3] = {
      {curr_edges[2]->vertex_d_data[0]->coordinate_xyz[0],
       curr_edges[2]->vertex_d_data[0]->coordinate_xyz[1],
       curr_edges[2]->vertex_d_data[0]->coordinate_xyz[2]},
      {curr_edges[2]->vertex_d_data[1]->coordinate_xyz[0],
       curr_edges[2]->vertex_d_data[1]->coordinate_xyz[1],
       curr_edges[2]->vertex_d_data[1]->coordinate_xyz[2]},
      {curr_edges[1]->vertex_d_data[1]->coordinate_xyz[0],
       curr_edges[1]->vertex_d_data[1]->coordinate_xyz[1],
       curr_edges[1]->vertex_d_data[1]->coordinate_xyz[2]}};

    // vector which are tangents of the sphere in the middle points of the edges
    // and are perpendicular to the edges
    double * g[3] = {
      &(curr_edges[0]->g[0]), &(curr_edges[1]->g[0]), &(curr_edges[2]->g[0])};

    // spherical barycentric coordinates of g
    double b_g[3][3] = {{g[0][0], g[0][1], g[0][2]},
                        {g[1][0], g[1][1], g[1][2]},
                        {g[2][0], g[2][1], g[2][2]}};
    compute_sb_coords(&(b_g[0][0]), 3, corner_coordinate_xyz);

    // computing quadratic spherical Bernstein-Bezier polynomials for the edge
    // middle points (e.g. B_020(w0)) using w0 as an example
    //   I. w0 is in the middle of the edge opposing corner 0
    //  II. for the barycentric coordinates if w0 we can say the following
    //    a) b_0(w0) = 0 -> because it is on the edge opposing v0
    //    b) b_1(w0) = b_2(w0) -> because it is in the middle between v1 and v2
    // III. B_ijk(p) = d!/(i!*j!*k!) * b_0(p)^i * b_1(p)^j * b_2(p)^k
    //      -> for all (i + j + k) = d
    //      -> p = w0 and d = 2 in our case
    //
    // from II.a) it follows: B_200 = B_110 = B_101 = 0
    // from II.b) it follows: B_020 = B_002 = 2 * B_011 = b_1(w0) = b_2(w0) = bw

    double bw[3] = {
      curr_edges[0]->sb_coord_middle_point *
      curr_edges[0]->sb_coord_middle_point,
      curr_edges[1]->sb_coord_middle_point *
      curr_edges[1]->sb_coord_middle_point,
      curr_edges[2]->sb_coord_middle_point *
      curr_edges[2]->sb_coord_middle_point};

#define COPY_D_DATA(edge_idx) \
  { \
    struct derivative_data_middle_point * curr_d_data = \
      curr_edges[edge_idx]->middle_d_data; \
    get_derivative_weights( \
      curr_d_data->gauss_nnn_patch, curr_d_data->coordinate_xyz, g[edge_idx], \
      weight_vector_data_buffer + n, 1.0 / 3.0); \
    n += curr_d_data->gauss_nnn_patch->num_src_points; \
  }
#define COPY_COEFF(c_ijk, edge_idx, b_idx, bw_factor) \
  { \
    struct weight_vector_data * curr_buffer = weight_vector_data_buffer + n; \
    size_t curr_n = c_ijk->n; \
    memcpy(curr_buffer, c_ijk->data, curr_n * sizeof(*curr_buffer)); \
    n += curr_n; \
    double factor = - bw[edge_idx] * b_g[edge_idx][b_idx] * bw_factor; \
    for (size_t i_ = 0; i_ < curr_n; ++i_) curr_buffer[i_].weight *= factor; \
  }
#define MULT_COEFF(edge_idx) \
  { \
    double factor = 1.0 / (2.0 * bw[edge_idx] * b_g[edge_idx][edge_idx]); \
    for (size_t i_ = 0; i_ < n; ++i_) \
      weight_vector_data_buffer[i_].weight *= factor; \
  }
#define COMPACT_WEIGHTS(edge_idx) \
  { \
    compact_weight_vector_data( \
      weight_vector_data_buffer, &n, weight_vector_data_pos_buffer); \
    curr_triangle->c_111_a[edge_idx].data = \
      xmalloc(n * sizeof(*(curr_triangle->c_111_a[edge_idx].data))); \
    memcpy(curr_triangle->c_111_a[edge_idx].data, weight_vector_data_buffer, \
           n * sizeof(*weight_vector_data_buffer)); \
    curr_triangle->c_111_a[edge_idx].n = n; \
  }

    // a_0 = (D_g f(w0)/3 -
    //        B_020(w0) * (b_0(g0)*c_120 + b_1(g0)*c_030 + b_2(g0)*c_021) -
    //        B_011(w0) * (                b_1(g0)*c_021 + b_2(g0)*c_012) -
    //        B_002(w0) * (b_0(g0)*c_102 + b_1(g0)*c_012 + b_2(g0)*c_003)) /
    //       (B_011(w0) * b_0(g0))
    size_t n = 0;
    COPY_D_DATA(0); // D_g f(w0)/3
    COPY_COEFF(c_120, 0, 0, 1.0); // - B_020(w0) * b_0(g0) * c_120
    COPY_COEFF(c_030, 0, 1, 1.0); // - B_020(w0) * b_1(g0) * c_030
    COPY_COEFF(c_021, 0, 2, 1.0); // - B_020(w0) * b_2(g0) * c_021
    COPY_COEFF(c_021, 0, 1, 2.0); // - B_011(w0) * b_1(g0) * c_021
    COPY_COEFF(c_012, 0, 2, 2.0); // - B_011(w0) * b_2(g0) * c_012
    COPY_COEFF(c_102, 0, 0, 1.0); // - B_002(w0) * b_0(g0) * c_102
    COPY_COEFF(c_012, 0, 1, 1.0); // - B_002(w0) * b_1(g0) * c_012
    COPY_COEFF(c_003, 0, 2, 1.0); // - B_002(w0) * b_2(g0) * c_003
    MULT_COEFF(0); // (...) / (B_011(w0) * b_0(g0))
    COMPACT_WEIGHTS(0); // a_0 = ...

    // a_1 = (D_g f(w1)/3 -
    //        B_200(w1) * (b_0(g1)*c_300 + b_1(g1)*c_210 + b_2(g1)*c_201) -
    //        B_101(w1) * (b_0(g1)*c_201 +                 b_2(g1)*c_102) -
    //        B_002(w1) * (b_0(g1)*c_102 + b_1(g1)*c_012 + b_2(g1)*c_003)) /
    //       (B_101(w1) * b_1(g1))
    n = 0;
    COPY_D_DATA(1); // D_g f(w1)/3
    COPY_COEFF(c_300, 1, 0, 1.0); // - B_200(w1) * b_0(g1) * c_300
    COPY_COEFF(c_210, 1, 1, 1.0); // - B_200(w1) * b_1(g1) * c_210
    COPY_COEFF(c_201, 1, 2, 1.0); // - B_200(w1) * b_2(g1) * c_201
    COPY_COEFF(c_201, 1, 0, 2.0); // - B_101(w1) * b_0(g1) * c_201
    COPY_COEFF(c_102, 1, 2, 2.0); // - B_101(w1) * b_2(g1) * c_102
    COPY_COEFF(c_102, 1, 0, 1.0); // - B_002(w1) * b_0(g1) * c_102
    COPY_COEFF(c_012, 1, 1, 1.0); // - B_002(w1) * b_1(g1) * c_012
    COPY_COEFF(c_003, 1, 2, 1.0); // - B_002(w1) * b_2(g1) * c_003
    MULT_COEFF(1); // (...) / (B_101(w1) * b_1(g1))
    COMPACT_WEIGHTS(1); // a_1 = ...

    // a_2 = (D_g f(w2)/3 -
    //        B_200(w2) * (b_0(g2)*c_300 + b_1(g2)*c_210 + b_2(g2)*c_201) -
    //        B_110(w2) * (b_0(g2)*c_210 + b_1(g2)*c_120                ) -
    //        B_020(w2) * (b_0(g2)*c_120 + b_1(g2)*c_030 + b_2(g2)*c_021)) /
    //       (B_110(w2) * b_2(g2))
    n = 0;
    COPY_D_DATA(2); // D_g f(w2)/3
    COPY_COEFF(c_300, 2, 0, 1.0); // - B_200(w2) * b_0(g2) * c_300
    COPY_COEFF(c_210, 2, 1, 1.0); // - B_200(w2) * b_1(g2) * c_210
    COPY_COEFF(c_201, 2, 2, 1.0); // - B_200(w2) * b_2(g2) * c_201
    COPY_COEFF(c_210, 2, 0, 2.0); // - B_110(w2) * b_0(g2) * c_210
    COPY_COEFF(c_120, 2, 1, 2.0); // - B_110(w2) * b_1(g2) * c_120
    COPY_COEFF(c_120, 2, 0, 1.0); // - B_020(w2) * b_0(g2) * c_120
    COPY_COEFF(c_030, 2, 1, 1.0); // - B_020(w2) * b_1(g2) * c_030
    COPY_COEFF(c_021, 2, 2, 1.0); // - B_020(w2) * b_2(g2) * c_021
    MULT_COEFF(2); // (...) / (B_110(w2) * b_2(g2))
    COMPACT_WEIGHTS(2); // a_2 = ...

#undef COMPACT_WEIGHTS
#undef MULT_COEFF
#undef COPY_COEFF
#undef COPY_D_DATA

    if (!(i&1024)) yac_comm_check(comm);
  }
  free(weight_vector_data_buffer);
  free(weight_vector_data_pos_buffer);
}

static void compute_weights(
  struct tgt_point_search_data * tgt_point_data, size_t num_interpolated_points,
  struct basic_patch_data * patch_data, struct search_data * search_data,
  struct communicator * comm) {

  //--------------------------------------------------------------------------
  // prepare patch data for the computation of the weights for each individual
  // target point
  //--------------------------------------------------------------------------

  // compute edge coefficients
  compute_edge_coefficients(patch_data, comm);

  // compute triangle coefficients
  compute_triangle_coefficients(patch_data, comm);

  struct weight_vector_data * weight_vector_data = NULL;
  size_t weight_vector_data_array_size = 0;
  size_t total_num_weights = 0;
  unsigned * num_weights_per_tgt =
    xmalloc(num_interpolated_points * sizeof(*num_weights_per_tgt));
  struct weight_vector_data_pos * compact_buffer =
    xmalloc((3 * 1 + // c_300, c_030, c_003
             3 * (1 + HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN) + // c_012, c_021, c_102, ...
             3 * (HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN + 2 +
                  6 * (1 + HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN))) * // c_111_a[3]
             sizeof(*compact_buffer));

  //--------------------------------------
  // compute weights for the target points
  //--------------------------------------

  for (size_t i = 0; i < (num_interpolated_points + 127) / 128; ++i) {

    size_t curr_num_interpolated_points =
      MIN(128, num_interpolated_points - i * 128);
    struct tgt_point_search_data * curr_tgt_point_data =
      tgt_point_data + i * 128;
    unsigned * curr_num_weights_per_tgt = num_weights_per_tgt + i * 128;

    size_t curr_max_num_weights = 0;
    for (size_t j = 0; j < curr_num_interpolated_points; ++j)
      curr_max_num_weights += get_max_num_weights(curr_tgt_point_data + j);

    ENSURE_ARRAY_SIZE(weight_vector_data, weight_vector_data_array_size,
                      total_num_weights + curr_max_num_weights);

    for (size_t j = 0; j < curr_num_interpolated_points; ++j) {
      compute_hcsbb_weights(
        curr_tgt_point_data + j, weight_vector_data + total_num_weights,
        curr_num_weights_per_tgt + j, compact_buffer);
      total_num_weights += curr_num_weights_per_tgt[j];
    }

    // allow the communicator to handle open requests
    yac_comm_check(comm);
  }

  free(compact_buffer);
  search_data->weights =
    xrealloc(
      weight_vector_data, total_num_weights * sizeof(*weight_vector_data));
  search_data->num_weights_per_tgt = num_weights_per_tgt;
  search_data->total_num_weights = total_num_weights;

  // free memory associated with the gauss_nnn_patches
  free_gauss_nnn_patches(
    patch_data->gauss_nnn_patches, patch_data->num_gauss_nnn_patches);
  patch_data->gauss_nnn_patches = NULL;
  patch_data->num_gauss_nnn_patches = 0;
}

static void adjust_src_point_data(
  struct tgt_point_search_data * tgt_point_data, size_t num_points,
  struct interpolation_method_data_request data_request) {

  int * global_ids = data_request.global_ids;
  int * tmp_global_ids = xmalloc(3 * 256 * sizeof(*tmp_global_ids));
  int * orig_pos = xmalloc(3 * 256 * sizeof(*orig_pos));
  for (size_t i = 0; i < 3 * 256; ++i) orig_pos[i] = (int)i;

  for (size_t i = 0; i < (num_points + 255)/256; ++i) {

    size_t curr_num_points = MIN(256, num_points - i * 256);
    size_t num_tmp_global_ids = 0;

    for (size_t j = 0; j < curr_num_points; ++j) {
      struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data + j;
      size_t num_src_points = (size_t)(curr_tgt_point_data->type_flag) + 1;
      for (size_t k = 0; k < num_src_points; ++k)
        if (curr_tgt_point_data->ranks[k] != (unsigned)-1)
          tmp_global_ids[num_tmp_global_ids++] =
            curr_tgt_point_data->src_point_global_ids[k];
    }

    // sort global ids
    yac_quicksort_index(tmp_global_ids, num_tmp_global_ids, orig_pos);

    // match the tmp_global_ids with the global_ids from the data_request
    for (size_t j = 0, k = 0; j < num_tmp_global_ids; ++j) {

      int curr_tmp_global_id = tmp_global_ids[j];
      while (global_ids[k] < curr_tmp_global_id) ++k;

      // at this points (global_ids[k] == tmp_global_ids[j])
      // -> we store the position
      tmp_global_ids[j] = (int)k;
    }

    // get results into original order
    yac_quicksort_index(orig_pos, num_tmp_global_ids, tmp_global_ids);

    num_tmp_global_ids = 0;

    // write back found positions
    for (size_t j = 0; j < curr_num_points; ++j) {
      struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data + j;
      size_t num_src_points = (size_t)(curr_tgt_point_data->type_flag) + 1;
      for (size_t k = 0; k < num_src_points; ++k)
        if (curr_tgt_point_data->ranks[k] != (unsigned)-1)
          curr_tgt_point_data->src_point_local_ids[k] =
            (unsigned)(tmp_global_ids[num_tmp_global_ids++]);
    }
    tgt_point_data += 256;
  }
  free(tmp_global_ids);
  free(orig_pos);
}

static void adjust_patch_data(
  struct basic_patch_data * patch_data,
  struct interpolation_method_data_request data_request) {

  int * global_ids = data_request.global_ids;

  size_t const max_src_points = 16 * HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN;
  int * tmp_global_ids = xmalloc(max_src_points * sizeof(*tmp_global_ids));
  int * orig_pos = xmalloc(max_src_points * sizeof(*orig_pos));

  for (size_t k = 0; k < max_src_points; ++k) orig_pos[k] = (int)k;

  struct gauss_nnn_patch * gauss_nnn_patches = patch_data->gauss_nnn_patches;
  size_t num_gauss_nnn_patches = patch_data->num_gauss_nnn_patches;
  for (size_t i = 0; i < (num_gauss_nnn_patches + 15)/16; ++i) {

    size_t curr_num_gauss_nnn_patches = MIN(16, num_gauss_nnn_patches - i * 16);
    struct gauss_nnn_patch * curr_gauss_nnn_patches =
      gauss_nnn_patches + i * 16;
    size_t num_tmp_global_ids = 0;

    for (size_t j = 0; j < curr_num_gauss_nnn_patches; ++j) {
      size_t num_src_points = curr_gauss_nnn_patches[j].num_src_points;
      struct src_point * curr_src_points =
        curr_gauss_nnn_patches[j].src_points;
      for (size_t k = 0; k < num_src_points; ++k)
        if (curr_src_points[k].subdomain_id == -1)
          tmp_global_ids[num_tmp_global_ids++] = curr_src_points[k].global_id;
    }

    // sort global ids
    yac_quicksort_index(tmp_global_ids, num_tmp_global_ids, orig_pos);

    // match the tmp_global_ids with the global_ids from the data_request
    for (size_t j = 0, k = 0; j < num_tmp_global_ids; ++j) {

      int curr_tmp_global_id = tmp_global_ids[j];
      while (global_ids[k] < curr_tmp_global_id) ++k;

      // at this points (global_ids[k] == tmp_global_ids[j])
      // -> we store the position
      tmp_global_ids[j] = (int)k;
    }

    // get results into original order
    yac_quicksort_index(orig_pos, num_tmp_global_ids, tmp_global_ids);

    num_tmp_global_ids = 0;

    // write back found positions
    for (size_t j = 0; j < curr_num_gauss_nnn_patches; ++j) {
      size_t num_src_points = curr_gauss_nnn_patches[j].num_src_points;
      struct src_point * curr_src_points =
        curr_gauss_nnn_patches[j].src_points;
      for (size_t k = 0; k < num_src_points; ++k)
        if (curr_src_points[k].subdomain_id == -1)
          curr_src_points[k].local_id =
            (unsigned)(tmp_global_ids[num_tmp_global_ids++]);
    }
  }

  struct derivative_data_src_point * d_data_src_point =
    patch_data->d_data_src_point;
  size_t d_data_src_point_count = patch_data->d_data_src_point_count;
  for (size_t i = 0;
       i < (d_data_src_point_count + max_src_points - 1)/max_src_points; ++i) {

    size_t curr_d_data_src_point_count =
      MIN(max_src_points, d_data_src_point_count - i * max_src_points);
    struct derivative_data_src_point * curr_d_data_src_point =
      d_data_src_point + i * max_src_points;
    size_t num_tmp_global_ids = 0;

    for (size_t j = 0; j < curr_d_data_src_point_count; ++j)
      if (curr_d_data_src_point[j].src_point.subdomain_id == -1)
        tmp_global_ids[num_tmp_global_ids++] =
          curr_d_data_src_point[j].src_point.global_id;

    // sort global ids
    yac_quicksort_index(tmp_global_ids, num_tmp_global_ids, orig_pos);

    // match the tmp_global_ids with the global_ids from the data_request
    for (size_t j = 0, k = 0; j < num_tmp_global_ids; ++j) {

      int curr_tmp_global_id = tmp_global_ids[j];
      while (global_ids[k] < curr_tmp_global_id) ++k;

      // at this points (global_ids[k] == tmp_global_ids[j])
      // -> we store the position
      tmp_global_ids[j] = (int)k;
    }

    // get results into original order
    yac_quicksort_index(orig_pos, num_tmp_global_ids, tmp_global_ids);

    num_tmp_global_ids = 0;

    // write back found positions
    for (size_t j = 0; j < curr_d_data_src_point_count; ++j)
      if (curr_d_data_src_point[j].src_point.subdomain_id == -1)
        curr_d_data_src_point[j].src_point.local_id =
          (unsigned)(tmp_global_ids[num_tmp_global_ids++]);
  }

  free(tmp_global_ids);
  free(orig_pos);
}

static void finish_hcsbb_interpolation(
  void * user_data, int * global_ids_to_remove, unsigned num_global_ids) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;
  struct search_data search_data = data->search_data;
  struct tgt_point_search_data * tgt_point_data = search_data.tgt_point_data;
  size_t num_tgt_points = search_data.num_tgt_points;
  struct basic_patch_data * patch_data = &(data->patch_data);
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb *)(data->method_stack[0]);

  // mark all points that are interpolated by a remote process
  mark_tgt_points(
    num_global_ids, global_ids_to_remove, search_data, REMOVE);

  // mark all points that are assigned to a remote process but are not actually
  // interpolated by it as NOT
  for (size_t i = 0; i < num_tgt_points; ++i)
    if (tgt_point_data[i].status_flag == REMOTE)
      tgt_point_data[i].status_flag = NOT;

  // reorder target points first by status flag (LOCAL < REMOTE < NOT < REMOVE)
  // and second by global target id
  qsort(tgt_point_data, num_tgt_points, sizeof(*tgt_point_data),
      compare_tgt_point_data_flag_global_id);

  // count the number of target points interpolated by the local process
  size_t num_interpolated_points = 0;
  while ((num_interpolated_points < num_tgt_points) &&
         (tgt_point_data[num_interpolated_points].status_flag == LOCAL))
    ++num_interpolated_points;

  // count the number of points that are not interpolated by the local or a
  // remote process
  size_t num_remaining_points = num_interpolated_points;
  while ((num_remaining_points < num_tgt_points) &&
         (tgt_point_data[num_remaining_points].status_flag != REMOVE))
    ++num_remaining_points;
  num_remaining_points -= num_interpolated_points;

  // generate data_request
  generate_data_request(data, num_interpolated_points);

  struct interpolation_method_data_request data_request =
    method_hcsbb->data_request[0][method_hcsbb->points_idx];

  // adjust src_point_data in tgt_point_data to data_request
  adjust_src_point_data(
    tgt_point_data, num_interpolated_points, data_request);

  // adjust patch_data to data_request
  adjust_patch_data(patch_data, data_request);

  // compute the actual weights
  // remark: after calling compute_weights the order of tgt_point_data cannot
  //         be changed
  compute_weights(
    tgt_point_data, num_interpolated_points, patch_data, &(data->search_data),
    yac_get_domain_comm(method_hcsbb->domain));

  struct interpolation_plan_element ** plan_stack = data->plan_stack;
  // generate interpolation plan
  *plan_stack = generate_interpolation_plan(data, num_interpolated_points);
  if (*plan_stack != NULL) plan_stack++;

  // retrieve coordinates_xyz and global_ids from tgt_point_data
  retrieve_tgt_point_data(
    tgt_point_data, num_interpolated_points + num_remaining_points,
    data->coordinates_xyz, data->global_ids);

  struct interpolation_method ** method_stack = data->method_stack+1;
  unsigned * p_num_interpolated_points = data->num_interpolated_points;
  double (*coordinates_xyz)[3] = data->coordinates_xyz;
  int * global_ids = data->global_ids;
  enum yac_location tgt_location = data->tgt_location;

  p_num_interpolated_points[0] = num_interpolated_points;
  p_num_interpolated_points++;

  global_ids += num_interpolated_points;
  coordinates_xyz += num_interpolated_points;

  free_intermediate_method_data(data);

  // call next interpolation method on stack
  if (*method_stack != NULL)
     (*method_stack)->vtable->do_search(
        method_stack, plan_stack, global_ids, coordinates_xyz, tgt_location,
        num_remaining_points, p_num_interpolated_points,
        method_hcsbb->data_request+1);

  if (method_hcsbb->finish_callback != NULL)
     method_hcsbb->finish_callback(method_hcsbb->finish_user_data);
}

static int compare_result_points_distance(const void * a, const void * b) {

  struct result_point const * result_a = (struct result_point const *)a;
  struct result_point const * result_b = (struct result_point const *)b;

  double cos_angle_a = result_a->cos_angle;
  double cos_angle_b = result_b->cos_angle;

  // bigger cos_angle -> small angle
  if (fabs(cos_angle_a - cos_angle_b) > 0.0)
    return (cos_angle_a > cos_angle_b)?1:-1;

  return result_a->global_id - result_b->global_id;
}

static int compare_result_points_global_id (const void * a, const void * b) {
  return (((struct result_point const *)a)->global_id -
          ((struct result_point const *)b)->global_id);
}

static void compute_gauss_nnn_patch(
  struct gauss_nnn_patch * gauss_nnn_patch,
  struct request_point * request_points,
  struct result_point result_point_buffer[][HCSBB_GAUSS_NNN],
  double w_nnn[][HCSBB_NUM_GAUSS_POINTS]) {

  // check and prepare the search results
  {
    for (size_t i = 0; i < HCSBB_NUM_GAUSS_POINTS; ++i) {

      size_t num_found_points = (size_t)(request_points[i].num_found_points);

      // if there are not enough source points for the current gauss integration
      // point -> gauss patch can not be generated
      if (num_found_points < HCSBB_GAUSS_NNN) {
        gauss_nnn_patch->src_points = NULL;
        gauss_nnn_patch->src_point_ranks = NULL;
        gauss_nnn_patch->num_src_points = 0;
        free(gauss_nnn_patch->weights);
        gauss_nnn_patch->weights = NULL;
        return;
      }

      // if there are more than HCSBB_GAUSS_NNN point for the current gauss
      // integration point
      if (num_found_points > HCSBB_GAUSS_NNN) {

        // sort the results be distance first and global source id second
        qsort(request_points[i].results, num_found_points,
              sizeof(*(request_points[i].results)),
              compare_result_points_distance);
        // we only need the first HCSBB_GAUSS_NNN source points
        request_points[i].num_found_points = HCSBB_GAUSS_NNN;
      }

      // sort the result points by their global source ids
      qsort(request_points[i].results, HCSBB_GAUSS_NNN,
            sizeof(*(request_points[i].results)),
            compare_result_points_global_id);
    }
  }

  size_t num_unique_result_points;
  struct result_point * result_point_buffer_ = &(result_point_buffer[0][0]);

  // gather all unique required source points
  {
    // copy all result points into the result_point_buffer
    for (size_t i = 0; i < HCSBB_NUM_GAUSS_POINTS; ++i)
      memcpy(&(result_point_buffer[i][0]), request_points[i].results,
             HCSBB_GAUSS_NNN * sizeof(result_point_buffer[0][0]));


    // sort result_point_buffer by global ids
    qsort(result_point_buffer_, HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN,
          sizeof(*result_point_buffer_), compare_result_points_global_id);

    // count number of unique result points
    num_unique_result_points = 1;
    int prev_global_id = result_point_buffer_[0].global_id;
    for (size_t i = 1; i < HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN; ++i) {
      int curr_global_id = result_point_buffer_[i].global_id;
      if (prev_global_id != curr_global_id) {
        prev_global_id = curr_global_id;
        ++num_unique_result_points;
      }
    }

    // get unique source points
    struct src_point * curr_src_point =
      xmalloc(num_unique_result_points * sizeof(*curr_src_point));
    unsigned * curr_src_point_rank =
      xmalloc(num_unique_result_points * sizeof(*curr_src_point_rank));
    gauss_nnn_patch->src_points = curr_src_point;
    gauss_nnn_patch->src_point_ranks = curr_src_point_rank;
    gauss_nnn_patch->num_src_points = num_unique_result_points;
    prev_global_id = result_point_buffer_[0].global_id - 1;
    for (size_t i = 0; i < HCSBB_NUM_GAUSS_POINTS * HCSBB_GAUSS_NNN; ++i) {
      if (prev_global_id != result_point_buffer_[i].global_id) {
        prev_global_id = result_point_buffer_[i].global_id;
        curr_src_point->global_id = result_point_buffer_[i].global_id;
        if (result_point_buffer_[i].is_local) {
          curr_src_point->subdomain_id =
            result_point_buffer_[i].data.local.subdomain_id;
          curr_src_point->local_id =
            result_point_buffer_[i].data.local.local_id;
          *curr_src_point_rank = (unsigned)-1;
        } else {
          curr_src_point->subdomain_id = -1;
          curr_src_point->local_id = -1;
          *curr_src_point_rank = result_point_buffer_[i].data.remote.rank;
        }
        ++curr_src_point;
        ++curr_src_point_rank;
      }
    }
  }

  // compute w_nnn
  {
    double inv_distances[num_unique_result_points];
    for (size_t i = 0; i < HCSBB_NUM_GAUSS_POINTS; ++i) {

      struct result_point * curr_result_points = request_points[i].results;
      struct src_point * unique_result_points = gauss_nnn_patch->src_points;
      double distance_sum = 0.0;
      size_t j, k;

      for (size_t l = 0; l < num_unique_result_points; ++l)
        inv_distances[l] = 0.0;

      // for all results of the current request point
      for (j = 0, k = 0; j < HCSBB_GAUSS_NNN; ++j) {

        int curr_global_id = curr_result_points[j].global_id;

        // get the index of the unique result points that matches the
        // current global id (it has to be in there; unique_result_points and
        // curr_result_points are sorted by global ids)
        while (unique_result_points[k].global_id != curr_global_id) ++k;

        double distance = acos(curr_result_points[j].cos_angle);

        // if src and tgt point are at the same location
        if (distance < yac_angle_tol) break;

        double inv_distance = 1.0 / (distance * distance);
        inv_distances[k] = inv_distance;
        distance_sum += inv_distance;
      }

      // if there is no source point that exactly matches the target point
      if (j == HCSBB_GAUSS_NNN) {

        distance_sum = 1.0 / distance_sum;
        for (size_t l = 0; l < num_unique_result_points; ++l)
          w_nnn[l][i] = inv_distances[l] * distance_sum;

      } else {

        for (size_t l = 0; l < num_unique_result_points; ++l)
          w_nnn[l][i] = 0.0;
        w_nnn[k][i] = 1.0;
      }
    }
  }

  double (*w_g)[HCSBB_NUM_GAUSS_POINTS] =
    (double (*)[HCSBB_NUM_GAUSS_POINTS])(gauss_nnn_patch->weights);

  // compute and store w_g * w_nnn
  double (*weights)[HCSBB_NUM_SBB_COEFFS] =
    xmalloc(num_unique_result_points * sizeof(*weights));
  gauss_nnn_patch->weights = (double*)weights;
  for (size_t i = 0; i < num_unique_result_points; ++i) {
    for (size_t j = 0; j < HCSBB_NUM_SBB_COEFFS; ++j) {
      double accu = 0.0;
      for (size_t l = 0; l < HCSBB_NUM_GAUSS_POINTS; ++l)
        accu += w_g[j][l] * w_nnn[i][l];
      weights[i][j] = accu;
    }
  }

  free(w_g);
}

static void generate_gauss_legendre_points(
  double gauss_vertices[HCSBB_NUM_GAUSS_POINTS][3],
  double gauss_sb_coords[HCSBB_NUM_GAUSS_POINTS][3],
  double bnd_triangle[3][3]) {

#if HCSBB_GAUSS_ORDER == 3
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.333333333333333-0.333333333333333,0.333333333333333,0.333333333333333},
    {1.0-0.200000000000000-0.200000000000000,0.200000000000000,0.200000000000000},
    {1.0-0.600000000000000-0.200000000000000,0.600000000000000,0.200000000000000},
    {1.0-0.200000000000000-0.600000000000000,0.200000000000000,0.600000000000000}
  };
#elif HCSBB_GAUSS_ORDER == 4
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.091576213509771-0.091576213509771,0.091576213509771,0.091576213509771},
    {1.0-0.816847572980459-0.091576213509771,0.816847572980459,0.091576213509771},
    {1.0-0.091576213509771-0.816847572980459,0.091576213509771,0.816847572980459},
    {1.0-0.445948490915965-0.445948490915965,0.445948490915965,0.445948490915965},
    {1.0-0.108103018168070-0.445948490915965,0.108103018168070,0.445948490915965},
    {1.0-0.445948490915965-0.108103018168070,0.445948490915965,0.108103018168070}
  };
#elif HCSBB_GAUSS_ORDER == 5
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.333333333333333-0.333333333333333,0.333333333333333,0.333333333333333},
    {1.0-0.101286507323456-0.101286507323456,0.101286507323456,0.101286507323456},
    {1.0-0.797426985353087-0.101286507323456,0.797426985353087,0.101286507323456},
    {1.0-0.101286507323456-0.797426985353087,0.101286507323456,0.797426985353087},
    {1.0-0.470142064105115-0.470142064105115,0.470142064105115,0.470142064105115},
    {1.0-0.059715871789770-0.470142064105115,0.059715871789770,0.470142064105115},
    {1.0-0.470142064105115-0.059715871789770,0.470142064105115,0.059715871789770}
  };
#elif HCSBB_GAUSS_ORDER == 6
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.063089014491502-0.063089014491502,0.063089014491502,0.063089014491502},
    {1.0-0.873821971016996-0.063089014491502,0.873821971016996,0.063089014491502},
    {1.0-0.063089014491502-0.873821971016996,0.063089014491502,0.873821971016996},
    {1.0-0.249286745170910-0.249286745170910,0.249286745170910,0.249286745170910},
    {1.0-0.501426509658179-0.249286745170910,0.501426509658179,0.249286745170910},
    {1.0-0.249286745170910-0.501426509658179,0.249286745170910,0.501426509658179},
    {1.0-0.310352451033785-0.053145049844816,0.310352451033785,0.053145049844816},
    {1.0-0.053145049844816-0.310352451033785,0.053145049844816,0.310352451033785},
    {1.0-0.636502499121399-0.053145049844816,0.636502499121399,0.053145049844816},
    {1.0-0.053145049844816-0.636502499121399,0.053145049844816,0.636502499121399},
    {1.0-0.636502499121399-0.310352451033785,0.636502499121399,0.310352451033785},
    {1.0-0.310352451033785-0.636502499121399,0.310352451033785,0.636502499121399}
  };
#elif HCSBB_GAUSS_ORDER == 7
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.333333333333333-0.333333333333333,0.333333333333333,0.333333333333333},
    {1.0-0.260345966079038-0.260345966079038,0.260345966079038,0.260345966079038},
    {1.0-0.479308067841923-0.260345966079038,0.479308067841923,0.260345966079038},
    {1.0-0.260345966079038-0.479308067841923,0.260345966079038,0.479308067841923},
    {1.0-0.065130102902216-0.065130102902216,0.065130102902216,0.065130102902216},
    {1.0-0.869739794195568-0.065130102902216,0.869739794195568,0.065130102902216},
    {1.0-0.065130102902216-0.869739794195568,0.065130102902216,0.869739794195568},
    {1.0-0.312865496004874-0.048690315425316,0.312865496004874,0.048690315425316},
    {1.0-0.048690315425316-0.312865496004874,0.048690315425316,0.312865496004874},
    {1.0-0.638444188569809-0.048690315425316,0.638444188569809,0.048690315425316},
    {1.0-0.048690315425316-0.638444188569809,0.048690315425316,0.638444188569809},
    {1.0-0.638444188569809-0.312865496004874,0.638444188569809,0.312865496004874},
    {1.0-0.312865496004874-0.638444188569809,0.312865496004874,0.638444188569809}
  };
#elif HCSBB_GAUSS_ORDER == 8
  static double const abscissa[HCSBB_NUM_GAUSS_POINTS][3] = {
    {1.0-0.333333333333333-0.333333333333333,0.333333333333333,0.333333333333333},
    {1.0-0.081414823414554-0.459292588292723,0.081414823414554,0.459292588292723},
    {1.0-0.459292588292723-0.081414823414554,0.459292588292723,0.081414823414554},
    {1.0-0.459292588292723-0.459292588292723,0.459292588292723,0.459292588292723},
    {1.0-0.658861384496480-0.170569307751760,0.658861384496480,0.170569307751760},
    {1.0-0.170569307751760-0.658861384496480,0.170569307751760,0.658861384496480},
    {1.0-0.170569307751760-0.170569307751760,0.170569307751760,0.170569307751760},
    {1.0-0.898905543365938-0.050547228317031,0.898905543365938,0.050547228317031},
    {1.0-0.050547228317031-0.898905543365938,0.050547228317031,0.898905543365938},
    {1.0-0.050547228317031-0.050547228317031,0.050547228317031,0.050547228317031},
    {1.0-0.008394777409958-0.263112829634638,0.008394777409958,0.263112829634638},
    {1.0-0.263112829634638-0.008394777409958,0.263112829634638,0.008394777409958},
    {1.0-0.008394777409958-0.728492392955404,0.008394777409958,0.728492392955404},
    {1.0-0.728492392955404-0.008394777409958,0.728492392955404,0.008394777409958},
    {1.0-0.263112829634638-0.728492392955404,0.263112829634638,0.728492392955404},
    {1.0-0.728492392955404-0.263112829634638,0.728492392955404,0.263112829634638}
  };
#else
  static double const abscissa[1][3];
#error "interpolation_method_hcsbb: the specified gauss order is currently not supported."
#endif

  for (size_t i = 0; i < HCSBB_NUM_GAUSS_POINTS; ++i) {

    gauss_vertices[i][0] = bnd_triangle[0][0] * abscissa[i][0] +
                           bnd_triangle[1][0] * abscissa[i][1] +
                           bnd_triangle[2][0] * abscissa[i][2];
    gauss_vertices[i][1] = bnd_triangle[0][1] * abscissa[i][0] +
                           bnd_triangle[1][1] * abscissa[i][1] +
                           bnd_triangle[2][1] * abscissa[i][2];
    gauss_vertices[i][2] = bnd_triangle[0][2] * abscissa[i][0] +
                           bnd_triangle[1][2] * abscissa[i][1] +
                           bnd_triangle[2][2] * abscissa[i][2];
    double scale = 1.0 / sqrt(gauss_vertices[i][0] * gauss_vertices[i][0] +
                              gauss_vertices[i][1] * gauss_vertices[i][1] +
                              gauss_vertices[i][2] * gauss_vertices[i][2]);
    gauss_vertices[i][0] *= scale;
    gauss_vertices[i][1] *= scale;
    gauss_vertices[i][2] *= scale;
    gauss_sb_coords[i][0] = abscissa[i][0] * scale;
    gauss_sb_coords[i][1] = abscissa[i][1] * scale;
    gauss_sb_coords[i][2] = abscissa[i][2] * scale;
  }
}

static void inverse(double A[HCSBB_NUM_SBB_COEFFS][HCSBB_NUM_SBB_COEFFS]) {

// LAPACKE_dsytrf_work and LAPACKE_dsytri might not be available (see yac_lapack_interface.h).
#ifdef YAC_LAPACK_NO_DSYTR
  lapack_int ipiv[HCSBB_NUM_SBB_COEFFS + 1] = {0};
  double work[HCSBB_NUM_SBB_COEFFS*HCSBB_NUM_SBB_COEFFS] = {0};

  if (LAPACKE_dgetrf(
        LAPACK_COL_MAJOR, (lapack_int) HCSBB_NUM_SBB_COEFFS,
        (lapack_int) HCSBB_NUM_SBB_COEFFS, &A[0][0], (lapack_int) HCSBB_NUM_SBB_COEFFS, ipiv) != 0 )
    yac_internal_abort_message("internal ERROR: dgetrf", __FILE__, __LINE__);

  if (LAPACKE_dgetri_work(
        LAPACK_COL_MAJOR, (lapack_int) HCSBB_NUM_SBB_COEFFS, &A[0][0],
        (lapack_int) HCSBB_NUM_SBB_COEFFS, ipiv, work,
        (lapack_int) (HCSBB_NUM_SBB_COEFFS * HCSBB_NUM_SBB_COEFFS)) != 0 )
    yac_internal_abort_message("internal ERROR: dgetri", __FILE__, __LINE__);
#else
  lapack_int ipiv[HCSBB_NUM_SBB_COEFFS] = {0};
  double work[HCSBB_NUM_SBB_COEFFS] = {0};

  if (LAPACKE_dsytrf_work(
        LAPACK_COL_MAJOR, 'L', (lapack_int) HCSBB_NUM_SBB_COEFFS , &A[0][0],
        (lapack_int) HCSBB_NUM_SBB_COEFFS, ipiv, work, (lapack_int) HCSBB_NUM_SBB_COEFFS) != 0)
    yac_internal_abort_message("internal ERROR: dsytrf", __FILE__, __LINE__);

  if (LAPACKE_dsytri_work(
        LAPACK_COL_MAJOR, 'L', (lapack_int) HCSBB_NUM_SBB_COEFFS , &A[0][0],
        (lapack_int) HCSBB_NUM_SBB_COEFFS, ipiv, work) != 0)
    yac_internal_abort_message("internal ERROR: dsytri", __FILE__, __LINE__);

  for (size_t i = 0; i < HCSBB_NUM_SBB_COEFFS; ++i)
    for (size_t j = i + 1; j < HCSBB_NUM_SBB_COEFFS; ++j)
      A[j][i] = A[i][j];
#endif
}

static void init_gauss_nnn_patch(
  struct gauss_nnn_patch * gauss_nnn_patch, double * bnd_triangle,
  double * gauss_vertices,
  double gauss_sbb_polynomials[HCSBB_NUM_SBB_COEFFS][HCSBB_NUM_GAUSS_POINTS],
  double C[HCSBB_NUM_SBB_COEFFS][HCSBB_NUM_SBB_COEFFS]) {

  // compute bounding triangle
  memcpy(&(gauss_nnn_patch->bnd_triangle[0][0]), bnd_triangle,
         9 * sizeof(*bnd_triangle));

  double gauss_sb_coords[HCSBB_NUM_GAUSS_POINTS][3];

  // compute gauss points
  generate_gauss_legendre_points(
    (double(*)[3])gauss_vertices, gauss_sb_coords, (double(*)[3])bnd_triangle);

  double (*w_g)[HCSBB_NUM_GAUSS_POINTS] =
    xmalloc(HCSBB_NUM_SBB_COEFFS * sizeof(w_g[0]));
  gauss_nnn_patch->weights = (double*)w_g;

  // compute w_g
  {
    // compute the spherical Bernstein basis polynomials for the
    // gauss points (P_g)
    compute_sbb_polynomials_gauss_3d(gauss_sb_coords, gauss_sbb_polynomials);

    // P_g^T * P_g
    // (actually this is a triangular matrix -> computing halve would be enough)
    for (size_t i = 0; i < HCSBB_NUM_SBB_COEFFS; ++i) {
      for (size_t j = 0; j < HCSBB_NUM_SBB_COEFFS; ++j) {
        double accu = 0;
        for (size_t k = 0; k < HCSBB_NUM_GAUSS_POINTS; ++k) {
          accu += gauss_sbb_polynomials[i][k] * gauss_sbb_polynomials[j][k];
        }
        C[i][j] = accu;
      }
    }

    // (P_p^T * P_p)^-1
    inverse(C);

    // (P_p^T * P_p)^-1 * P_p^T
    for (size_t i = 0; i < HCSBB_NUM_SBB_COEFFS; ++i) {
      for (size_t j = 0; j < HCSBB_NUM_GAUSS_POINTS; ++j) {
        double accu = 0;
        for (size_t k = 0; k < HCSBB_NUM_SBB_COEFFS; ++k) {
          accu += C[i][k] * gauss_sbb_polynomials[k][j];
        }
        w_g[i][j] = accu;
      }
    }
  }
}

// check whether all required gauss nnn patches have valid data for the
// interpolation of the given target point
static int check_interp_data(struct tgt_point_search_data * tgt_point_data) {

  struct edge_interp_data * edge;
  struct edge_interp_data * edges[3];

  switch (tgt_point_data->type_flag) {

    case (ON_VERTEX):
      return 1;
    case (ON_EDGE):
      edge = tgt_point_data->interp_data.edge;
      // check derivative data of both vertices
      return ((edge->vertex_d_data[0]->gauss_nnn_patch != NULL) &&
              (edge->vertex_d_data[0]->gauss_nnn_patch->num_src_points > 0)) &&
             ((edge->vertex_d_data[1]->gauss_nnn_patch != NULL) &&
              (edge->vertex_d_data[1]->gauss_nnn_patch->num_src_points > 0));
    case (ON_TRIANGLE):
      edges[0] = tgt_point_data->interp_data.triangle->edges[0];
      edges[1] = tgt_point_data->interp_data.triangle->edges[1];
      edges[2] = tgt_point_data->interp_data.triangle->edges[2];
      return
        // check derivative data of the three vertices and the three edge middle
        // points
        ((edges[0]->vertex_d_data[0]->gauss_nnn_patch != NULL) &&
         (edges[0]->vertex_d_data[0]->gauss_nnn_patch->num_src_points > 0)) &&
        ((edges[0]->vertex_d_data[1]->gauss_nnn_patch != NULL) &&
         (edges[0]->vertex_d_data[1]->gauss_nnn_patch->num_src_points > 0)) &&
        ((edges[2]->vertex_d_data[0]->gauss_nnn_patch != NULL) &&
         (edges[2]->vertex_d_data[0]->gauss_nnn_patch->num_src_points > 0)) &&
        ((edges[0]->middle_d_data->gauss_nnn_patch != NULL) &&
         (edges[0]->middle_d_data->gauss_nnn_patch->num_src_points > 0)) &&
        ((edges[1]->middle_d_data->gauss_nnn_patch != NULL) &&
         (edges[1]->middle_d_data->gauss_nnn_patch->num_src_points > 0)) &&
        ((edges[2]->middle_d_data->gauss_nnn_patch != NULL) &&
         (edges[2]->middle_d_data->gauss_nnn_patch->num_src_points > 0));
    default:
      yac_internal_abort_message(
        "ERROR(check_interp_data): invalid interp_type_flag\n",
        __FILE__, __LINE__);
      return 0;
  }
}

// processes results for the nnn search for the gauss integration points
static void nnn_finish_callback_2(
  void * user_data, struct request_point * request_points) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb *)(data->method_stack[0]);
  struct communicator * comm = yac_get_domain_comm(method_hcsbb->domain);
  struct basic_patch_data * patch_data = &(data->patch_data);
  struct search_data * search_data = &(data->search_data);
  size_t num_tgt_points = search_data->num_tgt_points;
  struct tgt_point_search_data * tgt_point_data = search_data->tgt_point_data;
  struct gauss_nnn_patch * curr_gauss_nnn_patch = patch_data->gauss_nnn_patches;
  size_t num_gauss_nnn_patches = patch_data->num_gauss_nnn_patches;

  // generate the nnn weights for the interpolation of the gauss integration
  // points which are required for the derivative estimates
  struct result_point (*result_point_buffer)[HCSBB_GAUSS_NNN] =
    xmalloc(HCSBB_NUM_GAUSS_POINTS * sizeof(*result_point_buffer));
  double (*w_nnn_buffer)[HCSBB_NUM_GAUSS_POINTS] =
    xmalloc(HCSBB_GAUSS_NNN * HCSBB_NUM_GAUSS_POINTS * sizeof(*w_nnn_buffer));
  for (size_t i = 0; i < num_gauss_nnn_patches; ++i)
    compute_gauss_nnn_patch(
      curr_gauss_nnn_patch++, request_points + i * HCSBB_NUM_GAUSS_POINTS,
      result_point_buffer, w_nnn_buffer);
  free(w_nnn_buffer);
  free(result_point_buffer);

  yac_nnn_search_free_request_points(
    request_points, num_gauss_nnn_patches * HCSBB_NUM_GAUSS_POINTS);
  free(request_points);

  size_t num_interpolated_points = 0;

  // check which target points actually can be interpolated
  {
    size_t i = 0, flag = 0;
    while ((i < num_tgt_points) && (tgt_point_data[i].status_flag == LOCAL)) {

      if (check_interp_data(tgt_point_data + i))
        ++num_interpolated_points;
      else {
        tgt_point_data[i].status_flag = NOT;
        flag = 1;
      }
      ++i;
    }

    // if there was a point, which was marked as local but can not be
    // interpolated
    if (flag)
      // sort the target point data by their interpolation status
      // (first LOCAL then REMOTE and at last NOT)
      qsort(tgt_point_data, num_tgt_points, sizeof(*tgt_point_data),
            compare_tgt_point_data_flag_global_id);
  }


  // gather all local and remote global ids
  // (the ones that can be interpolated at the front)
  int * global_ids = data->global_ids;
  size_t i = 0;
  for (; (i < num_tgt_points) && (tgt_point_data[i].status_flag == LOCAL); ++i) {
    global_ids[i] = tgt_point_data[i].global_id;
  }
  size_t num_local_points = i;
  for (; (i < num_tgt_points) && (tgt_point_data[i].status_flag == REMOTE); ++i) {
    global_ids[i] = tgt_point_data[i].global_id;
  }
  size_t num_remote_points = i - num_local_points;

  // sort the local and remote target point data by global ids
  // (this is required by mark_tgt_points that is called in
  // finish_hcsbb_interpolation)
  qsort(tgt_point_data, i, sizeof(*tgt_point_data),
        compare_tgt_point_data_global_id);

  // exchange final results with other source processes
  yac_exchange_results(global_ids, num_local_points, num_local_points +
                       num_remote_points, method_hcsbb->tags.exchange_results,
                       comm, finish_hcsbb_interpolation, data);
}

static int compare_d_request_points_results(const void * a, const void * b) {

  struct request_point const * request_point_a =
    (struct request_point const *)a;
  struct request_point const * request_point_b =
    (struct request_point const *)b;

  size_t num_found_points_a = request_point_a->num_found_points;
  size_t num_found_points_b = request_point_b->num_found_points;

  int insufficient_src_points_a = num_found_points_a != HCSBB_D_NNN;
  int insufficient_src_points_b = num_found_points_b != HCSBB_D_NNN;

  // at least one request point as an insufficient number of source points
  if (insufficient_src_points_a || insufficient_src_points_b)
    // put the one with insufficient number of source points to the end
    return insufficient_src_points_a - insufficient_src_points_b;

  struct result_point * results_a = request_point_a->results;
  struct result_point * results_b = request_point_b->results;

  for (size_t i = 0; i < HCSBB_D_NNN; ++i) {
    int global_id_a = results_a[i].global_id;
    int global_id_b = results_b[i].global_id;
    if (global_id_a != global_id_b) return global_id_a - global_id_b;
  }

  return request_point_a->global_id - request_point_b->global_id;
}

static int compare_bnd_triangle_gid(void const * a, void const * b) {

  double * bnd_triangle_a = ((struct bnd_triangle_gid const *)a)->bnd_triangle;
  double * bnd_triangle_b = ((struct bnd_triangle_gid const *)b)->bnd_triangle;

  if (a == bnd_triangle_a) return 1;
  if (b == bnd_triangle_b) return -1;
  if (bnd_triangle_a == bnd_triangle_b) return 0;

  for (size_t i = 0; i < 9; ++i)
    if (fabs(bnd_triangle_a[i] - bnd_triangle_b[i]) > 0.0)
      return (bnd_triangle_a[i] > bnd_triangle_b[i]) -
             (bnd_triangle_a[i] < bnd_triangle_b[i]);

  return 0;
}

// processes results for the nnn search for the points at which we need to
// estimate the derivative of the source field
static void nnn_finish_callback(
  void * user_data, struct request_point * request_points) {

  struct intermediate_method_data * data = 
    (struct intermediate_method_data *)user_data;
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb *)(data->method_stack[0]);
  struct basic_patch_data * patch_data = &(data->patch_data);
  size_t num_requested_points =
    patch_data->d_data_src_point_count + patch_data->d_data_middle_point_count;

  size_t new_num_requested_points = 0;

  size_t * request_point_to_gauss_nnn_patch =
    xmalloc(num_requested_points * sizeof(*request_point_to_gauss_nnn_patch));

  for (size_t i = 0; i < num_requested_points; ++i) {
    // if enough source points were found for the current request point
    if (request_points[i].num_found_points >= HCSBB_D_NNN) {

      // if there are more than HCSBB_D_NNN results
      if (request_points[i].num_found_points > HCSBB_D_NNN)
        // sort by distance first and global source ids second
        qsort(request_points[i].results, request_points[i].num_found_points,
              sizeof(*(request_points[i].results)), compare_result_points_distance);

      // sort the first HCSBB_D_NNN result points by global source ids
      // (independent of distance)
      qsort(
        request_points[i].results, request_points[i].num_found_points,
        sizeof(*(request_points[i].results)), compare_result_points_global_id);

      // we only need the first HCSBB_D_NNN results
      request_points[i].num_found_points = HCSBB_D_NNN;
      new_num_requested_points++;
    } else {
      request_point_to_gauss_nnn_patch[i] = -1;
    }
  }

  // sort request points based on their search results
  // (put the ones with insufficient number of source points to the end)
  qsort(request_points, num_requested_points, sizeof(*request_points),
        compare_d_request_points_results);

  struct request_point dummy_request_point = {
    .global_id = -1, .results = NULL, .num_found_points = 0};
  struct request_point * prev_request_point = &dummy_request_point;

  // generate bounding triangle for all unqiue search results
  struct bnd_triangle_gid * bnd_triangle_gid =
    xmalloc(new_num_requested_points * sizeof(*bnd_triangle_gid));
  double * bnd_triangle_buffer =
    xmalloc(9 * new_num_requested_points * sizeof(*bnd_triangle_buffer));
  double * prev_bnd_triangle = bnd_triangle_buffer - 9;
  for (size_t i = 0; i < new_num_requested_points; ++i) {

    struct request_point * curr_request_point = request_points + i;
    struct bnd_triangle_gid * curr_bnd_triangle_gid = bnd_triangle_gid + i;

    // of the current results are different for the last unique result
    if (compare_d_request_points_results(
          prev_request_point, curr_request_point)) {

      // get coordinates of result points
      double coordinate_xyz[HCSBB_D_NNN][3];
      for (size_t j = 0; j < HCSBB_D_NNN; ++j) {
        coordinate_xyz[j][0] = curr_request_point->results[j].coordinate_xyz[0];
        coordinate_xyz[j][1] = curr_request_point->results[j].coordinate_xyz[1];
        coordinate_xyz[j][2] = curr_request_point->results[j].coordinate_xyz[2];
      }

      // compute bounding triangle for current result points
      double * curr_bnd_triangle = prev_bnd_triangle + 9;
      yac_compute_bnd_triangle(
        &(coordinate_xyz[0][0]), HCSBB_D_NNN, curr_bnd_triangle, 16);
      curr_bnd_triangle_gid->bnd_triangle = curr_bnd_triangle;

      prev_bnd_triangle = curr_bnd_triangle;
      prev_request_point = curr_request_point;
    } else {
      curr_bnd_triangle_gid->bnd_triangle = prev_bnd_triangle;
    }
    curr_bnd_triangle_gid->global_id = curr_request_point->global_id;
  }

  // sort the bnd_triangles
  qsort(bnd_triangle_gid, new_num_requested_points, sizeof(*bnd_triangle_gid),
        compare_bnd_triangle_gid);

  // determine which bounding triangles are unique (even though search results
  // are different, triangles can still be the same)
  // generate mapping, that is used to assign gauss patches to d_data
  // generate a gauss_nnn_patch for each unique bounding triangle
  struct bnd_triangle_gid * prev_bnd_triangle_gid = bnd_triangle_gid;
  size_t num_unique_bnd_triangle_gids = (new_num_requested_points > 0);
  for (size_t i = 0; i < new_num_requested_points; ++i) {
    struct bnd_triangle_gid * curr_bnd_triangle_gid = bnd_triangle_gid + i;
    if (compare_bnd_triangle_gid(
          prev_bnd_triangle_gid, curr_bnd_triangle_gid)) {
      prev_bnd_triangle_gid++;
      *prev_bnd_triangle_gid = *curr_bnd_triangle_gid;
      ++num_unique_bnd_triangle_gids;
    }
    request_point_to_gauss_nnn_patch[curr_bnd_triangle_gid->global_id] =
      num_unique_bnd_triangle_gids-1;
  }

  size_t num_gauss_nnn_patches = num_unique_bnd_triangle_gids;
  struct gauss_nnn_patch * gauss_nnn_patches =
    xmalloc(num_gauss_nnn_patches * sizeof(*gauss_nnn_patches));
  patch_data->gauss_nnn_patches = gauss_nnn_patches;
  patch_data->num_gauss_nnn_patches = num_gauss_nnn_patches;
  double (*gauss_vertices_xyz)[3] =
    xmalloc(num_gauss_nnn_patches * HCSBB_NUM_GAUSS_POINTS *
            sizeof(*gauss_vertices_xyz));

  // initialise a gauss integration point patch for each unique bounding triangle
  {
    double (*gauss_sbb_polynomials_buffer)[HCSBB_NUM_GAUSS_POINTS] =
      xmalloc(HCSBB_NUM_SBB_COEFFS * sizeof(*gauss_sbb_polynomials_buffer));
    double (*work_buffer)[HCSBB_NUM_SBB_COEFFS] =
      xmalloc(HCSBB_NUM_SBB_COEFFS * sizeof(*work_buffer));
    for (size_t i = 0; i < num_gauss_nnn_patches; ++i)
      init_gauss_nnn_patch(
        gauss_nnn_patches + i, bnd_triangle_gid[i].bnd_triangle,
        gauss_vertices_xyz[i * HCSBB_NUM_GAUSS_POINTS],
        gauss_sbb_polynomials_buffer, work_buffer);
    free(work_buffer);
    free(gauss_sbb_polynomials_buffer);
  }
  free(bnd_triangle_buffer);
  free(bnd_triangle_gid);

  // set gauss_nnn_patch links in derivative data
  struct derivative_data_src_point * d_data_src_point =
    patch_data->d_data_src_point;
  size_t d_data_src_point_count = patch_data->d_data_src_point_count;
  for (size_t i = 0; i < d_data_src_point_count; ++i)
    if (request_point_to_gauss_nnn_patch[i] != -1)
      d_data_src_point[i].gauss_nnn_patch =
        gauss_nnn_patches + request_point_to_gauss_nnn_patch[i];
  request_point_to_gauss_nnn_patch += d_data_src_point_count;
  struct derivative_data_middle_point * d_data_middle_point =
    patch_data->d_data_middle_point;
  size_t d_data_middle_point_count = patch_data->d_data_middle_point_count;
  for (size_t i = 0; i < d_data_middle_point_count; ++i)
    if (request_point_to_gauss_nnn_patch[i] != -1)
      d_data_middle_point[i].gauss_nnn_patch =
        gauss_nnn_patches + request_point_to_gauss_nnn_patch[i];
  request_point_to_gauss_nnn_patch -= d_data_src_point_count;

  int * global_ids =
    xmalloc(num_gauss_nnn_patches * HCSBB_NUM_GAUSS_POINTS *
            sizeof(*global_ids));
  for (size_t i = 0; i < num_gauss_nnn_patches * HCSBB_NUM_GAUSS_POINTS; ++i)
    global_ids[i] = (int)i;

  yac_nnn_search_free_request_points(request_points, num_requested_points);
  free(request_points);

  // do nnn search for all gauss integration points
  yac_nnn_search_do_search(
    method_hcsbb->nnn_search, HCSBB_GAUSS_NNN,
    num_gauss_nnn_patches * HCSBB_NUM_GAUSS_POINTS, gauss_vertices_xyz,
    global_ids, data, nnn_finish_callback_2);

  free(global_ids);
  free(gauss_vertices_xyz);
  free(request_point_to_gauss_nnn_patch);
}

/**
 * callback routine passed to yac_exchange_results by
 * do_search_interpolation_method_hcsbb
 */
static void do_search_continue_hcsbb(
  void * user_data, int * global_ids_to_mark, unsigned num_global_ids) {

  struct intermediate_method_data * data =
    (struct intermediate_method_data *)user_data;
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(data->method_stack[0]);
  struct tgt_point_search_data * tgt_point_data =
    data->search_data.tgt_point_data;

  //-------------------------------------------------------------------
  // mark target points that are assigned to another source process
  //-------------------------------------------------------------------
  mark_tgt_points(
    (size_t)num_global_ids, global_ids_to_mark, data->search_data, REMOTE);

  //--------------------------------------------------------------
  // sort tgt_point_data by status_flag then by:
  // if LOCAL:
  //    type_flag, src global ids, and tgt global ids
  // if not LOCAL
  //    tgt global_ids
  // (this is done to determine tgt points that are interpolated
  // by the same source points)
  //--------------------------------------------------------------
  qsort(tgt_point_data, data->search_data.num_tgt_points,
        sizeof(*tgt_point_data),
        compare_tgt_point_data_flag_LOCAL);

  //------------------------------------------------
  // count number of points that are marked as local
  //------------------------------------------------
  size_t num_local_points = 0;
  size_t num_tgt_points = data->search_data.num_tgt_points;
  while ((num_local_points < num_tgt_points) &&
         (tgt_point_data[num_local_points].status_flag == LOCAL))
    num_local_points++;

  size_t num_subdomains = (size_t)yac_get_num_subdomains(method_hcsbb->domain);
  size_t points_idx = method_hcsbb->points_idx;
  struct points *** points = method_hcsbb->points;
  struct grid ** point_grids = xmalloc(num_subdomains * sizeof(*point_grids));
  for (size_t i = 0; i < num_subdomains; ++i)
    point_grids[i] = yac_get_point_grid(points[i][points_idx]);

  double (*required_points_xyz)[3];
  size_t num_required_points;

  init_basic_patch_data(
    &(data->patch_data), tgt_point_data, num_local_points,
    &required_points_xyz, &num_required_points, point_grids,
    (unsigned)num_subdomains);

  free(point_grids);

  int * dummy_global_ids =
    xmalloc(num_required_points * sizeof(*dummy_global_ids));
  for (size_t i = 0; i < num_required_points; ++i) dummy_global_ids[i] = (int)i;

  yac_nnn_search_do_search(
    method_hcsbb->nnn_search, HCSBB_D_NNN, num_required_points,
    required_points_xyz, dummy_global_ids, user_data, nnn_finish_callback);

  free(dummy_global_ids);
  free(required_points_xyz);
}

static void check_src_point_cell(
  struct tgt_point_search_data * tgt_point_data, struct grid * point_grid,
  int const * global_ids, unsigned src_point_cell_id, unsigned subdomain_id,
  unsigned ** tmp_buffer, size_t * tmp_buffer_size) {

  unsigned * src_point_cell_ids;
  size_t num_src_point_cells_to_check;

  // in case the cell contains other types of edges, we might have to check
  // the neighbour cells
  if (yac_grid_cell_contains_non_gc_edge(point_grid, src_point_cell_id)) {

    struct dep_list * src_point_cell_neighs =
      yac_get_cell_neigh_dep_list(point_grid);

    num_src_point_cells_to_check =
      yac_get_num_dependencies_of_element(
         src_point_cell_neighs, (size_t)src_point_cell_id) + 1;

    ENSURE_ARRAY_SIZE(
      *tmp_buffer, *tmp_buffer_size, num_src_point_cells_to_check);
    src_point_cell_ids = *tmp_buffer;

    src_point_cell_ids[0] = src_point_cell_id;
    memcpy(
      src_point_cell_ids+1,
        yac_get_dependencies_of_element(
          src_point_cell_neighs, (size_t)src_point_cell_id),
      (num_src_point_cells_to_check-1)*sizeof(*src_point_cell_ids));

  } else {

    src_point_cell_ids = &src_point_cell_id;
    num_src_point_cells_to_check = 1;
  }

  enum interp_status_flag status_flag = tgt_point_data->status_flag;

  for (size_t i = 0; (i < num_src_point_cells_to_check) && (status_flag == NOT);
       ++i) {

    unsigned curr_src_point_cell_id = src_point_cell_ids[i];
    unsigned const * src_point_corner_local_ids =
      yac_get_cell_corner_indices(point_grid, curr_src_point_cell_id);
    size_t num_cell_corners =
      (size_t)yac_get_num_cell_corners(point_grid, curr_src_point_cell_id);

    size_t lowest_global_id_idx = 0;
    {
      int lowest_global_id = global_ids[src_point_corner_local_ids[0]];
      for (size_t j = 0; j < (size_t)(num_cell_corners); ++j) {
        if (global_ids[src_point_corner_local_ids[j]] < lowest_global_id) {
          lowest_global_id = global_ids[src_point_corner_local_ids[j]];
          lowest_global_id_idx = j;
        }
      }
    }

    size_t num_triangles = num_cell_corners - 2;
    size_t triangle_indices[num_triangles][3];
    yac_triangulate_cell_indices(
      src_point_corner_local_ids, num_cell_corners, lowest_global_id_idx,
      &(triangle_indices[0]));

    for (size_t j = 0; j < num_triangles; ++j) {

      unsigned local_ids[3] =
        {(unsigned)(triangle_indices[j][0]),
         (unsigned)(triangle_indices[j][1]),
         (unsigned)(triangle_indices[j][2])};
      int curr_global_ids[3] =
        {global_ids[local_ids[0]],
         global_ids[local_ids[1]],
         global_ids[local_ids[2]]};

      // this is done to ensure bit-reproducibility
      yac_quicksort_index(&(curr_global_ids[0]), 3, (int*)&(local_ids[0]));

      double triangle_vertices[3][3];
      double x_coord[3] = {yac_get_corner_x_coord(point_grid, local_ids[0]),
                           yac_get_corner_x_coord(point_grid, local_ids[1]),
                           yac_get_corner_x_coord(point_grid, local_ids[2])};
      double y_coord[3] = {yac_get_corner_y_coord(point_grid, local_ids[0]),
                           yac_get_corner_y_coord(point_grid, local_ids[1]),
                           yac_get_corner_y_coord(point_grid, local_ids[2])};
      LLtoXYZ(x_coord[0], y_coord[0], &(triangle_vertices[0][0]));
      LLtoXYZ(x_coord[1], y_coord[1], &(triangle_vertices[1][0]));
      LLtoXYZ(x_coord[2], y_coord[2], &(triangle_vertices[2][0]));

      double sb_coords[3] = {
        tgt_point_data->coordinate_xyz[0],
        tgt_point_data->coordinate_xyz[1],
        tgt_point_data->coordinate_xyz[2]};
      compute_sb_coords(&(sb_coords[0]), 1, triangle_vertices);

      double const tol = 1.0e-12;

      // if the point is in the current triangle
      if ((sb_coords[0] > -tol) && (sb_coords[1] > -tol) &&
          (sb_coords[2] > -tol)) {

        tgt_point_data->subdomain_id = subdomain_id;

        status_flag = LOCAL;

        int tmp_type_flag = -1;
        for (size_t k = 0; k < 3; ++k) {
          if (fabs(sb_coords[k]) > tol) {
            tmp_type_flag++;
            tgt_point_data->src_point_local_ids[tmp_type_flag] = local_ids[k];
            tgt_point_data->src_point_global_ids[tmp_type_flag] =
              curr_global_ids[k];
            tgt_point_data->sb_coords[tmp_type_flag] = sb_coords[k];
          }
        }
        tgt_point_data->type_flag = (enum interp_type_flag)tmp_type_flag;
        break;
      }
    }
  }
  tgt_point_data->status_flag = status_flag;
}

static void do_search_interpolation_method_hcsbb(
  struct interpolation_method ** method,
  struct interpolation_plan_element ** plan_stack, int * global_ids,
  double (*coordinates_xyz)[3], enum yac_location tgt_location,
  unsigned num_points, unsigned * num_interpolated_points,
  struct interpolation_method_data_request ** data_request) {

  if ( ( coordinates_xyz == NULL ) && num_points > 0 )
    yac_internal_abort_message("ERROR: missing target point coordinates "
                  "(coordinates_xyz == NULL)", __FILE__, __LINE__);

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(method[0]);

  method_hcsbb->data_request = data_request;

  struct domain * domain = method_hcsbb->domain;
  struct communicator * domain_comm = yac_get_domain_comm(domain);

  // initialise data that is required for the hybrid cubic spherical
  // Bernstein-Bezier patch interpolation
  struct intermediate_method_data * data = xmalloc(1 * sizeof(*data));
  data->method_stack = method;
  data->plan_stack = plan_stack;
  data->num_interpolated_points = num_interpolated_points;
  *num_interpolated_points = 0;
  data->coordinates_xyz = coordinates_xyz;
  data->global_ids = global_ids;
  data->tgt_location = tgt_location;
  struct tgt_point_search_data * tgt_point_data =
    tgt_point_search_data_new(
      num_points, coordinates_xyz, global_ids);
  size_t num_tgt_points = (size_t)num_points;
  data->search_data.tgt_point_data = tgt_point_data;
  data->search_data.num_tgt_points = num_tgt_points;
  data->search_data.weights = NULL;
  data->search_data.num_weights_per_tgt = NULL;
  data->search_data.total_num_weights = 0;

  size_t num_subdomains = (size_t)yac_get_num_subdomains(domain);

  unsigned * tmp_buffer = NULL;
  size_t tmp_buffer_size = 0;

  // do an initial search that tries to find for each target point a matching
  // local cell, that is made up of (vertices, cell points, or edge points,
  // depending on source location)
  for (size_t subdomain_idx = 0; subdomain_idx < num_subdomains;
       ++subdomain_idx) {

    struct points * points =
      method_hcsbb->points[subdomain_idx][method_hcsbb->points_idx];
    unsigned * mask =
      (method_hcsbb->use_source_mask)?
      method_hcsbb->masks[subdomain_idx][method_hcsbb->points_idx]->mask:NULL;
    struct grid * point_grid = yac_get_point_grid(points);
    enum yac_location src_location = method_hcsbb->src_location;
    int const * subdomain_global_ids =
      yac_get_global_ids(domain, (unsigned)subdomain_idx, src_location);

    struct dep_list * tgt_to_src_point_cell =
      yac_do_point_search_c3(
        yac_get_subdomain_grid_search(
          method_hcsbb->search, (unsigned)subdomain_idx), coordinates_xyz,
          (size_t)num_points, points);

    // for each target point
    for (size_t i = 0; i < (size_t)num_points; ++i) {

      struct tgt_point_search_data * curr_tgt_point_data = tgt_point_data + i;
      unsigned const * src_point_cell_ids =
        yac_get_dependencies_of_element(tgt_to_src_point_cell, i);
      size_t num_src_point_cell_ids =
        yac_get_num_dependencies_of_element(tgt_to_src_point_cell, i);

      for (size_t j = 0; j < num_src_point_cell_ids; ++j) {

        check_src_point_cell(
          curr_tgt_point_data, point_grid, subdomain_global_ids,
          src_point_cell_ids[j], (unsigned)subdomain_idx, &tmp_buffer,
          &tmp_buffer_size);

        // if some matching source points were found
        if (curr_tgt_point_data->status_flag == LOCAL) {

          size_t num_found_source_points =
            (size_t)(curr_tgt_point_data->type_flag)+1;

          if (mask != NULL) {
            // check halo and mask for these source points
            for (size_t k = 0; k < num_found_source_points; ++k) {
              if (mask[curr_tgt_point_data->src_point_local_ids[k]]) {
                curr_tgt_point_data->status_flag = NOT;
                break;
              }
            }
          }

          // if status is still LOCAL
          // check for remote points
          if (curr_tgt_point_data->status_flag == LOCAL) {
            size_t flag = 1;
            for (size_t k = 0; k < num_found_source_points; ++k)
              if (!yac_is_halo_value(
                    domain, (unsigned)subdomain_idx, src_location,
                    curr_tgt_point_data->src_point_local_ids[k], 
                    &(curr_tgt_point_data->ranks[k]))) {

                // mark the point as being local
                curr_tgt_point_data->ranks[k] = (unsigned)(-1);
                flag = 0;
              }

            // if there is at least one local source point for this target point
            if (flag) curr_tgt_point_data->status_flag = NOT;
          }
          break;
        }
      }
    }

    yac_dep_list_delete(tgt_to_src_point_cell);

    // give the communicator layer the chance to process messages
    yac_comm_check(domain_comm);
  }

  free(tmp_buffer);

  // sort the target point data by global ids
  qsort(tgt_point_data, num_tgt_points, sizeof(*tgt_point_data),
        compare_tgt_point_data_global_id);

  // count the number of target points for which at least one matching local
  // source cell was found
  unsigned num_matching_tgt_points = 0;
  for (unsigned i = 0; i < num_points; ++i)
    if (tgt_point_data[i].status_flag == LOCAL)
      global_ids[num_matching_tgt_points++] = tgt_point_data[i].global_id;
  for (unsigned i = 0, j = num_matching_tgt_points; i < num_points; ++i)
    if (tgt_point_data[i].status_flag != LOCAL)
      global_ids[j++] = tgt_point_data[i].global_id;

  // exchange the cells results of the initial search with the other source
  // processes and determine points for which multiple source processes have
  // found matching cells once this exchange is finished the algorithms
  // continues in the routine do_search_continue_nnn
  yac_exchange_results(global_ids, num_matching_tgt_points, num_points,
                       method_hcsbb->tags.exchange_initial_results,
                       domain_comm, do_search_continue_hcsbb, data);
}

/**
 * blocking routine that ensures that the search is finished and everything
 * is cleaned up
 */
static void enddef_interpolation_method_hcsbb(
  struct interpolation_method ** method) {

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(method[0]);

  struct communicator * domain_comm = yac_get_domain_comm(method_hcsbb->domain);
  struct interpolation_method_hcsbb_tags tags = method_hcsbb->tags;
  struct nnn_search * nnn_search = method_hcsbb->nnn_search;

  while(yac_comm_wait(domain_comm, tags.exchange_initial_results[0]) ||
        yac_comm_wait(domain_comm, tags.exchange_initial_results[1]) ||
        yac_comm_wait(domain_comm, tags.exchange_results[0]) ||
        yac_comm_wait(domain_comm, tags.exchange_results[1]) ||
        yac_nnn_search_wait(nnn_search));

  if (method[1] != NULL) method[1]->vtable->enddef(method + 1);
}

static void get_source_exchange_data_interpolation_method_hcsbb (
  struct interpolation_method ** method,
  struct interpolation_method_data_request *** data_request) {
 
  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)(method[0]);
 
  *data_request = method_hcsbb->data_request;
  method_hcsbb->data_request = NULL;
}

static void delete_interpolation_method_hcsbb(
  struct interpolation_method * method) {

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)method;
  
  yac_nnn_search_delete(method_hcsbb->nnn_search);
  free(method);
}

static struct global_search *
get_global_search(struct interpolation_method * method) {

  return ((struct interpolation_method_hcsbb*)(method))->search;
}

static void set_finish_callback(
  struct interpolation_method * method, void (*callback)(void * user_data),
  void * user_data) {

  struct interpolation_method_hcsbb * method_hcsbb =
    (struct interpolation_method_hcsbb*)method;
  
  method_hcsbb->finish_callback = callback;
  method_hcsbb->finish_user_data = user_data;
}

