/**
 * @file patch_recovery.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "yac_lapack_interface.h"

#include "geometry.h"
#include "utils.h"
#include "patch_recovery.h"

/* number of gauss integration points in a unit triangle */

// TODO: move to a central place or include file and make it accessible for more routines

static unsigned const num_gauss_points[13] = {1,1,3,4,6,7,12,13,16,19,25,27,33};
static void inverse ( double* A, int n );
static void transposed ( double* A, double* B, int m, int n );

static int mmdotprod ( double* A, int d_a1, int d_a2, double* B, int d_b1, int d_b2,  double* C, int d_c1, int d_c2 );

unsigned yac_get_number_of_quadrature_points ( unsigned order ) {
  if ( order < 1 || order > 12 ) {
    printf ( "Quadrature of order %u not supported.\n", order );
    yac_internal_abort_message ( "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }
  return num_gauss_points[order];
}

/* number of coefficients for a 3D polynomial */

static unsigned const num_polynomial_coefficients[4] = {1,4,10,20};

unsigned yac_get_num_polynomial_coefficients ( unsigned order ) {
  if ( order > 3 ) {
    printf ( "Polynom order %u not supported.\n", order );
    yac_internal_abort_message ( "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }
  return num_polynomial_coefficients[order];
}

/* Abscissa (and weights) for 2D elements */

void yac_linear_gauss_legendre ( unsigned order, double * abscissa ) {

  double const epsilon = 1.0e-14; // relative precision 

  double p1, p2, p3, pp, z, z1;

  /* roots are symmetric in the interval so we only have to find half of them */

  unsigned m = (order+1)/2;

  for ( unsigned i = 0; i < m; ++i ) {

    z = cos ( M_PI * ( i + 0.75 ) / ( order + 0.5 ) );

    /* Starting with the above approximation to the ith root,
       we enter the main loop of refinement by Newtons's method */

    do {

      p1 = 1.0;
      p2 = 0.0;

      for ( unsigned j = 0; j < order; ++j ) {
        p3 = p2;
        p2 = p1;
        p1 = (( 2.0 * j + 1.0  ) * z * p2 - j * p3 ) / ( j + 1 );
      }

      /* p1 is now the desired Legendre polynomial. We next compute pp,
         its derivative, by a standard relation involving also p2, the
         polynomial of one lower order */

      pp = order * ( z * p1 - p2 ) / ( z*z - 1.0 );
      z1 = z;
      z  = z1 - p1/pp;

    } while ( fabs(z-z1) > epsilon );

    /* Abscissa and its symmetric counterpart already scaled to [0,1] */

    abscissa[i]         = 0.5*(1.0-z);
    abscissa[order-1-i] = 0.5*(1.0+z);

  }

}

void yac_triangle_gauss_legendre ( unsigned gauss_order, double * abscissa ) {

  // TODO: move to a central place or include file and make it accessible for more routines

  if ( gauss_order == 1 ) {
    abscissa[0]  =  0.33333333333333333;
    abscissa[1]  =  0.33333333333333333;
  }
  else if ( gauss_order == 2 ) {
    abscissa[0]  =  0.16666666666666667;
    abscissa[1]  =  0.16666666666666667;
    abscissa[2]  =  0.66666666666666667;
    abscissa[3]  =  0.16666666666666667;
    abscissa[4]  =  0.16666666666666667;
    abscissa[5]  =  0.66666666666666667;
  }
  else if ( gauss_order == 3 ) {
    abscissa[0]  =  0.33333333333333333;
    abscissa[1]  =  0.33333333333333333;
    abscissa[2]  =  0.20000000000000000;
    abscissa[3]  =  0.20000000000000000;
    abscissa[4]  =  0.60000000000000000;
    abscissa[5]  =  0.20000000000000000;
    abscissa[6]  =  0.20000000000000000;
    abscissa[7]  =  0.60000000000000000;
  }
  else if ( gauss_order == 4 ) {
    abscissa[0]  =  0.091576213509771;
    abscissa[1]  =  0.091576213509771;
    abscissa[2]  =  0.816847572980459;
    abscissa[3]  =  0.091576213509771;
    abscissa[4]  =  0.091576213509771;
    abscissa[5]  =  0.816847572980459;
    abscissa[6]  =  0.445948490915965;
    abscissa[7]  =  0.445948490915965;
    abscissa[8]  =  0.108103018168070;
    abscissa[9]  =  0.445948490915965;
    abscissa[10] =  0.445948490915965;
    abscissa[11] =  0.108103018168070;
  }
  else if ( gauss_order == 5 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.101286507323456;
    abscissa[3]  =  0.101286507323456;
    abscissa[4]  =  0.797426985353087;
    abscissa[5]  =  0.101286507323456;
    abscissa[6]  =  0.101286507323456;
    abscissa[7]  =  0.797426985353087;
    abscissa[8]  =  0.470142064105115;
    abscissa[9]  =  0.470142064105115;
    abscissa[10] =  0.059715871789770;
    abscissa[11] =  0.470142064105115;
    abscissa[12] =  0.470142064105115;
    abscissa[13] =  0.059715871789770;
  }
  else if ( gauss_order == 6 ) {
    abscissa[0]  =  0.063089014491502;
    abscissa[1]  =  0.063089014491502;
    abscissa[2]  =  0.873821971016996;
    abscissa[3]  =  0.063089014491502;
    abscissa[4]  =  0.063089014491502;
    abscissa[5]  =  0.873821971016996;
    abscissa[6]  =  0.249286745170910;
    abscissa[7]  =  0.249286745170910;
    abscissa[8]  =  0.501426509658179;
    abscissa[9]  =  0.249286745170910;
    abscissa[10] =  0.249286745170910;
    abscissa[11] =  0.501426509658179;
    abscissa[12] =  0.310352451033785;
    abscissa[13] =  0.053145049844816;
    abscissa[14] =  0.053145049844816;
    abscissa[15] =  0.310352451033785;
    abscissa[16] =  0.636502499121399;
    abscissa[17] =  0.053145049844816;
    abscissa[18] =  0.053145049844816;
    abscissa[19] =  0.636502499121399;
    abscissa[20] =  0.636502499121399;
    abscissa[21] =  0.310352451033785;
    abscissa[22] =  0.310352451033785;
    abscissa[23] =  0.636502499121399;
  }
  else if ( gauss_order == 7 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.260345966079038;
    abscissa[3]  =  0.260345966079038;
    abscissa[4]  =  0.479308067841923;
    abscissa[5]  =  0.260345966079038;
    abscissa[6]  =  0.260345966079038;
    abscissa[7]  =  0.479308067841923;
    abscissa[8]  =  0.065130102902216;
    abscissa[9]  =  0.065130102902216;
    abscissa[10] =  0.869739794195568;
    abscissa[11] =  0.065130102902216;
    abscissa[12] =  0.065130102902216;
    abscissa[13] =  0.869739794195568;
    abscissa[14] =  0.312865496004874;
    abscissa[15] =  0.048690315425316;
    abscissa[16] =  0.048690315425316;
    abscissa[17] =  0.312865496004874;
    abscissa[18] =  0.638444188569809;
    abscissa[19] =  0.048690315425316;
    abscissa[20] =  0.048690315425316;
    abscissa[21] =  0.638444188569809;
    abscissa[22] =  0.638444188569809;
    abscissa[23] =  0.312865496004874;
    abscissa[24] =  0.312865496004874;
    abscissa[25] =  0.638444188569809;
  }
  else if ( gauss_order == 8 ) {
	abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.081414823414554;
    abscissa[3]  =  0.459292588292723;
    abscissa[4]  =  0.459292588292723;
    abscissa[5]  =  0.081414823414554;
    abscissa[6]  =  0.459292588292723;
    abscissa[7]  =  0.459292588292723;
    abscissa[8]  =  0.658861384496480;
    abscissa[9]  =  0.170569307751760;
    abscissa[10] =  0.170569307751760;
    abscissa[11] =  0.658861384496480;
    abscissa[12] =  0.170569307751760;
    abscissa[13] =  0.170569307751760;
    abscissa[14] =  0.898905543365938;
    abscissa[15] =  0.050547228317031;
    abscissa[16] =  0.050547228317031;
    abscissa[17] =  0.898905543365938;
    abscissa[18] =  0.050547228317031;
    abscissa[19] =  0.050547228317031;
    abscissa[20] =  0.008394777409958;
    abscissa[21] =  0.263112829634638;
    abscissa[22] =  0.263112829634638;
    abscissa[23] =  0.008394777409958;
    abscissa[24] =  0.008394777409958;
    abscissa[25] =  0.728492392955404;
    abscissa[26] =  0.728492392955404;
    abscissa[27] =  0.008394777409958;
    abscissa[28] =  0.263112829634638;
    abscissa[29] =  0.728492392955404;
    abscissa[30] =  0.728492392955404;
    abscissa[31] =  0.263112829634638;
  }
  else if ( gauss_order == 9 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.020634961602525;
    abscissa[3]  =  0.489682519198738;
    abscissa[4]  =  0.489682519198738;
    abscissa[5]  =  0.020634961602525;
    abscissa[6]  =  0.489682519198738;
    abscissa[7]  =  0.489682519198738;
    abscissa[8]  =  0.125820817014127;
    abscissa[9]  =  0.437089591492937;
    abscissa[10] =  0.437089591492937;
    abscissa[11] =  0.125820817014127;
    abscissa[12] =  0.437089591492937;
    abscissa[13] =  0.437089591492937;
    abscissa[14] =  0.623592928761935;
    abscissa[15] =  0.188203535619033;
    abscissa[16] =  0.188203535619033;
    abscissa[17] =  0.623592928761935;
    abscissa[18] =  0.188203535619033;
    abscissa[19] =  0.188203535619033;
    abscissa[20] =  0.910540973211095;
    abscissa[21] =  0.044729513394453;
    abscissa[22] =  0.044729513394453;
    abscissa[23] =  0.910540973211095;
    abscissa[24] =  0.044729513394453;
    abscissa[25] =  0.044729513394453;
    abscissa[26] =  0.036838412054736;
    abscissa[27] =  0.221962989160766;
    abscissa[28] =  0.036838412054736;
    abscissa[29] =  0.741198598784498;
    abscissa[30] =  0.221962989160766;
    abscissa[31] =  0.036838412054736;
    abscissa[32] =  0.741198598784498;
    abscissa[33] =  0.036838412054736;
    abscissa[34] =  0.221962989160766;
    abscissa[35] =  0.741198598784498;
    abscissa[36] =  0.741198598784498;
    abscissa[37] =  0.221962989160766;
  }
  else if ( gauss_order == 10 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.028844733232685;
    abscissa[3]  =  0.485577633383657;
    abscissa[4]  =  0.485577633383657;
    abscissa[5]  =  0.028844733232685;
    abscissa[6]  =  0.485577633383657;
    abscissa[7]  =  0.485577633383657;
    abscissa[8]  =  0.781036849029926;
    abscissa[9]  =  0.109481575485037;
    abscissa[10] =  0.109481575485037;
    abscissa[11] =  0.781036849029926;
    abscissa[12] =  0.109481575485037;
    abscissa[13] =  0.109481575485037;
    abscissa[14] =  0.141707219414880;
    abscissa[15] =  0.307939838764121;
    abscissa[16] =  0.141707219414880;
    abscissa[17] =  0.550352941820999;
    abscissa[18] =  0.307939838764121;
    abscissa[19] =  0.141707219414880;
    abscissa[20] =  0.550352941820999;
    abscissa[21] =  0.141707219414880;
    abscissa[22] =  0.307939838764121;
    abscissa[23] =  0.550352941820999;
    abscissa[24] =  0.550352941820999;
    abscissa[25] =  0.307939838764121;
    abscissa[26] =  0.025003534762686;
    abscissa[27] =  0.246672560639903;
    abscissa[28] =  0.025003534762686;
    abscissa[29] =  0.728323904597411;
    abscissa[30] =  0.246672560639903;
    abscissa[31] =  0.025003534762686;
    abscissa[32] =  0.728323904597411;
    abscissa[33] =  0.025003534762686;
    abscissa[34] =  0.246672560639903;
    abscissa[35] =  0.728323904597411;
    abscissa[36] =  0.728323904597411;
    abscissa[37] =  0.246672560639903;
    abscissa[38] =  0.009540815400299;
    abscissa[39] =  0.066803251012200;
    abscissa[40] =  0.009540815400299;
    abscissa[41] =  0.923655933587500;
    abscissa[42] =  0.066803251012200;
    abscissa[43] =  0.009540815400299;
    abscissa[44] =  0.923655933587500;
    abscissa[45] =  0.009540815400299;
    abscissa[46] =  0.066803251012200;
    abscissa[47] =  0.923655933587500;
    abscissa[48] =  0.923655933587500;
    abscissa[49] =  0.066803251012200;
  }
  else if ( gauss_order == 11 ) {
    abscissa[0]  = -0.069222096541517;
    abscissa[1]  =  0.534611048270758;
    abscissa[2]  =  0.534611048270758;
    abscissa[3]  = -0.069222096541517;
    abscissa[4]  =  0.534611048270758;
    abscissa[5]  =  0.534611048270758;
    abscissa[6]  =  0.202061394068290;
    abscissa[7]  =  0.398969302965855;
    abscissa[8]  =  0.398969302965855;
    abscissa[9]  =  0.202061394068290;
    abscissa[10] =  0.398969302965855;
    abscissa[11] =  0.398969302965855;
    abscissa[12] =  0.593380199137435;
    abscissa[13] =  0.203309900431282;
    abscissa[14] =  0.203309900431282;
    abscissa[15] =  0.593380199137435;
    abscissa[16] =  0.203309900431282;
    abscissa[17] =  0.203309900431282;
    abscissa[18] =  0.761298175434837;
    abscissa[19] =  0.119350912282581;
    abscissa[20] =  0.119350912282581;
    abscissa[21] =  0.761298175434837;
    abscissa[22] =  0.119350912282581;
    abscissa[23] =  0.119350912282581;
    abscissa[24] =  0.935270103777448;
    abscissa[25] =  0.032364948111276;
    abscissa[26] =  0.032364948111276;
    abscissa[27] =  0.935270103777448;
    abscissa[28] =  0.032364948111276;
    abscissa[29] =  0.032364948111276;
    abscissa[30] =  0.050178138310495;
    abscissa[31] =  0.356620648261293;
    abscissa[32] =  0.050178138310495;
    abscissa[33] =  0.593201213428213;
    abscissa[34] =  0.356620648261293;
    abscissa[35] =  0.050178138310495;
    abscissa[36] =  0.593201213428213;
    abscissa[37] =  0.050178138310495;
    abscissa[38] =  0.356620648261293;
    abscissa[39] =  0.593201213428213;
    abscissa[40] =  0.593201213428213;
    abscissa[41] =  0.356620648261293;
    abscissa[42] =  0.021022016536166;
    abscissa[43] =  0.171488980304042;
    abscissa[44] =  0.021022016536166;
    abscissa[45] =  0.807489003159792;
    abscissa[46] =  0.171488980304042;
    abscissa[47] =  0.021022016536166;
    abscissa[48] =  0.807489003159792;
    abscissa[49] =  0.021022016536166;
    abscissa[50] =  0.171488980304042;
    abscissa[51] =  0.807489003159792;
    abscissa[52] =  0.807489003159792;
    abscissa[53] =  0.171488980304042;
  } else if ( gauss_order == 12 ) {
    abscissa[0]  = 0.023565220452390;
    abscissa[1]  = 0.488217389773805;
    abscissa[2]  = 0.488217389773805;
    abscissa[3]  = 0.023565220452390;
    abscissa[4]  = 0.488217389773805;
    abscissa[5]  = 0.488217389773805;
    abscissa[6]  = 0.120551215411079;
    abscissa[7]  = 0.439724392294460;
    abscissa[8]  = 0.439724392294460;
    abscissa[9]  = 0.120551215411079;
    abscissa[10] = 0.439724392294460;
    abscissa[11] = 0.439724392294460;
    abscissa[12] = 0.457579229975768;
    abscissa[13] = 0.271210385012116;
    abscissa[14] = 0.271210385012116;
    abscissa[15] = 0.457579229975768;
    abscissa[16] = 0.271210385012116;
    abscissa[17] = 0.271210385012116;
    abscissa[18] = 0.744847708916828;
    abscissa[19] = 0.127576145541586;
    abscissa[20] = 0.127576145541586;
    abscissa[21] = 0.744847708916828;
    abscissa[22] = 0.127576145541586;
    abscissa[23] = 0.127576145541586;
    abscissa[24] = 0.957365299093579;
    abscissa[25] = 0.021317350453210;
    abscissa[26] = 0.021317350453210;
    abscissa[27] = 0.957365299093579;
    abscissa[28] = 0.021317350453210;
    abscissa[29] = 0.021317350453210;
    abscissa[30] = 0.115343494534698;
    abscissa[31] = 0.275713269685514;
    abscissa[32] = 0.115343494534698;
    abscissa[33] = 0.608943235779788;
    abscissa[34] = 0.275713269685514;
    abscissa[35] = 0.115343494534698;
    abscissa[36] = 0.608943235779788;
    abscissa[37] = 0.115343494534698;
    abscissa[38] = 0.275713269685514;
    abscissa[39] = 0.608943235779788;
    abscissa[40] = 0.608943235779788;
    abscissa[41] = 0.275713269685514;
    abscissa[42] = 0.022838332222257;
    abscissa[43] = 0.281325580989940;
    abscissa[44] = 0.022838332222257;
    abscissa[45] = 0.695836086787803;
    abscissa[46] = 0.281325580989940;
    abscissa[47] = 0.022838332222257;
    abscissa[48] = 0.695836086787803;
    abscissa[49] = 0.022838332222257;
    abscissa[50] = 0.281325580989940;
    abscissa[51] = 0.695836086787803;
    abscissa[52] = 0.695836086787803;
    abscissa[53] = 0.281325580989940;
    abscissa[54] = 0.025734050548330;
    abscissa[55] = 0.116251915907597;
    abscissa[56] = 0.025734050548330;
    abscissa[57] = 0.858014033544073;
    abscissa[58] = 0.116251915907597;
    abscissa[59] = 0.025734050548330;
    abscissa[60] = 0.858014033544073;
    abscissa[61] = 0.025734050548330;
    abscissa[62] = 0.116251915907597;
    abscissa[63] = 0.858014033544073;
    abscissa[64] = 0.858014033544073;
    abscissa[65] = 0.116251915907597;
  }
  else {
    printf ( "Quadrature of order %u not supported.\n", gauss_order );
    yac_internal_abort_message (
      "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }

}

/* ------------- unit quadrilateral --------------- */

void yac_quad_2d_locations ( unsigned order, double * abscissa ) {

  unsigned i, j;
  double absc[order];
 
  unsigned const dim = 2;

  yac_linear_gauss_legendre ( order, absc );

  for ( i = 0; i < order; ++i ) {
    for ( j = 0; j < order; ++j ) {
      abscissa[(i*order+j)*dim]   = absc[i];
      abscissa[(i*order+j)*dim+1] = absc[j];
    }
  }
}

/* ------------ transformation to triangle ------------- */

void yac_reference_to_xyz_triangle ( unsigned num_points, struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location ) {
  unsigned i;
  static unsigned const dim3 = 3;

  for ( i = 0; i < num_points; ++i ) {

    transformed_location[i*dim3]   = ( triangle.a[0] - triangle.c[0] ) * reference_location[i*dim3]
                                   + ( triangle.b[0] - triangle.c[0] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[0]   * reference_location[i*dim3+2]
                                   +                   triangle.c[0];

    transformed_location[i*dim3+1] = ( triangle.a[1] - triangle.c[1] ) * reference_location[i*dim3]
                                   + ( triangle.b[1] - triangle.c[1] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[1]   * reference_location[i*dim3+2]
                                   +                   triangle.c[1];

    transformed_location[i*dim3+2] = ( triangle.a[2] - triangle.c[2] ) * reference_location[i*dim3]
                                   + ( triangle.b[2] - triangle.c[2] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[2]   * reference_location[i*dim3+2]
                                   +                   triangle.c[2];
  }
}

void yac_xyz_triangle_to_reference ( unsigned num_points, struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location ) {
  unsigned i, j;
  static unsigned const dim = 3;

  double A[3*3] = { triangle.a[0] - triangle.c[0] , triangle.b[0] - triangle.c[0] , -triangle.c[0],
                    triangle.a[1] - triangle.c[1] , triangle.b[1] - triangle.c[1] , -triangle.c[1],
                    triangle.a[2] - triangle.c[2] , triangle.b[2] - triangle.c[2] , -triangle.c[2] };

  inverse ( (double *) A, dim );

  for ( i = 0; i < num_points; ++i )
     for ( j = 0; j < dim; ++j )
        reference_location[i*dim+j] = A[j*dim+0] * ( transformed_location[i*dim+0] - triangle.c[0] )
                                    + A[j*dim+1] * ( transformed_location[i*dim+1] - triangle.c[1] )
                                    + A[j*dim+2] * ( transformed_location[i*dim+2] - triangle.c[2] );
}

/* ------------ transformation to quadrilateral ------------- */

void yac_reference_to_xy_quadrilateral ( unsigned num_points, struct grid_cell cell,
                                         double * reference_location,
                                         double * transformed_location ) {
  unsigned i;
  static unsigned const dim = 2;

  double const xsi[4] = {-1.0,  1.0, 1.0, -1.0 };
  double const eta[4] = {-1.0, -1.0, 1.0,  1.0 };

  for ( i = 0; i < num_points; ++i ) {

    transformed_location[i*dim+0] = 0.0;
    transformed_location[i*dim+1] = 0.0;

    for ( size_t n = 0; n < cell.num_corners; ++ n ) {

      transformed_location[i*dim+0] += 0.25
        * ( 1 + xsi[n] * reference_location[i*dim+0] )
        * ( 1 + eta[n] * reference_location[i*dim+1] ) * cell.coordinates_x[n];

      transformed_location[i*dim+1] += 0.25
        * ( 1 + xsi[n] * reference_location[i*dim+0] )
        * ( 1 + eta[n] * reference_location[i*dim+1] ) * cell.coordinates_y[n];
    }
  }  
}

void yac_reference_to_xyz_quadrilateral ( unsigned gauss_order, struct grid_cell cell,
                                          double * reference_location,
                                          double * transformed_location ) {

   /* quadrilaterals (as well as triangles) could also be handled by xyz_polygon
      the resulting positions of integrations points would then be different. */

   static unsigned const dim2 = 2;
   static unsigned const dim3 = 3;

   static unsigned num_lowerhalf = 0;
   static unsigned num_upperhalf = 0;

   struct xyz_triangle xyz_triangle;

   double reference_3Dlocation[3*gauss_order*gauss_order];
   double buffer_3d[3*gauss_order*gauss_order];

   unsigned i, j;

   static double const tol = 1.0e-15;
 
   LLtoXYZ( cell.coordinates_x[0], cell.coordinates_y[0], xyz_triangle.a );
   LLtoXYZ( cell.coordinates_x[1], cell.coordinates_y[1], xyz_triangle.b );
   LLtoXYZ( cell.coordinates_x[2], cell.coordinates_y[2], xyz_triangle.c );

   num_lowerhalf = 0;

   for ( i = 0; i < gauss_order; ++i ) {
      for (j = 0; j < gauss_order; ++j ) {
         if ( reference_location[(i*gauss_order+j)*dim2]+reference_location[(i*gauss_order+j)*dim2+1] <= 1.0+tol ) {
            reference_3Dlocation[num_lowerhalf*dim3+0] = reference_location[(i*gauss_order+j)*dim2];
            reference_3Dlocation[num_lowerhalf*dim3+1] = reference_location[(i*gauss_order+j)*dim2+1];
            reference_3Dlocation[num_lowerhalf*dim3+2] = 0.0;
            ++num_lowerhalf;
         }
      }
   }

  yac_reference_to_xyz_triangle ( num_lowerhalf, xyz_triangle, reference_3Dlocation, buffer_3d );

  for ( i = 0; i < 3*num_lowerhalf; ++i )
     transformed_location[i] = buffer_3d[i];

  LLtoXYZ( cell.coordinates_x[2], cell.coordinates_y[2], xyz_triangle.a );
  LLtoXYZ( cell.coordinates_x[3], cell.coordinates_y[3], xyz_triangle.b );
  LLtoXYZ( cell.coordinates_x[0], cell.coordinates_y[0], xyz_triangle.c );

  num_upperhalf = 0;

  for ( i = 0; i < gauss_order; ++i ) {
     for (j = 0; j < gauss_order; ++j ) {
        if ( reference_location[(i*gauss_order+j)*dim2]+reference_location[(i*gauss_order+j)*dim2+1] < 1.0-tol ) {
           reference_3Dlocation[num_upperhalf*dim3+0] = reference_location[(i*gauss_order+j)*dim2];
           reference_3Dlocation[num_upperhalf*dim3+1] = reference_location[(i*gauss_order+j)*dim2+1];
           reference_3Dlocation[num_upperhalf*dim3+2] = 0.0;
           ++num_upperhalf;
        }
     }
  }

  yac_reference_to_xyz_triangle ( num_upperhalf, xyz_triangle, reference_3Dlocation, buffer_3d );

  for ( i = 0; i < 3*num_upperhalf; ++i )
     transformed_location[3*num_lowerhalf+i] = buffer_3d[i];

  if ( num_lowerhalf+num_upperhalf != gauss_order*gauss_order ) {
     printf("compared %u %u %u \n",  num_lowerhalf, num_upperhalf,
            gauss_order * gauss_order );
     yac_internal_abort_message ( "ERROR: Quadrature points do not match", __FILE__, __LINE__ );
  }

}

/* ------------ transformation to polygons ------------- */

static
void get_barycenter ( unsigned num_vertices, struct xyz_polygon polygon, double barycenter[] ) {

   unsigned n;
   unsigned const dim3 = 3;

   for ( n = 0; n < dim3; ++n )
	  barycenter[n] = 0.0;

   for ( n = 0; n < num_vertices; ++n ) {
      barycenter[0] += polygon.x[n];
      barycenter[1] += polygon.y[n];
      barycenter[2] += polygon.z[n];
   }

   for ( n = 0; n < dim3; ++n )
      barycenter[n] /= (double) num_vertices;

   /* project barycenter to surface of unit sphere */

   double length =
         sqrt ( barycenter[0]*barycenter[0] +
                barycenter[1]*barycenter[1] +
                barycenter[2]*barycenter[2] );

   for ( n = 0; n < dim3; ++n )
      barycenter[n] /= length;
}

void yac_reference_to_xyz_polygon ( unsigned num_gauss_points_,
                                    struct xyz_polygon polygon,
                                    double * reference_3Dlocation,
                                    double * transformed_location ) {

   /* the total number of transformed locations to be returned
      equals num_gauss_points_*num_vertices */

   unsigned n, ngp;

   double barycenter[3];
   struct xyz_triangle xyz_triangle;

   double buffer_3d[3 * num_gauss_points_];

   /* the barycenter should always give us a point inside a convex polygon */

   unsigned num_vertices = polygon.num_vertices;

   get_barycenter ( num_vertices, polygon, barycenter );

   for ( n = 0; n < 3; ++n )
      xyz_triangle.c[n] = barycenter[n];

   for ( n = 0; n < num_vertices; ++n ) {

      unsigned n0 = n%num_vertices;
      unsigned n1 = (n+1)%num_vertices;

      xyz_triangle.a[0] = polygon.x[n0];
      xyz_triangle.a[1] = polygon.y[n0];
      xyz_triangle.a[2] = polygon.z[n0];

      xyz_triangle.b[0] = polygon.x[n1];
      xyz_triangle.b[1] = polygon.y[n1];
      xyz_triangle.b[2] = polygon.z[n1];

      yac_reference_to_xyz_triangle(
        num_gauss_points_, xyz_triangle, reference_3Dlocation, buffer_3d);

      unsigned offset = n * num_gauss_points_ * 3;

      for ( ngp = 0; ngp < 3 * num_gauss_points_; ++ngp )
         transformed_location[offset+ngp] = buffer_3d[ngp];
   }
}


/* --------------- polynomial treatment ---------------- */
void yac_set_polynomial_terms ( unsigned order, unsigned num_rows,
                                double * coordinates, double * reference, double *PT ) {

  unsigned i;

  if ( order > 3 )
     yac_internal_abort_message ( "ERROR: polynomial order not in the range 0 <= order <= 3 ", __FILE__, __LINE__ );

  for ( i = 0; i < num_rows; ++i )
      PT[0*num_rows+i] = 1.0;

  if ( order == 0 ) return;

  if ( num_rows > 65536 ) yac_internal_abort_message("internal ERROR: too much memory required", __FILE__, __LINE__);

  double x[num_rows], y[num_rows], z[num_rows];

  /* actually this give us the transposed of the matrix if we think of P
     containing only line for each integration points. */

  for ( i = 0; i < num_rows; ++i ) {
    x[i] = coordinates[i*3+0] - reference[0];
    y[i] = coordinates[i*3+1] - reference[1];
    z[i] = coordinates[i*3+2] - reference[2];
  }

  for ( i = 0; i < num_rows; ++i ) {
      PT[1*num_rows+i] = x[i];
      PT[2*num_rows+i] = y[i];
      PT[3*num_rows+i] = z[i];
  }

  if ( order > 1 ) {
    for ( i = 0; i < num_rows; ++i ) {
      PT[4*num_rows+i] = x[i] * x[i];
      PT[5*num_rows+i] = y[i] * y[i];
      PT[6*num_rows+i] = z[i] * z[i];
      PT[7*num_rows+i] = x[i] * y[i];
      PT[8*num_rows+i] = x[i] * z[i];
      PT[9*num_rows+i] = y[i] * z[i];
    }
  }

  if ( order > 2 ) {
    for ( i = 0; i < num_rows; ++i ) {
      PT[10*num_rows+i] = x[i] * x[i] * x[i];
      PT[11*num_rows+i] = y[i] * y[i] * y[i];
      PT[12*num_rows+i] = z[i] * z[i] * z[i];
      PT[13*num_rows+i] = x[i] * x[i] * y[i];
      PT[14*num_rows+i] = x[i] * x[i] * z[i];
      PT[15*num_rows+i] = y[i] * y[i] * x[i];
      PT[16*num_rows+i] = y[i] * y[i] * z[i];
      PT[17*num_rows+i] = z[i] * z[i] * x[i];
      PT[18*num_rows+i] = z[i] * z[i] * y[i];
      PT[19*num_rows+i] = x[i] * y[i] * z[i];
    }
  }
}

void yac_get_polynomial_coefficients ( unsigned num_points, unsigned num_coeff,
                                       double *PT, double *b, double*a ) {

  lapack_int lwork = num_points * 64;
  lapack_int nrhs  = 1;

  /* int lwork = MIN(num_coeff, num_points) * 64;
     work = xmalloc ( lwork * sizeof(*work) ); */

  double work[num_points*64];

  /* Fortran ordering */

  lapack_int info = LAPACKE_dgels_work(LAPACK_COL_MAJOR, 'N',
                                       (lapack_int) num_points, (lapack_int) num_coeff,
                                       nrhs, PT, (lapack_int) num_points,
                                       b, (lapack_int) num_points, work, lwork);
  if ( info != 0 ) yac_internal_abort_message("internal ERROR: dgels", __FILE__, __LINE__);

  memcpy ( a , b, num_coeff * sizeof(*b) );

}

/* ----------- methods to calculate patch values ----------------- */

void yac_patch_P0weights ( unsigned num_points, unsigned num_coeff,
                           unsigned num_cells, unsigned * num_gausspoints,
                           double *Pt, double *PT, double *weight ) {

  unsigned i, n, k;
  unsigned mi, mj;
  unsigned offset[num_cells];

  double Minv[num_coeff*num_coeff];
  double MinvPT[num_coeff*num_points];

  // PT already depends on the individual target point, thus we have to
  // perform all operations for each target point individually.

  // Calculate elements (mi,mj) of M as sum over all n points
  //
  // Mathematically it is the matrix product of PT * P, where
  // PT is the transposed of P. Here we do two steps in one
  // and use PT directly but step through such that it is equivalent
  // to PT * P.

  for ( mi = 0; mi < num_coeff; ++mi ) {
     for ( mj = 0; mj < num_coeff; ++mj ) {
        Minv[mi*num_coeff+mj] = 0;
        for ( n = 0; n < num_points; ++n )
            Minv[mi*num_coeff+mj] += PT[mi*num_points+n]*PT[mj*num_points+n];
     }
  }

// #define YAC_VERBOSE_INT_PATCH
#ifdef YAC_VERBOSE_INT_PATCH
  //
  // print out matrix in Matlab format.
  //
  // In Matlab,
  //
  // B = inv(A)
  // C = A*B
  //
  // should deliver the identity matrix C
  //
  for ( mi = 0; mi < num_coeff; ++mi ) {
     for ( mj = 0; mj < num_coeff; ++mj ) {
    	 if ( mj == 0  && mi == 0 ) {
             printf ("A = [ %4.8f ", Minv[mi*num_coeff+mj]);
    	 }
    	 else if ( mj == num_coeff-1 && mi == num_coeff-1 ) {
             printf (" %4.8f ]\n", Minv[mi*num_coeff+mj]);
    	 }
    	 else if ( mj == num_coeff-1 && mi != num_coeff-1 ) {
             printf (" %4.8f ;\n", Minv[mi*num_coeff+mj]);
    	 }
    	 else {
             printf (" %4.8f ", Minv[mi*num_coeff+mj]);
    	 }
     }
  }
#endif

  inverse ( Minv, num_coeff );

  for ( n = 0; n < num_points; ++n ) {
     for ( mi = 0; mi < num_coeff; ++mi ) {
        MinvPT[n*num_coeff+mi] = 0;
        for ( mj = 0; mj < num_coeff; ++mj )
           MinvPT[n*num_coeff+mi] += Minv[mi*num_coeff+mj] * PT[mj*num_points+n];
     }
  }

  offset[0] = 0;
  for ( n = 1; n < num_cells; ++n )
	  offset[n] = offset[n-1]+num_gausspoints[n];

  for ( n = 0; n < num_cells; ++n ) {
     weight[n] = 0.0;
     for ( i = 0; i < num_gausspoints[n]; ++i )
        for ( k = 0; k < num_coeff; ++k )
           weight[n] += Pt[k] * MinvPT[(offset[n]+i)*num_coeff+k];
  }

}

void yac_patch_P0weights_opt ( unsigned num_points, unsigned num_coeff,
                               unsigned num_cells, unsigned * num_gausspoints,
                               double *PT, double *weight ) {

  unsigned i, n;
  unsigned mi, mj;

  unsigned offset[num_cells];
  double Minv[num_coeff*num_coeff];
  double MinvPT[num_points];

  // PT already depends on the individual target point, thus we
  // have to perform all operations for each target point individually.

  // Calculate elements (mi,mj) of M as sum over all n points
  //
  // Mathematically it is the matrix product of PT * P, where
  // PT is the transposed of P. Here we do two steps in one
  // and use PT directly but step through such that it is equivalent
  // to PT * P.

  for ( mi = 0; mi < num_coeff; ++mi ) {
     for ( mj = 0; mj < num_coeff; ++mj ) {
        Minv[mi*num_coeff+mj] = 0;
        for ( n = 0; n < num_points; ++n )
            Minv[mi*num_coeff+mj] += PT[mi*num_points+n]*PT[mj*num_points+n];
     }
  }

#ifdef YAC_VERBOSE_INT_PATCH
  //
  // print out matrix in Matlab format.
  //
  // In Matlab,
  //
  // B = inv(A)
  // C = A*B
  //
  // should deliver the identity matrix C
  //
  for ( mi = 0; mi < num_coeff; ++mi ) {
	  for ( mj = 0; mj < num_coeff; ++mj ) {
		  if ( num_coeff > 1 ) {
			  if ( mj == 0  && mi == 0 ) {
	              printf ("A = [ %4.8f ", Minv[mi*num_coeff+mj]);
			  }
			  else if ( mj == num_coeff-1 && mi == num_coeff-1 ) {
	              printf (" %4.8f ]\n", Minv[mi*num_coeff+mj]);
			  }
			  else if ( mj == num_coeff-1 && mi != num_coeff-1 ) {
				  printf (" %4.8f ;\n", Minv[mi*num_coeff+mj]);
			  }
			  else {
	              printf (" %4.8f ", Minv[mi*num_coeff+mj]);
			  }
		  }
	  }
  }
#endif

  inverse ( Minv, num_coeff );

  // By construction the Pt[0] = 1.0 and Pt[n] = 0.0 for all n > 0
  // in the non-optimised version above. Pt, the polynomial evaluated
  // at the target position is different in meaning from PT, the transposed
  // of the polynomial P.
  //
  // Compared to the full version above we can therefore drop some
  // loops. We cannot do much above these lines as we still need
  // the full matrices to calculate the inverse.

  for ( n = 0; n < num_points; ++n ) MinvPT[n] = 0;

  for ( mj = 0; mj < num_coeff; ++mj )
     for ( n = 0; n < num_points; ++n )
        MinvPT[n] += Minv[mj] * PT[mj*num_points+n];

  offset[0] = 0;
  for ( n = 1; n < num_cells; ++n )
	  offset[n] = offset[n-1]+num_gausspoints[n];

  for ( n = 0; n < num_cells; ++n ) {
     weight[n] = 0.0;
#ifdef _CRAYC
#if (_RELEASE_MAJOR == 8) && (_RELEASE_MINOR == 3)
#pragma _CRI novector
#endif
#endif
     for ( i = 0; i < num_gausspoints[n]; ++i )
        weight[n] += MinvPT[offset[n]+i];
  }
}

double yac_patch_P0value ( unsigned num_cells, double *PtMinvPT, double *fi ) {

   unsigned n;
   double value = 0.0;

   for ( n = 0; n < num_cells; ++n ) value += PtMinvPT[n] * fi[n];

   return value;
}

double yac_patch_value_alt1 ( unsigned num_points, unsigned num_coeff,
                              double *Pt, double *PT, double *fi ) {

  unsigned k, n;
  int info;
  double value = 0.0;

  double P[num_points*num_coeff];
  double PTPinv[num_coeff*num_coeff];
  double PTf[num_coeff];
  double PTPinv_PTf[num_coeff];

  transposed ( PT, P, num_coeff, num_points );

  info = mmdotprod ( PT, num_coeff, num_points, P, num_points, num_coeff, PTPinv, num_coeff, num_coeff );
  if ( info < 0 ) yac_internal_abort_message("internal ERROR: wrong dimensions for mmdotprod", __FILE__, __LINE__);

  inverse ( PTPinv, num_coeff );

  for ( k = 0; k < num_coeff; ++k ) {
    PTf[k] = 0.0;
    for ( n = 0; n < num_points; ++n )
       PTf[k] += PT[k*num_points+n]*fi[n];
  }

  for ( k = 0; k < num_coeff; ++k ) {
     PTPinv_PTf[k] = 0.0;
     for ( n = 0; n < num_coeff; ++n )
        PTPinv_PTf[k] += PTPinv[k*num_coeff+n] * PTf[n];
  }

  for ( k = 0; k < num_coeff; ++k )
     value += Pt[k] * PTPinv_PTf[k];

  return value;
}

double yac_patch_value_alt2 ( unsigned num_coeff, double *P, double *a ) {

  unsigned n;
  double value = 0.0;

  for ( n = 0; n < num_coeff; ++n ) value += P[n] * a[n];

  return value;
}

/* ----------- matrix operations ----------------- */

/**  Linear algebra: calculates the inverse of a matrix a of dim n x n.
  *
  *  @param[in] A           transposed of matrix P containing the polynomial terms
  *  @param[in] n           number of rows an columns of A
  *
 **/

static void inverse ( double* A, int n ) {

   lapack_int ipiv[n + 1];
   lapack_int info;
   lapack_int lwork = (lapack_int) n * n;

   if ( n > 512 ) yac_internal_abort_message("internal ERROR: to much memory required", __FILE__, __LINE__);

   double work[lwork];

   info = LAPACKE_dgetrf(LAPACK_COL_MAJOR, (lapack_int) n,
                         (lapack_int) n, A, (lapack_int) n, ipiv);
   if ( info != 0 )
       yac_internal_abort_message("internal ERROR: dgetrf", __FILE__, __LINE__);

   info = LAPACKE_dgetri_work(LAPACK_COL_MAJOR, (lapack_int) n,
                              A, (lapack_int) n, ipiv, work, lwork);
   if ( info != 0 )
       yac_internal_abort_message("internal ERROR: dgetri", __FILE__, __LINE__);

}
/**  Linear algebra: calculates the transposed matrix a of dim m x n.
  *
  *  @param[in]  A          matrix A
  *  @param[out] B          transposed of A
  *  @param[in]  M          number of rows (columns) of A (B)
  *  @param[in]  N          number of columns (rows) of A (B)
  *
 **/

static void transposed ( double* A, double* B, int M, int N ) {

  int m, n;

  for ( n = 0; n < N; ++n )
    for ( m = 0; m < M; ++m )
      B[n*M+m] = A[m*N+n];

}

/**  Linear algebra: calculates the transposed matrix a of dim m x n.
  *
  *  @param[in]  A          matrix A
  *  @param[in]  d_a1       number of rows of A
  *  @param[in]  d_a2       number of columns of A, needs to be equal to d_b1
  *  @param[in]  B          matrix A
  *  @param[in]  d_b1       number of rows of B, needs to be equal to d_a2
  *  @param[in]  d_b2       number of columns of B
  *  @param[out] C          matrix C = A B
  *  @param[in]  d_c1       number of rows of C, needs to be equal to d_a1
  *  @param[in]  d_c2       number of columns of C, needs to be equal to d_b2
  *
 **/

static int mmdotprod ( double* A, int d_a1, int d_a2, double* B, int d_b1, int d_b2,  double* C, int d_c1, int d_c2 ) {

  /* could be replaced by

     void BLAS_dgemm( enum blas_order_type order, enum blas_transpose_type transa,
                      enum blas_transpose_type transb, int m, int n, int k,
                      SCALAR_IN alpha, const ARRAY a, int lda, const ARRAY b,
                      int ldb, SCALAR_IN beta, ARRAY c, int ldc );

     http://en.wikipedia.org/wiki/General_Matrix_Multiply

     only needed for alternative 1 to calculate patch value at target point

  */

  int i, j, k;

  if ( d_a2 != d_b1 ) return -1;
  if ( d_c1 != d_a1 ) return -2;
  if ( d_c2 != d_b2 ) return -3;

  for ( i = 0; i < d_a1; ++i ) {
    for ( j = 0; j < d_b2; j++ ) {
       C[i*d_a1+j] = 0.0;
       for ( k = 0; k < d_a2; k++ )
          C[i*d_a1+j] += A[i*d_a2+k]*B[k*d_b2+j];
    }
  }

  return 0;

}

