/**
 * @file test_cell_search.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <string.h>

#include "tests.h"
#include "grid_search.h"
#include "bucket_search.h"
#include "sphere_part.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_unstruct.h"
#include "ensure_array_size.h"
#include "test_common.h"

typedef struct grid_search * (*grid_search_constructor)(struct grid *);

static void check_dep_lists(struct dep_list * lists[2]);

int main (void) {

   grid_search_constructor grid_search_construct[] =
      {yac_bucket_search_new, yac_sphere_part_search_new};
   unsigned num_grid_search_construct = sizeof(grid_search_construct) /
                                        sizeof(grid_search_construct[0]);

   {
      //---------------
      // setup
      //---------------
      size_t num_corners_per_cell[4] = {4,4,4,4};
      size_t * p_num_corners_per_cell;
      unsigned cell_vertices[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7}; 
      unsigned * p_cell_vertices;
      double vertex_coordinates_x[2][9] = {{0,1,2, 0,1,2, 0,1,2},
                                           {0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5}};
      double vertex_coordinates_y[2][9] = {{0,0,0, 1,1,1, 2,2,2},
                                           {0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5}};

      for ( unsigned i = 0; i < 2; ++i ) {
         for ( unsigned j = 0; j < 9; ++j ) {
            vertex_coordinates_x[i][j] *= YAC_RAD;
            vertex_coordinates_y[i][j] *= YAC_RAD;
         }
      }

      struct grid * grid_data_a, * grid_data_b;

      // set up grid data
      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_vertices = TO_POINTER(cell_vertices);
      struct dep_list * cell_to_vertex_a =
         yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

      grid_data_a = yac_unstruct_grid_new(vertex_coordinates_x[0], vertex_coordinates_y[0], 
                                          9, cell_to_vertex_a);

      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_vertices = TO_POINTER(cell_vertices);
      struct dep_list * cell_to_vertex_b =
         yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

      grid_data_b = yac_unstruct_grid_new(vertex_coordinates_x[1], vertex_coordinates_y[1], 
                                          9, cell_to_vertex_b);

      //---------------
      // testing
      //---------------

      for (unsigned i = 0; i < num_grid_search_construct; ++i) {

         struct grid_search * search = grid_search_construct[i](grid_data_a);
      
         struct dep_list * deps = yac_do_cell_search(search, grid_data_b);

         // test whether the search returned something in deps
         if ((deps == NULL) || (yac_get_num_elements(deps) == 0))
            PUT_ERR("do_cell_search returned no result\n\n")

         // test whether the number of found cells is correct
         if (yac_get_num_elements(deps) != 4)
            PUT_ERR("wrong number of elements in deps\n")

         { // test whether the number of overlaps per cell is correct
            size_t ref_num_deps_per_element[4] = {4,2,2,1};

            for (unsigned j = 0; j < 4; ++j)
               if (yac_get_num_dependencies_of_element(deps, j) !=
                   ref_num_deps_per_element[j])
                  PUT_ERR("wrong number of matching cells in "
                          "deps.num_deps_per_element[j]\n")
         }
         { // test whether the right cells were found
            unsigned ref_dependencies[4][4] = {{0,1,2,3}, {1,3,-1,-1},
                                             {2,3,-1,-1}, {3,-1,-1,-1}};

            for (size_t j = 0; j < 4; ++j) {

               size_t num_matches = 0;
               unsigned const * curr_deps =
                  yac_get_dependencies_of_element(deps, j);
               size_t curr_num_deps =
                  yac_get_num_dependencies_of_element(deps, j);

               for (size_t k = 0; k < curr_num_deps; ++k)
                  for (size_t l = 0; l < curr_num_deps; ++l)
                     if (curr_deps[k] == ref_dependencies[j][l]) ++num_matches;

               if (num_matches != curr_num_deps)
                  PUT_ERR("wrong dependencies in deps\n")
            }
         }

      //---------------
      // cleanup
      //---------------

         yac_dep_list_delete(deps);
         yac_delete_grid_search(search);
      }

      yac_delete_grid(grid_data_a);
      yac_delete_grid(grid_data_b);
      yac_dep_list_delete(cell_to_vertex_a);
      yac_dep_list_delete(cell_to_vertex_b);
   }

   {
      //---------------
      // setup
      //---------------
      size_t num_corners_per_cell[4] = {4,4,4,4};
      size_t * p_num_corners_per_cell;
      unsigned cell_vertices[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7}; 
      unsigned * p_cell_vertices;
      double vertex_coordinates_x[2][9] = {{0,1,2, 0,1,2, 0,1,2},
                                           {0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5}};
      double vertex_coordinates_y[2][9] = {{0,0,0, 1,1,1, 2,2,2},
                                           {0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5}};

      for ( unsigned i = 0; i < 2; ++i ) {
         for ( unsigned j = 0; j < 9; ++j ) {
            vertex_coordinates_x[i][j] *= YAC_RAD;
            vertex_coordinates_y[i][j] *= YAC_RAD;
         }
      }

      struct grid * grid_data_a, * grid_data_b;

      // set up grid data
      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_vertices = TO_POINTER(cell_vertices);
      struct dep_list * cell_to_vertex_a =
         yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

      grid_data_a = yac_unstruct_grid_new(vertex_coordinates_x[0], vertex_coordinates_y[0], 
                                          9, cell_to_vertex_a);

      p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
      p_cell_vertices = TO_POINTER(cell_vertices);
      struct dep_list * cell_to_vertex_b =
         yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

      grid_data_b = yac_unstruct_grid_new(vertex_coordinates_x[1], vertex_coordinates_y[1], 
                                          9, cell_to_vertex_b);

      //---------------
      // testing
      //---------------

      for (unsigned i = 0; i < num_grid_search_construct; ++i) {

         struct grid_search * search = grid_search_construct[i](grid_data_a);

         struct grid_cell cell;
         size_t *cells = NULL;
         size_t n_cells = 0, cells_size = 0;

         yac_init_grid_cell(&cell);

         for (unsigned j = 0; j < yac_get_num_grid_cells(grid_data_b); ++j) {

            yac_get_grid_cell(grid_data_b, j, &cell);

            yac_do_cell_search_single(
               search, cell, &n_cells, &cells_size, &cells);

            { // test whether the number of overlaps is correct
               size_t ref_num_deps_per_element[4] = {4,2,2,1};

               if (n_cells != ref_num_deps_per_element[j])
                  PUT_ERR("wrong number of matching cells in n_cells\n")
            }
            { // test whether the right cells were found
               size_t ref_dependencies[4][4] = {{0,1,2,3}, {1,3,-1,-1},
                                                {2,3,-1,-1}, {3,-1,-1,-1}};
               size_t curr_num_deps = 0;

               for (size_t k = 0; k < n_cells; ++k)
                  for (size_t l = 0; l < n_cells; ++l)
                     if (cells[k] == ref_dependencies[j][l]) ++curr_num_deps;

               if (curr_num_deps != n_cells)
                  PUT_ERR("wrong dependencies in deps\n")
            }

         }

      //---------------
      // cleanup
      //---------------

         yac_delete_grid_search(search);
         free(cells);
         yac_free_grid_cell(&cell);
      }

      yac_delete_grid(grid_data_a);
      yac_delete_grid(grid_data_b);
      yac_dep_list_delete(cell_to_vertex_a);
      yac_dep_list_delete(cell_to_vertex_b);
   }

   { // test cell search for cyclic grids
      //---------------
      // setup
      //---------------
      struct grid * grid_a;
      double coords_x_a[12] = {-180, -150, -120, -90, -60, -30, 0, 30, 60, 90, 120, 150};
      double coords_y_a[7] = {-90, -60, -30, 0, 30, 60, 90};

      for ( unsigned i = 0; i < 12; ++i ) coords_x_a[i] *= YAC_RAD;
      for ( unsigned i = 0; i <  7; ++i ) coords_y_a[i] *= YAC_RAD;

      unsigned num_cells_a[2] = {12,6};
      unsigned cyclic_a[2] = {1,0};
   
      grid_a = yac_reg2d_grid_new(coords_x_a, coords_y_a, num_cells_a, cyclic_a);

      struct grid * grid_b;
      double coords_x_b[11] = {-50, -40, -30, -20, -10, 0, 10, 20, 30, 40, 50};
      double coords_y_b[7] = {-30, -20, -10, 0, 10, 20, 30};

      for ( unsigned i = 0; i < 11; ++i ) coords_x_b[i] *= YAC_RAD;
      for ( unsigned i = 0; i <  7; ++i ) coords_y_b[i] *= YAC_RAD;

      unsigned num_cells_b[2] = {10,6};
      unsigned cyclic_b[2] = {0,0};
   
      grid_b = yac_reg2d_grid_new(coords_x_b, coords_y_b, num_cells_b, cyclic_b);

      //---------------
      // testing
      //---------------

      struct dep_list * deps[2];

      for (unsigned i = 0; i < num_grid_search_construct; ++i) {

         struct grid_search * search = grid_search_construct[i](grid_a);

         deps[i] = yac_do_cell_search(search, grid_b);

         // test whether the number of found cells is correct
         if (yac_get_num_elements(deps[i]) != 60)
            PUT_ERR("error in get_num_elements(deps)\n");

         yac_delete_grid_search(search);
      }

      check_dep_lists(deps);

      yac_dep_list_delete(deps[0]);
      yac_dep_list_delete(deps[1]);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid_a);
      yac_delete_grid(grid_b);
   }

   return TEST_EXIT_CODE;
}

static int compare_uint(const void * a, const void * b) {

  return ((*(unsigned*)a) > (*(unsigned*)b)) -
         ((*(unsigned*)a) < (*(unsigned*)b));
}

static void check_dep_lists(struct dep_list * lists[2]) {

   unsigned *results[2] = {NULL, NULL};
   unsigned results_size[2] = {0,0};

   size_t num_elements[2];
   num_elements[0] = yac_get_num_elements(lists[0]);
   num_elements[1] = yac_get_num_elements(lists[1]);

   if (num_elements[0] != num_elements[1])
      PUT_ERR("num_elements does not match\n");

   for (size_t i = 0; i < num_elements[0]; ++i) {

      size_t curr_num_deps[2];
      curr_num_deps[0] = yac_get_num_dependencies_of_element(lists[0], i);
      curr_num_deps[1] = yac_get_num_dependencies_of_element(lists[1], i);

      if (curr_num_deps[0] != curr_num_deps[1])
         PUT_ERR("num_deps_per_element does not match\n");

      for (size_t j = 0; j < 2; ++j) {

         ENSURE_ARRAY_SIZE(results[j], results_size[j], curr_num_deps[j]);
         memcpy(results[j], yac_get_dependencies_of_element(lists[j],i),
                curr_num_deps[j] * sizeof(**results));
         qsort(results[j], curr_num_deps[j], sizeof(**results), compare_uint);
      }

      for (size_t j = 0; j < curr_num_deps[0]; ++j)
         if (results[0][j] != results[1][j])
            PUT_ERR("dependencies do not match\n");
   }

   free(results[0]), free(results[1]);
}

