/**
 * @file test_field_interpolation_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mpi.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "test_common.h"

#define MAX_COLLECTION_SIZE (10)

static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm);

static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm);

int mapping_on_source;

int main(int argc, char *argv[]) {

   if (argc != 2) {
      PUT_ERR("wrong number of arguments\n");
      return TEST_EXIT_CODE;
   }

   mapping_on_source = strcmp(argv[1], "src") == 0;
   if (!mapping_on_source && strcmp(argv[1], "tgt")) {
      PUT_ERR("invalid argument (has to be either \"src\" or \"tgt\")\n");
      return TEST_EXIT_CODE;
   }

   struct communicator * global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 4)
   {

      int my_global_rank;

      my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source an target

      unsigned target = my_global_rank < 2;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
}

/*
 * The source domain is distributed among 2 processes, of which one process has
 * one and the other two subdomains.
 *
 * The global source domain is a 5x4 grid:
 *
 *    24--44--25--45--26--46--27--47--28--48--29
 *    |       |       |       |       |       |
 *    34  15  36  16  38  17  40  18  42  19  43
 *    |       |       |       |       |       |
 *    18--33--19--35--20--37--21--39--22--41--23
 *    |       |       |       |       |       |
 *    23  10  25  11  27  12  29  13  31  14  32
 *    |       |       |       |       |       |
 *    12--22--13--24--14--26--15--28--16--30--17
 *    |       |       |       |       |       |
 *    12  05  14  06  16  07  18  08  20  09  21
 *    |       |       |       |       |       |
 *    06--11--07--13--08--15--09--17--10--19--11
 *    |       |       |       |       |       |
 *    01  00  03  01  05  02  07  03  09  04  10
 *    |       |       |       |       |       |
 *    00--01--01--02--02--04--03--06--04--08--05
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    01--01--01--01--01--01--01--01--01--01--01
 *    |       |       |       |       |       |
 *    01  01  01  01  01  01  01  01  01  01  01
 *    |       |       |       |       |       |
 *    01--01--01--01--01--01--01--01--01--01--10
 *    |       |       |       |       |       |
 *    01  01  01  01  01  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    01--01--00--01--01--10--10--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  00  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--00--10--10--10--10--10--10
 *    |       |       |       |       |       |
 *    00  00  00  00  10  10  10  10  10  10  10
 *    |       |       |       |       |       |
 *    00--00--00--00--10--10--10--10--10--10--10
 */
static void source_main(struct communicator * global_comm,
                        struct communicator * source_comm) {

   unsigned my_rank = yac_get_comm_rank(source_comm);

   unsigned num_subdomains;

   if (my_rank == 0)
      num_subdomains = 2;
   else
      num_subdomains = 1;

   struct grid * grid[2];
   unsigned num_cells_p0[2][2] = {{2,2},{5,2}},
            num_cells_p1[2] =  {3,3};
   unsigned cyclic[2] = {0,0};
   double coordinates_x_p0[2][6] = {{0*YAC_RAD,1*YAC_RAD,2*YAC_RAD},
                                    {0*YAC_RAD,1*YAC_RAD,2*YAC_RAD,
                                     3*YAC_RAD,4*YAC_RAD,5*YAC_RAD}},
          coordinates_x_p1[4] = {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD,5*YAC_RAD};
   double coordinates_y_p0[2][3] = {{0,1*YAC_RAD,2*YAC_RAD},
                                    {2*YAC_RAD,3*YAC_RAD,4*YAC_RAD}};
   double coordinates_y_p1[4] = {0,1*YAC_RAD,2*YAC_RAD,3*YAC_RAD};

   if (my_rank == 0)
      for (unsigned i = 0; i < 2; ++i)
         grid[i] = yac_reg2d_grid_new(coordinates_x_p0[i], coordinates_y_p0[i],
                                      num_cells_p0[i], cyclic);
   else
      grid[0] = yac_reg2d_grid_new(coordinates_x_p1, coordinates_y_p1,
                                   num_cells_p1, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids_p0[2][18] =
      {{0,1,2,6,7,8,12,13,14},
       {12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
   unsigned corner_is_on_rank_p0[2][18] =
      {{-1,-1,1,-1,-1,-1,0,-1,0},
       {-1, 0,-1, 1, 1, 1,-1,-1,-1,-1,-1, 1,-1,-1,-1,-1,-1,-1}};
   int global_cell_ids_p0[2][10] =
      {{0,1,2,5,6,7},
       {10,11,12,13,14,15,16,17,18,19}};
   unsigned cell_is_on_rank_p0[2][10] =
      {{-1,-1,-1,-1,-1,-1},
       {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

   int global_corner_ids_p1[16] =
      { 2, 3, 4, 5,  8, 9,10,11, 14,15,16,17, 20,21,22,23};
   unsigned corner_is_on_rank_p1[16] =
      {-1,-1,-1,-1,  0,-1,-1,-1,  0,-1,-1,-1,  0, 0, 0,-1};
   int global_cell_ids_p1[9] =
      {2,3,4,7,8,9,12,13,14};
   unsigned cell_is_on_rank_p1[9] =
      {-1,-1,-1,-1,-1,-1,-1,-1,-1};

   if (my_rank == 0)
      for (unsigned i = 0; i < 2; ++i)
         yac_init_subdomain_unstruct(subdomain+i, grid[i], source_comm,
                                     global_corner_ids_p0[i], corner_is_on_rank_p0[i],
                                     global_cell_ids_p0[i], cell_is_on_rank_p0[i], NULL, NULL);
   else
      yac_init_subdomain_unstruct(subdomain+0, grid[0], source_comm,
                                  global_corner_ids_p1, corner_is_on_rank_p1,
                                  global_cell_ids_p1, cell_is_on_rank_p1, NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0, subdomain+1};

   yac_init_domain(&domain, source_comm, subdomains, num_subdomains, "source_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "target_domain", global_comm, 1, 0,
                                  "source_component", "target_component");
   yac_enddef_global_search(search);

   struct points points[2];

   if (my_rank == 0)
      for (unsigned i = 0; i < 2; ++i)
         yac_init_points(points+i, grid[i], CORNER,
                         coordinates_x_p0[i], coordinates_y_p0[i]);
   else
      yac_init_points(points+0, grid[0], CORNER,
                      coordinates_x_p1, coordinates_y_p1);

   struct coupling_field * field;
   struct points * points_per_subdomain[2][1] = {{points+0},{points+1}};
   struct points ** points_set[2] = {points_per_subdomain[0], points_per_subdomain[1]};

   struct component * component = NULL;
   unsigned local_comp_id = yac_pointer_to_unique_id((void*)component);

   if (my_rank == 0)
      field = yac_coupling_field_new("source field", local_comp_id, &domain, points_set, NULL, 1);
   else
      field = yac_coupling_field_new("source field", local_comp_id, &domain, points_set, NULL, 1);

   struct interpolation_method * method_fixed, * method_avg;
   double fixed_value = 1337;

   method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
   method_avg = yac_new_interpolation_method_avg(search, &domain, "avg stack",
                                                 points_set, NULL, 1, 0, 0, 0);

   //---------------------------
   // set up field interpolation
   //---------------------------

   struct field_interpolation_gen * interpolation_gen;
   struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                 method_fixed,
                                                                 NULL};

   interpolation_gen =
    yac_new_field_interpolation_gen_src(points_set, 1, "source field", search,
                                        "target field",
                                        interpolation_method_stack, NULL, 1,
                                        mapping_on_source);

   yac_enddef_field_interpolation_gen(interpolation_gen);

   struct field_interpolation * interpolation =
      yac_get_field_interpolation(interpolation_gen);

   //---------------------
   // do the interpolation
   //---------------------

   double *** source_data[MAX_COLLECTION_SIZE];

   double source_field_p0_subdomain0[MAX_COLLECTION_SIZE][9] =
      {{ 0, 1,-1, 6, 7, 8, -1,13,-1}};
   double * source_field_p0_subdomain0_[MAX_COLLECTION_SIZE][1];
   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
      source_field_p0_subdomain0_[i][0] = source_field_p0_subdomain0[i];

   double source_field_p0_subdomain1[MAX_COLLECTION_SIZE][18] =
      {{12,-1,14,-1,-1,-1, 18,19,20,21,22,-1, 24,25,26,27,28,29}};

   double * source_field_p0_subdomain1_[MAX_COLLECTION_SIZE][1];
   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
      source_field_p0_subdomain1_[i][0] = source_field_p0_subdomain1[i];

   double source_field_p1[MAX_COLLECTION_SIZE][16] =
      {{ 2, 3, 4, 5, -1, 9,10,11, -1,15,16,17, -1,-1,-1,23}};
   double * source_field_p1_[MAX_COLLECTION_SIZE][1];
   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
      source_field_p1_[i][0] = source_field_p1[i];

   double ** source_field_p0_subdomain0__[MAX_COLLECTION_SIZE][2];

   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i) {
      if (my_rank == 0) {
         source_field_p0_subdomain0__[i][0] = source_field_p0_subdomain0_[i];
         source_field_p0_subdomain0__[i][1] = source_field_p0_subdomain1_[i];
      } else {
         source_field_p0_subdomain0__[i][0] = source_field_p1_[i];
         source_field_p0_subdomain0__[i][1] = NULL;
      }
   }

   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
      source_data[i] = source_field_p0_subdomain0__[i];

   for (unsigned i = 1; i < MAX_COLLECTION_SIZE; ++i) {
      for (size_t j = 0;
           j < sizeof(source_field_p0_subdomain0[0]) /
               sizeof(source_field_p0_subdomain0[0][0]); ++j)
         if (double_are_unequal(source_field_p0_subdomain0[0][j], -1))
            source_field_p0_subdomain0[i][j] =
               source_field_p0_subdomain0[0][j] + i * 100;
         else source_field_p0_subdomain0[i][j] = -1;
      for (size_t j = 0;
           j < sizeof(source_field_p0_subdomain1[0]) /
               sizeof(source_field_p0_subdomain1[0][0]); ++j)
         if (double_are_unequal(source_field_p0_subdomain1[0][j], -1))
            source_field_p0_subdomain1[i][j] =
               source_field_p0_subdomain1[0][j] + i * 100;
         else source_field_p0_subdomain1[i][j] = -1;
      for (size_t j = 0;
           j < sizeof(source_field_p1[0]) /
               sizeof(source_field_p1[0][0]); ++j)
         if (double_are_unequal(source_field_p1[0][j], -1.0))
            source_field_p1[i][j] =
               source_field_p1[0][j] + i * 100;
         else source_field_p1[i][j] = -1;
   }


   for (unsigned collection_size = 1; collection_size <= MAX_COLLECTION_SIZE;
        ++collection_size) {
   // for (unsigned collection_size = 1; collection_size <= 1; ++collection_size) {

      yac_put_field(interpolation, source_data, collection_size);
   }

   //--------
   // cleanup
   //--------

   yac_delete_field_interpolation(interpolation);
   yac_delete_field_interpolation_gen(interpolation_gen);
   yac_delete_interpolation_method(method_avg);
   yac_delete_interpolation_method(method_fixed);
   yac_coupling_field_delete(field);
   yac_free_points(&(points[0]));
   if (my_rank == 0) yac_free_points(&(points[1]));
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (unsigned i = 0; i < num_subdomains; ++i)
      yac_delete_grid(grid[i]);

   yac_reset_interpolation_method();
}

/*
 * The target domain is distributed among 2 processes with one subdomain each
 *
 * The global target domain is a 5x3 grid:
 *
 *    21--39--22--40--23--41--24--42--25--43--26--44--27
 *    |       |       |       |       |       |       |
 *    27  12  29  13  31  14  33  15  35  16  37  17  38
 *    |       |       |       |       |       |       |
 *    14--26--15--28--16--30--17--32--18--34--19--36--20
 *    |       |       |       |       |       |       |
 *    14  06  16  07  18  08  20  09  22  10  24  11  25
 *    |       |       |       |       |       |       |
 *    07--13--08--15--09--17--10--19--11--21--12--23--13
 *    |       |       |       |       |       |       |
 *    01  00  03  01  05  02  07  03  09  04  11  05  12
 *    |       |       |       |       |       |       |
 *    00--00--01--02--02--04--03--06--04--08--05--10--06
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    00--00--00--00--00--00--00--10--10--10--10--10--10
 *    |       |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10  10  10
 *    |       |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10--10--10
 *    |       |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10  10  10
 *    |       |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10--10--10
 *    |       |       |       |       |       |       |
 *    00  00  00  00  00  00  00  10  10  10  10  10  10
 *    |       |       |       |       |       |       |
 *    00--00--00--00--00--00--00--10--10--10--10--10--10
 */
static void target_main(struct communicator * global_comm,
                        struct communicator * target_comm) {

   int my_rank = yac_get_comm_rank(target_comm);

   struct grid * grid;
   unsigned num_cells[2] = {3,3};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[2][4] = {{0.5,1.5,2.5,3.5},{3.5,4.5,5.5,6.5}};
   double coordinates_y[4] = {0.5,1.5,2.5,3.5};

   for ( unsigned j = 0; j < 4; ++j ) {
      coordinates_y[j] *= YAC_RAD;
      for ( unsigned i = 0; i < 2; ++i ) {
         coordinates_x[i][j] *= YAC_RAD;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_rank], coordinates_y,
                             num_cells, cyclic);

   struct subdomain subdomain;

   int global_corner_ids[2][16] =
      {{0,1,2,3, 7,8,9,10, 14,15,16,17, 21,22,23,24},
       {3,4,5,6, 10,11,12,13, 17,18,19,20, 24,25,26,27}};
   unsigned corner_is_on_rank[2][16] =
      {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
       {0,-1,-1,-1,0,-1,-1,-1,0,-1,-1,-1,0,-1,-1,-1}};
   int global_cell_ids[2][9] = {{0,1,2,6,7,8,12,13,14}, {3,4,5,9,10,11,15,16,17}};
   unsigned cell_is_on_rank[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                     {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids[my_rank], corner_is_on_rank[my_rank],
                               global_cell_ids[my_rank], cell_is_on_rank[my_rank], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[1] = {&subdomain};

   yac_init_domain(&domain, target_comm, subdomains, 1, "target_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "source_domain", global_comm, 0, 1,
                                  "target_component", "source_component");
   yac_enddef_global_search(search);

   struct points points;

   yac_init_points(&points, grid, CORNER, coordinates_x[my_rank], coordinates_y);

   struct component * component = NULL;
   unsigned local_comp_id = yac_pointer_to_unique_id((void*)component);

   struct coupling_field * field;
   struct points * points_per_subdomain[1] = {&points};
   struct points ** points_set[1] = {points_per_subdomain};

   field = yac_coupling_field_new("target field", local_comp_id, &domain,
                                  points_set, NULL, 1);

   //---------------------------
   // set up field interpolation
   //---------------------------

   struct field_interpolation_gen * interpolation_gen;

   interpolation_gen =
    yac_new_field_interpolation_gen_tgt(points_set, 1, "target field", search,
                                        "source field", 0, mapping_on_source);

   yac_enddef_field_interpolation_gen(interpolation_gen);

   struct field_interpolation * interpolation =
      yac_get_field_interpolation(interpolation_gen);

   //---------------------
   // do the interpolation
   //---------------------

   double target_field[MAX_COLLECTION_SIZE][16];
   double * target_field_[MAX_COLLECTION_SIZE][1];
   for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
     target_field_[i][0] = target_field[i];
   double ** target_data[MAX_COLLECTION_SIZE];
      for (unsigned i = 0; i < MAX_COLLECTION_SIZE; ++i)
         target_data[i] = target_field_[i];

   for (unsigned collection_size = 1; collection_size <= MAX_COLLECTION_SIZE;
        ++collection_size) {
   // for (unsigned collection_size = 1; collection_size <= 1; ++collection_size) {

      for (unsigned i = 0; i < collection_size; ++i)
        for (unsigned j = 0; j < 16; ++j)
          target_field[i][j] = -1;

      yac_get_field(interpolation, target_data, collection_size);

   //----------------------------
   // check interpolation results
   //----------------------------

      double ref_target_data[2][16] = {{3.5,4.5,5.5,6.5,
                                        9.5,10.5,11.5,12.5,
                                        15.5,16.5,17.5,18.5,
                                        21.5,22.5,23.5,24.5},
                                       {-1,7.5,1337,-1,
                                        -1,13.5,1337,-1,
                                        -1,19.5,1337,-1,
                                        -1,25.5,1337,-1}};
      unsigned field_size[2] = {16,16};

      for (unsigned i = 0; i < collection_size; ++i) {
        for (unsigned j = 0; j < field_size[my_rank]; ++j) {
          if ((double_are_equal(ref_target_data[my_rank][j], -1.0)) ||
              (double_are_equal(ref_target_data[my_rank][j], 1337.0))) {
            if (double_are_unequal(target_data[i][0][j],
                                   ref_target_data[my_rank][j]))
               PUT_ERR("error in interpolated data on target side\n");
          } else {
            if (double_are_unequal(
                  target_data[i][0][j],
                  ref_target_data[my_rank][j] + (double)(i * 100)))
               PUT_ERR("error in interpolated data on target side\n");
          }
        }
      }
   }

   //---------
   // clean up
   //---------

   yac_delete_field_interpolation_gen(interpolation_gen);
   yac_delete_field_interpolation(interpolation);
   yac_coupling_field_delete(field);
   yac_free_points(&points);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

