/**
 * @file test_interpolation_method_file.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>

#include "tests.h"
#include "test_common.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_file.h"
#include "utils.h"
#include "weight_file_common.h"
#include "test_interpolation_method_common.h"

double const tol = 1e-7;
double const err_tol = 1e-14;

static void finish_callback(void * user_data);

int main (void) {

   { // simple weight file interpolation (one link per target point) with two
     // source processes and a single target process the source fields have two
     // struct points per subdomain (the test check how the interpolation
     // methods handle NULL target coordinates )

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      char const file_name[] = "weights.nc";

      int src_indices[] = {0,1,2,3,4,5,6,7,8,9,10,11};
      int tgt_indices[] = {0,1,2,3,4,5,6,7,8,9,10,11};
      double weights[] = {0,1,2,3,4,5,6,7,8,9,10,11};
      unsigned num_links = 12;
      enum yac_location src_locations[2] = {CORNER, CELL};
      enum yac_location tgt_location = CORNER;
      unsigned num_pointsets = 2;
      int num_links_per_pointset[2] = {num_links, 0};
      int * tgt_id_fixed = NULL;
      unsigned num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      unsigned num_fixed_values = 0;

      // create weight file
      write_weight_file(file_name, src_indices, tgt_indices, weights, num_links,
                        src_locations, num_pointsets, num_links_per_pointset,
                        tgt_id_fixed, num_fixed_tgt, fixed_values,
                        num_tgt_per_fixed_value, num_fixed_values,
                        tgt_location, "domain_src", "domain_tgt");

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt",
                                           comm_p1_src, 1,0, "component_src",
                                           "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt",
                                           comm_p2_src, 1,0, "component_src",
                                           "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src",
                                           comm_p1_tgt, 0,1, "component_tgt",
                                           "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_file[2];

      method_file[0] =
        yac_new_interpolation_method_file(search[0], &l_domain_p1_src,
                                          "file stack", src_locations, 2,
                                          file_name, 1);
      method_file[1] =
        yac_new_interpolation_method_file(search[1], &l_domain_p2_src,
                                          "file stack", src_locations, 2,
                                          file_name, 1);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_file[0],
                                                           NULL},
                                                          {method_file[1],
                                                           NULL}};
      struct interpolation_plan_element *
        interpolation_plan_element_stack[2][2] = {{NULL, NULL}, {NULL, NULL}};
      int global_ids[2][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}};
      unsigned num_points = 9;
      unsigned num_interpolated_points[2];

      unsigned finish_callback_flag[2] = {0, 0};

      yac_set_finish_callback_interpolation_method(
        method_stack[0], finish_callback, &finish_callback_flag[0]);
      yac_set_finish_callback_interpolation_method(
        method_stack[1], finish_callback, &finish_callback_flag[1]);

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
         global_ids[0], NULL, CORNER,
         num_points, num_interpolated_points+0, 2);
      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
         global_ids[1], NULL, CORNER,
         num_points, num_interpolated_points+1, 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,-1,4,5,-1,8,9,10}},
                                     {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {-1,2,3,-1,6,7,-1,-1,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      if ((finish_callback_flag[0] != 1) || (finish_callback_flag[1] != 1))
        PUT_ERR("finish_callback has not been called\n");

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{0, NULL, NULL}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2] = {5,4};
      unsigned ref_num_interpolated_points_per_process[2] =
         {ref_num_interpolated_points[0], ref_num_interpolated_points[1]};
      int ref_global_ids[2][5] = {{0,1,4,5,8},{2,3,6,7}};
      double ref_out_data[2][5] = {{0*0,1*1,4*4,5*5,8*8},{2*2,3*3,6*6,7*7}};

      for (unsigned i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points_per_process[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (double_are_unequal(send_buffer[i][j], ref_out_data[i][j]))
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();

      // delete weight file
      unlink(file_name);
   }

   { // simple weight file interpolation (multiple links per target point, not
     // all required source points are available) with two source processes and
     // a single target process the source fields have two struct points per
     // subdomain (the test check how the interpolation methods handle NULL
     // target coordiantes)

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      char const file_name[] = "weights.nc";

      int tgt_indices[] = { 0, 0, 0, 0,
                            1, 1, 1, 1,
                            2, 2, 2, 2,
                            3, 3, 3, 3,
                            4, 4, 4, 4,
                            5, 5, 5, 5,
                            6, 6, 6, 6};
      int src_indices[] = { 0, 1, 4, 5,
                            1, 2, 5, 6,
                            2, 3, 6, 7,
                            4, 5, 8, 9,
                            5, 6, 9,10,
                            6, 7,10,11,
                            7, 8,12,11};
      double weights[] = {0.1,0.2,0.3,0.4,
                          0.5,0.6,0.7,0.8,
                          0.9,1.0,1.1,1.2,
                          1.3,1.4,1.5,1.6,
                          1.7,1.8,1.9,2.0,
                          2.1,2.2,2.3,2.4,
                          2.5,2.6,2.7,2.8};
      unsigned num_links = 28;
      unsigned num_pointsets = 2;
      int num_links_per_pointset[2] = {num_links, 0};
      enum yac_location src_locations[2] = {CORNER, CELL};
      enum yac_location tgt_location = CORNER;
      int * tgt_id_fixed = NULL;
      unsigned num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      unsigned num_fixed_values = 0;

      // create weight file
      write_weight_file(file_name, src_indices, tgt_indices, weights, num_links,
                        src_locations, num_pointsets, num_links_per_pointset,
                        tgt_id_fixed, num_fixed_tgt, fixed_values,
                        num_tgt_per_fixed_value, num_fixed_values,
                        tgt_location, "domain_src", "domain_tgt");

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt",
                                           comm_p1_src, 1,0, "component_src",
                                           "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt",
                                           comm_p2_src, 1,0, "component_src",
                                           "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src",
                                           comm_p1_tgt, 0,1, "component_tgt",
                                           "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_file[2];

      method_file[0] =
        yac_new_interpolation_method_file(search[0], &l_domain_p1_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 2);
      method_file[1] =
        yac_new_interpolation_method_file(search[1], &l_domain_p2_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 2);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_file[0],
                                                           NULL},
                                                          {method_file[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] =
        {{NULL, NULL}, {NULL, NULL}};
      int global_ids[2][4] = {{0,3,4,6}, {1,2,5,6}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[2];

      yac_do_search_interpolation_method(method_stack[0],
                                         interpolation_plan_element_stack[0],
                                         global_ids[0], NULL, CORNER,
                                         num_points, &num_interpolated_points[0], 2);
      yac_do_search_interpolation_method(method_stack[1],
                                         interpolation_plan_element_stack[1],
                                         global_ids[1], NULL, CORNER,
                                         num_points, &num_interpolated_points[1], 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{6,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,-1,4,5,-1,8,9,10}},
                                     {{1,5,10,-1,-1,-1,-1,-1,-1},
                                      {-1,2,3,-1,6,7,-1,-1,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][3];

      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{1, (int[]){6}, (int[]){1}}, {0, NULL, NULL}},
          {{3, (int[]){1,5,10}, (int[]){0,0,0}}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2] = {3, 3};
      int ref_global_ids[2][3] = {{0,3,4},{1,2,5}};
      double ref_out_data[2][3] = {{0.1*0+0.2*1+0.3*4+0.4*5,
                                    1.3*4+1.4*5+1.5*8+1.6*9,
                                    1.7*5+1.8*6+1.9*9+2.0*10},
                                   {0.5*1+0.6*2+0.7*5+0.8*6,
                                    0.9*2+1.0*3+1.1*6+1.2*7,
                                    2.1*6+2.2*7+2.3*10+2.4*11}};

      for (unsigned i = 0; i < 2; ++i)
        if (ref_num_interpolated_points[i] != num_interpolated_points[i])
           PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (fabs(send_buffer[i][j]-ref_out_data[i][j]) > err_tol)
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();

      // delete weight file
      unlink(file_name);
   }

   { // simple weight file interpolation (weight file contains no weights and
     // not fixed target points) with two source processes and
     // a single target process the source fields have two struct points per
     // subdomain

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      char const file_name[] = "weights.nc";

      int * tgt_indices = NULL;
      int * src_indices = NULL;
      double * weights = NULL;
      unsigned num_links = 0;
      unsigned num_pointsets = 2;
      int num_links_per_pointset[2] = {0, 0};
      enum yac_location src_locations[2] = {CORNER, CELL};
      enum yac_location tgt_location = CORNER;
      int * tgt_id_fixed = NULL;
      unsigned num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      unsigned num_fixed_values = 0;

      // create weight file
      write_weight_file(file_name, src_indices, tgt_indices, weights, num_links,
                        src_locations, num_pointsets, num_links_per_pointset,
                        tgt_id_fixed, num_fixed_tgt, fixed_values,
                        num_tgt_per_fixed_value, num_fixed_values,
                        tgt_location, "domain_src", "domain_tgt");

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt",
                                           comm_p1_src, 1,0, "component_src",
                                           "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt",
                                           comm_p2_src, 1,0, "component_src",
                                           "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src",
                                           comm_p1_tgt, 0,1, "component_tgt",
                                           "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_file[2];

      method_file[0] =
        yac_new_interpolation_method_file(search[0], &l_domain_p1_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 1);
      method_file[1] =
        yac_new_interpolation_method_file(search[1], &l_domain_p2_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 1);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_file[0],
                                                           NULL},
                                                          {method_file[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] =
        {{NULL, NULL}, {NULL, NULL}};
      int global_ids[2][4] = {{0,3,4,6}, {1,2,5,6}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[2];

      yac_do_search_interpolation_method(method_stack[0],
                                         interpolation_plan_element_stack[0],
                                         global_ids[0], NULL, CORNER,
                                         num_points, &num_interpolated_points[0], 2);
      yac_do_search_interpolation_method(method_stack[1],
                                         interpolation_plan_element_stack[1],
                                         global_ids[1], NULL, CORNER,
                                         num_points, &num_interpolated_points[1], 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,-1,4,5,-1,8,9,10}},
                                     {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {-1,2,3,-1,6,7,-1,-1,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][3];

      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{0, NULL, NULL}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2] = {0, 0};

      for (unsigned i = 0; i < 2; ++i)
        if (ref_num_interpolated_points[i] != num_interpolated_points[i])
           PUT_ERR("wrong number of interpolated points\n");

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();

      // delete weight file
      unlink(file_name);
   }

   { // simple weight file interpolation (weight file contains only fixed target
     // points) with two source processes and a single target process the source
     // fields have two struct points per subdomain

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       | 
      //       08  03  10  04  12  05  13
      //       |       |       |       | 
      //       04--07--05--09--06--11--07
      //       |       |       |       | 
      //       01  00  03  01  05  02  06
      //       |       |       |       | 
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      char const file_name[] = "weights.nc";

      int * tgt_indices = NULL;
      int * src_indices = NULL;
      double * weights = NULL;
      unsigned num_links = 0;
      unsigned num_pointsets = 2;
      int num_links_per_pointset[2] = {0, 0};
      enum yac_location src_locations[2] = {CORNER, CELL};
      enum yac_location tgt_location = CORNER;
      int tgt_id_fixed[] = {0, 2, 4, 6};
      unsigned num_fixed_tgt = 4;
      double fixed_values[] = {-1.0, -2.0};
      int num_tgt_per_fixed_value[] = {1,3};
      unsigned num_fixed_values = 2;

      // create weight file
      write_weight_file(file_name, src_indices, tgt_indices, weights, num_links,
                        src_locations, num_pointsets, num_links_per_pointset,
                        tgt_id_fixed, num_fixed_tgt, fixed_values,
                        num_tgt_per_fixed_value, num_fixed_values,
                        tgt_location, "domain_src", "domain_tgt");

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();
      
      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt",
                                           comm_p1_src, 1,0, "component_src",
                                           "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt",
                                           comm_p2_src, 1,0, "component_src",
                                           "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src",
                                           comm_p1_tgt, 0,1, "component_tgt",
                                           "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_file[2];

      method_file[0] =
        yac_new_interpolation_method_file(search[0], &l_domain_p1_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 2);
      method_file[1] =
        yac_new_interpolation_method_file(search[1], &l_domain_p2_src,
                                          "file stack", src_locations,
                                          num_pointsets, file_name, 2);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_file[0],
                                                           NULL},
                                                          {method_file[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] =
        {{NULL, NULL}, {NULL, NULL}};
      int global_ids[2][4] = {{0,3,4,6}, {1,2,5,6}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[2];

      yac_do_search_interpolation_method(method_stack[0],
                                         interpolation_plan_element_stack[0],
                                         global_ids[0], NULL, CORNER,
                                         num_points, &num_interpolated_points[0], 2);
      yac_do_search_interpolation_method(method_stack[1],
                                         interpolation_plan_element_stack[1],
                                         global_ids[1], NULL, CORNER,
                                         num_points, &num_interpolated_points[1], 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,-1,4,5,-1,8,9,10}},
                                     {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {-1,2,3,-1,6,7,-1,-1,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][3];

      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{0, NULL, NULL}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2] = {2, 2};
      int ref_global_ids[2][2] = {{0,4}, {2,6}};
      double ref_out_data[2][2] = {{-1,-2}, {-2,-2}};

      for (unsigned i = 0; i < 2; ++i)
        if (ref_num_interpolated_points[i] != num_interpolated_points[i])
           PUT_ERR("wrong number of interpolated points\n");

      for (unsigned i = 0; i < 2; ++i)
         for (unsigned j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (fabs(send_buffer[i][j]-ref_out_data[i][j]) > err_tol)
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (unsigned i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();

      // delete weight file
      unlink(file_name);
   }

   yac_comm_finalize();

   return TEST_EXIT_CODE;
}

static void finish_callback(void * user_data) {
  ++*((unsigned*)user_data);
}

