/**
 * @file test_point_search.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <string.h>

#include "tests.h"
#include "grid_search.h"
#include "bucket_search.h"
#include "sphere_part.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_unstruct.h"
#include "test_common.h"

typedef struct grid_search * (*grid_search_constructor)(struct grid *);

int main (void) {

   grid_search_constructor grid_search_construct[] =
      {yac_bucket_search_new, yac_sphere_part_search_new};
   size_t num_grid_search_construct = sizeof(grid_search_construct) /
                                      sizeof(grid_search_construct[0]);

   //---------------
   // setup
   //---------------
   size_t num_corners_per_cell[4] = {4,4,4,4};
   unsigned cell_vertices[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7};
   size_t * p_num_corners_per_cell;
   unsigned * p_cell_vertices;
   double vertex_coordinates_x[2][9] = {{0,1,2, 0,1,2, 0,1,2},
                                        {0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5}};
   double vertex_coordinates_y[2][9] = {{0,0,0, 1,1,1, 2,2,2},
                                        {0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5}};

   for ( unsigned i = 0; i < 2; ++i ) {
      for ( unsigned j = 0; j < 9; ++j ) {
         vertex_coordinates_x[i][j] *= YAC_RAD;
         vertex_coordinates_y[i][j] *= YAC_RAD;
      }
   }

   struct grid * grid_data_a, * grid_data_b;
   struct dep_list * cell_to_vertex_a, * cell_to_vertex_b;

   { // set up grid data
   p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
   p_cell_vertices = TO_POINTER(cell_vertices);
   cell_to_vertex_a =
      yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

   grid_data_a =
      yac_unstruct_grid_new(vertex_coordinates_x[0], vertex_coordinates_y[0],
                            9, cell_to_vertex_a);

   p_num_corners_per_cell = TO_POINTER(num_corners_per_cell);
   p_cell_vertices = TO_POINTER(cell_vertices);
   cell_to_vertex_b =
      yac_dep_list_new(4, p_num_corners_per_cell, p_cell_vertices);

   grid_data_b =
      yac_unstruct_grid_new(vertex_coordinates_x[1], vertex_coordinates_y[1],
                            9, cell_to_vertex_b);
   }

   //---------------
   // testing
   //---------------
   { // test do_point_search

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         struct grid_search * search = grid_search_construct[n](grid_data_a);
         struct dep_list * deps = yac_do_point_search_c(search, grid_data_b);

         {
            size_t ref_num_deps[9] = {1,1,0,1,1,0,0,0,0};
            unsigned ref_deps[9] = {0,1,-1,2,3,-1,-1,-1,-1};

            if (yac_get_total_num_dependencies(deps) != 4)
               PUT_ERR("error in do_point_search_c(get_total_num_dependencies(deps))\n");

            for (size_t i = 0; i < 9; ++i)
               if (yac_get_num_dependencies_of_element(deps, i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_c(deps.num_deps_per_element[i])\n");

            for (size_t i = 0; i < 9; ++i) {

               if (yac_get_num_dependencies_of_element(deps, i) == 0) continue;

               if (yac_get_dependencies_of_element(deps,i)[0] != ref_deps[i])
                  PUT_ERR("error in do_point_search_c(get_dependencies_of_element)\n");
            }
         }

         yac_dep_list_delete(deps);
         yac_delete_grid_search(search);
      }
   }

   { // test do_point_search_p

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         struct grid_search * search = grid_search_construct[n](grid_data_a);
         struct dep_list * deps = yac_do_point_search_p(search, grid_data_b);

         {
            size_t ref_num_deps[9] = {4,4,0,4,4,0,0,0,0};
            unsigned ref_deps[9][4] = {{0,1,4,3}, {1,2,5,4}, {0,0,0,0},
                                       {3,4,7,6}, {4,5,8,7}, {0,0,0,0},
                                       {0,0,0,0}, {0,0,0,0}, {0,0,0,0}};

            if (yac_get_total_num_dependencies(deps) != 16)
               PUT_ERR("error in do_point_search_p(get_total_num_dependencies(deps))\n");

            for (unsigned i = 0; i < 9; ++i)
               if (yac_get_num_dependencies_of_element(deps, i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p(deps.num_deps_per_element[i])\n");

            for (size_t i = 0; i < 9; ++i) {

               unsigned const * curr_src_corners =
                  yac_get_dependencies_of_element(deps,i);

               unsigned num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_corners[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p(get_dependencies_of_element)\n");
            }
         }

         yac_dep_list_delete(deps);
         yac_delete_grid_search(search);
      }
   }

   yac_delete_grid(grid_data_a);
   yac_delete_grid(grid_data_b);
   yac_dep_list_delete(cell_to_vertex_a);
   yac_dep_list_delete(cell_to_vertex_b);

   { // test do_point_search_p3 and do_point_search_c3 for YAC_LOCATION_CELL

      //-------------------
      // set up
      //-------------------

      // set up points structure

      double grid_coord_x[] = {0,1,2,3,4,5};
      double grid_coord_y[] = {0,1,2,3,4,5};
      double points_coord_x[] = {0.5, 1.5, 2.5, 3.5, 4.5};
      double points_coord_y[] = {0.5, 1.5, 2.5, 3.5, 4.5};

      for ( unsigned i = 0; i < 6; ++i ) {
         grid_coord_x[i] *= YAC_RAD;
         grid_coord_y[i] *= YAC_RAD;
      }

      for ( unsigned i = 0; i < 5; ++i ) {
         points_coord_x[i] *= YAC_RAD;
         points_coord_y[i] *= YAC_RAD;
       }

      unsigned num_cells[2] = {5,5};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);
      yac_init_points(&points, reg_grid, CELL, points_coord_x, points_coord_y);

      // define points that are to be seached for

      double search_points_coord_x[] = {4.75,1.25,3.5,0.25,2.25,0.75,3.25,5.25};
      double search_points_coord_y[] = {0.75,1.25,1.5, 2.5,2.75,4.25,4.25,4.25};
      size_t num_search_points = sizeof(search_points_coord_x) /
                                 sizeof(search_points_coord_x[0]);
      double search_points_coord_xyz[num_search_points][3];

      for ( size_t i = 0; i < num_search_points; ++i )
         LLtoXYZ_deg(search_points_coord_x[i],
                     search_points_coord_y[i],
                     search_points_coord_xyz[i]);


      // init search data structure

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         struct grid_search * search = grid_search_construct[n](reg_grid);

         //--------------
         // testing
         //--------------

         struct dep_list * search_results[2];

         // do search
         search_results[0] = yac_do_point_search_p3(
            search, search_points_coord_xyz, num_search_points, &points);
         search_results[1] = yac_do_point_search_c3(
            search, search_points_coord_xyz, num_search_points, &points);

         { // test p3 data
            size_t ref_num_deps[8] = {0,4,4,0,4,4,4,0};
            unsigned ref_deps[8][4] = {{-1}, {0,1,6,5}, {2,3,8,7},{-1},
                                       {11,12,17,16}, {15,16,21,20},{17,18,23,22},{-1}};

            if (yac_get_total_num_dependencies(search_results[0]) != 20)
               PUT_ERR("error in do_point_search_p3(get_total_num_dependencies(search_results))\n");

            for (unsigned i = 0; i < 8; ++i)
               if (yac_get_num_dependencies_of_element(search_results[0], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(search_results[0].num_deps_per_element[i])\n");

            for (size_t i = 0; i < 8; ++i) {

               unsigned const * curr_src_corners =
                  yac_get_dependencies_of_element(search_results[0],i);

               unsigned num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_corners[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(get_dependencies_of_element)\n");
            }
         }

         { // test c3 data
            size_t ref_num_deps[8] = {0,1,4,0,1,1,1,0};
            unsigned ref_deps[8][4] = {{-1}, {0}, {2,3,6,7},{-1},
                                       {9}, {12},{14},{-1}};

            if (yac_get_total_num_dependencies(search_results[1]) != 8)
               PUT_ERR("error in do_point_search_c3(get_total_num_dependencies(search_results))\n");

            for (unsigned i = 0; i < 8; ++i)
               if (yac_get_num_dependencies_of_element(search_results[1], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_c3(search_results[1].num_deps_per_element[i])\n");

            unsigned num_matches;
            for (size_t i = 0; i < 8; ++i) {

               unsigned const * curr_src_cells =
                  yac_get_dependencies_of_element(search_results[1], i);

               num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_cells[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_c3(get_dependencies_of_element)\n");
            }
         }

         yac_dep_list_delete(search_results[0]);
         yac_dep_list_delete(search_results[1]);
         yac_delete_grid_search(search);
      }

      yac_delete_grid(reg_grid);
      yac_free_points(&points);
   }

   { // test do_point_search_p3 and do_point_search_c3 for YAC_LOCATION_CORNER

      //-------------------
      // set up
      //-------------------

      // set up points structure

      double grid_coord_x[] = {0,1,2,3,4,5};
      double grid_coord_y[] = {0,1,2,3,4,5};
      double points_coord_x[] = {0,1,2,3,4,5};
      double points_coord_y[] = {0,1,2,3,4,5};

      for ( unsigned i = 0; i < 6; ++i ) {
         grid_coord_x[i] *= YAC_RAD;
         grid_coord_y[i] *= YAC_RAD;
      }

      for ( unsigned i = 0; i < 6; ++i ) {
         points_coord_x[i] *= YAC_RAD;
         points_coord_y[i] *= YAC_RAD;
       }

      unsigned num_cells[2] = {5,5};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);
      yac_init_points(&points, reg_grid, CORNER, points_coord_x, points_coord_y);

      // define points that are to be seached for

      double search_points_coord_x[] = {4.75,1.25,3.5,0.25,2.25,0.75,3.25,5.25,4.0,2};
      double search_points_coord_y[] = {0.75,1.25,1.5, 2.5,2.75,4.25,4.25,4.25,0.0,3};
      size_t num_search_points = sizeof(search_points_coord_x) /
                                 sizeof(search_points_coord_x[0]);
      double search_points_coord_xyz[num_search_points][3];

      for ( size_t i = 0; i < num_search_points; ++i )
         LLtoXYZ_deg(search_points_coord_x[i],
                     search_points_coord_y[i],
                     search_points_coord_xyz[i]);


      // init search data structure

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         struct grid_search * search = grid_search_construct[n](reg_grid);

         //--------------
         // testing
         //--------------

         struct dep_list * search_results[2];

         // do search
         search_results[0] = yac_do_point_search_p3(
            search, search_points_coord_xyz, num_search_points, &points);
         search_results[1] = yac_do_point_search_c3(
            search, search_points_coord_xyz, num_search_points, &points);

         { // test p3 data
            size_t ref_num_deps[10] = {4,4,4,4,4,4,4,0,4,4};
            unsigned ref_deps[4][10][4] = {{{4,5,11,10}, {7,8,14,13}, {9,10,16,15},
                                            {12,13,19,18},{14,15,21,20},
                                            {24,25,31,30},{27,28,34,33},{-1},
                                            {3,4,10,9}, {13,14,20,19}},
                                           {{-1}, {-1}, {-1},{-1},{-1},{-1},{-1},{-1},
                                            {4,5,11,10},{14,15,21,20}},
                                           {{-1}, {-1}, {-1},{-1},{-1},{-1},{-1},{-1},
                                            {-1},{19,20,26,25}},
                                           {{-1}, {-1}, {-1},{-1},{-1},{-1},{-1},{-1},
                                            {-1},{20,21,27,26}}};

            if (yac_get_total_num_dependencies(search_results[0]) != 36)
               PUT_ERR("error in do_point_search_p3(get_total_num_dependencies(search_results[0]))\n");

            for (unsigned i = 0; i < 10; ++i)
               if (yac_get_num_dependencies_of_element(search_results[0], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(search_results[0].num_deps_per_element[i])\n");

            for (size_t i = 0; i < 10; ++i) {

               unsigned const * curr_src_corners =
                  yac_get_dependencies_of_element(search_results[0],i);

               int match_flag = 0;
               // for all possible solutions
               for (unsigned l = 0; l < 4; ++l) {
                  unsigned num_matches = 0;
                  for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                     for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                        if (ref_deps[l][i][j] == curr_src_corners[k])
                           ++num_matches;
                  if (num_matches == ref_num_deps[i]) {
                    match_flag = 1;
                    break;
                  }
               }

               if (!match_flag)
                  PUT_ERR("error in do_point_search_p3(get_dependencies_of_element)\n");
            }
         }

         { // test c3 data
            size_t ref_num_deps[10] = {1,1,1,1,1,1,1,0,2,4};
            unsigned ref_deps[10][4] = {{4}, {6}, {8},{10},
                                        {12}, {20},{23},{-1},{3,4},
                                        {11,12,16,17}};

            if (yac_get_total_num_dependencies(search_results[1]) != 13)
               PUT_ERR("error in do_point_search_c3(get_total_num_dependencies(search_results[1]))\n");

            for (unsigned i = 0; i < 10; ++i)
               if (yac_get_num_dependencies_of_element(search_results[1], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_c3(search_results[1].num_deps_per_element[i])\n");

            unsigned num_matches;
            for (size_t i = 0; i < 10; ++i) {

               unsigned const * curr_src_cells =
                  yac_get_dependencies_of_element(search_results[1], i);

               num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_cells[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_c3(get_dependencies_of_element)\n");
            }
         }

         yac_dep_list_delete(search_results[0]);
         yac_dep_list_delete(search_results[1]);
         yac_delete_grid_search(search);
      }

      yac_delete_grid(reg_grid);
      yac_free_points(&points);
   }

   { // test do_point_search_p3 and do_point_search_c3 for YAC_LOCATION_EDGE

      //-------------------
      // set up
      //-------------------

      // set up points structure

      double grid_coord_x[] = {0,1,2,3,4};
      double grid_coord_y[] = {0,1,2,3,4};
      double points_coord_x[] = {0.5,0,1.5,1,2.5,2,3.5,3,4,
                                 0.5,0,1.5,1,2.5,2,3.5,3,4,
                                 0.5,0,1.5,1,2.5,2,3.5,3,4,
                                 0.5,0,1.5,1,2.5,2,3.5,3,4,
                                 0.5,1.5,2.5,3.5};
      double points_coord_y[] = {0,0.5,0,0.5,0,0.5,0,0.5,0.5,
                                 1,1.5,1,1.5,1,1.5,1,1.5,1.5,
                                 2,2.5,2,2.5,2,2.5,2,2.5,2.5,
                                 3,3.5,3,3.5,3,3.5,3,3.5,3.5,
                                 4,4,4,4};

      for ( unsigned i = 0; i < 5; ++i ) {
         grid_coord_x[i] *= YAC_RAD;
         grid_coord_y[i] *= YAC_RAD;
      }

      for ( unsigned i = 0;
            i < sizeof(points_coord_x)/sizeof(points_coord_x[0]); ++i ) {
         points_coord_x[i] *= YAC_RAD;
         points_coord_y[i] *= YAC_RAD;
       }

      unsigned num_cells[2] = {4,4};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;
      struct points points;

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);
      yac_init_points(&points, reg_grid, EDGE, points_coord_x, points_coord_y);

      // define points that are to be seached for

      double search_points_coord_x[] = {0.5,2,0.5,4,3,2.5,4.5,3.5};
      double search_points_coord_y[] = {0.5,2,3.5,3,1,3.5,3.5,3};
      size_t num_search_points = sizeof(search_points_coord_x) /
                                 sizeof(search_points_coord_x[0]);
      double search_points_coord_xyz[num_search_points][3];

      for ( size_t i = 0; i < num_search_points; ++i )
         LLtoXYZ_deg(search_points_coord_x[i],
                     search_points_coord_y[i],
                     search_points_coord_xyz[i]);

      // init search data structure

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         struct grid_search * search = grid_search_construct[n](reg_grid);

         //--------------
         // testing
         //--------------

         struct dep_list * search_results[2];

         // do search
         search_results[0] = yac_do_point_search_p3(
            search, search_points_coord_xyz, num_search_points, &points);
         search_results[1] = yac_do_point_search_c3(
            search, search_points_coord_xyz, num_search_points, &points);

         { // test p3 data
            size_t ref_num_deps[8] = {4,4,4,0,4,4,0,4};
            unsigned ref_deps[8][4] = {{0,3,9,1}, {14,22,23,20}, {27,30,36,28}, {-1},
                                       {7,15,16,13}, {31,34,38,32},{-1},{24,26,33,25}};

            if (yac_get_total_num_dependencies(search_results[0]) != 24)
               PUT_ERR("error in do_point_search_p3(get_total_num_dependencies(search_results[0]))\n");

            for (unsigned i = 0; i < 8; ++i)
               if (yac_get_num_dependencies_of_element(search_results[0], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(search_results[0].num_deps_per_element[i])\n");

            for (size_t i = 0; i < 8; ++i) {

               unsigned const * curr_src_corners =
                  yac_get_dependencies_of_element(search_results[0], i);

               unsigned num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_corners[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(get_dependencies_of_element)\n");
            }
         }

         { // test c3 data
            size_t ref_num_deps[8] = {1,1,1,0,1,1,0,3};
            unsigned ref_deps[8][3] = {{0}, {20}, {12}, {-1},
                                       {18}, {14},{-1},{11,15,24}};

            if (yac_get_total_num_dependencies(search_results[1]) != 8)
               PUT_ERR("error in do_point_search_p3(get_total_num_dependencies(search_results[1]))\n");

            for (unsigned i = 0; i < 8; ++i)
               if (yac_get_num_dependencies_of_element(search_results[1], i) !=
                   ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(search_results[1].num_deps_per_element[i])\n");

            unsigned num_matches;
            for (size_t i = 0; i < 8; ++i) {

               unsigned const * curr_src_cells =
                  yac_get_dependencies_of_element(search_results[1], i);

               num_matches = 0;

               for (unsigned j = 0; j < ref_num_deps[i]; ++j)
                  for (unsigned k = 0; k < ref_num_deps[i]; ++k)
                     if (ref_deps[i][j] == curr_src_cells[k])
                        ++num_matches;

               if (num_matches != ref_num_deps[i])
                  PUT_ERR("error in do_point_search_p3(get_dependencies_of_element)\n");
            }
         }

         yac_dep_list_delete(search_results[0]);
         yac_dep_list_delete(search_results[1]);
         yac_delete_grid_search(search);
      }

      yac_delete_grid(reg_grid);
      yac_free_points(&points);
   }

   { // test do_point_search_p4

      //-------------------
      // set up
      //-------------------

      // set up points structure

      double grid_coord_x[] = {0.5*YAC_RAD, 1.5*YAC_RAD, 2.5*YAC_RAD, 3.5*YAC_RAD, 4.5*YAC_RAD};
      double grid_coord_y[] = {0.5*YAC_RAD, 1.5*YAC_RAD, 2.5*YAC_RAD, 3.5*YAC_RAD, 4.5*YAC_RAD};

      unsigned num_cells[2] = {4,4};
      unsigned cyclic[2] = {0,0};
      struct grid * reg_grid;

      reg_grid = yac_reg2d_grid_new(grid_coord_x, grid_coord_y, num_cells, cyclic);

      // define points that are to be seached for

      double search_points_coord_x[] = {4.75,1.25,3.5,0.25,2.25,0.75,3.25,5.25};
      double search_points_coord_y[] = {0.75,1.25,1.5,2.5,2.75,4.25,4.25,4.25};
      unsigned num_search_points = sizeof(search_points_coord_x) /
                                   sizeof(search_points_coord_x[0]);
      double search_points_coord_xyz[num_search_points][3];

      for ( unsigned i = 0; i < num_search_points; ++i )
         LLtoXYZ_deg(search_points_coord_x[i],
                     search_points_coord_y[i],
                     search_points_coord_xyz[i]);

      for (size_t n = 0; n < num_grid_search_construct; ++n) {

         // init search data structure
         struct grid_search * search = grid_search_construct[n](reg_grid);

         //--------------
         // testing
         //--------------

         size_t * deps = NULL;
         size_t deps_size = 0;

         for (unsigned j = 0; j < num_search_points; ++j) {

            size_t num_deps;

            // do search
            yac_do_point_search_p4(
               search, search_points_coord_xyz[j], &num_deps, &deps_size, &deps);

            {
               size_t ref_num_deps[8] = {0,4,4,0,4,4,4,0};
               size_t ref_deps[8][4] = {{-1}, {0,1,6,5}, {2,3,8,7},{-1},
                                        {11,12,17,16}, {15,16,21,20},{17,18,23,22},{-1}};

               if (num_deps != ref_num_deps[j])
                  PUT_ERR("error in do_point_search_p3(search_results.num_deps_per_element[j])\n");

               size_t num_matches = 0;

               for (size_t k = 0; k < ref_num_deps[j]; ++k)
                  for (size_t l = 0; l < ref_num_deps[j]; ++l)
                     if (ref_deps[j][k] == deps[l])
                        ++num_matches;

               if (num_matches != ref_num_deps[j])
                  PUT_ERR("error in do_point_search_p3(get_dependencies_of_element)\n");
            }
         }

         free(deps);
         yac_delete_grid_search(search);
      }

      yac_delete_grid(reg_grid);
   }

   return TEST_EXIT_CODE;
}

