/**
 * @file test_read_fesom.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <netcdf.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "tests.h"
#include "read_fesom_grid.h"
#include "grid.h"
#include "grid2vtk.h"

static void write_dummy_grid_file(char * name);
static void check_grid(struct grid * grid);

int main(void) {

   write_dummy_grid_file("fesom_grid.nc");

   struct grid * fesom_grid;

   fesom_grid = read_fesom_grid("fesom_grid.nc");

   unlink("fesom_grid.nc");

   check_grid(fesom_grid);

#define WRITE_VTK_GRID_FILE
#ifdef WRITE_VTK_GRID_FILE
   yac_write_grid_to_file(fesom_grid, "fesom");
#endif // WRITE_VTK_GRID_FILE

   yac_delete_grid(fesom_grid);

   return TEST_EXIT_CODE;
}

static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

static void write_dummy_grid_file(char * file_name) {

  int ncid, status;

  // create file
  status = nc_create(file_name, NC_CLOBBER, &ncid);
  if (status != NC_NOERR) handle_error(status);

  int dim_ncells_id;
  int dim_nv_id;
  int dim_ids[2];

  // define dimensions
  status = nc_def_dim(ncid, "ncells", 8, &dim_ncells_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_dim(ncid, "nv", 9, &dim_nv_id);
  if (status != NC_NOERR) handle_error(status);

  dim_ids[0] = dim_ncells_id;
  dim_ids[1] = dim_nv_id;

  int var_lon_id, var_lon_v_id, var_lat_id, var_lat_v_id;

  // define grid arrays
  status = nc_def_var(ncid, "lon", NC_DOUBLE, 1, &dim_ncells_id, &var_lon_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "lon_vertices", NC_DOUBLE, 2, dim_ids, &var_lon_v_id);
  status = nc_def_var(ncid, "lat", NC_DOUBLE, 1, &dim_ncells_id, &var_lat_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "lat_vertices", NC_DOUBLE, 2, dim_ids, &var_lat_v_id);
  if (status != NC_NOERR) handle_error(status);

  // end definition
  status = nc_enddef(ncid);
  if (status != NC_NOERR) handle_error(status);

  // write grid data

  double lon[8] = {1,2,3,4,3,3,5.5,5.5};
  double lon_vertices[8*9] = {0,1,2,2,1,0,0,0,0,
                              1,3,2,2,2,2,2,2,2,
                              3,4,2,2,2,2,2,2,2,
                              3,5,4,4,4,4,4,4,4,
                              2,4,4,2,2,2,2,2,2,
                              1,2,4,5,4,2,2,2,2,
                              4,6,7,6,5,4,4,4,4,
                              5,6,4,4,4,4,4,4,4};
  double lat[8] = {1.5,0.5,0.5,0.5,1.5,3,1.5,0.5};
  double lat_vertices[8*9] = {1,0,1,2,3,2,2,2,2,
                              0,0,1,1,1,1,1,1,1,
                              0,1,1,1,1,1,1,1,1,
                              0,0,1,1,1,1,1,1,1,
                              1,1,2,2,2,2,2,2,2,
                              3,2,2,3,4,4,4,4,4,
                              1,1,2,3,3,2,2,2,2,
                              0,1,1,1,1,1,1,1,1};

  status = nc_put_var_double(ncid, var_lon_id, lon);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lon_v_id, lon_vertices);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lat_id, lat);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lat_v_id, lat_vertices);
  if (status != NC_NOERR) handle_error(status);

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

static void check_grid(struct grid * grid) {

  unsigned ref_num_cells = 8;

  if (yac_get_num_grid_cells(grid) != ref_num_cells)
    PUT_ERR("wrong number of grid cells");

  unsigned ref_num_corners_per_cell[8] = {6,3,3,3,4,6,6,3};

  for (unsigned i = 0; i < ref_num_cells; ++i)
    if (yac_get_num_cell_corners(grid, i) != ref_num_corners_per_cell[i])
      PUT_ERR("wrong number of corners per CELL");
}

